package com.ibm.stg.rtc.ext.common;

import java.sql.Timestamp;
import java.util.Calendar;
import java.util.Date;
import java.util.HashSet;
import java.util.Set;

import com.ibm.team.process.common.IIteration;
import com.ibm.team.process.common.IProcessConfigurationElement;
import com.ibm.team.workitem.common.ISaveParameter;
import com.ibm.team.workitem.common.model.WorkItemApprovals;

public class ApprovalContainer {

	private static final String TAG_APPROVAL = "approval";
	private static final String TAG_DUE_DATE = "dueDate";
	private static final String TAG_PREVENT_DELIVER = "preventDeliver";
	private static final String ATTR_REQUIRED_APPROVAL_STATE_TO_DELIVER = "requiredApprovalStateToDeliver";
	private static final String ATTR_WORKITEM_TYPE = "workItemType";
	private static final String ATTR_NAME = "name";
	private static final String ATTR_APPROVAL_TYPE = "approvalType";
	private static final String ATTR_INTERNAL_RELEASE_ID = "internalReleaseId";
	private static final String ATTR_DAYS = "days";
	private static final String ATTR_REFERENCE = "reference";
	private static final String AFTER_CREATION_REFERENCE = "afterCreation";
	private static final String BEFORE_ITERATION_END_REFERENCE = "beforeIterationEnd";

	private String workItemType;
	private String approvalType;
	private String name;
	private String internalReleaseId;
	private Set<ApprovalWorkItemCondition> workItemConditions;
	private Timestamp dueDate;
	private Set<String> requiredApprovalStatesToDeliver;
	private Boolean preventDeliver;
	private IProcessConfigurationElement approvalElement;

	public ApprovalContainer(IProcessConfigurationElement element){
		if(!element.getName().equals(TAG_APPROVAL)){
			throw new IllegalArgumentException("ApprovalContainer only supports elements with the tag <" + TAG_APPROVAL + ">.");
		}
		this.approvalElement = element;
		// call the required attributes here so an exception is thrown on create
		this.getName();
		this.getApprovalType();
		this.getWorkItemType();

		// wait until the get methods are called to parse the xml to improve performance 
	}

	public String getWorkItemType(){
		if(this.workItemType == null){
			this.workItemType = this.approvalElement.getAttribute(ATTR_WORKITEM_TYPE);
		}
		if(this.workItemType == null || this.workItemType.length() == 0){
			throw new IllegalStateException("Each <" + ApprovalContainer.TAG_APPROVAL + "> tag must contain the " + ApprovalContainer.ATTR_WORKITEM_TYPE + " attribute.");
		}
		return this.workItemType;
	}

	public String getApprovalType(){
		if(this.approvalType == null){
			this.approvalType = this.approvalElement.getAttribute(ATTR_APPROVAL_TYPE);
		}
		if(!WorkItemApprovals.getTypes().contains(WorkItemApprovals.getType(this.approvalType))){
			throw new IllegalStateException("Each <" + ApprovalContainer.TAG_APPROVAL + "> tag must contain a valid " + ApprovalContainer.ATTR_APPROVAL_TYPE + " attribute.");
		}
		return this.approvalType;
	}

	public String getName(){
		if(this.name == null){
			String configuredName = this.approvalElement.getAttribute(ATTR_NAME);
			if(configuredName == null || configuredName.length() == 0){
				throw new IllegalStateException("Each <" + ApprovalContainer.TAG_APPROVAL + "> tag must contain the " + ApprovalContainer.ATTR_NAME + " attribute.");
			}
			this.name = ApprovalsHelper.generateApprovalDisplayName(configuredName);
		}
		return this.name;
	}
	
	/**
	 * 
	 * @return internal release identifier if it is configured for this approval, null if not configured by internal release
	 */
	public String getInternalReleaseId() {
		if (this.internalReleaseId == null) {
			this.internalReleaseId = this.approvalElement.getAttribute(ATTR_INTERNAL_RELEASE_ID);
		}
		return internalReleaseId;
	}

	/**
	 * returns true if the approval has a configuration for the current save parameter state
	 * @param saveParameter
	 * @return
	 */
	public boolean hasConditionConfig(ISaveParameter saveParameter){
		if(this.workItemConditions == null){
			populateApprovalWorkItemConditions();
		}

		for(ApprovalWorkItemCondition condition : this.workItemConditions){
			if(condition.matches(saveParameter)){
				return true;
			}
		}
		return false;
	}

	public Set<ApprovalWorkItemCondition> getAllConditionConfigs(){
		if(this.workItemConditions == null){
			populateApprovalWorkItemConditions();
		}
		return this.workItemConditions;
	}

	/**
	 * returns all the configured conditions that match the current save parameter state
	 * @param saveParameter
	 * @return
	 */
	public Set<ApprovalWorkItemCondition> getConditionConfigs(ISaveParameter saveParameter){
		if(this.workItemConditions == null){
			populateApprovalWorkItemConditions();
		}

		// search for the matching conditions
		Set<ApprovalWorkItemCondition> matches = new HashSet<ApprovalWorkItemCondition>();
		for(ApprovalWorkItemCondition condition : this.workItemConditions){
			if(condition.matches(saveParameter)){
				matches.add(condition);
			}
		}
		return matches;
	}

	private void populateApprovalWorkItemConditions(){
		this.workItemConditions = new HashSet<ApprovalWorkItemCondition>();
		IProcessConfigurationElement[] children = this.approvalElement.getChildren();
		for(IProcessConfigurationElement child : children){
			if(child.getName().equals(WorkItemCondition.TAG_CONDITION)){
				this.workItemConditions.add(new ApprovalWorkItemCondition(child));
			}
		}
	}

	public Timestamp getDueDate(IIteration iteration){
		if(this.dueDate == null){

			// the due date is a child element
			IProcessConfigurationElement[] children = this.approvalElement.getChildren();
			for(IProcessConfigurationElement childElement : children){
				if(childElement.getName().equals(TAG_DUE_DATE)){
					String daysAttr = childElement.getAttribute(ATTR_DAYS);
					int days;
					try{
						days = Integer.parseInt(daysAttr);
					}catch(Exception e){
						throw new IllegalStateException("Error parsing days '" + daysAttr + "' to integer: " + e.getMessage());
					}
					String referenceAttr = childElement.getAttribute(ATTR_REFERENCE);

					Calendar cal = Calendar.getInstance();
					if(AFTER_CREATION_REFERENCE.equals(referenceAttr)){
						cal.add(Calendar.DATE, days);
						this.dueDate = new Timestamp(cal.getTimeInMillis());

					} else if(BEFORE_ITERATION_END_REFERENCE.equals(referenceAttr)){
						if(iteration == null){
							throw new IllegalArgumentException("The approval due date could not be calculated because an iteration could not be found. Please ensure this project has the project timeline set and it has a current iteration.");
						}
						Date iterEndDate = iteration.getEndDate();
						if(iterEndDate == null){
							throw new IllegalStateException("The iteration end date is not set.");
						}
						cal.setTime(iterEndDate);
						cal.add(Calendar.DATE, -days);
						this.dueDate = new Timestamp(cal.getTimeInMillis());
					} else{
						// no default date, just return null
					}
					break; // there can only be one due date child
				}
			}
		}
		return this.dueDate;
	}

	/**
	 * returns a set of all the auto actions configured for this approval, save parameter state, and approval state
	 * @param saveParameter
	 * @param approvalState
	 * @return
	 */
	public Set<String> getAutoActions(ISaveParameter saveParameter, String approvalState){
		Set<String> actions = new HashSet<String>();
		Set<ApprovalWorkItemCondition> conditions = this.getConditionConfigs(saveParameter);
		for(ApprovalWorkItemCondition condition : conditions){
			if(condition.hasAutoAction(approvalState)){
				actions.add(condition.getAutoAction(approvalState));
			}
		}
		return actions;
	}

	public boolean hasAutoAction(ISaveParameter saveParameter){
		Set<ApprovalWorkItemCondition> conditions = this.getConditionConfigs(saveParameter);
		for(ApprovalWorkItemCondition condition : conditions){
			if(condition.hasAutoAction()){
				return true;
			}
		}
		return false;
	}

	public boolean isCreate(ISaveParameter saveParameter){
		Set<ApprovalWorkItemCondition> conditions = this.getConditionConfigs(saveParameter);
		for(ApprovalWorkItemCondition condition : conditions){
			if(condition.isCreate()){
				return true;
			}
		}
		return false;
	}

	public Set<ApprovalWorkItemCondition> getCreateConditions(){
		Set<ApprovalWorkItemCondition> createConditions = new HashSet<ApprovalWorkItemCondition>();
		if(this.workItemConditions == null){
			populateApprovalWorkItemConditions();
		}
		
		for(ApprovalWorkItemCondition condition : this.workItemConditions){
			if(condition.isCreate()){
				createConditions.add(condition);
			}
		}
		return createConditions;
	}

	public boolean isRemove(ISaveParameter saveParameter){
		Set<ApprovalWorkItemCondition> conditions = this.getConditionConfigs(saveParameter);
		for(ApprovalWorkItemCondition condition : conditions){
			if(condition.isRemove()){
				return true;
			}
		}
		return false;
	}
	
	public Set<String> getRequiredApprovalStatesToDeliver() {
		if (this.requiredApprovalStatesToDeliver == null) {
			populateRequiredApprovalStatesToDeliver();
		}
		return this.requiredApprovalStatesToDeliver;
	}
	
	public boolean hasPreventDeliver() {
		if (this.preventDeliver == null) {
			populateRequiredApprovalStatesToDeliver();
		}
		return this.preventDeliver; 
	}
	
	private void populateRequiredApprovalStatesToDeliver() {
		this.requiredApprovalStatesToDeliver = new HashSet<String>();
		preventDeliver = false;
		
		// the preventDeliver is a child element
		IProcessConfigurationElement[] children = this.approvalElement.getChildren();
		for(IProcessConfigurationElement childElement : children){
			if(childElement.getName().equals(TAG_PREVENT_DELIVER)){
				preventDeliver = true;
				String state = childElement.getAttribute(ATTR_REQUIRED_APPROVAL_STATE_TO_DELIVER);
				this.requiredApprovalStatesToDeliver.add(state);
			}
		}
	}

	public Set<ApprovalWorkItemCondition> getRemoveConditions(){
		Set<ApprovalWorkItemCondition> removeConditions = new HashSet<ApprovalWorkItemCondition>();
		if(this.workItemConditions == null){
			populateApprovalWorkItemConditions();
		}
		
		for(ApprovalWorkItemCondition condition : this.workItemConditions){
			if(condition.isRemove()){
				removeConditions.add(condition);
			}
		}
		return removeConditions;
	}

	public Set<String> getRequiredApprovalStates(ISaveParameter saveParameter){
		Set<String> approvalStates = new HashSet<String>();
		Set<ApprovalWorkItemCondition> conditions = this.getConditionConfigs(saveParameter);
		for(ApprovalWorkItemCondition condition : conditions){
			approvalStates.addAll(condition.getRequiredApprovalStates());
		}
		return approvalStates;
	}

	public boolean hasRequiredApprovalState(ISaveParameter saveParameter){
		return (getRequiredApprovalStates(saveParameter).size() > 0);
	}

	@Override
	public boolean equals(Object o){
		if(o == this) return true;
		if(o instanceof ApprovalContainer){
			ApprovalContainer ac = (ApprovalContainer)o;
			try{
				if(ac.getName().equals(this.getName())){
					return true;
				}
			} catch(Exception e){
				return false;
			}
		}
		return false;
	}
}
