package com.ibm.stg.rtc.ext.common;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import com.ibm.team.process.common.IProcessConfigurationElement;
import com.ibm.team.workitem.common.model.WorkItemApprovals;

/**
 * @author Kyle Christianson
 *
 */
public class ApprovalWorkItemCondition extends WorkItemCondition{

	private static final String TAG_CREATE = "createApproval";
	private static final String TAG_REMOVE = "removeApproval";
	private static final String TAG_REQUIRED_APPROVAL_STATE = "requiredApprovalState";
	private static final String TAG_AUTO_ACTION = "autoWorkItemAction";
	private static final String ATTR_WORKITEM_ACTION = "workItemAction";
	private static final String ATTR_APPROVAL_STATE = "approvalState";
	private static final String TAG_ALLOW_MODIFY_ROLES = "allowModifyRoles";
	private static final String ATTR_REMOVE_APPROVER = "removeApprover";
	private static final String ATTR_ADD_APPROVER = "addApprover";
	private static final String ATTR_DUE_DATE = "dueDate";
	private static final String TAG_ROLE = "role";
	private static final String VALUE_FALSE = "false";

	private Set<String> requiredApprovalStates;
	private Boolean create;
	private Boolean remove;
	private Boolean modifyApprovals;
	private Map<String, String> autoWorkItemActions;
	private Set<String> addApproverRoles;
	private Set<String> removeApproverRoles;
	private Set<String> modifyDueDateRoles;

	public ApprovalWorkItemCondition(IProcessConfigurationElement element){
		super(element);
		// wait until the get methods are called to parse the xml to improve performance
	}

	public boolean isCreate(){
		if(this.create == null){
			this.create = ConfigurationHelper.getBooleanTagValue(super.conditionElement, TAG_CREATE);
		}
		return this.create;
	}

	public boolean isRemove(){
		if(this.remove == null){
			this.remove = ConfigurationHelper.getBooleanTagValue(super.conditionElement, TAG_REMOVE);
		}
		return this.remove;
	}
		
	public Set<String> getRemoveApproverRoles() {
		if (this.removeApproverRoles == null) {
			populateAllowModifyRoles();
		}
		return removeApproverRoles;
	}
	
	public Set<String> getAddApproverRoles() {
		if (this.addApproverRoles == null) {
			populateAllowModifyRoles();
		}
		return addApproverRoles;
	}
	
	public Set<String> getModifyDueDateRoles() {
		if (this.modifyDueDateRoles == null) {
			populateAllowModifyRoles();
		}
		return modifyDueDateRoles;
	}
	
	public boolean isModifyApprovals() {
		if (this.modifyApprovals == null) {
			populateAllowModifyRoles();
		}
		return this.modifyApprovals;
	}

	/**
	 * @return true if one or more approval states are required
	 */
	public boolean hasRequiredApprovalState(){
		return (this.getRequiredApprovalStates().size() > 0);
	}

	public Set<String> getRequiredApprovalStates(){
		if(this.requiredApprovalStates == null){
			populateRequiredApprovalStates();
		}
		return this.requiredApprovalStates;
	}

	/**
	 * @return true if this work item state config has one or more auto actions
	 */
	public boolean hasAutoAction(){
		if(this.autoWorkItemActions == null){
			populateAutoActions();
		}
		return !this.autoWorkItemActions.isEmpty();
	}

	public boolean hasAutoAction(String approvalState){
		if(this.autoWorkItemActions == null){
			populateAutoActions();
		}
		return this.autoWorkItemActions.containsKey(approvalState);
	}

	public String getAutoAction(String approvalState){
		if(this.autoWorkItemActions == null){
			populateAutoActions();
		}
		return this.autoWorkItemActions.get(approvalState);
	}

	private void populateRequiredApprovalStates(){
		this.requiredApprovalStates = new HashSet<String>();
		IProcessConfigurationElement[] children = super.conditionElement.getChildren();
		for(IProcessConfigurationElement child : children){
			if(child.getName().equals(TAG_REQUIRED_APPROVAL_STATE)){
				String approvalStateId = child.getAttribute(ATTR_APPROVAL_STATE);
				if(approvalStateId == null || approvalStateId.length() == 0){
					throw new IllegalStateException("Each " + TAG_REQUIRED_APPROVAL_STATE + " element must contain the " + ATTR_APPROVAL_STATE + " attribute.");
				}else if(!WorkItemApprovals.getStates().contains(WorkItemApprovals.getState(approvalStateId))){
					throw new IllegalStateException(approvalStateId + " is an invalid approval state");
				}
				else{
					this.requiredApprovalStates.add(approvalStateId);
				}
			}
		}
	}

	private void populateAutoActions() {
		this.autoWorkItemActions = new HashMap<String, String>();
		IProcessConfigurationElement[] children = super.conditionElement.getChildren();
		for(IProcessConfigurationElement child : children){
			if(child.getName().equals(TAG_AUTO_ACTION)){
				String approvalState = child.getAttribute(ATTR_APPROVAL_STATE);
				if(approvalState == null || approvalState.length() == 0){
					throw new IllegalStateException("Each " + TAG_AUTO_ACTION + " element must contain the " + ATTR_APPROVAL_STATE + " attribute.");
				} else if(!WorkItemApprovals.getStates().contains(WorkItemApprovals.getState(approvalState))){
					throw new IllegalStateException(approvalState + " is an invalid approval state.");
				}
				String actionId = child.getAttribute(ATTR_WORKITEM_ACTION);
				if(actionId == null || actionId.length() == 0){
					throw new IllegalStateException("Each " + TAG_AUTO_ACTION + " must contain the " + ATTR_WORKITEM_ACTION + " attribute.");
				}
				this.autoWorkItemActions.put(approvalState, actionId);
			}
		}
	}
	
	private void populateAllowModifyRoles() {
		this.addApproverRoles = new HashSet<String>();
		this.removeApproverRoles = new HashSet<String>();
		this.modifyDueDateRoles = new HashSet<String>();
		modifyApprovals = false;
		IProcessConfigurationElement[] children = super.conditionElement.getChildren();
		boolean allowModifyRolesTagFound = false;	
		for(IProcessConfigurationElement roles : children){			
			if(roles.getName().equals(TAG_ALLOW_MODIFY_ROLES)) {
				allowModifyRolesTagFound = true;
				for (IProcessConfigurationElement role : roles.getChildren()) {
					modifyApprovals = true;
					String roleId = role.getAttribute(ConfigurationHelper.ATTR_ID);
					if(roleId == null || roleId.length() == 0){
						throw new IllegalStateException("Each " + TAG_ROLE + " element must contain the " + ConfigurationHelper.ATTR_ID + " attribute.");
					}
					//the default for the attributes are true unless specified otherwise
					//if the value is false, do not add this role to the list
					if (!VALUE_FALSE.equals(role.getAttribute(ATTR_ADD_APPROVER))) {
						addApproverRoles.add(roleId);
					}
					if (!VALUE_FALSE.equals(role.getAttribute(ATTR_REMOVE_APPROVER))) {
						removeApproverRoles.add(roleId);
					}
					if (!VALUE_FALSE.equals(role.getAttribute(ATTR_DUE_DATE))) {
						modifyDueDateRoles.add(roleId);
					}					
				}				
			}
		}
		//default the roles if it was appropriate - allowModifyRoles tag was not found in the condition and this condition isCreate
		if (!allowModifyRolesTagFound && isCreate()) {
			modifyApprovals = true;
			addApproverRoles.add("Product Owner");
			removeApproverRoles.add("Product Owner");
			modifyDueDateRoles.add("Product Owner");
		}
	}
}
