package com.ibm.stg.rtc.ext.common;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;

import com.ibm.team.process.common.IProcessArea;
import com.ibm.team.process.common.IProcessAreaHandle;
import com.ibm.team.process.common.IProcessConfigurationData;
import com.ibm.team.process.common.IProcessConfigurationElement;
import com.ibm.team.process.common.IProjectAreaHandle;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.workitem.common.IAuditableCommon;
import com.ibm.team.workitem.common.ISaveParameter;
import com.ibm.team.workitem.common.IWorkItemCommon;
import com.ibm.team.workitem.common.model.IApproval;
import com.ibm.team.workitem.common.model.IApprovalDescriptor;
import com.ibm.team.workitem.common.model.IWorkItem;
import com.ibm.team.workitem.common.model.ItemProfile;

public class ApprovalsHelper {
	private static final String APPROVER_LISTS_CONFIGURATION = "com.ibm.stg.workitem.configuration.approverlists";
	private static final String TAG_APPROVALS = "approvals";
	private static final String TAG_APPROVER_LISTS = "approverLists";

	/**
	 * creates the full approval name by adding the AUTO_GENERATE_PREFIX
	 * 
	 * @param name
	 * @return
	 */
	public static String generateApprovalDisplayName(String name) {
		return STGCommonIDs.STG_AUTO_GENERATED_ID + " " + name;
	}

	public static String removeGeneratedApprovalName(String generatedName) {
		return generatedName.substring(STGCommonIDs.STG_AUTO_GENERATED_ID.length() + 1);
	}

	/**
	 * checks whether or not the approval was auto generated by looking to see
	 * if the name contains the STG_AUTO_GENERATED_ID string
	 * 
	 * @param approval
	 * @return
	 */
	public static boolean isApprovalSTGAutoGenerated(IApprovalDescriptor approval) {
		String name = approval.getName();
		if (name.contains(STGCommonIDs.STG_AUTO_GENERATED_ID)) {
			return true;
		}
		else {
			return false;
		}
	}

	/**
	 * Searches the work item's approvals for an approval descriptor with the
	 * given name.
	 * 
	 * @param workItem
	 *            the work item to search through the approvals
	 * @param name
	 *            of the approval to search for
	 * @return the descriptor matching the name otherwise null if not found
	 */
	public static IApprovalDescriptor findApprovalDescriptorByName(IWorkItem workItem, String name) {
		if (workItem == null) {
			return null;
		}
		List<IApprovalDescriptor> allDescriptors = workItem.getApprovals().getDescriptors();
		for (IApprovalDescriptor descriptor : allDescriptors) {
			if (descriptor.getName().equals(name)) {
				return descriptor;
			}
		}
		return null;
	}

	/**
	 * returns a list of all the stg auto generated approvals on the work item
	 * 
	 * @param workItem
	 * @return
	 */
	public static List<IApprovalDescriptor> getAllAutoGeneratedApprovals(IWorkItem workItem) {
		List<IApprovalDescriptor> generatedApprovals = new ArrayList<IApprovalDescriptor>();
		if (workItem == null) {
			return generatedApprovals;
		}

		List<IApprovalDescriptor> allApprovals = workItem.getApprovals().getDescriptors();

		for (IApprovalDescriptor approval : allApprovals) {
			if (isApprovalSTGAutoGenerated(approval) == true) {
				generatedApprovals.add(approval);
			}
		}

		return generatedApprovals;
	}

	/**
	 * returns all the approvals configured for the given work item type
	 * 
	 * @param config
	 * @param workItemType
	 * @return
	 * @throws TeamRepositoryException
	 * @throws Exception
	 *             if the configuration is invalid
	 */
	public static List<ApprovalContainer> getConfiguredApprovals(IProcessConfigurationElement config,
			IWorkItem workItem, IWorkItemCommon workItemCommon) throws TeamRepositoryException {
		HashMap<String, ApprovalContainer> approvalsByName = new HashMap<String, ApprovalContainer>();

		String internalReleaseId = SCMHelper.getInternalReleaseId(workItem, workItemCommon, null);

		for (IProcessConfigurationElement child : config.getChildren()) {
			if (child.getName().equals(TAG_APPROVALS)) {
				for (IProcessConfigurationElement approvalElement : child.getChildren()) {
					ApprovalContainer configuredApproval = new ApprovalContainer(approvalElement);
					if (configuredApproval.getWorkItemType().equals(workItem.getWorkItemType())) {

						// Compare the internal release id values to see if it is valid
						// to be included in this list. If the work item internal release
						// is set, then the approval must not have a internal release or
						// have a matching internal release.
						if (configuredApproval.getInternalReleaseId() == null
								|| (internalReleaseId != null && internalReleaseId.equals(configuredApproval.getInternalReleaseId()))) {

							// if there is already an approval with the same
							// approval name, should we keep it or replace it?
							if (approvalsByName.containsKey(configuredApproval.getName())) {
								ApprovalContainer approvalHolder = approvalsByName.get(configuredApproval.getName());
								if (approvalHolder.getInternalReleaseId() == null
										&& configuredApproval.getInternalReleaseId() != null
										&& configuredApproval.getInternalReleaseId().equals(internalReleaseId)) {
									// replace the existing configured approval
									// with this more specific configured
									// approval
									approvalsByName.put(configuredApproval.getName(), configuredApproval);
								}

							}
							else {
								// We haven't added a configured approval with this name
								approvalsByName.put(configuredApproval.getName(), configuredApproval);
							}
						}
					}
				}
			}
		}

		return new ArrayList<ApprovalContainer>(approvalsByName.values());
	}

	/**
	 * Returns all the approvals configured for this work item type and has a
	 * configuration for the work item state
	 * 
	 * @param config
	 * @param workItemType
	 * @param workItemState
	 * @return
	 * @throws TeamRepositoryException
	 */
	public static List<ApprovalContainer> getConfiguredApprovals(IProcessConfigurationElement config,
			ISaveParameter saveParameter, IWorkItem workItem, IWorkItemCommon workItemCommon)
					throws TeamRepositoryException {
		List<ApprovalContainer> approvals = getConfiguredApprovals(config, workItem, workItemCommon);
		// remove any approvals which are not configured for the given state
		Iterator<ApprovalContainer> i = approvals.iterator();
		while (i.hasNext()) {
			ApprovalContainer approval = i.next();
			if (approval.hasConditionConfig(saveParameter) == false) {
				i.remove();
			}
		}
		return approvals;
	}

	public static ApprovalContainer getConfiguredApproval(IProcessConfigurationElement config, IWorkItem workItem,
			String name, IWorkItemCommon workItemCommon) throws TeamRepositoryException {
		List<ApprovalContainer> approvals = getConfiguredApprovals(config, workItem, workItemCommon);
		String internalReleaseId = SCMHelper.getInternalReleaseId(workItem, workItemCommon, null);

		ApprovalContainer holderApproval = null;
		for (ApprovalContainer approval : approvals) {
			if (internalReleaseId == null && name.equals(approval.getName())) {
				return approval;
			}
			if (internalReleaseId != null && internalReleaseId.equals(approval.getInternalReleaseId()) && name.equals(approval.getName())) {
				return approval;
			}
			if (name.equals(approval.getName())) {
				holderApproval = approval;
			}
		}

		if (holderApproval != null) {
			return holderApproval;
		}
		return null;
	}

	/**
	 * Returns all approver lists configured for this repository
	 * 
	 * @param config
	 * @return
	 * @throws TeamRepositoryException
	 * @throws Exception
	 */
	private static List<ApproverListContainer> getAllApproverLists(IProjectAreaHandle projectAreaHandle,
			IAuditableCommon auditableCommon) throws TeamRepositoryException {
		IProcessConfigurationData configurationData = auditableCommon.getProcess(projectAreaHandle, null)
				.findProcessConfiguration(APPROVER_LISTS_CONFIGURATION, null);
		if (configurationData == null) {
			throw new IllegalStateException("The approver list configuration data could not be found.");
		}

		List<ApproverListContainer> approverLists = new ArrayList<ApproverListContainer>();
		for (IProcessConfigurationElement child : configurationData.getElements()) {
			if (child.getName().equals(TAG_APPROVER_LISTS)) {
				for (IProcessConfigurationElement approverListElement : child.getChildren()) {
					ApproverListContainer approverListContainer = new ApproverListContainer(approverListElement);
					// special generic case
					if (approverListContainer.getRepositoryURI() == null && approverListContainer.getTeam() == null) {
						approverLists.add(approverListContainer);
					}
					else if (ConfigurationHelper.hasThisRepositoryUriAttribute(approverListElement, auditableCommon)) {
						approverLists.add(approverListContainer);
					}
				}
			}
		}
		return approverLists;
	}

	/**
	 * Searches through the configured approver lists and returns the one that
	 * matches. To determine a match, the method first eliminates approver lists
	 * which specify a different approval name or process area which is not in
	 * the current team hierarchy. If the work item has the internal release attribute
	 * set, any approver lists that have a internal release set but are not for that
	 * internal release are removed. Starting at the lowest level team, it looks for an
	 * approver list configured specifically for the approval name, work item
	 * internal release, and the team. Then it looks for an approver list configured for
	 * the team and approval name without a internal release. Finally it will look
	 * for the approver list configured for the team. If an approver list does
	 * not exist for this team, it searches each parent team, the project area,
	 * and the generic approver list. null is returned if no match is found.
	 * 
	 * @param config
	 * @param processArea
	 * @param approvalName
	 * @param workItem
	 * @param auditableCommon
	 * @param workItemCommon
	 * @return
	 * @throws Exception
	 */
	public static ApproverListContainer getApproverList(IProcessAreaHandle processAreaHandle, String approvalName,
			IWorkItem workItem, IAuditableCommon auditableCommon, IWorkItemCommon workItemCommon)
					throws TeamRepositoryException {
		IProcessArea processArea = (IProcessArea) auditableCommon.resolveAuditable(processAreaHandle, ItemProfile
				.createFullProfile(IProcessArea.ITEM_TYPE), null);
		List<ApproverListContainer> approverLists = getAllApproverLists(processArea.getProjectArea(), auditableCommon);
		List<String> teamUUIDs = ConfigurationHelper.getTeamAreaHierarchyUUIDs(processArea, auditableCommon);

		// allows us to retrieve a generic approver list that is not specific to
		// a project area.
		// Example: automatic verify-approval with the submitter as the approver
		teamUUIDs.add("");

		String internalReleaseId = SCMHelper.getInternalReleaseId(workItem, workItemCommon, null);

		// remove the approver lists that do not pertain
		Iterator<ApproverListContainer> i = approverLists.iterator();
		while (i.hasNext()) {
			ApproverListContainer approverList = i.next();
			approverList.setCurrentApprovalName(approvalName);
			String configureName = removeGeneratedApprovalName(approvalName);
			// remove approver list where the approval name is set but doesn't
			// match the approval name parm
			if (approverList.getConfiguredApprovalName() != null
					&& !approverList.getConfiguredApprovalName().equals(configureName)) {
				i.remove();
			}
			// remove the approver list where the team area is set but doesn't
			// match the team area parm
			else if (approverList.getTeam() != null && !teamUUIDs.contains(approverList.getTeam())) {
				i.remove();
			}
			// remove the approver list if the internal release was a parm and it does not
			// match this approver list internal release
			else if (internalReleaseId != null && approverList.getInternalReleaseId() != null
					&& !approverList.getInternalReleaseId().equals(internalReleaseId)) {
				i.remove();
			}
			// remove the approver list if the internal release was not a parm and the
			// approver list has a internal release
			else if (internalReleaseId == null && approverList.getInternalReleaseId() != null) {
				i.remove();
			}
		}

		// if there is zero or one list left, return
		if (approverLists.size() == 1) {
			return approverLists.get(0);
		}
		else if (approverLists.size() == 0) {
			return null;
		}

		// there are more than one approver lists left so we need to find the
		// one at the lowest level in the hierarchy
		ApproverListContainer holder = null;
		for (String teamUUID : teamUUIDs) {
			for (ApproverListContainer list : approverLists) {
				if (teamUUID.equals(list.getTeam()) || teamUUID.equals("")) {
					if (list.getConfiguredApprovalName() != null) {
						// team + name + internal release = best match
						if (internalReleaseId == null || internalReleaseId.equals(list.getInternalReleaseId())) {
							// return immediately because we have a teamUUID,
							// approval name, and optional internal release match
							return list;
						}
						// team + name = 2nd best match
						if (internalReleaseId != null & list.getInternalReleaseId() == null) {
							// hold on to this one because it matches the
							// teamUUID but does not have a internal release
							// we will use it if we cannot find another one with
							// a internal release
							holder = list;
						}
					}
				}
			}

			// if holder was set then return it
			if (holder != null) {
				return holder;
			}
		}

		return null;
	}

	public static IApproval findApprover(List<IApproval> approverList, IApproval approverToFind) {
		for (IApproval approver : approverList) {
			if (approver.getApprover().sameItemId(approverToFind.getApprover())) {
				return approver;
			}
		}
		return null;
	}
}
