package com.ibm.stg.rtc.ext.common;

import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.eclipse.core.runtime.IProgressMonitor;

import com.ibm.team.process.common.IProcessAreaHandle;
import com.ibm.team.process.common.IProcessConfigurationElement;
import com.ibm.team.repository.common.IContributor;
import com.ibm.team.repository.common.IContributorHandle;
import com.ibm.team.repository.common.PermissionDeniedException;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.workitem.common.IAuditableCommon;
import com.ibm.team.workitem.common.IWorkItemCommon;
import com.ibm.team.workitem.common.model.IWorkItem;
import com.ibm.team.workitem.common.model.ItemProfile;
import com.ibm.team.workitem.common.model.WorkItemEndPoints;

/**
 * @author Kyle Christianson
 *
 */
public class ApproverListContainer {

	public static final String TAG_APPROVER_LIST = "approverList";
	public static final String TAG_APPROVER = "approver";
	public static final String ATTR_TEAM = "team";
	public static final String ATTR_INTERNAL_RELEASE_ID = "internalReleaseId";
	public static final String ATTR_APPROVAL_NAME = "approvalName";
	public static final String ATTR_USER_ID = "userId";
	public static final String ATTR_INCLUDE_OWNER = "includeWorkItemOwner";
	public static final String ATTR_INCLUDE_CREATOR = "includeWorkItemCreator";
	public static final String ATTR_INCLUDE_PARENT_WI_OWNER = "includeParentWorkItemOwner";
	public static final String ATTR_INCLUDE_PARENT_WI_CREATOR = "includeParentWorkItemCreator";
	public static final String ATTR_INCLUDE_DUPLICATE_WI_APPROVERS = "includeDuplicateWIApprovers";


	private String repositoryURI;
	private String team;
	private String internalReleaseId;
	private String configuredApprovalName;
	private Boolean creator;
	private Boolean owner;
	private Boolean duplicateApprovers;
	private Boolean parentCreator;
	private Boolean parentOwner;
	private Set<String> approverIds;
	private IProcessConfigurationElement approverListElement;
	private String currentApprovalName; // represents the current approval name that this list is getting applied to, needed for finding duplicate WI approvers

	public ApproverListContainer(IProcessConfigurationElement element){
		if(!element.getName().equals(TAG_APPROVER_LIST)){
			throw new IllegalArgumentException("ApproverListContainer only supports elements with the tag <" + TAG_APPROVER_LIST + ">.");
		}
		this.approverListElement = element;
		// wait until the get methods are called to parse the xml to improve performance 
	}

	public String getRepositoryURI(){
		if(this.repositoryURI == null){
			this.repositoryURI = this.approverListElement.getAttribute(ConfigurationHelper.ATTR_REPOSITORY_URI);
		}
		return this.repositoryURI;
	}

	public String getTeam(){
		if(this.team == null){
			this.team = this.approverListElement.getAttribute(ATTR_TEAM);
		}
		return this.team;
	}
	
	public String getInternalReleaseId() {
		if (this.internalReleaseId == null) {
			this.internalReleaseId = this.approverListElement.getAttribute(ATTR_INTERNAL_RELEASE_ID);
		}
		return this.internalReleaseId;
	}

	public String getConfiguredApprovalName(){
		if(this.configuredApprovalName == null){
			this.configuredApprovalName = this.approverListElement.getAttribute(ATTR_APPROVAL_NAME);
		}
		return this.configuredApprovalName;
	}

	public void setCurrentApprovalName(String name){
		this.currentApprovalName = name;
	}

	public String getCurrentApprovalName(){
		return this.currentApprovalName;
	}

	public boolean isCreator(){
		if(this.creator == null){
			this.creator = ConfigurationHelper.getBooleanAttributeValue(this.approverListElement, ATTR_INCLUDE_CREATOR);
		}
		return this.creator;
	}

	public boolean isOwner(){
		if(this.owner == null){
			this.owner = ConfigurationHelper.getBooleanAttributeValue(this.approverListElement, ATTR_INCLUDE_OWNER);
		}
		return this.owner;
	}

	public boolean isDuplicateWIApprovers(){
		if(this.duplicateApprovers == null){
			this.duplicateApprovers = ConfigurationHelper.getBooleanAttributeValue(this.approverListElement, ATTR_INCLUDE_DUPLICATE_WI_APPROVERS);
		}
		return this.duplicateApprovers;
	}
	
	public boolean isParentCreator() {
		if (this.parentCreator == null) {
			this.parentCreator = ConfigurationHelper.getBooleanAttributeValue(this.approverListElement, ATTR_INCLUDE_PARENT_WI_CREATOR);
		}
		return parentCreator;
	}
	
	public boolean isParentOwner() {
		if (this.parentOwner == null) {
			this.parentOwner = ConfigurationHelper.getBooleanAttributeValue(this.approverListElement, ATTR_INCLUDE_PARENT_WI_OWNER);
		}
		return parentOwner;
	}

	/**
	 * @return the static user ids configured in the xml for this approver list 
	 * @throws TeamRepositoryException
	 */
	public Set<String> getStaticApproverIds() throws TeamRepositoryException{
		if(this.approverIds == null){
			populateApproverIds();
		}
		return this.approverIds;
	}

	private void populateApproverIds() {
		this.approverIds = new HashSet<String>();
		IProcessConfigurationElement[] children = this.approverListElement.getChildren();
		for(IProcessConfigurationElement child : children){
			if(child.getName().equals(TAG_APPROVER)){
				String userId = child.getAttribute(ATTR_USER_ID);
				if(userId != null && userId.length() != 0){
					this.approverIds.add(userId);
				}
			}
		}
	}


	/**
	 * Returns all the approvals configured for this approver list and the given work item
	 * @param workItem used to retrieve dynamic configured approvers. must not be null
	 * @param workItemCommon
	 * @param monitor
	 * @return
	 * @throws TeamRepositoryException
	 */
	public Set<String> getAllApprovers(IWorkItem workItem, IWorkItemCommon workItemCommon, IProgressMonitor monitor) throws TeamRepositoryException{
		IAuditableCommon auditableCommon = workItemCommon.getAuditableCommon();
		WorkItemCommonTasks workItemCommonTasks = new WorkItemCommonTasks();
		Set<String> approvers = new HashSet<String>();

		// add the static configured approvers
		approvers.addAll(this.getStaticApproverIds());

		// add the creator if configured
		if(isCreator()){
			addApprover(approvers, workItem.getCreator(), workItemCommonTasks, auditableCommon, monitor);
		}

		// add the owner if configured
		if(isOwner()){
			addApprover(approvers, workItem.getOwner(), workItemCommonTasks, auditableCommon, monitor);
		}
		
		// add this work item parent's creator if configured
		if (isParentCreator()) {
			IWorkItem parentWorkItem = workItemCommonTasks.getParentWorkItem(workItem, auditableCommon, workItemCommon, monitor);
			// if this work item has a parent, check if any users from the parent need to be added to the approver list
			if (parentWorkItem != null) {
				addApprover(approvers, parentWorkItem.getCreator(), workItemCommonTasks, auditableCommon, monitor);
			}
		}

		// add this work item parent's owner if configured
		if (isParentOwner()) {
			IWorkItem parentWorkItem = workItemCommonTasks.getParentWorkItem(workItem, auditableCommon, workItemCommon, monitor);
			// if this work item has a parent, check if any users from the parent need to be added to the approver list
			if (parentWorkItem != null) {
				addApprover(approvers, parentWorkItem.getOwner(), workItemCommonTasks, auditableCommon, monitor);
			}
		}

		// add any duplicate work item approver lists
		if(isDuplicateWIApprovers()){
			List<IWorkItem> duplicates = workItemCommonTasks.getLinkedWorkItems(workItem, WorkItemEndPoints.DUPLICATE_WORK_ITEM, workItemCommon, monitor);

			for(IWorkItem duplicate : duplicates){
				IProcessAreaHandle processArea = workItemCommon.findProcessArea(duplicate, monitor);
				ApproverListContainer dupApproverList = ApprovalsHelper.getApproverList(processArea, this.getCurrentApprovalName(), duplicate, auditableCommon, workItemCommon);
				approvers.addAll(dupApproverList.getAllApprovers(duplicate, workItemCommon, monitor));
			}
		}

		return approvers;
	}

	/**
	 * Returns a list of approvers which are configured and have changed between the two work item states given
	 * @param newWorkItemState the new work item state we are saving. must not be null
	 * @param oldWorkItemState the state of the work item previous to the save. this may be null meaning everything has changed
	 * @param workItemCommon
	 * @param monitor
	 * @return
	 * @throws TeamRepositoryException
	 */
	public Set<String> getChangedApprovers(WorkItemChangeInfo workItemChangeInfo, IWorkItemCommon workItemCommon, IProgressMonitor monitor) throws TeamRepositoryException{
		IAuditableCommon auditableCommon = workItemCommon.getAuditableCommon();
		WorkItemCommonTasks workItemCommonTasks = new WorkItemCommonTasks();
		Set<String> approvers = new HashSet<String>();

		// if this is a new duplicate, add the static configured approver ids
		if(workItemChangeInfo.isNewDuplicate()){
			approvers.addAll(this.getStaticApproverIds());
		}

		// add the creator if configured and changed
		if(isCreator() && workItemChangeInfo.isCreatorChanged()){
			addApprover(approvers, workItemChangeInfo.getWorkItem().getCreator(), workItemCommonTasks, auditableCommon, monitor);
		}

		// add the owner if configured and changed
		if(isOwner() && workItemChangeInfo.isOwnerChanged()){
			addApprover(approvers, workItemChangeInfo.getWorkItem().getOwner(), workItemCommonTasks, auditableCommon, monitor);
		}
		
		// add the parent creator if configured and changed
		if (isParentCreator() && workItemChangeInfo.isParentCreatorChanged()) {
			try {
				IWorkItem parentWorkItem = workItemCommonTasks.getParentWorkItem(workItemChangeInfo.getWorkItem(), auditableCommon,
						workItemCommon, monitor);
				// if this work item has a parent, check if any users from the parent need to be added to the approver list
				if (parentWorkItem != null) {
					addApprover(approvers, parentWorkItem.getCreator(), workItemCommonTasks, auditableCommon, monitor);
				}
			}
			catch (PermissionDeniedException e) {
					// we shouldn't we get this far if the user doesn't have permission to the parent
			}
		}
		
		// add the parent owner if configured and changed
		if (isParentOwner() && workItemChangeInfo.isParentOwnerChanged()) {
			try {
				IWorkItem parentWorkItem = workItemCommonTasks.getParentWorkItem(workItemChangeInfo.getWorkItem(), auditableCommon,
						workItemCommon, monitor);
				// if this work item has a parent, check if any users from the parent need to be added to the approver list
				if (parentWorkItem != null) {
					addApprover(approvers, parentWorkItem.getOwner(), workItemCommonTasks, auditableCommon, monitor);
				}
			}
			catch (PermissionDeniedException e) {
				// we shouldn't we get this far if the user doesn't have permission to the parent
			}
		}

		// check for changed duplicate links
		if(isDuplicateWIApprovers()){
			for(WorkItemChangeInfo duplicateInfo : workItemChangeInfo.getChangedDuplicates()){
				// add the duplicate approvers to this approver list
				IProcessAreaHandle processArea = workItemCommon.findProcessArea(duplicateInfo.getWorkItem(), monitor);
				ApproverListContainer dupApproverList = ApprovalsHelper.getApproverList(processArea, this.getCurrentApprovalName(), duplicateInfo.getWorkItem(), auditableCommon, workItemCommon);
				if(duplicateInfo.isNewDuplicate()){
					approvers.addAll(dupApproverList.getAllApprovers(duplicateInfo.getWorkItem(), workItemCommon, monitor));
				}else{
					approvers.addAll(dupApproverList.getChangedApprovers(duplicateInfo, workItemCommon, monitor));
				}
			}
		}

		return approvers;
	}
	
	/**
	 * Resolves the userHandle and adds it to the set of approvers if it is not unassigned
	 * @param approvers
	 * @param userHandle
	 * @param workItemCommonTasks
	 * @param auditableCommon
	 * @param monitor
	 * @throws TeamRepositoryException
	 */
	private void addApprover(Set<String> approvers, IContributorHandle userHandle, WorkItemCommonTasks workItemCommonTasks, IAuditableCommon auditableCommon, IProgressMonitor monitor) throws TeamRepositoryException {
		IContributor user = (IContributor) auditableCommon.resolveAuditable(userHandle, ItemProfile.createFullProfile(IContributor.ITEM_TYPE), monitor);
		if(!workItemCommonTasks.isUnassignedContributor(user, auditableCommon, monitor)) {
			approvers.add(user.getUserId());
		}
		return;
	}
}
