package com.ibm.stg.rtc.ext.common;

import java.net.URI;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import com.ibm.team.process.common.IProcessArea;
import com.ibm.team.process.common.IProcessAreaHandle;
import com.ibm.team.process.common.IProcessConfigurationData;
import com.ibm.team.process.common.IProcessConfigurationElement;
import com.ibm.team.process.common.IProjectArea;
import com.ibm.team.process.common.IProjectAreaHandle;
import com.ibm.team.process.common.ITeamArea;
import com.ibm.team.process.common.ITeamAreaHandle;
import com.ibm.team.process.common.ITeamAreaHierarchy;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.workitem.common.IAuditableCommon;
import com.ibm.team.workitem.common.IWorkItemCommon;
import com.ibm.team.workitem.common.internal.workflow.WorkflowInfo;
import com.ibm.team.workitem.common.model.CategoryId;
import com.ibm.team.workitem.common.model.ICategory;
import com.ibm.team.workitem.common.model.ICategoryHandle;
import com.ibm.team.workitem.common.model.ItemProfile;

/**
 * @author Kyle Christianson
 *
 */
public class ConfigurationHelper {

	public static final String ITEM_PROPERTIES_CONFIGURATION = "com.ibm.stg.workitem.configuration.itemProperties";
	public static final String INTERNAL_RELEASE_CONFIGURATION = "com.ibm.stg.configuration.internalReleases";
	public static final String TAG_TEAM = "team";
	public static final String TAG_CATEGORY = "category";
	public static final String TAG_SHOW_WARNINGS = "showWarnings";
	public static final String TAG_INTERNAL_RELEASE = "internalRelease";
	public static final String TAG_STREAM = "stream";
	public static final String ATTR_REPOSITORY_URI = "repositoryURI";
	public static final String TAG_WORKITEM_TYPE = "workItemType";
	public static final String TAG_WORKITEM_STATE = "workItemState";
	public static final String ATTR_ID = "id";
	public static final String ATTR_DEFAULT_OWNER = "Owner";
	public static final String ATTR_UUID = "uuid";
	public static final String ATTR_VALUE = "value";
	public static final String ATTR_NAME = "name";
	public static final String VALUE_TRUE = "true";
	public static final String ATTR_PREVENT_USE = "preventUse";

	// the default ports to consider equal when comparing URIs. Include -1 because that means no port
	public static final List<Integer> DEFAULT_PORTS = Arrays.asList(80, 443, -1);

	/**
	 * Returns the configuration for the given work item type
	 * @param configuration
	 * @param workItemType
	 * @return
	 */
	public static IProcessConfigurationElement getWITypeConfiguration(IProcessConfigurationElement configuration, String workItemType){
		return getConfigurationElementById(configuration, TAG_WORKITEM_TYPE, workItemType);
	}

	public static List<String> getWITypesConfigured(IProcessConfigurationElement configuration){
		return getChildrenTagIds(configuration, TAG_WORKITEM_TYPE);
	}

	/**
	 * Returns the configuration for the given work item type and state
	 * @param configuration
	 * @param workItemType
	 * @param workItemState
	 * @return
	 */
	public static IProcessConfigurationElement getWIStateConfiguration(IProcessConfigurationElement configuration, String workItemType, String workItemState) {
		IProcessConfigurationElement wiTypeConfig = getWITypeConfiguration(configuration, workItemType);
		if(wiTypeConfig != null){
			// convert legacy id to the actual id. if this is not done, a state id of "s2" will return the string identifier of "2"
			// which does not match the configured states. so this method adds the "s" prefix to integer states. 
			// see com.ibm.team.workitem.common.internal.workflow.WorkflowInfo
			String workItemStateLegacy = convertLegacyId(workItemState, 's');
			return getConfigurationElementById(wiTypeConfig, TAG_WORKITEM_STATE, workItemStateLegacy);
		}else{
			return null;
		}
	}

	// copied from com.ibm.team.workitem.common.internal.workflow.WorkflowInfo
	public static String convertLegacyId(String id, char prefix) {
		// legacy: if id is an integer, convert it to legacy state id s<int>
		try {
			Integer.parseInt(id);
			return prefix + id;
		} catch (NumberFormatException e) {
			// ignore: not legacy format
		}
		return id;
	}

	/**
	 * Returns the list of work item states configured for the given work item type or null
	 * if the work item type is not configured.
	 * @param configuration
	 * @param workItemType
	 * @return
	 */
	public static List<String> getWIStatesConfigured(IProcessConfigurationElement configuration, String workItemType){
		IProcessConfigurationElement wiTypeConfig = getWITypeConfiguration(configuration, workItemType);
		if(wiTypeConfig != null){
			List<String> stateIds = getChildrenTagIds(wiTypeConfig, TAG_WORKITEM_STATE);


			// convert state ids to their legacy ids. if this is not done, a state id of "s2" will return the string
			// identifier of "2" which does not match the configured states. so this method removes the "s" prefix
			// to legacy states. see com.ibm.team.workitem.common.internal.workflow.WorkflowInfo
			List<String> convertedStateIds = new ArrayList<String>();
			for(String stateId : stateIds){
				convertedStateIds.add(WorkflowInfo.stripOffPrefix(stateId, 's'));
			}

			return convertedStateIds;
		}else{
			return null;
		}
	}

	/**
	 * Loops through the given configuration element and finds the child element with the given tag name and 
	 * an id attribute that matches the given id. Returns null if none found.
	 * @param configuration
	 * @param tagName
	 * @param id
	 * @return
	 */
	public static IProcessConfigurationElement getConfigurationElementById(IProcessConfigurationElement configuration, String tagName, String id){
		if(configuration == null){
			return null;
		}
		IProcessConfigurationElement[] configurationElements = configuration.getChildren();
		// loop through all the child xml elements looking for the correct tag name and id configuration
		for (IProcessConfigurationElement configurationElement : configurationElements) {
			// if the element is the correct name, check the id
			if (configurationElement.getName().equals(tagName)) {
				String attrId = configurationElement.getAttribute(ATTR_ID);
				// if the element is the correct id, return it
				if(attrId != null && attrId.equals(id)){
					return configurationElement;
				}
			}
		}
		return null;
	}

	/**
	 * Loops through the given configuration element and finds the child element with the given tag name. Returns null if none found.
	 * @param configuration
	 * @param tagName
	 * @param id
	 * @return
	 */
	public static IProcessConfigurationElement getConfigurationElementByTagName(IProcessConfigurationElement configuration, String tagName){
		if(configuration == null){
			return null;
		}
		IProcessConfigurationElement[] configurationElements = configuration.getChildren();
		// loop through all the child xml elements looking for the correct tag name and id configuration
		for (IProcessConfigurationElement configurationElement : configurationElements) {
			// if the element is the correct name, return it
			if (configurationElement.getName().equals(tagName)) {
				return configurationElement;
			}
		}
		return null;
	}

	/**
	 * Returns a list of string id attributes of children elements with the specified tag name.
	 * If an element does not specify the id attribute, nothing gets added to the list.
	 * @param configuration
	 * @param tagName
	 * @return
	 */
	public static List<String> getChildrenTagIds(IProcessConfigurationElement configuration, String tagName){
		List<String> idList = new ArrayList<String>();
		if(configuration == null){
			return idList;
		}

		IProcessConfigurationElement[] configurationElements = configuration.getChildren();

		// loop through all the child xml elements looking for the specified tag name
		for (IProcessConfigurationElement configurationElement : configurationElements) {
			// if the element matches the tag name, add the id to the list
			if (configurationElement.getName().equals(tagName)) {
				String attrId = configurationElement.getAttribute(ATTR_ID);
				if(attrId != null){
					idList.add(attrId);
				}
			}
		}
		return idList;
	}

	/**
	 * Searches for a configured category owner in the hierarchy of categories
	 * @param category
	 * @param workItemCommon
	 * @return
	 * @throws TeamRepositoryException
	 */
	public static String getCategoryOwner(ICategoryHandle categoryHandle, IWorkItemCommon workItemCommon) throws TeamRepositoryException{
		ICategory category = workItemCommon.getAuditableCommon().resolveAuditable(categoryHandle, ICategory.FULL_PROFILE, null);
		List<String> categoryUUIDs = getCategoryHierarchyUUIDs(category, workItemCommon);
		Map<String, Map<String,String>> allCategoryProperties = getItemProperties(category.getProjectArea(), workItemCommon.getAuditableCommon(), TAG_CATEGORY);

		for(String uuid : categoryUUIDs){
			Map<String,String> categoryProperties = allCategoryProperties.get(uuid);
			if(categoryProperties != null && categoryProperties.containsKey(ATTR_DEFAULT_OWNER)){
				return categoryProperties.get(ATTR_DEFAULT_OWNER);
			}
		}
		return null;
	}

	/**
	 * Searches for a configured category that should not allow usage in certain work item types
	 * @param category
	 * @param workItemType
	 * @param workItemCommon
	 * @return true if the category can be used
	 * @throws TeamRepositoryException
	 */
	public static boolean isCategoryUsable(ICategory category, String workItemType, IWorkItemCommon workItemCommon) throws TeamRepositoryException {
		Map<String, Map<String, String>> allCategoryProperties = getItemProperties(category.getProjectArea(), workItemCommon.getAuditableCommon(), TAG_CATEGORY);

		for (String uuid : allCategoryProperties.keySet()) {
			if (uuid.equals(category.getItemId().getUuidValue())) {
				Map<String, String> categoryProperties = allCategoryProperties.get(uuid);
				if (categoryProperties != null && categoryProperties.containsKey(ATTR_PREVENT_USE)) {
					String workItemTypes = categoryProperties.get(ATTR_PREVENT_USE);
					String[] workItemTypesArray = workItemTypes.split(",");
					List<String> workItemTypesList = Arrays.asList(workItemTypesArray);
					if (workItemTypesList.contains(workItemType)) {
						return false;
					}
				}
			}
		}
		return true;
	}

	/**
	 * Searches for a configured team owner in the hierarchy of teams
	 * @param processArea
	 * @param auditableCommon
	 * @return
	 * @throws TeamRepositoryException
	 */
	public static String getTeamOwner(IProcessAreaHandle processAreaHandle, IAuditableCommon auditableCommon) throws TeamRepositoryException{
		IProcessArea processArea = (IProcessArea)auditableCommon.resolveAuditable(processAreaHandle, ItemProfile.createFullProfile(IProcessArea.ITEM_TYPE), null);
		List<String> teamUUIDs = getTeamAreaHierarchyUUIDs(processArea, auditableCommon);
		Map<String, Map<String,String>> allTeamProperties = getItemProperties(processArea.getProjectArea(), auditableCommon, TAG_TEAM);

		for(String uuid : teamUUIDs){
			Map<String,String> teamProperties = allTeamProperties.get(uuid);
			if(teamProperties != null && teamProperties.containsKey(ATTR_DEFAULT_OWNER)){
				return teamProperties.get(ATTR_DEFAULT_OWNER);
			}
		}
		return null;
	}

	/**
	 * Looks up all the item properties of the item type for this project and repository.
	 * @param projectAreaHandle
	 * @param auditableCommon
	 * @param itemTypeName the item type or configuration tag name, such as "team" or "category"
	 * @return a map of item uuid to the map of name-value properties
	 * @throws TeamRepositoryException
	 */
	private static Map<String, Map<String,String>> getItemProperties(IProjectAreaHandle projectAreaHandle, IAuditableCommon auditableCommon, String itemTypeName) throws TeamRepositoryException{
		IProcessConfigurationData configurationData = auditableCommon.getProcess(projectAreaHandle, null).findProcessConfiguration(ITEM_PROPERTIES_CONFIGURATION, null);
		if(configurationData == null){
			throw new IllegalStateException("The item properties configuration data could not be found.");
		}

		Map<String, Map<String,String>> items = new HashMap<String, Map<String,String>>();

		for(IProcessConfigurationElement child : configurationData.getElements()){
			if(child.getName().equals(itemTypeName)){
				if(hasThisRepositoryUriAttribute(child, auditableCommon)){
					Map<String, String> properties = new HashMap<String,String>();
					for(String attribute : child.getAttributeNames()){
						properties.put(attribute, child.getAttribute(attribute));
					}
					String uuid = properties.get(ATTR_UUID);
					if(uuid == null){
						throw new IllegalStateException("Each item property must contain a '" + ATTR_UUID + "' attribute.");
					}
					items.put(uuid, properties);
				}

			}
		}
		return items;
	}

	public static List<InternalRelease> getAllInternalReleases(IProjectAreaHandle projectAreaHandle, IAuditableCommon auditableCommon) throws TeamRepositoryException{
		IProcessConfigurationData configurationData = auditableCommon.getProcess(projectAreaHandle, null).findProcessConfiguration(INTERNAL_RELEASE_CONFIGURATION, null);
		if(configurationData == null){
			throw new IllegalStateException("The internal release configuration data could not be found.");
		}
		List<InternalRelease> results = new ArrayList<InternalRelease>();
		for(IProcessConfigurationElement child : configurationData.getElements()){
			if(child.getName().equals(TAG_INTERNAL_RELEASE)){
				InternalRelease ir = new InternalRelease(child);
				results.add(ir);
			}
		}
		return results;
	}

	/**
	 * Traverses the Team Hierarchy and stores the UUIDs of the teams starting with the given processArea and 
	 * ending with the project area
	 * @param processArea
	 * @param auditableCommon
	 * @return
	 * @throws TeamRepositoryException
	 */
	public static List<String> getTeamAreaHierarchyUUIDs(IProcessArea processArea, IAuditableCommon auditableCommon) throws TeamRepositoryException{
		List<String> teamUUIDs = new ArrayList<String>();
		teamUUIDs.add(processArea.getItemId().getUuidValue());
		if(processArea instanceof ITeamArea){
			ITeamAreaHandle currentTeam = (ITeamArea)processArea;
			IProjectArea projectArea = (IProjectArea)auditableCommon.resolveAuditable(processArea.getProjectArea(), ItemProfile.createFullProfile(IProjectArea.ITEM_TYPE), null);
			ITeamAreaHierarchy teamHierarchy = projectArea.getTeamAreaHierarchy();
			boolean isRootTeam = false;
			while(isRootTeam == false){
				ITeamAreaHandle teamParent = teamHierarchy.getParent(currentTeam);
				if(teamParent == null){
					isRootTeam = true;
					teamUUIDs.add(projectArea.getItemId().getUuidValue());
				}else{
					teamUUIDs.add(teamParent.getItemId().getUuidValue());
					currentTeam = teamParent;
				}
			}
		}
		return teamUUIDs;
	}

	/**
	 * Traverses the Category Hierarchy and stores the UUIDs of the categories starting with the given category and 
	 * ending with the root
	 */
	public static List<String> getCategoryHierarchyUUIDs(ICategory category, IWorkItemCommon workItemCommon) throws TeamRepositoryException{
		List<String> categoryUUIDs = new ArrayList<String>();
		categoryUUIDs.add(category.getItemId().getUuidValue());

		CategoryId parentCategoryId = category.getParentId2();
		ICategory currentCategory;
		while(parentCategoryId != null){
			ICategoryHandle parentCategoryHandle = workItemCommon.findCategoryById2(category.getProjectArea(), parentCategoryId, null);
			currentCategory = workItemCommon.getAuditableCommon().resolveAuditable(parentCategoryHandle, ICategory.FULL_PROFILE, null);
			categoryUUIDs.add(currentCategory.getItemId().getUuidValue());
			parentCategoryId = currentCategory.getParentId2();
		} while(parentCategoryId != null);

		return categoryUUIDs;
	}

	public static boolean getBooleanAttributeValue(IProcessConfigurationElement element, String attribute){
		if(element == null){
			throw new IllegalArgumentException("getBooleanAttributeValue cannot accept a null element");
		}
		String value = element.getAttribute(attribute);
		if(VALUE_TRUE.equals(value)){
			return true;
		}else{
			return false;
		}
	}


	public static boolean getBooleanTagValue(IProcessConfigurationElement element, String tagName){
		if(element == null){
			throw new IllegalArgumentException("getBooleanTagValue cannot accept a null element");
		}
		IProcessConfigurationElement[] children = element.getChildren();
		Boolean result = null;
		for(IProcessConfigurationElement child : children){
			if(child.getName().equals(tagName)){
				if(result != null){
					throw new IllegalStateException("There can only be one " + tagName + " element for each " + element.getName() + " element.");
				}
				return getBooleanAttributeValue(child, ATTR_VALUE);
			}
		}

		if(result != null){
			return result;
		} else{
			// result is still null after searching all the children, so return false 
			return false;
		}
	}

	public static Set<WorkItemCondition> getAllConditions(IProcessConfigurationElement element){
		Set<WorkItemCondition> conditions = new HashSet<WorkItemCondition>();
		if(element == null){
			return conditions;
		}
		IProcessConfigurationElement[] children = element.getChildren();
		for(IProcessConfigurationElement child : children){
			if(child.getName().equals(WorkItemCondition.TAG_CONDITION)){
				conditions.add(new WorkItemCondition(child));
			}
		}
		return conditions;
	}

	/**
	 * returns true if the given element has a repositoryURI which matches the repository uri in the give auditableCommon
	 * @param element
	 * @param auditableCommon
	 * @return
	 * @throws TeamRepositoryException
	 * @throws IllegalStateException
	 */
	public static boolean hasThisRepositoryUriAttribute(IProcessConfigurationElement element, IAuditableCommon auditableCommon) throws TeamRepositoryException, IllegalStateException{
		URI repositoryUri;
		try{
			repositoryUri = new URI(auditableCommon.getPublicRepositoryURI());
		}catch(Exception e){
			throw new TeamRepositoryException("The repository URI could not be parsed: " + e.getMessage()); 
		}

		String uri = element.getAttribute(ConfigurationHelper.ATTR_REPOSITORY_URI);
		if(uri == null || uri.length() == 0){
			throw new IllegalStateException("The " + ConfigurationHelper.ATTR_REPOSITORY_URI + " attribute is required.");
		}

		URI configuredUri;
		try{
			configuredUri = new URI(uri); 
		}catch(Exception e){
			throw new IllegalStateException("The " + ConfigurationHelper.ATTR_REPOSITORY_URI + " attribute '" + uri + "' is not a valid uri.");
		}
		if(configuredUri.getHost() == null){
			throw new IllegalStateException("The " + ConfigurationHelper.ATTR_REPOSITORY_URI + " attribute '" + uri + "' must contain a host.");
		}
		if(sameURIHostAndPort(configuredUri, repositoryUri)){
			return true;
		}
		return false;
	}

	/**
	 * Returns true iff the two URIs contain the same host and port. This method treats no host
	 * unique and only equal to no host on the other URI. This method treats no port as unique
	 * as well with the exception of the DEFAULT_PORTS (80 and 443). So ports 80, 443, and no port
	 * will all be treated as equal.
	 * @param uri1
	 * @param uri2
	 * @return
	 */
	public static boolean sameURIHostAndPort(URI uri1, URI uri2){
		if((uri1.getHost() == null && uri2.getHost() == null) || (uri1.getHost().equals(uri2.getHost()))){
			// hosts are the same, now check port
			if(uri1.getPort() == uri2.getPort()){
				return true;
			}else if(DEFAULT_PORTS.contains(uri1.getPort()) && DEFAULT_PORTS.contains(uri2.getPort())){
				return true;
			}
		}
		return false;
	}
}
