package com.ibm.stg.rtc.ext.common;

import com.ibm.team.process.common.advice.IProcessReport;
import com.ibm.team.process.common.advice.IReportInfo;
import com.ibm.team.process.common.advice.TeamOperationCanceledException;
import com.ibm.team.process.internal.common.advice.ReportInfo;
import com.ibm.team.process.internal.common.advice.impl.ProcessReportImpl;
import com.ibm.team.repository.common.IItemHandle;

public class RepositoryCommonTasks {

	/**
	 * This method is useful when creating an error message to present to a user.
	 * Call this before suppressAllErrors() because after suppressAllErrors() this
	 * method will not find any errors. 
	 * 
	 * @param e
	 *        The exception to find the problem info from
	 * @return The description of the first problem info encountered. If a problem info was not found, this returns the
	 *         exception's getMessage()
	 */
	public String getFirstErrorDescription(TeamOperationCanceledException e) {

		String desc;

		try {
			IProcessReport report = e.getReport();
			IReportInfo error = findFirstError(report);
			desc = error.getDescription();
		} catch (Exception exception) {
			desc = e.getMessage();
		}

		if (desc == null) {
			desc = e.getMessage();
		}
		return desc;
	}

	/**
	 * This method suppresses all errors by setting any nested report status 
	 * values to OK. This includes ERROR and WARNING messages.
	 * This allows the caller to display an appropriate message without having
	 * nested error messages. It also suppresses all errors allowing the caller
	 * to determine if it wants to fail or proceed.
	 * @param callerName this is the name of the calling behavior to display to the user
	 * @param e
	 */
	public void suppressAllErrors(String callerName, TeamOperationCanceledException e){
		setAllInfosToOK(e.getReport(), callerName);
	}

	/**
	 * This function was copied from Rational's
	 * com.ibm.team.process.internal.common.advice.runtime.OperationAdviceManager because it is a useful recursive
	 * funtion used to find the first error message in a chain of report infos
	 * 
	 * @param report
	 * @return
	 */
	private IReportInfo findFirstError(IProcessReport report) {
		IReportInfo[] infos = report.getInfos();
		for (int i = 0; i < infos.length; i++) {
			IReportInfo info = infos[i];
			if (info.getSeverity() == IProcessReport.ERROR) {
				return info;
			}
		}
		IProcessReport[] nestedReports = report.getNestedReports();
		for (int i = 0; i < nestedReports.length; i++) {
			IReportInfo error = findFirstError(nestedReports[i]);
			if (error != null) {
				return error;
			}
		}
		return null;
	}

	private void setAllInfosToOK(IProcessReport report, String callerName){
		IReportInfo[] infos = report.getInfos();
		boolean recompute = false;
		for (int i = 0; i < infos.length; i++) {
			IReportInfo info = infos[i];
			if (info.getSeverity() != IProcessReport.OK) {

				// modify the description of the report info so anyone debugging will know
				// that this status has changed and which caller changed it
				((ReportInfo)info).setDescription("***This status has been modified from " + 
						getReportStatusString(info.getSeverity()) + " to " + getReportStatusString(IProcessReport.OK) + 
						" by " + callerName + "***\n\n" + info.getDescription());

				// it is not recommended to change the severity of a report info
				// after it has been added to a report which is why we have to call recomputeSeverity
				info.setSeverity(IProcessReport.OK);
				recompute = true;
			}
		}

		if(recompute){
			// recomputeSeverity() will recompute the severity of the calling report and any 
			// parent reports. It does not recompute the severity of any nested reports
			((ProcessReportImpl)report).recomputeSeverity();
		}

		IProcessReport[] nestedReports = report.getNestedReports();
		for (int i = 0; i < nestedReports.length; i++) {
			setAllInfosToOK(nestedReports[i], callerName);
		}
	}

	private String getReportStatusString(int status){
		switch(status){
		case IProcessReport.OK: return "OK";
		case IProcessReport.WARNING: return "WARNING";
		case IProcessReport.ERROR: return "ERROR";
		default: return "Unknown";
		}
	}

	/**
	 * Returns true if the two objects are equal. If the objects are items in the repository,
	 * this method will try to match the item ids. Otherwise it will call {@link Object#equals(Object)}
	 * on the objects. If either object is null, returns false
	 * @param left
	 * @param right
	 * @return false if the objects are not equal or one of them is null
	 */
	public static boolean itemsEqual(Object left, Object right) {

		if (left == right)
			return true;

		if (left == null || right == null)
			return false;

		if (left instanceof IItemHandle && right instanceof IItemHandle)
			return ((IItemHandle) left).sameItemId((IItemHandle) right);

		return left.equals(right);
	}
}
