package com.ibm.stg.rtc.ext.common;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;

import org.eclipse.core.runtime.IProgressMonitor;

import com.ibm.team.links.common.ILink;
import com.ibm.team.links.common.ILinkCollection;
import com.ibm.team.links.common.ILinkQueryPage;
import com.ibm.team.links.common.IReference;
import com.ibm.team.links.common.factory.IReferenceFactory;
import com.ibm.team.links.common.service.ILinkService;
import com.ibm.team.links.service.ILinkServiceLibrary;
import com.ibm.team.process.common.IProjectAreaHandle;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.scm.common.IChangeSetHandle;
import com.ibm.team.scm.common.IComponentHandle;
import com.ibm.team.scm.common.IWorkspaceHandle;
import com.ibm.team.scm.common.links.ILinkConstants;
import com.ibm.team.scm.common.process.IChangeHistoryAddOperandDelta;
import com.ibm.team.scm.common.process.IScmDeltaSource;
import com.ibm.team.workitem.common.IAuditableCommon;
import com.ibm.team.workitem.common.IWorkItemCommon;
import com.ibm.team.workitem.common.model.IAttribute;
import com.ibm.team.workitem.common.model.IEnumeration;
import com.ibm.team.workitem.common.model.ILiteral;
import com.ibm.team.workitem.common.model.IWorkItem;
import com.ibm.team.workitem.common.model.IWorkItemHandle;
import com.ibm.team.workitem.common.model.Identifier;

/**
 * @author Kyle Christianson
 *
 */
public class SCMHelper {


	/**
	 * This method should be used by deliver operations to get the deliver target workspaces and
	 * the change sets being delivered.
	 * Change sets in a baseline deliver are included in the list returned.
	 * 
	 * @param scmDeltaSource retrieved from AdvisableOperation.getOperationData() of a deliver operation
	 * @return A map of each target workspace to the change sets being delivered to that workspace
	 * @throws TeamRepositoryException
	 */
	public static Map<ItemHandleKey<IWorkspaceHandle>, List<IChangeSetHandle>> getDeliverInformation(IScmDeltaSource scmDeltaSource) throws TeamRepositoryException{

		Map<ItemHandleKey<IWorkspaceHandle>, List<IChangeSetHandle>> workspaceToChangeSets = new HashMap<ItemHandleKey<IWorkspaceHandle>, List<IChangeSetHandle>>();

		for (IChangeHistoryAddOperandDelta delta : scmDeltaSource.getDeltas(IChangeHistoryAddOperandDelta.class)) {
			IWorkspaceHandle targetWorkspace = delta.getTargetWorkspace();

			List<IChangeSetHandle> changeSets = getAllChangeSets(delta);
			workspaceToChangeSets.put(new ItemHandleKey<IWorkspaceHandle>(targetWorkspace), changeSets);
		}

		return workspaceToChangeSets;
	}

	private static List<IChangeSetHandle> getAllChangeSets(IChangeHistoryAddOperandDelta delta) throws TeamRepositoryException{
		List<IChangeSetHandle> changes = new LinkedList<IChangeSetHandle>();

		for (IComponentHandle comp : delta.getComponents()) {

			for (IChangeSetHandle cs : delta.getAdditionalBaselineChangeSets(comp)) {
				changes.add(cs);
			}

			for (IChangeSetHandle cs : delta.getAdditionalChangeSets(comp)) {
				changes.add(cs);
			}
		}
		return changes;
	}

	/**
	 * Returns a list of all work items associated to the give change set
	 * This will not return work items linked by remote change request links
	 * @param changeSet
	 * @param linkService
	 * @return
	 * @throws TeamRepositoryException
	 */
	public static List<IWorkItemHandle> getAssociatedWorkItems(IChangeSetHandle changeSet, ILinkService linkService) throws TeamRepositoryException{

		ILinkServiceLibrary linkServiceLibrary = (ILinkServiceLibrary)linkService.getServiceLibrary(ILinkServiceLibrary.class);
		List<IWorkItemHandle> results = new ArrayList<IWorkItemHandle>();

		List<ILink> allLinks = getAllLinks(linkServiceLibrary, changeSet);
		for(ILink link : allLinks){
			// this does not include remote change requests!
			if(link.getLinkTypeId().equals(ILinkConstants.CHANGESET_WORKITEM_LINKTYPE_ID)){
				Object targetObject = link.getTargetRef().resolve();
				if(targetObject instanceof IWorkItemHandle){
					results.add((IWorkItemHandle) targetObject);
				}
			}
		}
		return results;
	}

	private static List<ILink> getAllLinks(ILinkServiceLibrary linkServiceLibrary, IChangeSetHandle changeSet) throws TeamRepositoryException{
		List<ILink> links = new ArrayList<ILink>();
		IReference source = IReferenceFactory.INSTANCE.createReferenceToItem(changeSet, "", "");
		ILinkQueryPage results = linkServiceLibrary.findLinksBySource(source);
		ILinkCollection collection = results.getAllLinksFromHereOn();
		links.addAll(collection);
		return links;
	}

	/**
	 * Returns the internal release string identifier of the work item or
	 * null if the work item doesn't have the Internal Release attribute
	 * @param workItem
	 * @param workItemCommon
	 * @param monitor
	 * @return
	 * @throws TeamRepositoryException
	 */
	public static String getInternalReleaseId(IWorkItem workItem, IWorkItemCommon workItemCommon, IProgressMonitor monitor) throws TeamRepositoryException {
		Identifier<?> literal = getInternalReleaseIdLiteral(workItem, workItemCommon, monitor);
		if(literal == null){
			return null;
		}else{
			return literal.getStringIdentifier();
		}
	}

	/**
	 * Returns the internal release literal identifier of the work item or
	 * null if the work item doesn't have the Internal Release attribute
	 * @param workItem
	 * @param workItemCommon
	 * @param monitor
	 * @return
	 * @throws TeamRepositoryException
	 */
	public static Identifier<? extends ILiteral> getInternalReleaseIdLiteral(IWorkItem workItem, IWorkItemCommon workItemCommon, IProgressMonitor monitor) throws TeamRepositoryException {
		WorkItemCommonTasks workItemCommonTasks = new WorkItemCommonTasks();
		IAttribute intRelAttribute = workItemCommonTasks.getAttribute(STGCommonIDs.STG_ATTR_INTERNAL_RELEASE, workItem.getProjectArea(), workItemCommon, monitor);
		if(intRelAttribute == null){
			return null;
		}
		if (workItem.hasAttribute(intRelAttribute)) {
			return (Identifier<? extends ILiteral>) workItem.getValue(intRelAttribute);
		}else {
			return null;
		}
	}

	/**
	 * Returns the display name of the work item's internal release or
	 * null if the work item doesn't have the Internal Release attribute
	 * @param workItem
	 * @param workItemCommon
	 * @param monitor
	 * @return
	 * @throws TeamRepositoryException
	 */
	public static String getInternalReleaseName(IWorkItem workItem, IWorkItemCommon workItemCommon,
			IProgressMonitor monitor) throws TeamRepositoryException {
		Identifier<? extends ILiteral> literalIdentifier = getInternalReleaseIdLiteral(workItem, workItemCommon, monitor);
		if (literalIdentifier != null) {
			WorkItemCommonTasks workItemCommonTasks = new WorkItemCommonTasks();
			IAttribute intRelAttribute = workItemCommonTasks.getAttribute(STGCommonIDs.STG_ATTR_INTERNAL_RELEASE, workItem.getProjectArea(), workItemCommon, monitor);
			IEnumeration<? extends ILiteral> intRelEnumeration = workItemCommon.resolveEnumeration(intRelAttribute,
					monitor);
			ILiteral intRelLiteral = intRelEnumeration.findEnumerationLiteral(literalIdentifier);
			if(intRelLiteral == null){
				return literalIdentifier.getStringIdentifier();
			}else{
				return intRelLiteral.getName();
			}
		}

		return null;
	}

	/**
	 * Returns the metadata tracked Internal Release corresponding to the given Work Item.
	 * If the work item does not have the Internal Release attribute or the identifier is
	 * not found, null is returned.
	 * @throws TeamRepositoryException
	 */
	public static InternalRelease getInternalRelease(IWorkItem workItem, IWorkItemCommon workItemCommon, IProgressMonitor monitor) throws TeamRepositoryException {
		String internalReleaseId = getInternalReleaseId(workItem, workItemCommon, monitor);
		if(internalReleaseId == null){
			return null;
		}
		List<InternalRelease> internalReleases = ConfigurationHelper.getAllInternalReleases(workItem.getProjectArea(),
				workItemCommon.getAuditableCommon());
		for (InternalRelease intRel : internalReleases) {
			if (intRel.getIdentifier().equals(internalReleaseId)) {
				return intRel;
			}
		}
		return null;
	}

	/**
	 * Returns true iff workItem1 and workItem2 have the same internal release value
	 * if neither work item have the Internal Release attribute, true is returned
	 * @param workItem1
	 * @param workItem2
	 * @param workItemCommon
	 * @param monitor
	 * @return
	 * @throws TeamRepositoryException
	 */
	public static boolean sameInternalRelease(IWorkItem workItem1, IWorkItem workItem2, IWorkItemCommon workItemCommon,
			IProgressMonitor monitor) throws TeamRepositoryException {
		String internalRelease1 = getInternalReleaseId(workItem1, workItemCommon, monitor);
		String internalRelease2 = getInternalReleaseId(workItem2, workItemCommon, monitor);

		if (internalRelease1 == null) {
			return internalRelease2 == null;
		}
		else if (internalRelease1.equals(internalRelease2)) {
			return true;
		}
		else {
			return false;
		}
	}


	/**
	 * Returns the metadata tracked Internal Release corresponding to the given workspace. If there is
	 * no Internal Release corresponding to this workspace, null is returned. 
	 * @param workspace
	 * @param projectAreaHandle
	 * @param auditableCommon
	 * @return
	 * @throws TeamRepositoryException
	 */
	public static InternalRelease getInternalReleaseOfStream(IWorkspaceHandle workspace, IProjectAreaHandle projectAreaHandle, IAuditableCommon auditableCommon) throws TeamRepositoryException{
		List<InternalRelease> internalReleases = ConfigurationHelper.getAllInternalReleases(projectAreaHandle, auditableCommon);
		for(InternalRelease ir : internalReleases){
			Stream[] streams = ir.getStreams();
			for(Stream stream : streams){
				if(workspace.getItemId().getUuidValue().equals(stream.getUuid())){
					return ir;
				}
			}
		}
		return null;
	}



	/**
	 * Returns true if the work item has the same internal release value set as the given internal release.
	 * Returns false if the work item does not have an internal release or it is different from the given internal release.
	 * @param workItem
	 * @param internalRelease
	 * @param workItemCommon
	 * @param monitor
	 * @return
	 * @throws TeamRepositoryException
	 */
	public static boolean sameInternalRelease(IWorkItem workItem, InternalRelease internalRelease, IWorkItemCommon workItemCommon, IProgressMonitor monitor) throws TeamRepositoryException{
		String workItemInternalRelease = getInternalReleaseId(workItem, workItemCommon, monitor);
		return internalRelease.getIdentifier().equals(workItemInternalRelease);
	}
	
	/**
	 * Returns a mapping of change set to list of work items that have been filtered to only include work items related to the stream.
	 * If the change set has no associated work items related to the stream, the list of work items will be empty.
	 * 
	 * @param changeSetHandles
	 * @param streamHandle
	 * @param workItemCommon
	 * @param auditableCommon
	 * @param linkService
	 * @param monitor
	 * @return
	 * @throws TeamRepositoryException 
	 */
	public static Map<IChangeSetHandle, List<IWorkItem>> getAssociatedWorkItemsRelatedToStream(List<IChangeSetHandle> changeSetHandles, IWorkspaceHandle streamHandle, 
			IWorkItemCommon workItemCommon, IAuditableCommon auditableCommon, ILinkService linkService, IProgressMonitor monitor) throws TeamRepositoryException {
		Map<IChangeSetHandle, List<IWorkItem>> workItemsRelatedToStream = new HashMap<IChangeSetHandle, List<IWorkItem>>();

		for (IChangeSetHandle changeSetHandle : changeSetHandles) {			
			List<IWorkItem> workItems = new ArrayList<IWorkItem>();
			List<IWorkItemHandle> associatedWorkItemHandles = SCMHelper.getAssociatedWorkItems(changeSetHandle, linkService);		
			for (IWorkItemHandle associatedWorkItemHandle : associatedWorkItemHandles) {
				// resolve the handle for the work item
				IWorkItem workItem = auditableCommon.resolveAuditable(associatedWorkItemHandle, IWorkItem.FULL_PROFILE, monitor);
				
				InternalRelease intRelease = SCMHelper.getInternalRelease(workItem, workItemCommon, monitor);
				if (intRelease == null) {
					//work item doesn't have an internal release, skip over it
					continue;
				}
				for (Stream stream : intRelease.getStreams()) {
					if (stream.getUuid().equals(streamHandle.getItemId().getUuidValue())) {
						//work item is related to the stream
						workItems.add(workItem);
						//continue to the next work item
						break;
					}
				}
			}
			workItemsRelatedToStream.put(changeSetHandle, workItems);
		}
		return workItemsRelatedToStream;
	}
}