package com.ibm.stg.rtc.ext.common;

import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.eclipse.core.runtime.IProgressMonitor;

import com.ibm.team.links.common.IReference;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.workitem.common.ISaveParameter;
import com.ibm.team.workitem.common.IWorkItemCommon;
import com.ibm.team.workitem.common.model.IWorkItem;
import com.ibm.team.workitem.common.model.IWorkItemReferences;
import com.ibm.team.workitem.common.model.WorkItemEndPoints;

public class WorkItemChangeInfo {

	public static final String CHANGE_DUPLICATE_INFO_PARAMETER = "com.ibm.stg.saveParameter.changeDuplicateInfo";
	public static final String CHANGE_PARENT_INFO_PARAMETER = "com.ibm.stg.saveParameter.changeParentInfo";
	public static final String CHANGE_INFO_PARAMETER = "com.ibm.stg.saveParameter.changeInfo";
	public static final String NEW_DUPLICATE = "newDuplicate";
	public static final String CHANGED_OWNER = "changedOwner";
	public static final String CHANGED_CREATOR = "changedCreator";
	public static final String PARENT_CHANGED_OWNER = "changedParentOwner";
	public static final String PARENT_CHANGED_CREATOR = "changedParentCreator";

	private Set<WorkItemChangeInfo> changedDuplicates = new HashSet<WorkItemChangeInfo>();
	private boolean creatorChanged;
	private boolean ownerChanged;
	private boolean parentOwnerChanged;
	private boolean parentCreatorChanged;
	private boolean newDuplicate;
	private IWorkItem workItem;
	private String changeType;


	private WorkItemChangeInfo(String changeInfoStr, IWorkItemCommon workItemCommon) throws TeamRepositoryException{
		if(changeInfoStr != null) {
			if (changeInfoStr.startsWith(CHANGE_DUPLICATE_INFO_PARAMETER)) {
				changeType = CHANGE_DUPLICATE_INFO_PARAMETER;
			}
			else if (changeInfoStr.startsWith(CHANGE_PARENT_INFO_PARAMETER)) {
				changeType = CHANGE_PARENT_INFO_PARAMETER;
			}
			else if (changeInfoStr.startsWith(CHANGE_INFO_PARAMETER)) {
				changeType = CHANGE_INFO_PARAMETER;
			}

			if (changeType != null) {
				this.creatorChanged = changeInfoStr.contains(CHANGED_CREATOR);
				this.ownerChanged = changeInfoStr.contains(CHANGED_OWNER);
				this.newDuplicate = changeInfoStr.contains(NEW_DUPLICATE);

				String strId = changeInfoStr.substring(changeInfoStr.lastIndexOf(".") + 1);
				int id;
				try {
					id = Integer.parseInt(strId);
				}
				catch (NumberFormatException e) {
					throw new TeamRepositoryException("Invalid duplicate save parameter. Duplicate work item id must follow the last '.'",
							e);
				}

				this.workItem = workItemCommon.findWorkItemById(id, IWorkItem.FULL_PROFILE, null);
				this.parentCreatorChanged = changeInfoStr.contains(PARENT_CHANGED_CREATOR);
				this.parentOwnerChanged = changeInfoStr.contains(PARENT_CHANGED_OWNER);
			}
		}
		if (changeType == null) {
			throw new IllegalArgumentException("The stringParameter must not be null and must start with " + CHANGE_DUPLICATE_INFO_PARAMETER + " or " + CHANGE_PARENT_INFO_PARAMETER + " or " + CHANGE_INFO_PARAMETER);
		}
	}

	/**
	 * Creates a new WorkItemChangeInfo by comparing the newWorkItemState with the oldWorkItemState
	 *
	 * @param newWorkItemState must not be null
	 * @param oldWorkItemState use null to indicate everything has changed, use the same object as newWorkItemState
	 * @param newDuplicate true if this work item is a new duplicate of another work item
	 * to indicate nothing has changed
	 */
	private WorkItemChangeInfo(IWorkItem newWorkItemState, IWorkItem oldWorkItemState, boolean newDuplicate, String changeType){
		if(newWorkItemState == null){
			throw new IllegalArgumentException("newWorkItemState cannot be null");
		}

		if (changeType == null) {
			this.changeType = CHANGE_INFO_PARAMETER;
		}
		else {
			this.changeType = changeType;
		}
		
		this.workItem = newWorkItemState;
		this.newDuplicate = newDuplicate;

		if(newWorkItemState.equals(oldWorkItemState)){
			this.creatorChanged = false;
			this.ownerChanged = false;
		}else{
			this.creatorChanged = oldWorkItemState == null || (newWorkItemState.getCreator().sameItemId(oldWorkItemState.getCreator()) == false);
			this.ownerChanged = oldWorkItemState == null || (newWorkItemState.getOwner().sameItemId(oldWorkItemState.getOwner()) == false);
		}
	}

	public boolean isCreatorChanged() {
		return creatorChanged;
	}
	public boolean isOwnerChanged() {
		return ownerChanged;
	}
	public boolean isParentCreatorChanged() {
		return parentCreatorChanged;
	}
	public boolean isParentOwnerChanged() {
		return parentOwnerChanged;
	}
	public boolean isNewDuplicate() {
		return newDuplicate;
	}
	public IWorkItem getWorkItem() {
		return workItem;
	}
	public int getWorkItemId(){
		return workItem.getId();
	}
	public Set<WorkItemChangeInfo> getChangedDuplicates(){
		return changedDuplicates;
	}

	@Override
	public boolean equals(Object o) {
		if(o != null && o instanceof WorkItemChangeInfo){
			return this.getWorkItem().sameItemId(((WorkItemChangeInfo)o).getWorkItem());
		}
		return false;
	}
	@Override
	public int hashCode() {
		return this.getWorkItemId();
	}
	
	public String toString(String changeParameter){
		String info = changeParameter;

		if (this.isCreatorChanged()) {
			info += "." + CHANGED_CREATOR;
		}
		if (this.isOwnerChanged()) {
			info += "." + CHANGED_OWNER;
		}
		if (this.isNewDuplicate()) {
			info += "." + NEW_DUPLICATE;
		}

		if (this.isParentCreatorChanged()) {
			info += "." + PARENT_CHANGED_CREATOR;
		}
		if (this.isParentOwnerChanged()) {
			info += "." + PARENT_CHANGED_OWNER;
		}

		info += "." + this.getWorkItemId();
		return info;
	}

	public Set<String> getChangeStringParameters(){
		return getChangeStringParameters(changeType);
	}
	
	public Set<String> getChangeStringParameters(String changeParameter){
		Set<String> parameters = new HashSet<String>();

		if(this.containsChanges()){
			parameters.add(this.toString(changeParameter));
		}

		for(WorkItemChangeInfo dup : this.changedDuplicates){
			parameters.addAll(dup.getChangeStringParameters());
		}

		return parameters;
	}

	public boolean containsChanges() {
		return this.isNewDuplicate() || this.isCreatorChanged() || this.isOwnerChanged() || this.getChangedDuplicates().size() > 0
				|| this.isParentCreatorChanged() || this.isParentOwnerChanged();
	}

	public void addChangeInfoParameters(ISaveParameter saveParameter, IWorkItemCommon workItemCommon) throws TeamRepositoryException{
		Set<String> stringSaveParameters = saveParameter.getAdditionalSaveParameters();
		for(String stringSaveParameter : stringSaveParameters){
			if(stringSaveParameter.contains(CHANGE_DUPLICATE_INFO_PARAMETER)) {
				this.changedDuplicates.add(new WorkItemChangeInfo(stringSaveParameter, workItemCommon));
			}
			else if (stringSaveParameter.contains(CHANGE_PARENT_INFO_PARAMETER)) {
				WorkItemChangeInfo parentWorkItemChangeInfo = new WorkItemChangeInfo(stringSaveParameter, workItemCommon);
				if (parentWorkItemChangeInfo.isParentCreatorChanged()) {
					this.parentCreatorChanged = true;
				}
				else if (parentWorkItemChangeInfo.isCreatorChanged()) {
					this.parentCreatorChanged = true;
				}
				if (parentWorkItemChangeInfo.isParentOwnerChanged()) {
					this.parentOwnerChanged = true;
				}
				else if (parentWorkItemChangeInfo.isOwnerChanged()) {
					this.parentOwnerChanged = true;
				}
			}
			else if (stringSaveParameter.contains(CHANGE_INFO_PARAMETER)) {
				WorkItemChangeInfo workItemChangeInfo = new WorkItemChangeInfo(stringSaveParameter, workItemCommon);
				if (workItemChangeInfo.isCreatorChanged()) {
					this.creatorChanged = true;
				}
				if (workItemChangeInfo.isOwnerChanged()) {
					this.ownerChanged = true;
				}
				if (workItemChangeInfo.isParentCreatorChanged()) {
					this.parentCreatorChanged = true;
				}
				if (workItemChangeInfo.isParentOwnerChanged()) {
					this.parentOwnerChanged = true;
				}
				if (workItemChangeInfo.isNewDuplicate()) {
					this.newDuplicate = true;
				}
				
			}
		}
	}

	public void addNewDuplicates(IWorkItemReferences workItemReferences, IWorkItemCommon workItemCommon, IProgressMonitor monitor) throws TeamRepositoryException{
		WorkItemCommonTasks workItemCommonTasks = new WorkItemCommonTasks();
		if(workItemCommonTasks.hasNewDuplicateLink(workItemReferences)){
			List<IReference> references = workItemReferences.getCreatedReferences(WorkItemEndPoints.DUPLICATE_WORK_ITEM);
			List<IWorkItem> newDuplicates = workItemCommonTasks.getWorkItemsFromReferences(references, workItemCommon.getAuditableCommon(), monitor);
			for(IWorkItem dup : newDuplicates){
				// we don't know the old state of the work item but it doesn't matter because it is a new duplicate
				WorkItemChangeInfo dupInfo = new WorkItemChangeInfo(dup, dup, true, CHANGE_DUPLICATE_INFO_PARAMETER);
				this.changedDuplicates.add(dupInfo);
			}
		}
	}

	public void addNewParentChild(IWorkItemReferences workItemReferences, IWorkItemCommon workItemCommon, IProgressMonitor monitor)
			throws TeamRepositoryException {
		WorkItemCommonTasks workItemCommonTasks = new WorkItemCommonTasks();
		//need to look if there are new child links in addition to a new parent link for the case when adding an existing child to a parent
		if (workItemCommonTasks.hasNewParentLink(workItemReferences) || workItemCommonTasks.hasNewChildLink(workItemReferences)) {
			this.parentCreatorChanged = true;
			this.parentOwnerChanged = true;
		}
	}

	public void addChangedDuplicate(WorkItemChangeInfo workItemChangeInfo){
		this.changedDuplicates.add(workItemChangeInfo);
	}

	public static WorkItemChangeInfo getWorkItemChangeInfo(ISaveParameter saveParameter, IWorkItemCommon workItemCommon, IProgressMonitor monitor, String changeType) throws TeamRepositoryException {
		WorkItemCommonTasks workItemCommonTasks = new WorkItemCommonTasks();
		// create the change info for this work item
		IWorkItem savedWorkItem = (IWorkItem)saveParameter.getNewState();
		boolean isNewDuplicate = workItemCommonTasks.hasNewDuplicateOfLink(saveParameter.getNewReferences());
		WorkItemChangeInfo changeInfo = new WorkItemChangeInfo(savedWorkItem, (IWorkItem)saveParameter.getOldState(), isNewDuplicate, changeType);

		// now add the duplicates which have changed

		// first, check for changed duplicates on the string save parameters
		// and changed parent info
		changeInfo.addChangeInfoParameters(saveParameter, workItemCommon);

		// second, check for new duplicate links to this work item
		// note: we don't need to check for existing duplicates because if they were changed,
		// they would have had their own save and added string save parameters
		// note: this doesn't ever seem to return any results which is why we need the third check below
		changeInfo.addNewDuplicates(saveParameter.getNewReferences(), workItemCommon, monitor);
		
		//update changes to parent/child links
		changeInfo.addNewParentChild(saveParameter.getNewReferences(), workItemCommon, monitor);

		// third, check for other save parameters with "duplicate of" link to this work item.
		// this is kind of a hack because RTC triggers a additional save on the duplicate work item
		// when a "duplicate of" link is added, but the additional save does not include the 
		// "duplicate" link in the new references so it won't get picked up in the call above. Also, we cannot
		// trigger a new save on the duplicate work item with string save parameters because
		// RTC does not refresh the work item state causing a stale data exception
		List<ISaveParameter> allSaveParameters = saveParameter.getSaveOperationParameter().getSaveParameters();
		for(ISaveParameter additionalSaveParameter : allSaveParameters){

			if(workItemCommonTasks.hasDuplicateOfLink(additionalSaveParameter.getNewReferences(), null, workItemCommon.getAuditableCommon(), monitor)){
				// there is another work item saving which is marked as a "duplicate of" so see if it's a duplicate of is linked to this work item
				IWorkItem duplicateOfWI = workItemCommonTasks.getDuplicateOfWorkItem(additionalSaveParameter.getNewReferences(), workItemCommon.getAuditableCommon(), monitor);

				if(duplicateOfWI.sameItemId(savedWorkItem)){
					// use recursive calls here because there may be a chain of duplicates being saved at the same time
					WorkItemChangeInfo dupInfo = getWorkItemChangeInfo(additionalSaveParameter, workItemCommon, monitor, CHANGE_DUPLICATE_INFO_PARAMETER);

					changeInfo.addChangedDuplicate(dupInfo);
				}
			}
			// same for "Parent" link to this work item (case when adding existing child to a parent)
			if (additionalSaveParameter.getNewReferences().hasReferences(WorkItemEndPoints.PARENT_WORK_ITEM)) {
				changeInfo.addChangeInfoParameters(additionalSaveParameter, workItemCommon);
			}
		}

		return changeInfo;

	}
}
