package com.ibm.stg.rtc.ext.common;

import com.ibm.team.process.common.IProcessConfigurationElement;
import com.ibm.team.workitem.common.ISaveParameter;
import com.ibm.team.workitem.common.internal.workflow.WorkflowInfo;
import com.ibm.team.workitem.common.model.IResolution;
import com.ibm.team.workitem.common.model.IState;
import com.ibm.team.workitem.common.model.IWorkItem;
import com.ibm.team.workitem.common.model.Identifier;
import com.ibm.team.workitem.common.workflow.IWorkflowAction;
import com.ibm.team.workitem.common.workflow.IWorkflowInfo;

/**
 * @author Kyle Christianson
 *
 */
public class WorkItemCondition {

	protected static final String TAG_CONDITION = "condition";
	private static final String ATTR_WORKITEM_STATE = "workItemState";
	private static final String ATTR_WORKITEM_RESOLUTION = "workItemResolution";
	private static final String ATTR_WORKITEM_ACTION = "workItemAction";
	private static final String ATTR_NOT_WORKITEM_STATE = "notWorkItemState";
	private static final String ATTR_NOT_WORKITEM_RESOLUTION = "notWorkItemResolution";
	private static final String ATTR_NOT_WORKITEM_ACTION = "notWorkItemAction";

	private String workItemState;
	private String workItemResolution;
	private String workItemAction;

	private boolean notWorkItemState;
	private boolean notWorkItemResolution;
	private boolean notWorkItemAction;
	protected IProcessConfigurationElement conditionElement;

	public WorkItemCondition(IProcessConfigurationElement configuration){
		if(!configuration.getName().equals(TAG_CONDITION)){
			throw new IllegalArgumentException("WorkItemCondition only supports elements with the tag <" + TAG_CONDITION + ">.");
		}

		// convert state ids to their legacy ids. if this is not done, a state id of "s2" will return the string
		// identifier of "2" which does not match the configured states. so this method removes the "s" prefix
		// to legacy states. see com.ibm.team.workitem.common.internal.workflow.WorkflowInfo
		String stateId = configuration.getAttribute(ATTR_WORKITEM_STATE);
		this.workItemState = WorkflowInfo.stripOffPrefix(stateId, 's');

		// convert resolution ids to their legacy ids. if this is not done, a resolution id of "r2" will return the string
		// identifier of "2" which does not match the configured states. so this method removes the "r" prefix
		// to legacy states. see com.ibm.team.workitem.common.internal.workflow.WorkflowInfo
		String resolutionId = configuration.getAttribute(ATTR_WORKITEM_RESOLUTION);
		this.workItemResolution = WorkflowInfo.stripOffPrefix(resolutionId, 'r');

		this.workItemAction = configuration.getAttribute(ATTR_WORKITEM_ACTION);

		this.notWorkItemState = ConfigurationHelper.getBooleanAttributeValue(configuration, ATTR_NOT_WORKITEM_STATE);
		this.notWorkItemResolution = ConfigurationHelper.getBooleanAttributeValue(configuration, ATTR_NOT_WORKITEM_RESOLUTION);
		this.notWorkItemAction = ConfigurationHelper.getBooleanAttributeValue(configuration, ATTR_NOT_WORKITEM_ACTION);

		this.conditionElement = configuration;
	}

	/**
	 * @return the workItemState
	 */
	public String getWorkItemState() {
		return workItemState;
	}

	/**
	 * @return the workItemResolution
	 */
	public String getWorkItemResolution() {
		return workItemResolution;
	}

	/**
	 * @return the workItemAction
	 */
	public String getWorkItemAction() {
		return workItemAction;
	}

	/**
	 * @return the notWorkItemState
	 */
	public boolean isNotWorkItemState() {
		return notWorkItemState;
	}

	/**
	 * @return the notWorkItemResolution
	 */
	public boolean isNotWorkItemResolution() {
		return notWorkItemResolution;
	}

	/**
	 * @return the notWorkItemAction
	 */
	public boolean isNotWorkItemAction() {
		return notWorkItemAction;
	}

	/**
	 * @return the xml element this condition was created from
	 */
	public IProcessConfigurationElement getConditionElement(){
		return conditionElement;
	}

	/**
	 * Returns true if this condition matches the current save parameter.
	 * If the state, resolution, or action is null, that check is skipped.
	 * For example, if resolution is null, the match returns true for any 
	 * resolution value as long as the state and action match. If action is
	 * specified, state is the old source work item status. If action is not
	 * specified, state is the new target status.
	 * @param saveParameter
	 * @return
	 */
	public boolean matches(ISaveParameter saveParameter){
		IWorkItem newWorkItemState = (IWorkItem)saveParameter.getNewState();
		IWorkItem oldWorkItemState = (IWorkItem)saveParameter.getOldState();

		String action = saveParameter.getWorkflowAction();
		String oldState = (oldWorkItemState == null) ? null : oldWorkItemState.getState2().getStringIdentifier();
		String newState = (newWorkItemState == null) ? null : newWorkItemState.getState2().getStringIdentifier();

		Identifier<IResolution> resolutionId = newWorkItemState.getResolution2();
		String resolution = (resolutionId == null) ? null : resolutionId.getStringIdentifier();

		return matches(action, newState, oldState, resolution);
	}

	public boolean matches(String action, String newState, String fromState, String resolution){
		if(actionMatches(action) && stateMatches(newState, fromState) && resolutionMatches(resolution)){
			return true;
		} else{
			return false;
		}
	}


	private boolean actionMatches(String action){
		if(this.getWorkItemAction() != null){
			boolean equal = this.getWorkItemAction().equals(action);
			if(this.isNotWorkItemAction() == false){
				return equal;
			} else{
				return !equal;
			}
		} else{
			return true; // no action was specified so it always matches
		}
	}

	private boolean stateMatches(String newState, String oldState){
		if(this.getWorkItemState() != null){
			boolean equal;
			if(this.getWorkItemAction() == null){
				// no action was configured so check the new state
				equal = this.getWorkItemState().equals(newState);
			} else{
				// an action is configured, so check the old (from) state
				equal = this.getWorkItemState().equals(oldState); 
			}

			if(this.isNotWorkItemState() == false){
				return equal;
			} else{
				return !equal;
			}

		} else{
			return true; // no state was specified so it always matches
		}
	}

	private boolean resolutionMatches(String resolution){
		if(this.getWorkItemResolution() != null){
			boolean equal = this.getWorkItemResolution().equals(resolution);
			if(this.isNotWorkItemResolution() == false){
				return equal;
			} else{
				return !equal;
			}
		}else{
			return true; // no resolution was specified so it always matches
		}
	}

	@Override
	/**
	 * Returns true iff the two WorkItemConditions have the same state, resolution, and action (including null values).
	 */
	public boolean equals(Object o){
		if(o == this) return true;
		if(o instanceof WorkItemCondition){
			WorkItemCondition condition = (WorkItemCondition)o;
			if(isEqual(condition.getWorkItemState(), this.getWorkItemState()) 
					&& isEqual(condition.getWorkItemResolution(), this.getWorkItemResolution()) 
					&& isEqual(condition.getWorkItemAction(), this.getWorkItemAction())){
				return true;
			}
		}
		return false;
	}

	private boolean isEqual(String s1, String s2){
		if(s1 == null || s2 == null){
			return s1 == s2;
		}else{
			return s1.equals(s2);
		}
	}

	public String toString(IWorkflowInfo workflowInfo){
		final String ACTION = "perform the action ";
		final String NOT_ACTION = "perform any action except ";
		final String IN_STATE = "be in the state ";
		final String NOT_IN_STATE = "be in any state except ";
		final String FROM_STATE = "from the state ";
		final String NOT_FROM_STATE = "from any state except ";
		final String HAVE_RESOLUTION = "have the resolution ";
		final String NOT_HAVE_RESOLUTION = "have any resolution except ";
		final String WITH_RESOLUTION = "with the resolution ";
		final String NOT_WITH_RESOLUTION = "with any resolution except ";



		if(this.getWorkItemAction() == null && this.getWorkItemState() == null && this.getWorkItemResolution() == null){
			return "perform any save";
		}

		StringBuilder sb = new StringBuilder();
		if(this.getWorkItemAction() != null){
			String actionName = workflowInfo.getActionName(Identifier.create(IWorkflowAction.class, this.getWorkItemAction()));
			if(this.isNotWorkItemAction() == false){
				sb.append(ACTION);
			}else{
				sb.append(NOT_ACTION);
			}
			sb.append(actionName);
		}

		if(this.getWorkItemState() != null){
			if(sb.length() > 0){
				sb.append(" ");
			}
			String stateName = workflowInfo.getStateName(Identifier.create(IState.class, this.workItemState));
			if(this.getWorkItemAction() == null && this.isNotWorkItemState() == false){
				sb.append(IN_STATE);
			}else if(this.getWorkItemAction() == null && this.isNotWorkItemState() == true){
				sb.append(NOT_IN_STATE);
			}else if(this.getWorkItemAction() != null && this.isNotWorkItemState() == false){
				sb.append(FROM_STATE);
			}else if(this.getWorkItemAction() != null && this.isNotWorkItemState() == true){
				sb.append(NOT_FROM_STATE);
			}
			sb.append(stateName);
		}

		if(this.getWorkItemResolution() != null){
			if(sb.length() > 0){
				sb.append(" ");
			}
			String resolutionName = workflowInfo.getResolutionName(Identifier.create(IResolution.class, this.getWorkItemResolution()));
			if(this.getWorkItemAction() == null && this.getWorkItemState() == null && this.isNotWorkItemResolution() == false){
				sb.append(HAVE_RESOLUTION);
			}else if(this.getWorkItemAction() == null && this.getWorkItemState() == null && this.isNotWorkItemResolution() == true){
				sb.append(NOT_HAVE_RESOLUTION);
			}else if(this.isNotWorkItemResolution() == false){
				sb.append(WITH_RESOLUTION);
			}else{
				sb.append(NOT_WITH_RESOLUTION);
			}
			sb.append(resolutionName);
		}

		return sb.toString();
	}
}
