package com.ibm.stg.rtc.ext.stgdefect.advisors;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.eclipse.core.runtime.IProgressMonitor;

import com.ibm.stg.rtc.ext.common.ConfigurationHelper;
import com.ibm.stg.rtc.ext.common.InternalRelease;
import com.ibm.stg.rtc.ext.common.STGCommonIDs;
import com.ibm.stg.rtc.ext.common.WorkItemCommonTasks;
import com.ibm.stg.rtc.ext.stgdefect.common.STGDefectCommonIDs;
import com.ibm.team.process.common.IProcessConfigurationElement;
import com.ibm.team.process.common.advice.AdvisableOperation;
import com.ibm.team.process.common.advice.IAdvisorInfo;
import com.ibm.team.process.common.advice.IAdvisorInfoCollector;
import com.ibm.team.process.common.advice.runtime.IOperationAdvisor;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.repository.service.AbstractService;
import com.ibm.team.workitem.common.IAuditableCommon;
import com.ibm.team.workitem.common.ISaveParameter;
import com.ibm.team.workitem.common.IWorkItemCommon;
import com.ibm.team.workitem.common.model.IAttribute;
import com.ibm.team.workitem.common.model.IEnumeration;
import com.ibm.team.workitem.common.model.ILiteral;
import com.ibm.team.workitem.common.model.IWorkItem;
import com.ibm.team.workitem.common.model.IWorkItemType;
import com.ibm.team.workitem.common.model.Identifier;

/**
 * @author Karen Paone
 * 
 */
public class PreventCRModificationAdvisor extends AbstractService implements IOperationAdvisor {
	/*
	 * List the source control types that change records cannot be modified for. It is
	 * implicit that all actions for the source control type are not allowed. In
	 * the following example, any change record modifications for the source
	 * control types CMVC and CC are not allowed.
	 * 
	 * <sourceControlType id="CMVC"/>
	 * <sourceControlType id="CC" />
	 * 
	 * This example indicates that change record with any source control types
	 * cannot be edited:
	 * 
	 * <sourceControlType id="ALL"/>
	 */
	public static final String ID = "com.ibm.stg.rtc.ext.stgdefect.advisors.PreventCRModificationAdvisor";

	private static final String SOURCE_CONTROL_TYPE_ELEMENT = "sourceControlType";
	private static final String ID_ATTRIBUTE = "id";
	private static final String ALL_VALUE = "ALL";

	/**
	 * Prevents a user from modifying the change record
	 */
	public void run(AdvisableOperation operation, IProcessConfigurationElement advisorConfiguration,
			IAdvisorInfoCollector collector, IProgressMonitor monitor) throws TeamRepositoryException {
		ISaveParameter saveParameter = null;
		IWorkItemCommon workItemCommon = this.getService(IWorkItemCommon.class);
		IAuditableCommon auditableCommon = this.getService(IAuditableCommon.class);
		WorkItemCommonTasks workItemCommonTasks = new WorkItemCommonTasks();
		IWorkItem newWorkItemState = null;
		IWorkItem oldWorkItemState = null;

		Object tempObj = operation.getOperationData();
		if (workItemCommonTasks.isSaveParamterType(tempObj)) {
			saveParameter = (ISaveParameter) tempObj;
		}
		else {
			return;
		}

		tempObj = saveParameter.getNewState();
		if (workItemCommonTasks.isIWorkItemType(tempObj)) {
			newWorkItemState = (IWorkItem) tempObj;
		}
		else {
			return;
		}

		tempObj = saveParameter.getOldState();
		if (workItemCommonTasks.isIWorkItemType(tempObj)) {
			oldWorkItemState = (IWorkItem) tempObj;
		}
		else if (tempObj == null) {
			oldWorkItemState = null;
		}
		else {
			return;
		}

		IWorkItemType workItemType = workItemCommon.findWorkItemType(newWorkItemState.getProjectArea(),
				newWorkItemState.getWorkItemType(), monitor);
		if (workItemType == null)
			return;

		if (!workItemType.getIdentifier().equals(STGDefectCommonIDs.CHANGE_RECORD_ID)) {
			return;
		}

		IAttribute internalReleaseAttr = workItemCommonTasks.getAttribute(STGCommonIDs.STG_ATTR_INTERNAL_RELEASE, newWorkItemState.getProjectArea(), workItemCommon, monitor);
		if (internalReleaseAttr == null) {
			return;
		}

		if (!newWorkItemState.isAttributeSet(internalReleaseAttr)) {
			return;
		}

		//retrieve the CR's internal release literal value
		IEnumeration<? extends ILiteral> intRelEnumeration = workItemCommon.resolveEnumeration(internalReleaseAttr,
				monitor);
		ILiteral literal = intRelEnumeration.findEnumerationLiteral((Identifier<? extends ILiteral>) newWorkItemState
				.getValue(internalReleaseAttr));
		String internalReleaseLiteral = literal.getIdentifier2().getStringIdentifier();

		if (internalReleaseLiteral == null) {
			return;
		}

		InternalRelease crIntRel = null;

		//get the internal release with this CR's literal value
		List<InternalRelease> internalReleases = ConfigurationHelper.getAllInternalReleases(newWorkItemState
				.getProjectArea(), auditableCommon);
		for (InternalRelease internalRelease : internalReleases) {
			if (internalRelease.getIdentifier().equals(internalReleaseLiteral)) {
				crIntRel = internalRelease;
				break;
			}
		}

		if (crIntRel == null) {
			return;
		}

		//retrieve the source control types that modifications are prevented
		List<String> sourceControlTypes = readAdvisorConfiguration(advisorConfiguration);

		if (sourceControlTypes == null || sourceControlTypes.isEmpty()) {
			return;
		}

		// determine if this change record is for a source control type that does not allow modifications in RTC
		for (String sourceControlType : sourceControlTypes) {
			if (crIntRel.getSourceControlType().equals(sourceControlType) || sourceControlType.equals(ALL_VALUE)) {
				if (oldWorkItemState == null) {
					// error, the change record is being created
					IAdvisorInfo info = collector.createProblemInfo("Permission denied",
							"Cannot manually create a Change Record for internal release '" + crIntRel.getName() + "' because its source control type is "
									+ sourceControlType + ".", ID);
					info.setProblemObject(newWorkItemState.getItemHandle());
					collector.addInfo(info);
				}
				else {
					// error, CRs with internal releases for this source control
					// type cannot be modified.
					IAdvisorInfo info = collector.createProblemInfo("Permission denied",
							"Cannot modify a Change Record of internal release '" + crIntRel.getName() + "' because its source control type is "
									+ sourceControlType + ".", ID);
					info.setProblemObject(newWorkItemState.getItemHandle());
					collector.addInfo(info);
				}
				break;
			}
		}

	}

	public static List<String> readAdvisorConfiguration(IProcessConfigurationElement advisorConfiguration) {
		Set<String> sourceControlTypes = new HashSet<String>();

		IProcessConfigurationElement[] configurationElements = advisorConfiguration.getChildren();
		for (int i = 0; i < configurationElements.length; i++) {
			IProcessConfigurationElement configurationElement = configurationElements[i];
			if (SOURCE_CONTROL_TYPE_ELEMENT.equals(configurationElement.getName())) {
				sourceControlTypes.add(configurationElement.getAttribute(ID_ATTRIBUTE));
			}
		}

		return new ArrayList<String>(sourceControlTypes);
	}
}
