package com.ibm.stg.rtc.ext.stgdefect.advisors;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;

import org.eclipse.core.runtime.IProgressMonitor;

import com.ibm.stg.rtc.ext.common.InternalRelease;
import com.ibm.stg.rtc.ext.common.ItemHandleKey;
import com.ibm.stg.rtc.ext.common.SCMHelper;
import com.ibm.team.links.common.service.ILinkService;
import com.ibm.team.process.common.IProcessConfigurationElement;
import com.ibm.team.process.common.advice.AdvisableOperation;
import com.ibm.team.process.common.advice.IAdvisorInfoCollector;
import com.ibm.team.process.common.advice.IReportInfo;
import com.ibm.team.process.common.advice.runtime.IOperationAdvisor;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.repository.service.AbstractService;
import com.ibm.team.repository.service.IRepositoryItemService;
import com.ibm.team.scm.common.IChangeSet;
import com.ibm.team.scm.common.IChangeSetHandle;
import com.ibm.team.scm.common.IWorkspace;
import com.ibm.team.scm.common.IWorkspaceHandle;
import com.ibm.team.scm.common.process.IScmDeltaSource;
import com.ibm.team.workitem.common.IAuditableCommon;
import com.ibm.team.workitem.common.IWorkItemCommon;
import com.ibm.team.workitem.common.model.IWorkItem;
import com.ibm.team.workitem.common.model.IWorkItemHandle;
import com.ibm.team.workitem.common.model.ItemProfile;

/**
 * If the deliver target stream is tracked in metadata, this precondition will
 * validate if all change sets on the delivery are associated with a CR having
 * an Internal Release that is related to the deliver target stream in metadata.
 * If not, it prevents the user from delivering. Together with the 
 * RequireAllAssociateChangeSetsDelivered precondition on work item save, this 
 * precondition ensure change sets are delivered properly.
 * 
 * This precondition can be configured with both CreateCROnDeliverAdvisor and
 * WIActionOnChangeSetAssociation. CreateCRDeliverAdvisor will run before this
 * precondition. If WIActionOnChangeSetAssociation is configured, the new CR
 * will automatically change to the configured state.
 * 
 * 
 * @author Karen Paone
 * @see com.ibm.stg.rtc.ext.stgdefect.advisors.RequireAllAssociatedChangeSetsDeliveredAdvisor
 * 
 */
public class PreventDeliverWithoutCRAdvisor extends AbstractService implements IOperationAdvisor {
	private static final String ID = "com.ibm.stg.rtc.ext.stgdefect.advisors.PreventDeliverWithoutCRAdvisor";

	public void run(AdvisableOperation operation, IProcessConfigurationElement advisorConfiguration,
			IAdvisorInfoCollector collector, IProgressMonitor monitor) throws TeamRepositoryException {
		if (operation == null) {
			return;
		}
		
		String[] actions = operation.getActions();
		if (actions == null) {
			return;
		}
		
		IScmDeltaSource scmDeltaSource;

		Object tempObj = operation.getOperationData();
		if (tempObj instanceof IScmDeltaSource) {
			scmDeltaSource = (IScmDeltaSource) tempObj;
		} else {
			return;
		}
		
		IAuditableCommon auditableCommon = this.getService(IAuditableCommon.class);
		IWorkItemCommon workItemCommon = this.getService(IWorkItemCommon.class);
		ILinkService linkService = this.getService(ILinkService.class);

		//retrieve everything included in the delivery
		Map<ItemHandleKey<IWorkspaceHandle>, List<IChangeSetHandle>> deliverWorkspacesToChangeSets = SCMHelper.getDeliverInformation(scmDeltaSource);
		
		//go through through the change sets by workspace/stream
		for (ItemHandleKey<IWorkspaceHandle> workspaceKey : deliverWorkspacesToChangeSets.keySet()) {
			IWorkspaceHandle targetWorkspaceHandle = workspaceKey.getItemHandle();
			InternalRelease internalRelease = SCMHelper.getInternalReleaseOfStream(targetWorkspaceHandle, operation.getProcessArea().getProjectArea(), auditableCommon);
			if (internalRelease == null) {
				//not a tracked stream, skip it
				continue;
			}
			
			//retrieve all related work items for the target deliever stream associated to the change sets 
			Map<IChangeSetHandle, List<IWorkItem>> changeSetWorkItems = SCMHelper.getAssociatedWorkItemsRelatedToStream(deliverWorkspacesToChangeSets.get(workspaceKey),
					targetWorkspaceHandle, workItemCommon, auditableCommon, linkService, monitor);

			for (IChangeSetHandle changeSetHandle : changeSetWorkItems.keySet()) {
				//if the change set has an empty list of work items, then no related work items were found for the target deliver stream
				if (changeSetWorkItems.get(changeSetHandle).isEmpty()) {
					// if a matching CR wasn't found for this change set, any
					// problems will be added to the collector

					IChangeSet changeSet = (IChangeSet) auditableCommon
							.resolveAuditable(changeSetHandle, ItemProfile.createProfile(IChangeSet.ITEM_TYPE,
									new String[] { IChangeSet.COMMENT_PROPERTY }), monitor);
					IRepositoryItemService itemService = this.getService(IRepositoryItemService.class);
					IWorkspace targetWorkSpace = (IWorkspace) itemService.fetchItem(targetWorkspaceHandle,
							new String[] { IWorkspace.NAME_PROPERTY });
					String targetStreamName = "";
					if (targetWorkSpace != null) {
						targetStreamName = targetWorkSpace.getName();
					}

					List<String> workItemIDs = new ArrayList<String>();

					
					for (IWorkItemHandle workItemHandle : SCMHelper.getAssociatedWorkItems(changeSetHandle, linkService)) {
						IWorkItem workItem = auditableCommon.resolveAuditable(workItemHandle, IWorkItem.SMALL_PROFILE, monitor);
						workItemIDs.add(String.valueOf(workItem.getId()));
					}

					IReportInfo info = collector
							.createProblemInfo("Change Record required",
									"All change sets delivered to tracked stream '"
											+ targetStreamName
											+ "' must be associated to a Change Record with the Internal Release '"
											+ internalRelease.getName()
											+ "'.  Change Set "
											+ (changeSet.getComment() == null
													|| changeSet.getComment().trim().equals("") ? "<no comment>" : "'"
													+ changeSet.getComment() + "'") + " is associated to:  "
											+ workItemIDs + ".", ID);
					collector.addInfo(info);
				}
			}			
		}		
	}
}
