package com.ibm.stg.rtc.ext.stgdefect.advisors;

import java.util.ArrayList;
import java.util.List;

import org.eclipse.core.runtime.IProgressMonitor;

import com.ibm.stg.rtc.ext.common.WorkItemCommonTasks;
import com.ibm.stg.rtc.ext.stgdefect.common.STGDefectCommonIDs;
import com.ibm.team.links.common.IReference;
import com.ibm.team.process.common.IProcessConfigurationElement;
import com.ibm.team.process.common.advice.AdvisableOperation;
import com.ibm.team.process.common.advice.IAdvisorInfo;
import com.ibm.team.process.common.advice.IAdvisorInfoCollector;
import com.ibm.team.process.common.advice.runtime.IOperationAdvisor;
import com.ibm.team.repository.common.PermissionDeniedException;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.repository.service.AbstractService;
import com.ibm.team.workitem.common.IAuditableCommon;
import com.ibm.team.workitem.common.ISaveParameter;
import com.ibm.team.workitem.common.IWorkItemCommon;
import com.ibm.team.workitem.common.model.ICategoryHandle;
import com.ibm.team.workitem.common.model.IWorkItem;

public class PreventFiledAgainstEditAdvisor extends AbstractService implements IOperationAdvisor {

	/*
	 * Prevents the Filed Against field from being changed on a work item if it contains
	 * active Change Records. Prevents the Filed Against field from being changed on a
	 * Change Record unless it is being changed to match the parent work item.
	 * 
	 * Also if a new Change Record is being saved, the Filed Against is set to the same
	 * as the parent category. Note that advisors usually should not modify a work item, however this is a special 
	 * case because the CR Filed Against must be programmatically set on create and we need to avoid sending
	 * multiple notifications. From then on, the follow-up action SetCRCategory will keep the Filed Against
	 * value of the CR matching its parent.
	 */

	public static final String ID = "com.ibm.stg.rtc.ext.stgdefect.advisors.PreventFiledAgainstEditAdvisor";

	public void run(AdvisableOperation operation, IProcessConfigurationElement advisorConfiguration, IAdvisorInfoCollector collector, IProgressMonitor monitor) throws TeamRepositoryException {
		WorkItemCommonTasks workItemCommonTasks = new WorkItemCommonTasks();
		ISaveParameter saveParameter = null;
		IWorkItem newWorkItemState;
		IWorkItem oldWorkItemState;

		Object tempObj = operation.getOperationData();
		if (workItemCommonTasks.isSaveParamterType(tempObj)) {
			saveParameter = (ISaveParameter) tempObj;
		} else {
			return;
		}

		tempObj = saveParameter.getNewState();
		if (workItemCommonTasks.isIWorkItemType(tempObj)) {
			newWorkItemState = (IWorkItem) tempObj;
		} else {
			return;
		}

		tempObj = saveParameter.getOldState();
		if (workItemCommonTasks.isIWorkItemType(tempObj)) {
			oldWorkItemState = (IWorkItem) tempObj;
		} else if(tempObj == null){
			oldWorkItemState = null;
		} else{
			return;
		}

		IAuditableCommon auditableCommon = this.getService(IAuditableCommon.class);
		IWorkItemCommon workItemCommon = this.getService(IWorkItemCommon.class);
		
		ICategoryHandle newCategory = newWorkItemState.getCategory();
		ICategoryHandle oldCategory = (oldWorkItemState == null) ? null : oldWorkItemState.getCategory();
		boolean categoryChanging = isCategoryChanging(saveParameter.isCreation(), newCategory, oldCategory);	

		if (categoryChanging) {
			// get change records - do this after checking if the category is changing to prevent errors about permissions when it isn't necessary
			
			//STG Defect 14807: this will not fail if the user does not have permission to child work item
			List<IWorkItem> changeRecords = new ArrayList<IWorkItem>();
			List<IReference> childReferences = workItemCommonTasks.getAllChildWorkItemReferences(saveParameter.getNewReferences(), auditableCommon, workItemCommon, monitor);
			for (IReference childReference : childReferences) {
				try {
					IWorkItem workItem = workItemCommonTasks.getWorkItemFromReference(childReference,
							auditableCommon, monitor);
					if (workItem.getWorkItemType().equals(STGDefectCommonIDs.CHANGE_RECORD_ID)) {
						changeRecords.add(workItem);	
					}
				}
				catch (PermissionDeniedException e) {
					//code is only concerned with child change records in the same project
					//intentional no-op
				}
			}
		
			if (changeRecords.size() > 0) {
				// check for active change records
	
				for(IWorkItem changeRecord : changeRecords){
					if(STGDefectCommonIDs.CHANGE_RECORD_ACTIVE_STATES.contains(changeRecord.getState2().getStringIdentifier())){
						// the work item cannot change categories because it has active change records
						IAdvisorInfo info= collector.createProblemInfo("Cannot edit Filed Against", "Filed Against cannot be changed when a work item has active change records. Please close, cancel, or nochange all Change Records before changing Filed Against.", ID);
						info.setProblemObject(newWorkItemState.getItemHandle());
						collector.addInfo(info);
						return;
					}
				}
			}
		}


		if (newWorkItemState.getWorkItemType().equals(STGDefectCommonIDs.CHANGE_RECORD_ID)){
			// get the CR's parent defect's category:  this will get a permission denied error if the user does not have access to the parent work item
			IWorkItem parent = workItemCommonTasks.getParentWorkItem(saveParameter.getNewReferences(), auditableCommon, monitor);
			if(parent != null){
				ICategoryHandle parentCategory = parent.getCategory();

				// if this is a new change record, set its category the same as its parent
				// only need to set the category on create because the follow-up action SetCRCategory will keep the CR's category
				// in sync with its parent after creation
				if(saveParameter.isCreation()){
					newWorkItemState.setCategory(parentCategory);
				}else if(categoryChanging){
					// if category is changing and the new category is not the same as the parent category
					if(newCategory == null || !newCategory.sameItemId(parentCategory)){
						// then add an error
						IAdvisorInfo info= collector.createProblemInfo("Cannot edit Filed Against", "Filed Against cannot be edited on a Change Record. Change Records must have the same Filed Against Category as their parent work item.", ID);
						info.setProblemObject(newWorkItemState.getItemHandle());
						collector.addInfo(info);
					}
				}
			}
		}
	}

	/**
	 * Returns true iff the category is changing. A new work item means it is changing.
	 * @param creation
	 * @param newWorkItemState
	 * @param oldWorkItemState
	 * @return
	 */
	private boolean isCategoryChanging(boolean creation, ICategoryHandle newCategory, ICategoryHandle oldCategory) {
		if(creation == true){
			return true;
		}else if (newCategory == null){
			return oldCategory == null;
		}else{
			return !newCategory.sameItemId(oldCategory);
		}
	}
}
