package com.ibm.stg.rtc.ext.stgdefect.advisors;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.eclipse.core.runtime.IProgressMonitor;

import com.ibm.stg.rtc.ext.common.ConfigurationHelper;
import com.ibm.stg.rtc.ext.common.InternalRelease;
import com.ibm.stg.rtc.ext.common.SCMHelper;
import com.ibm.stg.rtc.ext.common.Stream;
import com.ibm.stg.rtc.ext.common.WorkItemCommonTasks;
import com.ibm.team.links.common.IReference;
import com.ibm.team.links.common.registry.IEndPointDescriptor;
import com.ibm.team.links.common.registry.ILinkTypeRegistry;
import com.ibm.team.process.common.IProcessConfigurationElement;
import com.ibm.team.process.common.advice.AdvisableOperation;
import com.ibm.team.process.common.advice.IAdvisorInfoCollector;
import com.ibm.team.process.common.advice.IReportInfo;
import com.ibm.team.process.common.advice.runtime.IOperationAdvisor;
import com.ibm.team.repository.common.ItemNotFoundException;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.repository.common.UUID;
import com.ibm.team.repository.service.AbstractService;
import com.ibm.team.scm.common.IChangeSet;
import com.ibm.team.scm.common.IChangeSetHandle;
import com.ibm.team.scm.common.IScmService;
import com.ibm.team.scm.common.IWorkspace;
import com.ibm.team.scm.common.IWorkspaceHandle;
import com.ibm.team.scm.common.links.ILinkConstants;
import com.ibm.team.workitem.common.ISaveParameter;
import com.ibm.team.workitem.common.IWorkItemCommon;
import com.ibm.team.workitem.common.model.IWorkItem;
import com.ibm.team.workitem.common.model.IWorkItemReferences;
import com.ibm.team.workitem.common.model.ItemProfile;

/**
 * This precondition requires all of the change record's associated change sets
 * to be delivered to AT LEAST ONE stream belonging to the CR's internal release.
 * Together with the PreventDeliverWithoutCRAdvisor precondition
 * on deliver, this precondition ensures change sets are delivered properly for
 * tracked streams.
 * 
 * @author Karen Paone
 * @see com.ibm.stg.rtc.ext.stgdefect.advisors.PreventDeliverWithoutCRAdvisor
 */

/*
 * Example - requires all associated change set to be delivered before allowing the change record to move to the Code Complete state:
 * 
 	<precondition id="com.ibm.stg.rtc.ext.stgdefect.advisors.RequireAllAssociatedChangeSetsDeliveredAdvisor" name="STG: Require all change sets delivered" optional="false">
        <workItemType id="com.ibm.stg.workItemType.changeRecord">
                <workItemState id="com.ibm.stg.workflow.changeRecord.state.codeComplete"/>
        </workItemType>
    </precondition>
 */

public class RequireAllAssociatedChangeSetsDeliveredAdvisor extends AbstractService implements IOperationAdvisor {
	public static final String ID = "com.ibm.stg.rtc.ext.stgdefect.advisors.RequireAllAssociatedChangeSetsDeliveredAdvisor";

	public void run(AdvisableOperation operation, IProcessConfigurationElement advisorConfiguration,
			IAdvisorInfoCollector collector, IProgressMonitor monitor) throws TeamRepositoryException {
		ISaveParameter saveParameter = null;
		IWorkItem newWorkItemState = null;
		WorkItemCommonTasks workItemCommonTasks = new WorkItemCommonTasks();

		Object tempObj = operation.getOperationData();
		if (workItemCommonTasks.isSaveParamterType(tempObj)) {
			saveParameter = (ISaveParameter) tempObj;
		}
		else {
			return;
		}

		tempObj = saveParameter.getNewState();
		if (workItemCommonTasks.isIWorkItemType(tempObj)) {
			newWorkItemState = (IWorkItem) tempObj;
		}
		else {
			return;
		}

		// check to see if the work item type has this advisor configured, if not return
		List<String> states = ConfigurationHelper.getWIStatesConfigured(advisorConfiguration, newWorkItemState.getWorkItemType());
		if(states == null){
			// this work item type was not configured
			return;
		}

		if(states.size() > 0 && !states.contains(newWorkItemState.getState2().getStringIdentifier())){
			// this state was not configured
			return;
		}

		IWorkItemCommon workItemCommon = this.getService(IWorkItemCommon.class);

		//retrieve the associated change sets
		Map<UUID, IChangeSetHandle> changeSetMap = getAssociatedChangeSets(newWorkItemState, workItemCommon, monitor);

		InternalRelease internalRelease = SCMHelper.getInternalRelease(newWorkItemState, workItemCommon, monitor);

		if (changeSetMap.size() > 0 && (internalRelease == null || internalRelease.getStreams().length == 0)) {
			//do not allow the CR to change states if the CR has change sets and the internal release is not set or has no streams
			String typeName = workItemCommonTasks.getTypeName(newWorkItemState, workItemCommon, monitor);
			String stateName = workItemCommonTasks.getStateName(newWorkItemState, workItemCommon, monitor);
			String summary = null;
			String errorMsg = null;
			//the CR does not have an internal release set
			if (internalRelease == null) {
				summary = "Invalid Internal Release";
				errorMsg = "The internal release selected was not found in metadata. The internal release must be setup in metadata to confirm all of the associated change sets have been delivered to at least one of the internal release's streams before changing "
						+ typeName + " to state " + stateName + ".";
			}
			// the CR has an internal release set, but does not have any streams
			else if (internalRelease.getStreams().length == 0) {
				summary = "Internal Release does not have any streams";
				errorMsg = "The "
						+ internalRelease.getName()
						+ " internal release does not have any streams.  The internal release must have one or more streams to confirm all of the associated change sets have been delivered to at least one of the streams before changing "
						+ typeName + " to state " + stateName + ".";
			}
			IReportInfo info = collector.createProblemInfo(summary, errorMsg, ID);
			info.setProblemObject(newWorkItemState.getItemHandle());
			collector.addInfo(info);
			return;
		}

		if (changeSetMap.size() == 0) {
			return;
		}

		IScmService scmService = getService(IScmService.class);		

		// iterate through the streams belonging the CR's internal release, see
		// which of the change sets have been delivered to these streams

		// as change sets are found in streams, they are removed from the list
		// of change sets since each change set only needs to be delivered to 1 of the streams
		for (Stream stream : internalRelease.getStreams()) {
			if (changeSetMap.size() == 0) {
				//all change sets have been located in streams
				break;
			}
			//create a handle to the stream
			IWorkspaceHandle streamHandle = (IWorkspaceHandle)IWorkspace.ITEM_TYPE.createItemHandle(UUID.valueOf(stream.getUuid()), null);
			// retrieve the change sets that are found in this stream's change history
			// this throws ItemNotFoundException if the streamHandle doesn't exist
			IChangeSetHandle[] foundChangeSets;
			try{
				foundChangeSets = scmService.changeSetsInWorkspace(streamHandle, changeSetMap
						.values().toArray(new IChangeSetHandle[changeSetMap.size()]), null, null);
			}catch(ItemNotFoundException e){
				// if a stream cannot be found, skip it as not delivered to
				continue;
			}

			if (foundChangeSets != null && foundChangeSets.length > 0) {
				// at least one change set was delivered to the stream
				// remove these change sets from the list, we no longer need to look for these change sets in other streams
				for (IChangeSetHandle changeSet : foundChangeSets) {
					changeSetMap.remove(changeSet.getItemId());
				}
			}				
		}

		if (changeSetMap.size() == 0) {
			// all change sets have been delivered to at least 1 stream belonging to the internal release set on the CR
			return;
		}

		// no hits, so this CS has not been delivered to any stream belonging to the CR's internal release
		
		String streams = "";
		for (int i = 0; i < internalRelease.getStreams().length; i++ ) {
			Stream stream = internalRelease.getStreams()[i];
			streams += stream.getName();
			if ((i + 1) < internalRelease.getStreams().length) {
				streams += ", ";
			}
		}

		for (IChangeSetHandle changeSetHandle : changeSetMap.values()) {
			String typeName = workItemCommonTasks.getTypeName(newWorkItemState, workItemCommon, monitor);
			String stateName = workItemCommonTasks.getStateName(newWorkItemState, workItemCommon, monitor);
			IChangeSet changeSet = (IChangeSet) workItemCommon.getAuditableCommon().resolveAuditable(changeSetHandle,
					ItemProfile.createFullProfile(IChangeSet.ITEM_TYPE), monitor);

			IReportInfo info = collector.createProblemInfo("Change set not delivered",
					"All associated change sets must be delivered to at least one stream belonging to "
							+ typeName
							+ "'s internal release '"
							+ internalRelease.getName()
							+ "' to move to the '"
							+ stateName
							+ "' state.  Change set "
							+ ((changeSet.getComment() != null && changeSet.getComment().trim().length() > 0) ? "'"
									+ changeSet.getComment() + "'" : "<no comment>")
							+ " needs to be delivered to one or more of these streams:  " + streams + ".", ID);
			info.setProblemObject(newWorkItemState.getItemHandle());
			collector.addInfo(info);
		}
	}

	/**
	 * 
	 * @param workItem
	 * @param workItemCommon
	 * @param monitor
	 * @return map of change set handles that is associated to a work item, key is the UUID
	 * @throws TeamRepositoryException
	 */
	private Map<UUID, IChangeSetHandle> getAssociatedChangeSets(IWorkItem workItem, IWorkItemCommon workItemCommon,
			IProgressMonitor monitor) throws TeamRepositoryException {
		IWorkItemReferences references = workItemCommon.resolveWorkItemReferences(workItem, monitor);
		IEndPointDescriptor changeSetEndPoint = ILinkTypeRegistry.INSTANCE.getLinkType(
				ILinkConstants.CHANGESET_WORKITEM_LINKTYPE_ID).getSourceEndPointDescriptor();
		List<IReference> changeSetReferences = references.getReferences(changeSetEndPoint);

		// create a list of change set handles
		Map<UUID, IChangeSetHandle> changeSetHandles = new HashMap<UUID, IChangeSetHandle>();
		for (IReference changeSetReference : changeSetReferences) {
			// resolve the change reference to retrieve the change set handle
			Object resolved = changeSetReference.resolve();
			if (resolved == null || !(resolved instanceof IChangeSetHandle)) {
				throw new TeamRepositoryException("A link of type " + ILinkConstants.CHANGESET_WORKITEM_LINKTYPE_ID
						+ " did not resolve to a change set.");
			}
			IChangeSetHandle handle = (IChangeSetHandle) resolved;
			changeSetHandles.put(handle.getItemId(), handle);
		}
		return changeSetHandles;
	}
}
