package com.ibm.stg.rtc.ext.stgdefect.advisors;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.eclipse.core.runtime.IProgressMonitor;

import com.ibm.stg.rtc.ext.common.ConfigurationHelper;
import com.ibm.stg.rtc.ext.common.WorkItemCommonTasks;
import com.ibm.team.process.common.IProcessConfigurationElement;
import com.ibm.team.process.common.advice.AdvisableOperation;
import com.ibm.team.process.common.advice.IAdvisorInfo;
import com.ibm.team.process.common.advice.IAdvisorInfoCollector;
import com.ibm.team.process.common.advice.runtime.IOperationAdvisor;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.repository.service.AbstractService;
import com.ibm.team.workitem.common.ISaveParameter;
import com.ibm.team.workitem.common.IWorkItemCommon;
import com.ibm.team.workitem.common.model.IAttribute;
import com.ibm.team.workitem.common.model.ILiteral;
import com.ibm.team.workitem.common.model.IWorkItem;
import com.ibm.team.workitem.common.model.Identifier;

public class RequireAttributeBasedOnOtherEnumValueAdvisor extends AbstractService implements IOperationAdvisor {
	/*
	 * This advisor requires the defined field to be filled in for the enumeration and values configured in the precondition xml
	 * 
	 * <precondition id="com.ibm.stg.rtc.ext.stgdefect.advisors.RequireAttributeBasedOnOtherEnumValueAdvisor" name="STG: Require Attribute Based on Other Enum Value" optional="false">
            <workItemType id="com.ibm.stg.workItemType.stgDefect">
                <requiredAttribute id="com.ibm.stg.attribute.severityJustification">
                    <requiredByAttribute id="com.ibm.team.workitem.attribute.severity" value="severity.literal.l4"/>
                    <requiredByAttribute id="com.ibm.team.workitem.attribute.severity" value="severity.literal.l5"/>
                </requiredAttribute>
                <requiredAttribute id="com.ibm.team.workitem.attribute.duedate">
                    <requiredByAttribute id="com.ibm.stg.attribute.phaseFound" value="com.ibm.stg.enumeration.phaseFound.literal.integrationTest"/>
                    <requiredByAttribute id="com.ibm.stg.attribute.subphase" value="com.ibm.stg.enumeration.subphase.literal.hardwareBringup"/>
                </requiredAttribute>
            </workItemType>
            <workItemType id="com.ibm.stg.workItemType.changeRecord">
                <requiredAttribute id="com.ibm.stg.attribute.targetFixDate">
                    <requiredByAttribute id="com.ibm.stg.attribute.operatingSystem" value="com.ibm.stg.enumeration.operatingSystem.literal.aix"/>
                </requiredAttribute>
            </workItemType>
        </precondition>
	 */


	public static final String ID = "com.ibm.stg.rtc.ext.stgdefect.advisors.RequireAttributeBasedOnOtherEnumValueAdvisor";

	private static final String TAG_REQUIRED_ATTR = "requiredAttribute";
	private static final String TAG_REQUIRED_BY_ATTR = "requiredByAttribute";
	private static final String ATTR_ID = "id";
	private static final String ATTR_VALUE = "value";

	public void run(AdvisableOperation operation, IProcessConfigurationElement advisorConfiguration, IAdvisorInfoCollector collector, IProgressMonitor monitor) throws TeamRepositoryException {
		IWorkItemCommon workItemCommon = this.getService(IWorkItemCommon.class);
		WorkItemCommonTasks workItemCommonTasks = new WorkItemCommonTasks();
		ISaveParameter saveParameter = null;
		IWorkItem newWorkItemState = null;

		Object tempObj = operation.getOperationData();
		if (workItemCommonTasks.isSaveParamterType(tempObj)) {
			saveParameter = (ISaveParameter) tempObj;
		} else {
			return;
		}

		tempObj = saveParameter.getNewState();
		if (workItemCommonTasks.isIWorkItemType(tempObj)) {
			newWorkItemState = (IWorkItem) tempObj;
		} else {
			return;
		}

		// determine what is configured for the parent work item type when children in states
		IProcessConfigurationElement typeConfiguration = ConfigurationHelper.getWITypeConfiguration(advisorConfiguration,
				newWorkItemState.getWorkItemType());
		
		if (typeConfiguration == null) {
			return;
		}

		for (String requiredAttributeId : getRequiredAttributes(typeConfiguration)) {
			IAttribute requiredAttribute = workItemCommonTasks.getAttribute(requiredAttributeId, newWorkItemState.getProjectArea(),
					workItemCommon, monitor);

			// see if the required attribute is set, if it does not have a value, see if an error is needed
			if (!workItemCommonTasks.attributeHasNonNullValue(newWorkItemState, requiredAttribute, workItemCommon.getAuditableCommon())) {
				// retrieve all of the attributes - see if the required by attribute value requires the attribute to be set and error
				List<IAttribute> requiredByAttributes = getRequiredByAttributes(requiredAttributeId, newWorkItemState, typeConfiguration, workItemCommonTasks,
						workItemCommon, monitor);

				for (IAttribute requiredByAttribute : requiredByAttributes) {
					// required attribute is not filled in so report a problem
					ILiteral requiredByValueLiteral = workItemCommon.resolveEnumeration(requiredByAttribute, monitor)
							.findEnumerationLiteral((Identifier<ILiteral>) newWorkItemState.getValue(requiredByAttribute));
					IAttribute requiredAttr = workItemCommonTasks.getAttribute(requiredAttributeId, newWorkItemState.getProjectArea(),
							workItemCommon, monitor);

					IAdvisorInfo info = collector.createProblemInfo(requiredAttr.getDisplayName() + " Required",
							requiredAttr.getDisplayName() + " field is required for " + requiredByAttribute.getDisplayName() + " value "
									+ requiredByValueLiteral.getName(), ID);
					info.setProblemObject(newWorkItemState.getItemHandle());
					collector.addInfo(info);
				}
			}
		}
	}

	/**
	 *  Loops through the child configuration xml and returns all the attributes and values which require the attribute to have a value
	 * 
	 * @param requiredAttributeId
	 * @param advisorConfiguration
	 * @return
	 */
	private Map<String, Set<String>> getAttributesAndValues(String requiredAttributeId, IProcessConfigurationElement advisorConfiguration) {
		Map<String, Set<String>> attrsAndValues = new HashMap<String, Set<String>>();
		
		IProcessConfigurationElement[] configurationElements = advisorConfiguration.getChildren();
		
		for (IProcessConfigurationElement configurationElement : configurationElements) {
			if (configurationElement.getAttribute(ATTR_ID).equals(requiredAttributeId)) {
				for (IProcessConfigurationElement childElement : configurationElement.getChildren()) {
					if (childElement.getName().equals(TAG_REQUIRED_BY_ATTR)) {
						String attribute = childElement.getAttribute(ATTR_ID);
						String value = childElement.getAttribute(ATTR_VALUE);

						Set<String> values = new HashSet<String>();

						if (attrsAndValues.containsKey(attribute)) {
							values = attrsAndValues.get(attribute);
						}
						values.add(value);
						attrsAndValues.put(attribute, values);
					}
				}
			}
		}
		
		return attrsAndValues;
	}
	
	/**
	 * retrieve the required attribute id
	 * 
	 * @param advisorConfiguration
	 * @return
	 * @throws TeamRepositoryException 
	 */
	private List<String> getRequiredAttributes(IProcessConfigurationElement configuration) throws TeamRepositoryException {
		List<String> requiredAttributes = new ArrayList<String>();

		for (IProcessConfigurationElement childElement : configuration.getChildren()) {
			if (childElement.getName().equals(TAG_REQUIRED_ATTR)) {
				requiredAttributes.add(childElement.getAttribute(ATTR_ID));
			}
			else {
				throw new TeamRepositoryException("Attribute " + ATTR_ID + " is required for " + TAG_REQUIRED_ATTR);
			}
		}
		
		return requiredAttributes;
	}
	
	/**
	 * Loops through the requiredBy attribute and value pairs in the process configuration and checks the work item for that value in the attribute, if it matches, the attribute is included in the list of required by attributes.
	 * 
	 * @param requiredAttributeId
	 * @param workItem
	 * @param condition
	 * @param workItemCommonTasks
	 * @param workItemCommon
	 * @param monitor
	 * @return required by attributes with matching values for the configuration
	 * @throws TeamRepositoryException
	 */
	private List<IAttribute> getRequiredByAttributes(String requiredAttributeId, IWorkItem workItem,
			IProcessConfigurationElement configuration, WorkItemCommonTasks workItemCommonTasks, IWorkItemCommon workItemCommon,
			IProgressMonitor monitor) throws TeamRepositoryException {
		List<IAttribute> attributes = new ArrayList<IAttribute>();
		Map<String, Set<String>> attrsAndValues = getAttributesAndValues(requiredAttributeId, configuration);

		for (String requiredByAttributeId : attrsAndValues.keySet()) {
			IAttribute requiredByAttribute = workItemCommonTasks.getAttribute(requiredByAttributeId, workItem.getProjectArea(),
					workItemCommon, monitor);

			if (workItemCommonTasks.attributeHasNonNullValue(workItem, requiredByAttribute, workItemCommon.getAuditableCommon())) {
				for (String value : attrsAndValues.get(requiredByAttributeId)) {
					if (value.equals(((Identifier<ILiteral>) workItem.getValue(requiredByAttribute)).getStringIdentifier())) {
						attributes.add(requiredByAttribute);
					}
				}
			}
		}
		return attributes;
	}
}
