package com.ibm.stg.rtc.ext.stgdefect.advisors;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;

import org.eclipse.core.runtime.IProgressMonitor;

import com.ibm.stg.rtc.ext.common.ConfigurationHelper;
import com.ibm.stg.rtc.ext.common.WorkItemCommonTasks;
import com.ibm.stg.rtc.ext.stgdefect.common.STGDefectCommonIDs;
import com.ibm.team.process.common.IProcessConfigurationElement;
import com.ibm.team.process.common.IProjectAreaHandle;
import com.ibm.team.process.common.advice.AdvisableOperation;
import com.ibm.team.process.common.advice.IAdvisorInfo;
import com.ibm.team.process.common.advice.IAdvisorInfoCollector;
import com.ibm.team.process.common.advice.runtime.IOperationAdvisor;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.repository.service.AbstractService;
import com.ibm.team.workitem.common.IAuditableCommon;
import com.ibm.team.workitem.common.ISaveParameter;
import com.ibm.team.workitem.common.IWorkItemCommon;
import com.ibm.team.workitem.common.model.IState;
import com.ibm.team.workitem.common.model.IWorkItem;
import com.ibm.team.workitem.common.model.IWorkItemType;
import com.ibm.team.workitem.common.model.Identifier;
import com.ibm.team.workitem.common.workflow.IWorkflowInfo;

/**
 * This precondition makes sure that when a new Change Record is created, it has a
 * parent work item in the correct project, type, and state. The configuration 
 * allows Change Records to be created on the workItemType in the given workItemStates.
 * If no workItemState elements exist under a workItemType element, Change Records 
 * will be allowed for all states.
 * 
 * Example:
 * 
		<precondition id="com.ibm.stg.rtc.ext.advisors.RequireXXXXX" name="STG: Require Change Sets Complete" optional="false">
			<workItemType id="com.ibm.stg.workItemType.stgDefect">
				<workItemState id="com.ibm.stg.workflow.stgDefect.state.working"/>
			</workItemType>
			<workItemType id="task">
				<workItemState id="s2"/>
			</workItemType>
		</precondition>
 * 
 * @author christik
 *
 */
public class RequireCRParentAdvisor extends AbstractService implements IOperationAdvisor {

	public static final String ID = "com.ibm.stg.rtc.ext.stgdefect.advisors.RequireCRParentAdvisor";

	public void run(AdvisableOperation operation, IProcessConfigurationElement advisorConfiguration, IAdvisorInfoCollector collector, IProgressMonitor monitor) throws TeamRepositoryException {
		ISaveParameter saveParameter = null;
		IWorkItem newWorkItemState = null;
		WorkItemCommonTasks workItemCommonTasks = new WorkItemCommonTasks();

		Object tempObj = operation.getOperationData();
		if (workItemCommonTasks.isSaveParamterType(tempObj)) {
			saveParameter = (ISaveParameter) tempObj;
		} else {
			return;
		}

		tempObj = saveParameter.getNewState();
		if (workItemCommonTasks.isIWorkItemType(tempObj)) {
			newWorkItemState = (IWorkItem) tempObj;
		} else {
			return;
		}

		// check to see if the work item is a Change Record, if not return
		if(!(newWorkItemState.getWorkItemType().equals(STGDefectCommonIDs.CHANGE_RECORD_ID))){
			return;
		}

		// if this is not a new Change Record return
		if(saveParameter.getOldState() != null){
			return;
		}

		IWorkItemCommon workItemCommon = this.getService(IWorkItemCommon.class);
		IAuditableCommon auditableCommon = workItemCommon.getAuditableCommon();


		IWorkItem parent = workItemCommonTasks.getParentWorkItem(saveParameter.getNewReferences(), auditableCommon, monitor);

		// if the Change Record doesn't have a parent work item, report a problem
		if(parent == null){
			List<String> types = ConfigurationHelper.getChildrenTagIds(advisorConfiguration, ConfigurationHelper.TAG_WORKITEM_TYPE);
			String typeNames = getTypeNames(types, workItemCommon, newWorkItemState.getProjectArea());
			IAdvisorInfo info= collector.createProblemInfo("No Change Record parent", "Change Records must have a parent work item of type " + typeNames, ID);
			info.setProblemObject(newWorkItemState.getItemHandle());
			collector.addInfo(info);
			return;
		}

		// check to see that the parent is in the same project
		if(parent.getProjectArea().sameItemId(newWorkItemState.getProjectArea()) == false){
			IAdvisorInfo info= collector.createProblemInfo("Invalid Change Record parent project", "Change Records must be linked to a parent work item in the same project area.", ID);
			info.setProblemObject(newWorkItemState.getItemHandle());
			collector.addInfo(info);
		}

		// check to see if change records are allowed for the parent work item type
		List<String> states = ConfigurationHelper.getWIStatesConfigured(advisorConfiguration, parent.getWorkItemType());
		if(states == null){
			List<String> types = ConfigurationHelper.getChildrenTagIds(advisorConfiguration, ConfigurationHelper.TAG_WORKITEM_TYPE);
			String typeNames = getTypeNames(types, workItemCommon, newWorkItemState.getProjectArea());
			IAdvisorInfo info= collector.createProblemInfo("Invalid Change Record parent type", "Change Records must have a parent work item of type " + typeNames, ID);
			info.setProblemObject(newWorkItemState.getItemHandle());
			collector.addInfo(info);
			return;
		}

		// check to see if the parent is in a valid state to create a change record
		String parentStateId = parent.getState2().getStringIdentifier();
		if(states.size() > 0 && !states.contains(parentStateId)){
			String typeName = getTypeNames(Collections.singletonList(parent.getWorkItemType()), workItemCommon, parent.getProjectArea());
			IWorkflowInfo workflowInfo = workItemCommon.findWorkflowInfo(parent, monitor);
			String stateNames = getStateNames(states, workflowInfo);
			IAdvisorInfo info= collector.createProblemInfo("Invalid Change Record parent state", "The parent " + typeName + " must be in the state " + stateNames + " in order to create a Change Record", ID);
			info.setProblemObject(newWorkItemState.getItemHandle());
			collector.addInfo(info);
			return;
		}
	}

	// copied from com.ibm.team.workitem.common.internal.workflow.WorkflowInfo
	private String convertLegacyId(String id, char prefix) {
		// legacy: if id is an integer, convert it to legacy state id s<int>
		try {
			Integer.parseInt(id);
			return prefix + id;
		} catch (NumberFormatException e) {
			// ignore: not legacy format
		}
		return id;
	}

	/**
	 * @param requiredStates
	 * @return
	 */
	private String getStateNames(List<String> stateIds, IWorkflowInfo workFlowInfo) {
		List<String> stateNames = new ArrayList<String>(stateIds.size());
		for(String stateId : stateIds){
			String stateName = workFlowInfo.getStateName(Identifier.create(IState.class, stateId));
			if(stateName == null){
				stateName = stateId;
			}
			stateNames.add(stateName);
		}
		return concatStrings(stateNames);
	}

	private String getTypeNames(List<String> typeIds, IWorkItemCommon workItemCommon, IProjectAreaHandle projectArea) throws TeamRepositoryException{
		List<String> typeNames = new ArrayList<String>(typeIds.size());
		for(String typeId : typeIds){
			String typeName = null;
			IWorkItemType type = workItemCommon.findWorkItemType(projectArea, typeId, null);
			if(type != null){
				typeName = type.getDisplayName();
			}
			if(typeName == null){
				typeName = typeId;
			}
			typeNames.add(typeName);
		}
		return concatStrings(typeNames);
	}

	private String concatStrings(List<String> strings){
		String result = "";
		for(int i = 0; i < strings.size(); i++){
			result = result + strings.get(i);
			// if there are more, add the ',' and 'or'
			if(i < strings.size() - 1){
				if(strings.size() > 2){
					result = result + ", ";
				}else{
					result = result + " ";
				}
				if(i == strings.size() - 2){
					result = result + "or ";
				}
			}
		}
		return result;
	}

}
