package com.ibm.stg.rtc.ext.stgdefect.advisors;

import org.eclipse.core.runtime.IProgressMonitor;

import com.ibm.stg.rtc.ext.common.WorkItemCommonTasks;
import com.ibm.stg.rtc.ext.stgdefect.common.STGDefectCommonIDs;
import com.ibm.stg.rtc.ext.stgdefect.common.STGDefectCommonTasks;
import com.ibm.team.process.common.IProcessConfigurationElement;
import com.ibm.team.process.common.advice.AdvisableOperation;
import com.ibm.team.process.common.advice.IAdvisorInfo;
import com.ibm.team.process.common.advice.IAdvisorInfoCollector;
import com.ibm.team.process.common.advice.runtime.IOperationAdvisor;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.repository.service.AbstractService;
import com.ibm.team.workitem.common.ISaveParameter;
import com.ibm.team.workitem.common.IWorkItemCommon;
import com.ibm.team.workitem.common.model.IResolution;
import com.ibm.team.workitem.common.model.IWorkItem;
import com.ibm.team.workitem.common.model.Identifier;
import com.ibm.team.workitem.common.workflow.IWorkflowInfo;

/**
 * This precondition only runs for STG Defects and makes sure there is a 'Duplicate Of' link whenever
 * the defect's resolution is 'Duplicate'. It also prevents users from adding a duplicate work item
 * when the resolution is not 'Duplicate'.
 * 
 * @author Kyle Christianson
 *
 */
public class RequireDuplicateLinkAdvisor extends AbstractService implements IOperationAdvisor {

	public static final String ID = "com.ibm.stg.rtc.ext.stgdefect.advisors.RequireDuplicateLinkAdvisor";

	public void run(AdvisableOperation operation, IProcessConfigurationElement advisorConfiguration, IAdvisorInfoCollector collector, IProgressMonitor monitor) throws TeamRepositoryException {
		ISaveParameter saveParameter = null;
		IWorkItemCommon workItemCommon = this.getService(IWorkItemCommon.class);
		WorkItemCommonTasks workItemCommonTasks = new WorkItemCommonTasks();
		IWorkItem newWorkItemState = null;

		Object tempObj = operation.getOperationData();
		if (workItemCommonTasks.isSaveParamterType(tempObj)) {
			saveParameter = (ISaveParameter) tempObj;
		} else {
			return;
		}

		tempObj = saveParameter.getNewState();
		if (workItemCommonTasks.isIWorkItemType(tempObj)) {
			newWorkItemState = (IWorkItem) tempObj;
		} else {
			return;
		}

		// check to see if the work item is an STG Defect, if not return
		if(!(newWorkItemState.getWorkItemType().equals(STGDefectCommonIDs.STG_DEFECT_ID))){
			return;
		}

		boolean duplicateResolution = STGDefectCommonTasks.isResolutionDuplicate(newWorkItemState);

		// require a duplicate of link to any work item type if the resolution is duplicate
		if(duplicateResolution == true){

			// leaving the duplicateWIType parameter null means that this will be true if any duplicate work item (of any type) exists 
			boolean duplicateOfWIExists = workItemCommonTasks.hasDuplicateOfLink(saveParameter.getNewReferences(), null, workItemCommon.getAuditableCommon(), monitor);

			if(duplicateOfWIExists == false){
				// there is no duplicate of link to any work item
				IWorkflowInfo wfInfo = workItemCommon.findWorkflowInfo(newWorkItemState, monitor);
				String name = wfInfo.getResolutionName(newWorkItemState.getResolution2());
				IAdvisorInfo info= collector.createProblemInfo("'Duplicate Of' link required", "A 'Duplicate Of' link is required on STG Defect " + newWorkItemState.getId() + " because its resolution is marked '" + name + "'.", ID);
				info.setProblemObject(newWorkItemState.getItemHandle());
				collector.addInfo(info);
				return;
			}
		}

		// prevent adding duplicate of link if the resolution is not duplicate

		if(duplicateResolution == false){

			boolean newDuplicateWIExists = workItemCommonTasks.hasNewDuplicateOfLink(saveParameter.getNewReferences());
			if(newDuplicateWIExists == true){
				// there is no duplicate of link to any work item
				IWorkflowInfo wfInfo = workItemCommon.findWorkflowInfo(newWorkItemState, monitor);
				String name = wfInfo.getResolutionName(Identifier.create(IResolution.class, STGDefectCommonIDs.STG_DEFECT_RESOLUTION_DUPLICATE));
				IAdvisorInfo info= collector.createProblemInfo("Invalid 'Duplicate Of' link", "STG Defect " + newWorkItemState.getId() + " cannot have a 'Duplicate Of' link because its resolution is not '" + name + "'.", ID);
				info.setProblemObject(newWorkItemState.getItemHandle());
				collector.addInfo(info);
				return;
			}
		}
	}



}
