package com.ibm.stg.rtc.ext.stgdefect.participants;

import java.util.List;

import org.eclipse.core.runtime.IProgressMonitor;

import com.ibm.stg.rtc.ext.common.RepositoryCommonTasks;
import com.ibm.stg.rtc.ext.common.WorkItemCommonTasks;
import com.ibm.stg.rtc.ext.stgdefect.common.STGDefectCommonIDs;
import com.ibm.team.links.common.IReference;
import com.ibm.team.links.common.registry.IEndPointDescriptor;
import com.ibm.team.process.common.IProcessConfigurationElement;
import com.ibm.team.process.common.advice.AdvisableOperation;
import com.ibm.team.process.common.advice.IProcessReport;
import com.ibm.team.process.common.advice.IReportInfo;
import com.ibm.team.process.common.advice.TeamOperationCanceledException;
import com.ibm.team.process.common.advice.runtime.IOperationParticipant;
import com.ibm.team.process.common.advice.runtime.IParticipantInfoCollector;
import com.ibm.team.repository.common.IContributorHandle;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.repository.service.AbstractService;
import com.ibm.team.workitem.common.IAuditableCommon;
import com.ibm.team.workitem.common.ISaveParameter;
import com.ibm.team.workitem.common.model.IWorkItem;
import com.ibm.team.workitem.common.model.IWorkItemReferences;
import com.ibm.team.workitem.common.model.WorkItemEndPoints;
import com.ibm.team.workitem.service.IWorkItemServer;

public class SubscribeOwnerToParentOrChildrenParticipant extends AbstractService implements IOperationParticipant {

	public static final String ID = "com.ibm.stg.stgdefect.participants.SubscribeOwnerToParentOrChildrenParticipant";
	public static final String NAME = "STG: Subscribe parent owner to child work items";

	public void run(AdvisableOperation operation, IProcessConfigurationElement participantConfig, IParticipantInfoCollector collector, IProgressMonitor monitor) throws TeamRepositoryException {
		IWorkItemServer workItemServer = this.getService(IWorkItemServer.class);
		IAuditableCommon auditableCommon = this.getService(IAuditableCommon.class);
		WorkItemCommonTasks workItemCommonTasks = new WorkItemCommonTasks();
		ISaveParameter saveParameter = null;
		IWorkItem newWorkItemState = null;
		IWorkItem oldWorkItemState = null;

		Object tempObj = operation.getOperationData();
		if (workItemCommonTasks.isSaveParamterType(tempObj)) {
			saveParameter = (ISaveParameter) tempObj;
		} else {
			return;
		}

		tempObj = saveParameter.getNewState();
		if (workItemCommonTasks.isIWorkItemType(tempObj)) {
			newWorkItemState = (IWorkItem) tempObj;
		} else {
			return;
		}

		tempObj = saveParameter.getOldState();
		if (workItemCommonTasks.isIWorkItemType(tempObj)) {
			oldWorkItemState = (IWorkItem) tempObj;
		} else if(tempObj == null){
			oldWorkItemState = null;
		} else{
			return;
		}

		// this follow-up action only applied to STG Defect and Change Record work items
		if(!(newWorkItemState.getWorkItemType().equals(STGDefectCommonIDs.STG_DEFECT_ID) || newWorkItemState.getWorkItemType().equals(STGDefectCommonIDs.CHANGE_RECORD_ID))){
			return;
		}


		IContributorHandle newStateOwner = newWorkItemState.getOwner();
		IContributorHandle oldStateOwner = (oldWorkItemState == null) ? null : oldWorkItemState.getOwner();

		// if the owner has changed, subscribe the new owner to the children and parent
		if(!newStateOwner.sameItemId(oldStateOwner)){

			IWorkItem parent = workItemCommonTasks.getParentWorkItem(saveParameter.getNewReferences(), auditableCommon, monitor);
			List<IWorkItem> children = workItemCommonTasks.getChildWorkItems(STGDefectCommonIDs.CHANGE_RECORD_ID, saveParameter.getNewReferences(), auditableCommon, monitor);

			if(parent != null){
				try{
					workItemCommonTasks.subscribeContributor(parent.getId(), newStateOwner, workItemServer, monitor);
				}catch(TeamOperationCanceledException e){
					RepositoryCommonTasks repoCommon = new RepositoryCommonTasks();
					IReportInfo info = collector.createInfo("Follow-up action failed", "Could not subscribe owner to parent work item " + parent.getId() + " because a precondition failed with the following message: \n" + repoCommon.getFirstErrorDescription(e));
					info.setProblemObject(parent.getItemHandle());
					info.setSeverity(IProcessReport.WARNING);
					collector.addInfo(info);
					// suppress the follow-up save's errors so this save can continue
					repoCommon.suppressAllErrors(NAME, e);
				}
			}

			// Flip through the child work items and subscribe the owner to them
			for(IWorkItem child : children){
				try{
					workItemCommonTasks.subscribeContributor(child.getId(), newStateOwner, workItemServer, monitor);
				}catch(TeamOperationCanceledException e){
					RepositoryCommonTasks repoCommon = new RepositoryCommonTasks();
					IReportInfo info = collector.createInfo("Follow-up action failed", "Could not subscribe owner to child work item " + child.getId() + " because a precondition failed with the following message: \n" + repoCommon.getFirstErrorDescription(e));
					info.setProblemObject(child.getItemHandle());
					info.setSeverity(IProcessReport.WARNING);
					collector.addInfo(info);
					// suppress the follow-up save's errors so this save can continue
					repoCommon.suppressAllErrors(NAME, e);
				}
			}
		} 

		//check for changed parent and child links
		IWorkItemReferences workItemReferences = saveParameter.getNewReferences();
		List<IEndPointDescriptor> changedTypes = workItemReferences.getChangedReferenceTypes();

		// if the parent has changed, subscribe the parent owner to this work item and this work item's owner to the parent
		if(changedTypes.contains(WorkItemEndPoints.PARENT_WORK_ITEM)){
			IWorkItem newParent = workItemCommonTasks.getParentWorkItem(workItemReferences, auditableCommon, monitor);
			if(newParent != null && newParent.getWorkItemType().equals(STGDefectCommonIDs.STG_DEFECT_ID)){
				try{
					workItemCommonTasks.subscribeContributor(newWorkItemState.getId(), newParent.getOwner(), workItemServer, monitor);
				}catch(TeamOperationCanceledException e){
					RepositoryCommonTasks repoCommon = new RepositoryCommonTasks();
					IReportInfo info = collector.createInfo("Follow-up action failed", "Could not subscribe parent owner to work item " + newWorkItemState.getId() + " because a precondition failed with the following message: \n" + repoCommon.getFirstErrorDescription(e));
					info.setProblemObject(newWorkItemState.getItemHandle());
					info.setSeverity(IProcessReport.WARNING);
					collector.addInfo(info);
					// suppress the follow-up save's errors so this save can continue
					repoCommon.suppressAllErrors(NAME, e);
				}
				try{
					workItemCommonTasks.subscribeContributor(newParent.getId(), newStateOwner, workItemServer, monitor);
				}catch(TeamOperationCanceledException e){
					RepositoryCommonTasks repoCommon = new RepositoryCommonTasks();
					IReportInfo info = collector.createInfo("Follow-up action failed", "Could not subscribe owner to parent work item " + newParent.getId() + " because a precondition failed with the following message: \n" + repoCommon.getFirstErrorDescription(e));
					info.setProblemObject(newParent.getItemHandle());
					info.setSeverity(IProcessReport.WARNING);
					collector.addInfo(info);
					// suppress the follow-up save's errors so this save can continue
					repoCommon.suppressAllErrors(NAME, e);
				}
			}
		}

		// if children have changed, subscribe this owner to any new children and the children's owner to this work item
		if(changedTypes.contains(WorkItemEndPoints.CHILD_WORK_ITEMS)){

			List<IReference> newChildrenRef = workItemReferences.getCreatedReferences(WorkItemEndPoints.CHILD_WORK_ITEMS);
			List<IWorkItem> newChildrenWI = workItemCommonTasks.getWorkItemsFromReferences(STGDefectCommonIDs.CHANGE_RECORD_ID, newChildrenRef, auditableCommon, monitor);
			for(IWorkItem child : newChildrenWI){
				try{
					workItemCommonTasks.subscribeContributor(child.getId(), newStateOwner, workItemServer, monitor);
				}catch(TeamOperationCanceledException e){
					RepositoryCommonTasks repoCommon = new RepositoryCommonTasks();
					IReportInfo info = collector.createInfo("Follow-up action failed", "Could not subscribe owner to child work item " + child.getId() + " because a precondition failed with the following message: \n" + repoCommon.getFirstErrorDescription(e));
					info.setProblemObject(child.getItemHandle());
					info.setSeverity(IProcessReport.WARNING);
					collector.addInfo(info);
					// suppress the follow-up save's errors so this save can continue
					repoCommon.suppressAllErrors(NAME, e);
				}
				try{
					workItemCommonTasks.subscribeContributor(newWorkItemState.getId(), child.getOwner(), workItemServer, monitor);
				}catch(TeamOperationCanceledException e){
					RepositoryCommonTasks repoCommon = new RepositoryCommonTasks();
					IReportInfo info = collector.createInfo("Follow-up action failed", "Could not subscribe child owner to work item " + newWorkItemState.getId() + " because a precondition failed with the following message: \n" + repoCommon.getFirstErrorDescription(e));
					info.setProblemObject(newWorkItemState.getItemHandle());
					info.setSeverity(IProcessReport.WARNING);
					collector.addInfo(info);
					// suppress the follow-up save's errors so this save can continue
					repoCommon.suppressAllErrors(NAME, e);
				}
			}
		}
	}
}
