package com.ibm.stg.rtc.ext.advisors;

import java.sql.Timestamp;
import java.util.Date;
import java.util.GregorianCalendar;
import java.util.HashMap;
import java.util.Map;
import java.util.Set;

import org.eclipse.core.runtime.IProgressMonitor;

import com.ibm.stg.rtc.ext.common.ConfigurationHelper;
import com.ibm.stg.rtc.ext.common.WorkItemCommonTasks;
import com.ibm.stg.rtc.ext.common.WorkItemCondition;
import com.ibm.team.process.common.IProcessConfigurationElement;
import com.ibm.team.process.common.advice.AdvisableOperation;
import com.ibm.team.process.common.advice.IAdvisorInfoCollector;
import com.ibm.team.process.common.advice.runtime.IOperationAdvisor;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.repository.service.AbstractService;
import com.ibm.team.workitem.common.IAuditableCommon;
import com.ibm.team.workitem.common.ISaveParameter;
import com.ibm.team.workitem.common.IWorkItemCommon;
import com.ibm.team.workitem.common.model.IAttribute;
import com.ibm.team.workitem.common.model.IWorkItem;

/**
 * This precondition defaults a date field value a specified number of days from now. By default it will only set the date if it doesn't
 * already have a value, but there is an option to force the date field to be set.
 * 
 * To configure a date to be defaulted the same for all work item types that have that attribute, it can be at the top level.
 * 
 * A date can also be configured for certain conditions. If a work item meets the condition, the precondition uses that default date
 * configuration if the same date is configured at both the top level and in a condition.
 * 
 * Example 1:
 * <precondition id="com.ibm.stg.rtc.ext.advisors.DefaultDateFieldsAdvisor" name="STG: Default Dates" optional="false">
	 <autoDate id="com.ibm.stg.attribute.requiredDate" value="450" force="true"/>
	 <autoDate id="com.ibm.team.workitem.attribute.duedate" value="1" />
   </precondition>

	Example 2:
   <precondition id="com.ibm.stg.rtc.ext.advisors.DefaultDateFieldAdvisor" name="STG: Default Dates" optional="false">
      <autoDate id="com.ibm.team.workitem.attribute.duedate" value="1" />
      <autoDate id="com.ibm.stg.attribute.requiredDate" value="100" force="true" />
	  <workItemType id="com.ibm.stg.workItemType.stgDefect">
		  <condition workItemState="com.ibm.stg.workflow.stgDefect.state.open" >
		  	<autoDate id="com.ibm.stg.attribute.duedate" value="7" />
		    <autoDate id="com.ibm.stg.attribute.requiredDate" value="1" force="true" />
		  </condition>
		  <condition workItemState="com.ibm.stg.workflow.stgDefect.state.working" >
		     <autoDate id="com.ibm.stg.attribute.duedate" value="12" />
		  </condition>
	  </workItemType>
	  <workItemType id="task">
	  	<condition>
	  		<autoDate id="com.ibm.team.workitem.attribute.duedate" value="10" />
	  	</condition>
	  </workItemType>
   </precondition>
 * 
 * @author Karen Paone
 * 
 */
public class DefaultDateFieldsAdvisor extends AbstractService implements IOperationAdvisor {

	public static final String ID = "com.ibm.stg.rtc.ext.advisors.DefaultDateFieldsAdvisor";

	private static final String TAG_AUTO_DATE = "autoDate";
	private static final String ATTR_ID = "id";
	private static final String ATTR_VALUE = "value";
	private static final String ATTR_FORCE = "force";
	private static final String VALUE_FALSE = "false";
	private static final String VALUE_TRUE = "true";

	private Map<String, String> dateProperties = null;
	private Map<String, String> forceDateProperties = null;
	private WorkItemCommonTasks workItemCommonTasks = null;

	public void run(AdvisableOperation operation, IProcessConfigurationElement advisorConfiguration, IAdvisorInfoCollector collector,
			IProgressMonitor monitor) throws TeamRepositoryException {
		workItemCommonTasks = new WorkItemCommonTasks();
		ISaveParameter saveParameter = null;
		IWorkItem newWorkItemState = null;

		Object tempObj = operation.getOperationData();
		if (workItemCommonTasks.isSaveParamterType(tempObj)) {
			saveParameter = (ISaveParameter) tempObj;
		}
		else {
			return;
		}

		tempObj = saveParameter.getNewState();
		if (workItemCommonTasks.isIWorkItemType(tempObj)) {
			newWorkItemState = (IWorkItem) tempObj;
		}
		else {
			return;
		}

		IWorkItemCommon workItemCommon = this.getService(IWorkItemCommon.class);

		dateProperties = new HashMap<String, String>();
		forceDateProperties = new HashMap<String, String>();

		// these are date fields that should be defaulted for all work item types and conditions if configured at this level
		for (IProcessConfigurationElement element : advisorConfiguration.getChildren()) {
			if (element.getName().equals(TAG_AUTO_DATE)) {
				addDateProperty(element);
			}
		}

		// determine what is configured for the work item type
		IProcessConfigurationElement typeConfiguration = ConfigurationHelper.getWITypeConfiguration(advisorConfiguration,
				newWorkItemState.getWorkItemType());

		// if there are work item types configured but not this work item, no action
		// if there are no work item types configured, then it is OK
		if (!ConfigurationHelper.getWITypesConfigured(advisorConfiguration).isEmpty() && typeConfiguration == null) {
			return;
		}
		// this type is not configured and there are not date properties configured for all types, no action
		if (typeConfiguration == null && dateProperties.isEmpty() && forceDateProperties.isEmpty()) {
			return;
		}

		boolean conditionMet = false;
		if (typeConfiguration == null) {
			// if there are no types configured, then the dates can be defaulted
			conditionMet = true;
		}
		else {
			Set<WorkItemCondition> conditions = ConfigurationHelper.getAllConditions(typeConfiguration);
			if (conditions.isEmpty()) {
				// if there are no conditions for this type, then the dates can be defaulted for this work item for any condition
				conditionMet = true;
			}

			// if there are conditions, then the dates should only be defaulted specifically during these conditions
			for (WorkItemCondition condition : conditions) {
				if (condition.matches(saveParameter)) {
					conditionMet = true;
					// add more date properties that are specific to the work item condition
					for (IProcessConfigurationElement element : condition.getConditionElement().getChildren()) {
						addDateProperty(element);
					}
				}
			}
		}

		if (!conditionMet) {
			return;
		}

		Date currentTime = new Date();

		// iterate through the forceDateProperties map, defaults each date included in the map
		for (String dateAttributeId : forceDateProperties.keySet()) {
			IAttribute dateAttribute = workItemCommonTasks.getAttribute(dateAttributeId, newWorkItemState.getProjectArea(), workItemCommon,
					monitor);
			setDateField(newWorkItemState, dateAttribute, forceDateProperties.get(dateAttributeId), true, currentTime,
					workItemCommon.getAuditableCommon());
		}
		// iterate through the dateProperties map, for each dates in the properties, see if they have a value on the work item, if not,
		// default it
		for (String dateAttributeId : dateProperties.keySet()) {
			IAttribute dateAttribute = workItemCommonTasks.getAttribute(dateAttributeId, newWorkItemState.getProjectArea(), workItemCommon,
					monitor);
			setDateField(newWorkItemState, dateAttribute, dateProperties.get(dateAttributeId), false, currentTime,
					workItemCommon.getAuditableCommon());
		}

	}

	// if the work item has the attribute and it is either forced or the attribute is not set, then set the date the specified number of
	// days from now
	private void setDateField(IWorkItem workItem, IAttribute dateAttribute, String value, boolean forced, Date date,
			IAuditableCommon auditableCommon) throws NumberFormatException, TeamRepositoryException {
		if (dateAttribute != null && workItem.hasAttribute(dateAttribute)
				&& (forced || !workItemCommonTasks.attributeHasNonNullValue(workItem, dateAttribute, auditableCommon))) {
			// this date field should be defaulted
			// the value is the number of days from now the date should be defaulted
			GregorianCalendar cal = new GregorianCalendar();
			// NOTE: this is sufficient for days spanning next year
			cal.setTime(date);
			cal.add(GregorianCalendar.DAY_OF_YEAR, Integer.valueOf(value).intValue());
			Timestamp newDate = new Timestamp(cal.getTimeInMillis());
			workItem.setValue(dateAttribute, newDate);
		}
	}

	// adds the date property to the appropriate date property map; distinguished by if it is forced
	private void addDateProperty(IProcessConfigurationElement element) {
		String field = element.getAttribute(ATTR_ID);
		String value = element.getAttribute(ATTR_VALUE);
		String force = element.getAttribute(ATTR_FORCE);
		if (force == null || force.equals(VALUE_FALSE)) {
			dateProperties.put(field, value);
		}
		else if (force.equals(VALUE_TRUE)) {
			forceDateProperties.put(field, value);
		}
	}
}
