package com.ibm.stg.rtc.ext.advisors;

import java.util.ArrayList;
import java.util.Collections;
import java.util.Iterator;
import java.util.List;

import org.eclipse.core.runtime.IProgressMonitor;

import com.ibm.stg.rtc.ext.common.ConfigurationHelper;
import com.ibm.stg.rtc.ext.common.WorkItemCommonTasks;
import com.ibm.team.process.common.IProcessConfigurationElement;
import com.ibm.team.process.common.advice.AdvisableOperation;
import com.ibm.team.process.common.advice.IAdvisorInfoCollector;
import com.ibm.team.process.common.advice.IProcessReport;
import com.ibm.team.process.common.advice.IReportInfo;
import com.ibm.team.process.common.advice.runtime.IOperationAdvisor;
import com.ibm.team.repository.common.IContributorHandle;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.repository.common.service.IContributorService;
import com.ibm.team.repository.service.AbstractService;
import com.ibm.team.workitem.common.IAuditableCommon;
import com.ibm.team.workitem.common.ISaveParameter;
import com.ibm.team.workitem.common.IWorkItemCommon;
import com.ibm.team.workitem.common.model.IWorkItem;

public class DefaultOwnerAdvisor extends AbstractService implements IOperationAdvisor {

	/* This precondition lets teams define a default owner of their work items. It only performs changes on work items
	 * when the Owned By is left Unassigned.
	 * 
	 * Note that preconditions usually should not modify a work item, however this is a special 
	 * case because the default owner is set on create and we need to avoid sending
	 * multiple notifications and performing nested saving.
	 * 
	 * The default owner is looked up in the configuration data xml
	 * (see example below). Options for the default owner include category owner, team owner, parent owner, creator,
	 * and current user. The order in which the default owner is searched for is specified by the "order" attribute
	 * or the order the come in the xml if "order" attribute is not specified.
	 * Additional options include allowUnassigned which prevents the save from occurring if a default owner could not
	 * be found. CreationOnly which checks sets the owner only for newly created work items. And showWarnings which 
	 * can prevent warnings from being displayed if.
	 * 
		<configuration-data id="com.ibm.stg.workitem.configuration.itemProperties">
			<team repositoryURI="" uuid="" Owner=""/>
		    <team repositoryURI="" uuid="" Owner=""/>
		    <category repositoryURI="" uuid="" Owner=""/>
		    <category repositoryURI="" uuid="" Owner=""/>
		    <category repositoryURI="" uuid="" Owner=""/>
		</configuration-data>

		<followup-action>
			<workItemType id="com.ibm.stg.workItemType.stgDefect" allowUnassigned="false" creationOnly="false" showWarnings="true">
				<categoryOwner order="1"/>
				<teamOwner order="2"/>
			    <parentOwner order="5"/>
				<creator order="3"/>
				<currentUser order="4"/>
			</workItemType>
		</followup-action>
	 */

	public static final String ID = "com.ibm.stg.rtc.ext.advisors.DefaultOwnerAdvisor";

	private static final String TAG_TEAM_OWNER = "teamOwner";
	private static final String TAG_CATEGORY_OWNER = "categoryOwner";
	private static final String TAG_PARENT_OWNER = "parentOwner";
	private static final String TAG_CREATOR = "creator";
	private static final String TAG_CURRENT_USER = "currentUser";
	private static final String ATTR_CREATION_ONLY = "creationOnly";
	private static final String ATTR_ALLOW_UNASSIGNED = "allowUnassigned";
	private static final String ATTR_SHOW_WARNINGS = "showWarnings";
	private static final String ATTR_ORDER = "order";

	public void run(AdvisableOperation operation, IProcessConfigurationElement participantConfig, IAdvisorInfoCollector collector, IProgressMonitor monitor) throws TeamRepositoryException {
		WorkItemCommonTasks workItemCommonTasks = new WorkItemCommonTasks();
		ISaveParameter saveParameter = null;
		IWorkItem newWorkItemState;

		Object tempObj = operation.getOperationData();
		if (workItemCommonTasks.isSaveParamterType(tempObj)) {
			saveParameter = (ISaveParameter) tempObj;
		} else {
			return;
		}

		tempObj = saveParameter.getNewState();
		if (workItemCommonTasks.isIWorkItemType(tempObj)) {
			newWorkItemState = (IWorkItem) tempObj;
		} else {
			return;
		}


		IWorkItemCommon workItemCommon = this.getService(IWorkItemCommon.class);
		IAuditableCommon auditableCommon = workItemCommon.getAuditableCommon();

		if(workItemCommonTasks.isUnassignedContributor(newWorkItemState.getOwner(), auditableCommon, monitor)){
			// the owner was left unassigned so we need to assign the default owner


			// see if there is a configuration for this work item type
			IProcessConfigurationElement config = ConfigurationHelper.getWITypeConfiguration(participantConfig, newWorkItemState.getWorkItemType());
			if(config == null){
				return; // no configuration for this work item type
			}

			boolean creationOnly = ConfigurationHelper.getBooleanAttributeValue(config, ATTR_CREATION_ONLY);
			boolean allowUnassigned = ConfigurationHelper.getBooleanAttributeValue(config, ATTR_ALLOW_UNASSIGNED);
			boolean showWarnings = ConfigurationHelper.getBooleanAttributeValue(config, ATTR_SHOW_WARNINGS);
			if(creationOnly == true && saveParameter.isCreation() == false){
				return; // default owner only configured for work item creation
			}

			IContributorService contributorService = this.getService(IContributorService.class);

			List<OwnerLookup> ownerLookups = getDefaultOwnerLookupOrder(config);
			IContributorHandle defaultOwner = null;
			Iterator<OwnerLookup> i = ownerLookups.iterator();
			while(defaultOwner == null && i.hasNext()){
				OwnerLookup ownerLookup = i.next();
				if(TAG_TEAM_OWNER.equals(ownerLookup.getId())){
					String userId = ConfigurationHelper.getTeamOwner(workItemCommon.findProcessArea(newWorkItemState, monitor), auditableCommon);
					if(userId != null){
						defaultOwner = contributorService.fetchContributorByUserId(userId);
						if(defaultOwner == null && showWarnings){
							IReportInfo info = collector.createProblemInfo("Follow-up action contains invalid user id", "The user id '" + userId + "' specified in " + ID + " xml configuration could not be found in the repository.", ID);
							info.setSeverity(IProcessReport.WARNING);
							collector.addInfo(info);
							break; // stop searching because this is supposed to be the owner but there is a configuration error
						}
					}
				}else if(TAG_CATEGORY_OWNER.equals(ownerLookup.getId())){
					String userId = ConfigurationHelper.getCategoryOwner(newWorkItemState.getCategory(), workItemCommon);
					if(userId != null){
						defaultOwner = contributorService.fetchContributorByUserId(userId);
						if(defaultOwner == null && showWarnings){
							IReportInfo info = collector.createProblemInfo("Follow-up action contains invalid user id", "The user id '" + userId + "' specified in " + ID + " xml configuration could not be found in the repository.", ID);
							info.setSeverity(IProcessReport.WARNING);
							collector.addInfo(info);
							break; // stop searching because this is supposed to be the owner but there is a configuration error
						}
					}
				}else if(TAG_PARENT_OWNER.equals(ownerLookup.getId())){
					IWorkItem parent = workItemCommonTasks.getParentWorkItem(saveParameter.getNewReferences(), auditableCommon, monitor);
					defaultOwner = parent == null ? null : parent.getOwner();
				}else if(TAG_CURRENT_USER.equals(ownerLookup.getId())){
					defaultOwner = this.getAuthenticatedContributor();
				}else if(TAG_CREATOR.equals(ownerLookup.getId())){
					defaultOwner = newWorkItemState.getCreator();
				}
			}

			if(defaultOwner != null && !workItemCommonTasks.isUnassignedContributor(defaultOwner, auditableCommon, monitor)){
				// set the new owner!
				newWorkItemState.setOwner(defaultOwner);
			}else if(allowUnassigned == false){
				IReportInfo info = collector.createProblemInfo("Attribute 'Owned By' not set", "The 'Owned By' attribute needs to be set but a configured default owner was not found.", ID);
				info.setSeverity(IProcessReport.ERROR);
				info.setProblemObject(newWorkItemState.getItemHandle());
				collector.addInfo(info);
			}else if(showWarnings){
				IReportInfo info = collector.createProblemInfo("No default owner configured", "No configured default owner was found, therefore this work item will remain Owned By Unassigned.", ID);
				info.setSeverity(IProcessReport.WARNING);
				collector.addInfo(info);
			}
		} 
	}

	/**
	 * @param config
	 * @return
	 */
	private List<OwnerLookup> getDefaultOwnerLookupOrder(IProcessConfigurationElement config) {
		List<OwnerLookup> lookupOrder = new ArrayList<OwnerLookup>();
		IProcessConfigurationElement[] children = config.getChildren();
		for(int i = 0; i < children.length; i++){
			lookupOrder.add(new OwnerLookup(children[i], i));
		}
		Collections.sort(lookupOrder);
		return lookupOrder;
	}

	private class OwnerLookup implements Comparable<OwnerLookup>{
		private String ownerLookupId;
		private Integer attrOrder;
		private Integer xmlOrder;

		OwnerLookup(IProcessConfigurationElement element, int xmlOrder){
			this.xmlOrder = xmlOrder;
			ownerLookupId = element.getName();
			String orderStr = element.getAttribute(ATTR_ORDER);
			if(orderStr != null){
				try{
					attrOrder = Integer.parseInt(orderStr);
				}catch(Exception e){
					attrOrder = null;
				}
			}
		}

		String getId(){
			return this.ownerLookupId;
		}

		/**
		 * @see java.lang.Comparable#compareTo(java.lang.Object)
		 */
		public int compareTo(OwnerLookup o) {
			if(this.attrOrder == null){
				return 1;
			}else if(o.attrOrder == null){
				return -1;
			}else if(this.attrOrder == o.attrOrder){
				return this.xmlOrder.compareTo(o.xmlOrder);
			}else{
				return this.attrOrder.compareTo(o.attrOrder);
			}
		}

		/**
		 * @see java.lang.Object#toString()
		 */
		@Override
		public String toString() {
			return this.ownerLookupId;
		}

	}
}
