package com.ibm.stg.rtc.ext.advisors;

import java.sql.Timestamp;

import org.eclipse.core.runtime.IProgressMonitor;

import com.ibm.stg.rtc.ext.common.ConfigurationHelper;
import com.ibm.stg.rtc.ext.common.WorkItemCommonTasks;
import com.ibm.team.foundation.common.text.XMLString;
import com.ibm.team.process.common.IIteration;
import com.ibm.team.process.common.IIterationHandle;
import com.ibm.team.process.common.IProcessConfigurationElement;
import com.ibm.team.process.common.IProjectArea;
import com.ibm.team.process.common.advice.AdvisableOperation;
import com.ibm.team.process.common.advice.IAdvisorInfoCollector;
import com.ibm.team.process.common.advice.IProcessReport;
import com.ibm.team.process.common.advice.IReportInfo;
import com.ibm.team.process.common.advice.runtime.IOperationAdvisor;
import com.ibm.team.repository.common.IContributor;
import com.ibm.team.repository.common.IContributorHandle;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.repository.service.AbstractService;
import com.ibm.team.workitem.common.IAuditableCommon;
import com.ibm.team.workitem.common.ISaveParameter;
import com.ibm.team.workitem.common.IWorkItemCommon;
import com.ibm.team.workitem.common.model.IAttribute;
import com.ibm.team.workitem.common.model.ICategory;
import com.ibm.team.workitem.common.model.ICategoryHandle;
import com.ibm.team.workitem.common.model.IDeliverable;
import com.ibm.team.workitem.common.model.IDeliverableHandle;
import com.ibm.team.workitem.common.model.IEnumeration;
import com.ibm.team.workitem.common.model.ILiteral;
import com.ibm.team.workitem.common.model.IWorkItem;
import com.ibm.team.workitem.common.model.Identifier;
import com.ibm.team.workitem.common.model.ItemProfile;

public class DefaultSummaryAdvisor extends AbstractService implements IOperationAdvisor {

	/* This precondition lets teams define a default summary of their work items. It only performs changes on work items
	 * when the Summary is left blank. This precondition will only fail if a default summary template cannot be found 
	 * for a configured work item type. If a template is specified, this advisor will always put the template in
	 * the work item even if it has problems resolving some of the attribute values.
	 * 
	 * Note that preconditions usually should not modify a work item, however this is a special 
	 * case because the default summary must be set before the requireProperties precondition 
	 * fails, and because the default summary is set on create and we need to avoid sending
	 * multiple notifications and performing nested saving.
	 * 
	 * The default summary template is looked up in the configuration data xml
	 * (see example below). There can be two templates, summaryTemplate is the 
	 * default template and parentExistsSummaryTemplate is the template that will be used
	 * if a parent exists. Note that the parentExistsSummaryTemplate is optional and if 
	 * not specified, the summaryTemplate will be used. 
	 * 
	 * The template may include attribute identifiers delimited by the $ character. Following
	 * the $ character, you must specify a valid attribute identifier no spaces allowed. The
	 * attribute identifier must be followed by a blank space or the end of the string. To 
	 * look up the value of a parent work item, use 'parent:' after the $ and before the
	 * attribute, again with no spaces. See examples below.
	 * 
	 * You can set showWarnings to false in order to suppress 
	 * warning messages if something goes wrong. 
	 * 
	 * 
		<precondition id="com.ibm.stg.rtc.ext.advisors.DefaultSummaryAdvisor" name="STG: Default Summary" optional="false">
			<workItemType id="com.ibm.stg.workItemType.changeRecord" showWarnings="true">
				<summaryTemplate value="ChangeRecord: $com.ibm.stg.attribute.internalRelease2 / $com.ibm.stg.attribute.stream : $parent:summary"/>
			</workItemType>
			<workItemType id="task" showWarnings="true">
				<summaryTemplate value="This is the default summary of a task"/>
				<parentExistsSummaryTemplate value="This is the default summary of a task with parent $parent:id"/>
			</workItemType>
		</precondition>

	 */

	public static final String ID = "com.ibm.stg.rtc.ext.advisors.DefaultSummaryAdvisor";

	private static final String ATTR_SHOW_WARNINGS = "showWarnings";

	private static final String TAG_SUMMARY_TEMPLATE = "summaryTemplate";
	private static final String TAG_PARENT_EXISTS_TEMPLATE = "parentExistsSummaryTemplate";

	private static final char ATTRIBUTE_DELIMITER = '$';
	private static final String PARENT_DELIMITER = "parent:";

	boolean showWarnings;
	IAdvisorInfoCollector collector;

	public void run(AdvisableOperation operation, IProcessConfigurationElement participantConfig, IAdvisorInfoCollector collector, IProgressMonitor monitor) throws TeamRepositoryException {
		WorkItemCommonTasks workItemCommonTasks = new WorkItemCommonTasks();
		ISaveParameter saveParameter = null;
		IWorkItem newWorkItemState;

		Object tempObj = operation.getOperationData();
		if (workItemCommonTasks.isSaveParamterType(tempObj)) {
			saveParameter = (ISaveParameter) tempObj;
		} else {
			return;
		}

		tempObj = saveParameter.getNewState();
		if (workItemCommonTasks.isIWorkItemType(tempObj)) {
			newWorkItemState = (IWorkItem) tempObj;
		} else {
			return;
		}

		this.collector = collector;

		IWorkItemCommon workItemCommon = this.getService(IWorkItemCommon.class);
		IAuditableCommon auditableCommon = workItemCommon.getAuditableCommon();

		XMLString currentSummary = newWorkItemState.getHTMLSummary();
		if(currentSummary == null || currentSummary.getXMLText() == null || currentSummary.getXMLText().length() == 0){
			// the summary was left blank so we need to assign the default summary


			// see if there is a configuration for this work item type
			IProcessConfigurationElement config = ConfigurationHelper.getWITypeConfiguration(participantConfig, newWorkItemState.getWorkItemType());
			if(config == null){
				return; // no configuration for this work item type
			}

			this.showWarnings = ConfigurationHelper.getBooleanAttributeValue(config, ATTR_SHOW_WARNINGS);

			IWorkItem parent = workItemCommonTasks.getParentWorkItem(saveParameter.getNewReferences(), auditableCommon, monitor);

			String summaryTemplate = getSummaryTemplate(config, parent != null);


			if(summaryTemplate == null || summaryTemplate.length() == 0){
				IReportInfo info = collector.createProblemInfo("Default Summary Configuration Problem", "A default summary was configured for this work item type, but the summary template could not be found. Please check the configuration xml for this advisor.", ID);
				info.setProblemObject(newWorkItemState.getItemHandle());
				collector.addInfo(info);
				return;
			}

			String defaultSummary = getDefaultSummary(summaryTemplate, newWorkItemState, parent, workItemCommon, monitor);


			// STG Defect 16260: use createFromXMLText instead of createFromPlainText
			newWorkItemState.setHTMLSummary(XMLString.createFromXMLText(defaultSummary));
		} 
	}

	/**	
	 * Turns the summary template into the default summary by replacing delimited attributes with real values
	 * @param summaryTemplate
	 * @param workItem
	 * @param parentWorkItem
	 * @param workItemCommon
	 * @param monitor
	 * @return
	 * @throws TeamRepositoryException
	 */
	private String getDefaultSummary(String summaryTemplate, IWorkItem workItem, IWorkItem parentWorkItem, IWorkItemCommon workItemCommon, IProgressMonitor monitor) throws TeamRepositoryException {
		if(summaryTemplate == null){
			return null;
		}

		StringBuilder summary = new StringBuilder();

		// curIndex is a pointer to the current position in the template
		int curIndex = 0;
		// nextIndex is the next position in the template which we need to process something
		int nextIndex;
		// attributeFound keeps track of whether or not we need to process an attribute
		boolean attributeFound;
		// loop through the entire template
		while(curIndex < summaryTemplate.length()){
			// the next thing to process is an attribute delimiter
			nextIndex = summaryTemplate.indexOf(ATTRIBUTE_DELIMITER, curIndex);
			if(nextIndex == -1){
				// the attribute delimiter was not found
				attributeFound = false;
				// instead of the next index being -1, set it to the end of the template
				nextIndex = summaryTemplate.length();
			}else{
				attributeFound = true;
			}

			// copy the template as-is until the next process point
			summary.append(summaryTemplate.substring(curIndex, nextIndex));
			curIndex = nextIndex;

			if(attributeFound){
				// processing an attribute, look for the end of the attribute identifier by a ' '
				nextIndex = summaryTemplate.indexOf(' ', curIndex);
				if(nextIndex == -1){
					// a space wasn't found so the end of the attribute identifier is the end of the template
					nextIndex = summaryTemplate.length();
				}

				// extract out the attribute identifier. use curIndex + 1 so we don't include the ATTRIBUTE_DELIMITER char
				String attributeId = summaryTemplate.substring(curIndex + 1, nextIndex);
				curIndex = nextIndex;

				// lookup the actual value of the identifier
				String attributeValue = getAttributeValue(attributeId, workItem, parentWorkItem, workItemCommon, monitor);
				// add the value 
				summary.append(attributeValue);
			}
		}

		return summary.toString();
	}

	/**
	 * @param attributeId
	 * @param workItem
	 * @param parentWorkItem
	 * @param workItemCommon
	 * @return
	 * @throws TeamRepositoryException 
	 */
	private String getAttributeValue(String attribute, IWorkItem workItem, IWorkItem parentWorkItem, IWorkItemCommon workItemCommon, IProgressMonitor monitor) throws TeamRepositoryException {
		String attributeId;
		IWorkItem workItemToUse;
		if(attribute.length() >= PARENT_DELIMITER.length() && attribute.substring(0, PARENT_DELIMITER.length()).equalsIgnoreCase(PARENT_DELIMITER)){
			// the attribute started with the parent delimiter so remove it from the attribute and use the parent work item
			attributeId = attribute.substring(PARENT_DELIMITER.length());
			if(parentWorkItem == null){
				// this is a parent attribute but the parent work item is null
				addWarning("A parent attribute is specified in the default summary template but no parent work item exists.");
				return "?" + attribute + "?";
			}
			workItemToUse = parentWorkItem;
		}else{
			attributeId = attribute;
			workItemToUse = workItem;
		}

		WorkItemCommonTasks workItemCommonTasks = new WorkItemCommonTasks();
		IAttribute attributeObj = workItemCommonTasks.getAttribute(attributeId, workItemToUse.getProjectArea(), workItemCommon, monitor);

		if(attributeObj == null){
			// an invalid attribute id was specified
			IProjectArea project = (IProjectArea)workItemCommon.getAuditableCommon().resolveAuditable(workItemToUse.getProjectArea(), ItemProfile.createProfile(IProjectArea.ITEM_TYPE, IProjectArea.NAME_PROPERTY_ID), monitor);
			addWarning("The attribute '" + attributeId + "' specified in the default summary template was not found in project " + project.getName());
			return "?" + attribute + "?";
		}
		if(workItemToUse.hasAttribute(attributeObj) == false){
			// the attribute exists, but this work item does not have it
			String workItemText;
			if(workItemToUse == parentWorkItem){
				workItemText = "the parent work item";
			}else{
				workItemText = "this work item";
			}
			addWarning("The attribute '" + attributeId + "' specified in the default summary template exists in the project, but does not exist on " + workItemText);
			return "?" + attribute + "?";
		}
		Object value = workItemToUse.getValue(attributeObj);

		// convert the value object to a string
		if(value instanceof String){
			return (String)value;
		}else if(value instanceof Identifier<?>){
			IEnumeration<? extends ILiteral> enumeration = workItemCommon.resolveEnumeration(attributeObj, monitor);
			ILiteral literal = enumeration.findEnumerationLiteral((Identifier<? extends ILiteral>)value);
			return literal.getName();
		}else if(value instanceof Boolean){
			return String.valueOf(((Boolean)value).booleanValue());
		}else if(value instanceof Integer){
			return String.valueOf(((Integer)value).intValue());
		}else if(value instanceof IContributorHandle){
			IContributor contributor = (IContributor)workItemCommon.getAuditableCommon().resolveAuditable((IContributorHandle)value, ItemProfile.createProfile(IContributor.ITEM_TYPE, IContributor.NAME_PROPERTY), monitor);
			return contributor.getName();
		}else if(value instanceof ICategoryHandle){
			ICategory category = (ICategory)workItemCommon.getAuditableCommon().resolveAuditable((ICategoryHandle)value, ItemProfile.createProfile(ICategory.ITEM_TYPE, ICategory.NAME_PROPERTY), monitor);
			return category.getName();
		}else if(value instanceof IDeliverableHandle){
			IDeliverable deliverable = (IDeliverable)workItemCommon.getAuditableCommon().resolveAuditable((IDeliverableHandle)value, ItemProfile.createProfile(IDeliverable.ITEM_TYPE, IDeliverable.NAME_PROPERTY), monitor);
			return deliverable.getName();
		}else if(value instanceof IIterationHandle){
			IIteration iteration = (IIteration)workItemCommon.getAuditableCommon().resolveAuditable((IIterationHandle)value, ItemProfile.createProfile(IIteration.ITEM_TYPE, IIteration.NAME_PROPERTY_ID), monitor);
			return iteration.getName();
		}else if(value instanceof Timestamp){
			return ((Timestamp)value).toString();
		}else if(value == null){
			addWarning("The attribute '" + attributeId + "' exists on the work item but has a null value");
			// for null, return an empty string
			return "";
		}else{
			// unable to convert attribute to a string
			addWarning("The attribute '" + attributeId + "' has a value of type " + value.getClass().getName() + " which is not supported");
			return "?" + attribute + "?";
		}

	}

	/**
	 * @param config
	 * @param hasParent
	 * @return
	 */
	private String getSummaryTemplate(IProcessConfigurationElement config, boolean hasParent) {
		String template = null;

		// first if a parent exists, see if there is a parent exists template
		if(hasParent){
			IProcessConfigurationElement templateElement = ConfigurationHelper.getConfigurationElementByTagName(config, TAG_PARENT_EXISTS_TEMPLATE);
			if(templateElement != null){
				template = templateElement.getAttribute(ConfigurationHelper.ATTR_VALUE);
			}
		}

		// now if the template is still null or empty, try to get it from the default summary template
		if(template == null || template.length() == 0){
			IProcessConfigurationElement templateElement = ConfigurationHelper.getConfigurationElementByTagName(config, TAG_SUMMARY_TEMPLATE);
			if(templateElement != null){
				template = templateElement.getAttribute(ConfigurationHelper.ATTR_VALUE);
			}
		}

		return template;
	}

	private void addWarning(String message){
		if(showWarnings){
			IReportInfo info = collector.createProblemInfo("Problem calculating default summary", message, ID);
			info.setSeverity(IProcessReport.WARNING);
			collector.addInfo(info);
		}
	}

}
