package com.ibm.stg.rtc.ext.advisors;

import java.util.List;

import org.eclipse.core.runtime.IProgressMonitor;

import com.ibm.stg.rtc.ext.common.ConfigurationHelper;
import com.ibm.stg.rtc.ext.common.WorkItemCommonTasks;
import com.ibm.team.process.common.IProcessConfigurationElement;
import com.ibm.team.process.common.advice.AdvisableOperation;
import com.ibm.team.process.common.advice.IAdvisorInfo;
import com.ibm.team.process.common.advice.IAdvisorInfoCollector;
import com.ibm.team.process.common.advice.runtime.IOperationAdvisor;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.repository.service.AbstractService;
import com.ibm.team.workitem.common.IAuditableCommon;
import com.ibm.team.workitem.common.ISaveParameter;
import com.ibm.team.workitem.common.IWorkItemCommon;
import com.ibm.team.workitem.common.model.IAttribute;
import com.ibm.team.workitem.common.model.IAttributeHandle;
import com.ibm.team.workitem.common.model.IWorkItem;

/**
 * This precondition checks the configured attributes for valid values. Valid values are determined by calling
 * {@link IAttribute#getValueSet(IAuditableCommon, IWorkItem, IProgressMonitor)} which does a real-time lookup
 * of the ValueSetProvider and determines what current values can be. If the current value is not part of this
 * value set, this precondition produces and error. The following xml is an example of how to configure 
 * attributes to test:
 * 
 * <workItemType id="com.ibm.stg.workItemType.stgDefect">
		<attribute id="com.ibm.stg.attribute.odc.trigger"/>
		<attribute id="com.ibm.stg.attribute.odc.activity"/>
	</workItemType>
 * 
 * 
 * @author Kyle Christianson
 */
public class EnforceValueSetFilterAdvisor extends AbstractService implements IOperationAdvisor {

	private static final String ID = "com.ibm.stg.rtc.ext.advisors.EnforceValueSetFilterAdvisor";

	private static final String TAG_ATTRIBUTE = "attribute";

	public void run(AdvisableOperation operation, IProcessConfigurationElement advisorConfiguration, IAdvisorInfoCollector collector, IProgressMonitor monitor) throws TeamRepositoryException {
		WorkItemCommonTasks workItemCommonTasks = new WorkItemCommonTasks();
		IWorkItemCommon workItemCommon = this.getService(IWorkItemCommon.class);
		ISaveParameter saveParameter = null;
		IWorkItem newWorkItemState = null;

		Object tempObj = operation.getOperationData();
		if (workItemCommonTasks.isSaveParamterType(tempObj)) {
			saveParameter = (ISaveParameter) tempObj;
		} else {
			return;
		}

		tempObj = saveParameter.getNewState();
		if (workItemCommonTasks.isIWorkItemType(tempObj)) {
			newWorkItemState = (IWorkItem) tempObj;
		} else {
			return;
		}

		IProcessConfigurationElement workItemTypeConfig = ConfigurationHelper.getWITypeConfiguration(advisorConfiguration, newWorkItemState.getWorkItemType());
		if(workItemTypeConfig != null){
			IAuditableCommon auditableCommon = workItemCommon.getAuditableCommon();
			List<String> attributes = ConfigurationHelper.getChildrenTagIds(workItemTypeConfig, TAG_ATTRIBUTE);
			for(String attributeId : attributes){
				IAttribute attribute = workItemCommonTasks.getAttribute(attributeId, newWorkItemState.getProjectArea(), workItemCommon, monitor);

				if(attribute == null){
					IAdvisorInfo info= collector.createProblemInfo("Invalid Configuration", attributeId + " is not a valid attribute." , ID);
					info.setProblemObject(newWorkItemState.getItemHandle());
					collector.addInfo(info);
				}else if(workItemCommonTasks.attributeHasValidValue(newWorkItemState, attribute, auditableCommon) == false){
					// the value is invalid so construct the error message
					String name = attribute.getDisplayName();
					String dependentAttrs = "";
					List<IAttributeHandle> dependencies = attribute.getDependencies();
					for(int i = 0; i < dependencies.size(); i++){
						IAttributeHandle attrHandle = dependencies.get(i);
						IAttribute depAttr = auditableCommon.resolveAuditable(attrHandle, IAttribute.FULL_PROFILE, monitor);
						dependentAttrs = dependentAttrs + depAttr.getDisplayName();
						// if this is not the last attribute
						if(i != dependencies.size() - 1){
							// if there are three or more add commas
							if(dependencies.size() > 3){
								dependentAttrs = dependentAttrs + ", ";
							}else{
								dependentAttrs = dependentAttrs + " ";
							}
							// if this is the second to last attribute then add 'and'
							if(i == dependencies.size() - 2){
								dependentAttrs = dependentAttrs + "and ";
							}
						}
					}
					IAdvisorInfo info= collector.createProblemInfo("Invalid " + name, name + " is invalid for the current value of " + dependentAttrs , ID);
					info.setProblemObject(newWorkItemState.getItemHandle());
					collector.addInfo(info);
				}
			}
		}
	}
}
