package com.ibm.stg.rtc.ext.advisors;

import org.eclipse.core.runtime.IProgressMonitor;

import com.ibm.stg.rtc.ext.common.WorkItemCommonTasks;
import com.ibm.team.process.common.IProcessConfigurationElement;
import com.ibm.team.process.common.advice.AdvisableOperation;
import com.ibm.team.process.common.advice.IAdvisorInfoCollector;
import com.ibm.team.process.common.advice.IReportInfo;
import com.ibm.team.process.common.advice.runtime.IOperationAdvisor;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.repository.service.AbstractService;
import com.ibm.team.workitem.common.ISaveParameter;
import com.ibm.team.workitem.common.model.CategoryId;
import com.ibm.team.workitem.common.model.ICategory;

/**
 * This operation advisor prevents category creation, name changes, description
 * changes, archival, and hierarchy changes during project area save.
 * 
 * 
 * @author Kyle Christianson
 * 
 */
public class PreventCategoryEditAdvisor extends AbstractService implements IOperationAdvisor {

	private static final String ID = "com.ibm.stg.rtc.ext.advisors.PreventCategoryEditAdvisor";

	public void run(AdvisableOperation operation, IProcessConfigurationElement advisorConfiguration, IAdvisorInfoCollector collector, IProgressMonitor monitor) throws TeamRepositoryException {

		WorkItemCommonTasks workItemCommonTasks = new WorkItemCommonTasks();
		ISaveParameter saveParameter = null;
		ICategory newCategoryState = null;
		ICategory oldCategoryState = null;

		Object tempObj = operation.getOperationData();
		if (workItemCommonTasks.isSaveParamterType(tempObj)) {
			saveParameter = (ISaveParameter) tempObj;
		} else {
			return;
		}

		tempObj = saveParameter.getNewState();
		if (tempObj instanceof ICategory) {
			newCategoryState = (ICategory) tempObj;
		} else {
			return;
		}

		tempObj = saveParameter.getOldState();
		if(tempObj instanceof ICategory){
			oldCategoryState = (ICategory)tempObj;
		} else if(tempObj == null){
			oldCategoryState = null;
		} else{
			return;
		}

		if(oldCategoryState == null){
			// error because a category is being created
			//IReportInfo info = collector.createProblemInfo("Permission Denied", "Categories must be created in Metadata Plus", ID);
			IReportInfo info = collector.createProblemInfo("Permission Denied", "To create categories you must submit an ITHelp Defect. Please see https://stgdoc.rchland.ibm.com/STGC/RTC_FAQ#whynotcreate for more information.", ID);
			info.setProblemObject(newCategoryState.getItemHandle());
			collector.addInfo(info);
		}else{
			if(oldCategoryState.getName().equals(newCategoryState.getName()) == false){
				// error because a category name is being edited
				//IReportInfo info = collector.createProblemInfo("Permission Denied", "Category names can only be edited in Metadata Plus", ID);
				IReportInfo info = collector.createProblemInfo("Permission Denied", "To rename categories you must submit an ITHelp Defect. Please see https://stgdoc.rchland.ibm.com/STGC/RTC_FAQ#whynotcreate for more information.", ID);
				info.setProblemObject(newCategoryState.getItemHandle());
				collector.addInfo(info);
			}

			if(oldCategoryState.getHTMLDescription().getPlainText().equals(newCategoryState.getHTMLDescription().getPlainText()) == false){
				// error because a category description is being edited
				//IReportInfo info = collector.createProblemInfo("Permission Denied", "Category descriptions can only be edited in Metadata Plus", ID);
				IReportInfo info = collector.createProblemInfo("Permission Denied", "To edit category descriptions you must submit an ITHelp Defect. Please see https://stgdoc.rchland.ibm.com/STGC/RTC_FAQ#whynotcreate for more information.", ID);
				info.setProblemObject(newCategoryState.getItemHandle());
				collector.addInfo(info);
			}

			// kpaone - added for task 12162
			if (oldCategoryState.isArchived() != newCategoryState.isArchived()) {
				//error because a category is being archived or unarchived
				IReportInfo info = collector.createProblemInfo("Permission Denied", "To archive or unarchive categories you must submit an ITHelp Defect. Please see https://stgdoc.rchland.ibm.com/STGC/RTC_FAQ#whynotcreate for more information.", ID);
				info.setProblemObject(newCategoryState.getItemHandle());
				collector.addInfo(info);
			}

			// Fix from ITHelp IT040511: 
			// The old and new category may be the root category with no parent so first check if they are null to avoid NullPointerException

			CategoryId oldParent = oldCategoryState.getParentId2();
			CategoryId newParent = newCategoryState.getParentId2();
			boolean hierarchyChanged = false;

			if(oldParent == null){
				if(newParent != null){
					hierarchyChanged = true;
				}
			}else if(oldParent.equals(newParent) == false){
				hierarchyChanged = true;
			}

			if(hierarchyChanged == true){
				// error because a category hierarchy is being edited
				//IReportInfo info = collector.createProblemInfo("Permission Denied", "Category hierarchy can only be edited in Metadata Plus", ID);
				IReportInfo info = collector.createProblemInfo("Permission Denied", "To edit the category hierarchy you must submit an ITHelp Defect. Please see https://stgdoc.rchland.ibm.com/STGC/RTC_FAQ#whynotcreate for more information.", ID);
				info.setProblemObject(newCategoryState.getItemHandle());
				collector.addInfo(info);
			}
		}

	}
}
