package com.ibm.stg.rtc.ext.advisors;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.eclipse.core.runtime.IProgressMonitor;

import com.ibm.stg.rtc.ext.common.ConfigurationHelper;
import com.ibm.team.process.common.IProcessConfigurationElement;
import com.ibm.team.process.common.advice.AdvisableOperation;
import com.ibm.team.process.common.advice.IAdvisorInfoCollector;
import com.ibm.team.process.common.advice.IReportInfo;
import com.ibm.team.process.common.advice.runtime.IOperationAdvisor;
import com.ibm.team.repository.common.IAuditableHandle;
import com.ibm.team.repository.common.IItem;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.repository.service.AbstractService;
import com.ibm.team.repository.service.IRepositoryItemService;
import com.ibm.team.scm.common.process.SaveClosedChangeSetOperationData;
import com.ibm.team.scm.common.process.SaveClosedChangeSetOperationData.ChangeType;
import com.ibm.team.workitem.common.IWorkItemCommon;
import com.ibm.team.workitem.common.model.IWorkItem;
import com.ibm.team.workitem.common.model.IWorkItemHandle;
import com.ibm.team.workitem.common.model.IWorkItemType;
import com.ibm.team.workitem.common.workflow.IWorkflowInfo;

/**
 * This advisor prevents change sets from being added to a work item in certain states. To configure it,
 * first specify the work item type to prevent change sets on. Then optionally specify the states
 * to prevent change set additions. If no states are specified, all states will not allow change sets.
 * 
 * Note: The com.ibm.team.scm.server.modifyChangeset operation runs the preconditions configured for the
 * Change Set's Component owner. If the component is owned by a user, this will not run.
 * 
 * Example:
 * 
		<precondition id="com.ibm.stg.rtc.ext.advisors.PreventChangeSetAdvisor" name="Prevent Change Sets" optional="false">
			<workItemType id="com.ibm.stg.workItemType.stgDefect"/>
			<workItemType id="com.ibm.stg.workItemType.changeRecord">
				<workItemState id="com.ibm.stg.workflow.changeRecord.state.open"/>
			</workItemType>
		</precondition>

 * @author Kyle Christianson
 *
 */
public class PreventChangeSetAdvisor extends AbstractService implements IOperationAdvisor {

	public static final String ID = "com.ibm.stg.rtc.ext.advisors.PreventChangeSetAdvisor";

	public void run(AdvisableOperation operation, IProcessConfigurationElement advisorConfiguration, IAdvisorInfoCollector collector, IProgressMonitor monitor) throws TeamRepositoryException {
		//IWorkItemCommon workItemCommon = this.getService(IWorkItemCommon.class);


		if(! (operation.getOperationData() instanceof SaveClosedChangeSetOperationData))
			return;
		SaveClosedChangeSetOperationData operationData = (SaveClosedChangeSetOperationData) operation.getOperationData();

		if(operationData.getChangeType() != ChangeType.ADD_LINK) {
			return;
		}

		List<IWorkItemHandle> workItemHandles = new ArrayList<IWorkItemHandle>(); 
		for (Iterator<?> iterator = operationData.getLinks().iterator(); iterator.hasNext(); ) {
			IAuditableHandle workItemHandle = (IAuditableHandle) iterator.next();
			if (workItemHandle instanceof IWorkItemHandle) {
				workItemHandles.add((IWorkItemHandle) workItemHandle);
			}
		}

		IRepositoryItemService itemService = getService(IRepositoryItemService.class);
		IWorkItemCommon workItemCommon = getService(IWorkItemCommon.class);
		IItem[] workItems = itemService.fetchItems(workItemHandles.toArray(new IWorkItemHandle[0]), null);

		for (int i = 0; i < workItems.length; i++) {
			IWorkItem workItem = (IWorkItem)workItems[i];

			List<String> states = ConfigurationHelper.getWIStatesConfigured(advisorConfiguration, workItem.getWorkItemType());
			if(states != null){
				// the work item type is configured so find the states
				// no states means prevent change sets for all states
				if(states.size() == 0 || states.contains(workItem.getState2().getStringIdentifier())){
					IWorkItemType workItemType = workItemCommon.findWorkItemType(workItem.getProjectArea(), workItem.getWorkItemType(), monitor);
					String typeName;
					if(workItemType != null){
						typeName = workItemType.getDisplayName();
					}else{
						typeName = workItem.getState2().getStringIdentifier();
					}

					IWorkflowInfo workflowInfo = workItemCommon.findWorkflowInfo(workItem, monitor);
					String stateName = null;
					if(workflowInfo != null){
						stateName = workflowInfo.getStateName(workItem.getState2());
					}
					if(stateName == null){
						stateName = workItem.getState2().getStringIdentifier();
					}
					IReportInfo info = collector.createProblemInfo("Change set not allowed", "Change sets are not allowed for " + typeName + " work items in the " + stateName + " state.", ID);
					info.setProblemObject(workItem.getItemHandle());
					collector.addInfo(info);
				}
			}
		}
	}
}
