package com.ibm.stg.rtc.ext.advisors;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.eclipse.core.runtime.IProgressMonitor;

import com.ibm.stg.rtc.ext.common.RepositoryCommonTasks;
import com.ibm.stg.rtc.ext.common.WorkItemCommonTasks;
import com.ibm.team.process.common.IProcessConfigurationElement;
import com.ibm.team.process.common.advice.AdvisableOperation;
import com.ibm.team.process.common.advice.IAdvisorInfo;
import com.ibm.team.process.common.advice.IAdvisorInfoCollector;
import com.ibm.team.process.common.advice.runtime.IOperationAdvisor;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.repository.service.AbstractService;
import com.ibm.team.workitem.common.IAuditableCommon;
import com.ibm.team.workitem.common.ISaveParameter;
import com.ibm.team.workitem.common.IWorkItemCommon;
import com.ibm.team.workitem.common.internal.model.WorkItemAttributes;
import com.ibm.team.workitem.common.internal.workflow.WorkflowInfo;
import com.ibm.team.workitem.common.model.IAttribute;
import com.ibm.team.workitem.common.model.IResolution;
import com.ibm.team.workitem.common.model.IState;
import com.ibm.team.workitem.common.model.IWorkItem;
import com.ibm.team.workitem.common.model.IWorkItemType;
import com.ibm.team.workitem.common.model.Identifier;
import com.ibm.team.workitem.common.workflow.IWorkflowAction;
import com.ibm.team.workitem.common.workflow.IWorkflowInfo;

@SuppressWarnings("restriction")
public class PreventFieldEditAdvisor extends AbstractService implements IOperationAdvisor {

	/* This precondition is configurable the same way the RequiredPropertiesAdvisor is configured except instead of requiring 
	 * properties, it is making them non-editable. The following is an example of how to configure this precondition. Use the <properties> tag with
	 * the workItemType attribute to prevent edit on all states including creation. Or use <workflowProperties> tag with the
	 * workItemType and stateId attributes to specify certain states to prevent edit. Use the keyword "creation" for the stateId if
	 * you want to prevent the fields from being edited when the work item is created. 

	<properties workItemTypeCategory="com.ibm.team.workitem.workItemType.task">
        <property id="com.ibm.team.workitem.attribute.summary"/>
        <property id="com.ibm.team.workitem.attribute.category"/>
    </properties>
	<workflowProperties stateId="com.ibm.stg.workflow.changeRecord.state.open" workItemType="com.ibm.stg.workItemType.changeRecord">
    	<property id="com.ibm.stg.attribute.languageID"/>
    	<property id="com.ibm.team.workitem.attribute.target"/>
    </workflowProperties>
	 */

	public static final String ID = "com.ibm.stg.stgdefect.advisors.PreventFieldEditAfterCreateAdvisor";
	private static final String CREATION = "creation";

	private static final String PROPERTIES_ELEMENT= "properties"; //$NON-NLS-1$
	private static final String WORKFLOW_PROPERTIES_ELEMENT= "workflowProperties"; //$NON-NLS-1$
	private static final String TYPE_ATTRIBUTE= "workItemType"; //$NON-NLS-1$
	private static final String CATEGORY_ATTRIBUTE= "workItemTypeCategory"; //$NON-NLS-1$
	private static final String STATE_ID_ATTRIBUTE= "stateId"; //$NON-NLS-1$
	private static final String RESOLUTION_ID_ATTRIBUTE= "resolutionId"; //$NON-NLS-1$
	private static final String PROPERTY_ELEMENT= "property"; //$NON-NLS-1$
	private static final String ID_ATTRIBUTE= "id"; //$NON-NLS-1$

	public void run(AdvisableOperation operation, IProcessConfigurationElement advisorConfiguration, IAdvisorInfoCollector collector, IProgressMonitor monitor) throws TeamRepositoryException {
		ISaveParameter saveParameter = null;
		IWorkItemCommon workItemCommon = this.getService(IWorkItemCommon.class);
		IAuditableCommon auditableCommon= this.getService(IAuditableCommon.class);
		WorkItemCommonTasks workItemCommonTasks = new WorkItemCommonTasks();
		IWorkItem newWorkItemState = null;
		IWorkItem oldWorkItemState = null;



		Object tempObj = operation.getOperationData();
		if (workItemCommonTasks.isSaveParamterType(tempObj)) {
			saveParameter = (ISaveParameter) tempObj;
		} else {
			return;
		}

		tempObj = saveParameter.getNewState();
		if (workItemCommonTasks.isIWorkItemType(tempObj)) {
			newWorkItemState = (IWorkItem) tempObj;
		} else {
			return;
		}

		tempObj = saveParameter.getOldState();
		if (workItemCommonTasks.isIWorkItemType(tempObj)) {
			oldWorkItemState = (IWorkItem) tempObj;
		} else if(tempObj == null){
			oldWorkItemState = null;
		} else{
			return;
		}

		IWorkItemType workItemType= workItemCommon.findWorkItemType(newWorkItemState.getProjectArea(), newWorkItemState.getWorkItemType(), monitor);
		if (workItemType == null)
			return;

		List<String> list= readAdvisorConfiguration(workItemType, advisorConfiguration);
		List<String> workflowProperties= Collections.emptyList();

		String stateId;
		if(saveParameter.isCreation()){
			stateId = CREATION;
		} else{
			stateId= findTargetStateId(newWorkItemState, saveParameter.getWorkflowAction(), workItemCommon, monitor);
		}

		Identifier<IResolution> resolution= newWorkItemState.getResolution2();
		String resolutionId= resolution != null ? resolution.getStringIdentifier() : null;

		if (stateId != null) {
			workflowProperties= readWorkflowAdvisorConfiguration(workItemType, stateId, resolutionId, advisorConfiguration);
		}
		if (!workflowProperties.isEmpty()) {
			list.addAll(workflowProperties);
		}
		String[] properties= list.toArray(new String[list.size()]);

		for (int i= 0; i < properties.length; i++) {

			IAttribute attribute= workItemCommonTasks.getAttribute(properties[i], newWorkItemState.getProjectArea(), workItemCommon, monitor);
			if (attribute == null)
				continue;

			if (newWorkItemState.isAttributeSet(attribute)) {
				if (!RepositoryCommonTasks.itemsEqual(getValue(newWorkItemState, attribute, auditableCommon), getValue(oldWorkItemState, attribute, auditableCommon))) {
					// the field has been edited so collect an error
					IWorkflowInfo wfInfo= workItemCommon.findWorkflowInfo(newWorkItemState, monitor);
					String stateName = wfInfo.getStateName(Identifier.create(IState.class, stateId));
					IAdvisorInfo info= collector.createProblemInfo("Cannot edit " + attribute.getDisplayName(), attribute.getDisplayName() + " cannot be edited on a " + workItemType.getDisplayName() + " when it is in the " + stateName + " state.", ID);
					info.setProblemObject(newWorkItemState.getItemHandle());
					collector.addInfo(info);
				}
			}
		}
	}

	public static String findTargetStateId(IWorkItem workItem, String actionId, IWorkItemCommon workItemCommon, IProgressMonitor monitor) throws TeamRepositoryException {
		Identifier<IState> state= workItem.getState2();
		IWorkflowInfo wfInfo= workItemCommon.findWorkflowInfo(workItem, monitor);
		if (state == null && actionId == null && wfInfo != null) {
			actionId= wfInfo.getStartActionId() == null?null:wfInfo.getStartActionId().getStringIdentifier();
		}
		if (wfInfo != null && state != null) {
			if (!Arrays.asList(wfInfo.getAllStateIds()).contains(state)) {
				actionId= wfInfo.getStartActionId() == null?null:wfInfo.getStartActionId().getStringIdentifier();
			}
		}
		if (actionId != null && wfInfo != null) {
			state= wfInfo.getActionResultState(Identifier.create(IWorkflowAction.class, actionId));
			if (state == null) {
				actionId= wfInfo.getStartActionId() == null?null:wfInfo.getStartActionId().getStringIdentifier();
				if (actionId != null) {
					state= wfInfo.getActionResultState(Identifier.create(IWorkflowAction.class, actionId));
				}
			} 
		}
		if (state != null) {
			String stateId= state.getStringIdentifier();
			try {
				Integer.parseInt(stateId);
				stateId= "s" + stateId;//$NON-NLS-1$
			} catch (NumberFormatException e) {
			}
			return stateId;
		}
		return null;
	}

	private Object getValue(IWorkItem workItem, IAttribute attribute, IAuditableCommon auditableCommon) throws TeamRepositoryException {
		if(workItem == null){
			return attribute.getDefaultValue(auditableCommon, workItem, null);
		}

		if(!workItem.hasAttribute(attribute)){
			return null;
		}

		Object value= workItem.getValue(attribute);
		/*if (AttributeTypes.STRING_TYPES.contains(attribute.getAttributeType()) && value instanceof String)
			value= ((String) value).trim();
		else if (AttributeTypes.HTML_TYPES.contains(attribute.getAttributeType()) && value instanceof String)
			value= XMLString.createFromXMLText(((String) value)).getPlainText().trim();*/
		return value;
	}

	public static List<String> readAdvisorConfiguration(IWorkItemType workItemType, IProcessConfigurationElement advisorConfiguration) {

		Set<String> properties= new HashSet<String>();

		IProcessConfigurationElement[] configurationElements= advisorConfiguration.getChildren();
		for (int i= 0; i < configurationElements.length; i++) {
			IProcessConfigurationElement configurationElement= configurationElements[i];
			if (PROPERTIES_ELEMENT.equals(configurationElement.getName())) {
				if (configurationElement.getAttribute(TYPE_ATTRIBUTE) == null && configurationElement.getAttribute(CATEGORY_ATTRIBUTE) == null) {
					readProperties(properties, configurationElement.getChildren());
				} else {
					if (workItemType.getIdentifier().equals(configurationElement.getAttribute(TYPE_ATTRIBUTE)))
						readProperties(properties, configurationElement.getChildren());
					if (workItemType.getCategory().equals(configurationElement.getAttribute(CATEGORY_ATTRIBUTE)))
						readProperties(properties, configurationElement.getChildren());
				}
			}
		}

		return new ArrayList<String>(properties);
	}

	public static List<String> readWorkflowAdvisorConfiguration(IWorkItemType workItemType, String stateId, String resolutionId, IProcessConfigurationElement advisorConfiguration) {

		Set<String> properties= new HashSet<String>();

		IProcessConfigurationElement[] configurationElements= advisorConfiguration.getChildren();
		for (int i= 0; i < configurationElements.length; i++) {
			IProcessConfigurationElement configurationElement= configurationElements[i];
			if (WORKFLOW_PROPERTIES_ELEMENT.equals(configurationElement.getName())) {
				if (stateId.equals(configurationElement.getAttribute(STATE_ID_ATTRIBUTE))) {
					String requiredResolutionId= WorkflowInfo.stripOffPrefix(configurationElement.getAttribute(RESOLUTION_ID_ATTRIBUTE), 'r');
					if (requiredResolutionId == null || requiredResolutionId.equals(resolutionId)) {
						if (workItemType.getIdentifier().equals(configurationElement.getAttribute(TYPE_ATTRIBUTE)))
							readProperties(properties, configurationElement.getChildren());
						if (workItemType.getCategory().equals(configurationElement.getAttribute(CATEGORY_ATTRIBUTE)))
							readProperties(properties, configurationElement.getChildren());
					}
				}
			}
		}

		return new ArrayList<String>(properties);
	}

	private static void readProperties(Set<String> properties, IProcessConfigurationElement[] configurationElements) {

		for (int i= 0; i < configurationElements.length; i++) {
			IProcessConfigurationElement configurationElement= configurationElements[i];
			if (PROPERTY_ELEMENT.equals(configurationElement.getName())) {
				String property= configurationElement.getAttribute(ID_ATTRIBUTE);
				if (property == null || property.length() == 0)
					continue;
				if ("version".equals(property)) //$NON-NLS-1$
					property= IWorkItem.FOUND_IN_PROPERTY;
				property= WorkItemAttributes.getAttributeId(Identifier.create(IAttribute.class, property));
				properties.add(property);
			}
		}
	}

}
