package com.ibm.stg.rtc.ext.advisors;

import java.util.List;

import org.eclipse.core.runtime.IProgressMonitor;

import com.ibm.stg.rtc.ext.common.WorkItemCommonTasks;
import com.ibm.team.links.common.IItemReference;
import com.ibm.team.links.common.IReference;
import com.ibm.team.process.common.IProcessConfigurationElement;
import com.ibm.team.process.common.advice.AdvisableOperation;
import com.ibm.team.process.common.advice.IAdvisorInfo;
import com.ibm.team.process.common.advice.IAdvisorInfoCollector;
import com.ibm.team.process.common.advice.runtime.IOperationAdvisor;
import com.ibm.team.repository.common.IItemHandle;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.repository.service.AbstractService;
import com.ibm.team.repository.service.IRepositoryItemService;
import com.ibm.team.workitem.common.IAuditableCommon;
import com.ibm.team.workitem.common.ISaveParameter;
import com.ibm.team.workitem.common.model.IAttachment;
import com.ibm.team.workitem.common.model.IAttachmentHandle;
import com.ibm.team.workitem.common.model.WorkItemEndPoints;

/**
 * checks the max size of an attachment. This precondition does not work in the eclipse client. Use Max Attachment Size RTC Server property instead.
 * @author Kyle Christianson
 *
 */
public class PreventLargeAttachmentAdvisor extends AbstractService implements IOperationAdvisor {

	public static final String ID = "com.ibm.stg.advisor.ValidateAttachmentSizeAdvisor";
	private static final String TAG_ATTACHMENTS = "attachments";
	private static final String ATTR_MAX_SIZE = "maxSizeMB";
	private static final long BYTES_IN_MEGABYTE = 1048576;

	public void run(AdvisableOperation operation, IProcessConfigurationElement advisorConfiguration, IAdvisorInfoCollector collector, IProgressMonitor monitor) throws TeamRepositoryException {
		ISaveParameter saveParameter = null;
		WorkItemCommonTasks workItemCommonTasks = new WorkItemCommonTasks();
		IAuditableCommon auditableCommon = this.getService(IAuditableCommon.class);
		IRepositoryItemService repositoryItemService = this.getService(IRepositoryItemService.class);

		Object tempObj = operation.getOperationData();
		if (workItemCommonTasks.isSaveParamterType(tempObj)) {
			saveParameter = (ISaveParameter) tempObj;
		} else {
			return;
		}

		long maxSize;
		try{
			maxSize = getMaxAttachmentSize(advisorConfiguration);
		} catch(NumberFormatException e){
			IAdvisorInfo info = collector.createProblemInfo("Configuration Error", "The PreventLargeAttachment precondition must contain a valid max size attribute in its xml configuration.", ID);
			collector.addInfo(info);
			return;
		}

		tempObj = saveParameter.getNewState();
		if (tempObj instanceof IAttachment) {
			IAttachment attachment = (IAttachment) tempObj;
			long contentLength = attachment.getContent().getRawLength();
			if (contentLength > maxSize) {
				IAdvisorInfo info = collector.createProblemInfo("Attachment too large", "The max attachment size is " + maxSize / BYTES_IN_MEGABYTE + "MB.", ID);
				info.setProblemObject(attachment.getItemHandle());
				collector.addInfo(info);
			}
		}else if (workItemCommonTasks.isIWorkItemType(tempObj)) {
			// performing the save on a work item
			// get all the attachment links
			List<IReference> attachmentRefs = saveParameter.getNewReferences().getReferences(WorkItemEndPoints.ATTACHMENT);
			for(IReference attachmentRef : attachmentRefs){
				// for every attachment links, get the attachment object
				if(attachmentRef.isItemReference()){
					IItemHandle itemHandle = ((IItemReference)attachmentRef).getReferencedItem();
					if(itemHandle instanceof IAttachmentHandle){
						IAttachmentHandle attHandle = (IAttachmentHandle)itemHandle;
						if(repositoryItemService.itemExists(attHandle)){ // this returns false when a new attachment is added in the eclipse client
							IAttachment attachment = (IAttachment)auditableCommon.findCachedAuditable(attHandle, IAttachment.DEFAULT_PROFILE);
							// check the size of the attachment
							long contentLength = attachment.getContent().getRawLength();
							if (contentLength > maxSize) {
								IAdvisorInfo info = collector.createProblemInfo("Attachment too large", "The max attachment size is " + maxSize / BYTES_IN_MEGABYTE + "MB.", ID);
								info.setProblemObject(attachment.getItemHandle());
								collector.addInfo(info);
							}
						}
					}
				}
			}
		}
	}



	/**
	 * Looks up the max attachment size from the configuration xml and converts it to bytes
	 * @param advisorConfiguration
	 * @return
	 */
	private long getMaxAttachmentSize(IProcessConfigurationElement advisorConfiguration) {
		String strSize = null;

		IProcessConfigurationElement[] configurationElements = advisorConfiguration.getChildren();

		// loop through all the child xml elements for this precondition
		for (IProcessConfigurationElement configurationElement : configurationElements) {
			// if the element is the attachments element, get the value
			if (configurationElement.getName().equals(TAG_ATTACHMENTS)) {
				strSize = configurationElement.getAttribute(ATTR_MAX_SIZE);
				break;
			}
		}

		long size = Long.parseLong(strSize);
		size = size * BYTES_IN_MEGABYTE;

		return size;
	}

}
