package com.ibm.stg.rtc.ext.advisors;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import org.eclipse.core.runtime.IProgressMonitor;

import com.ibm.stg.rtc.ext.common.ApprovalContainer;
import com.ibm.stg.rtc.ext.common.ApprovalWorkItemCondition;
import com.ibm.stg.rtc.ext.common.ApprovalsHelper;
import com.ibm.stg.rtc.ext.common.ApproverListContainer;
import com.ibm.stg.rtc.ext.common.PermissionsHelper;
import com.ibm.stg.rtc.ext.common.WorkItemChangeInfo;
import com.ibm.stg.rtc.ext.common.WorkItemCommonTasks;
import com.ibm.team.process.common.IProcessArea;
import com.ibm.team.process.common.IProcessConfigurationElement;
import com.ibm.team.process.common.advice.AdvisableOperation;
import com.ibm.team.process.common.advice.IAdvisorInfo;
import com.ibm.team.process.common.advice.IAdvisorInfoCollector;
import com.ibm.team.process.common.advice.runtime.IOperationAdvisor;
import com.ibm.team.repository.common.IAuditableHandle;
import com.ibm.team.repository.common.IContributor;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.repository.service.AbstractService;
import com.ibm.team.repository.service.IRepositoryItemService;
import com.ibm.team.workitem.common.IAuditableCommon;
import com.ibm.team.workitem.common.ISaveParameter;
import com.ibm.team.workitem.common.IWorkItemCommon;
import com.ibm.team.workitem.common.model.IApproval;
import com.ibm.team.workitem.common.model.IApprovalDescriptor;
import com.ibm.team.workitem.common.model.IApprovals;
import com.ibm.team.workitem.common.model.IResolution;
import com.ibm.team.workitem.common.model.IWorkItem;
import com.ibm.team.workitem.common.model.Identifier;
import com.ibm.team.workitem.common.model.ItemProfile;
import com.ibm.team.workitem.common.model.WorkItemApprovals;
import com.ibm.team.workitem.common.workflow.IWorkflowInfo;

public class PreventSTGApprovalChangeAdvisor extends AbstractService implements IOperationAdvisor {

	public static final String ID = "com.ibm.stg.rtc.ext.advisors.PreventSTGApprovalChangeAdvisor";

	private static final String VALUE_WORK_ITEM_CREATOR = "workItemCreator";
	private static final String VALUE_WORK_ITEM_OWNER = "workItemOwner";

	public void run(AdvisableOperation operation, IProcessConfigurationElement advisorConfiguration, IAdvisorInfoCollector collector, IProgressMonitor monitor) throws TeamRepositoryException {
		WorkItemCommonTasks workItemCommonTasks = new WorkItemCommonTasks();
		IRepositoryItemService repositoryItemService = this.getService(IRepositoryItemService.class);
		IWorkItemCommon workItemCommon = this.getService(IWorkItemCommon.class);
		IAuditableCommon auditableCommon = workItemCommon.getAuditableCommon();
		ISaveParameter saveParameter = null;
		IWorkItem newWorkItemState = null;
		IWorkItem oldWorkItemState = null;

		Object tempObj = operation.getOperationData();
		if (workItemCommonTasks.isSaveParamterType(tempObj)) {
			saveParameter = (ISaveParameter) tempObj;
		} else {
			return;
		}

		tempObj = saveParameter.getNewState();
		if (workItemCommonTasks.isIWorkItemType(tempObj)) {
			newWorkItemState = (IWorkItem) tempObj;
		} else {
			return;
		}

		tempObj = saveParameter.getOldState();
		if (workItemCommonTasks.isIWorkItemType(tempObj)) {
			oldWorkItemState = (IWorkItem) tempObj;
		} else if(tempObj == null){
			oldWorkItemState = null;
		} else{
			return;
		}

		List<IApprovalDescriptor> newApprovalDescriptors = ApprovalsHelper.getAllAutoGeneratedApprovals(newWorkItemState);

		// first see if there are duplicate approvals with the same name
		if(duplicateApprovals(newApprovalDescriptors)){
			IAdvisorInfo info = collector.createProblemInfo("Approval Error", "An auto-generated approval cannot be manually created.", ID);
			info.setProblemObject(newWorkItemState.getItemHandle());
			collector.addInfo(info);
		}

		List<IApprovalDescriptor> oldApprovalDescriptors = ApprovalsHelper.getAllAutoGeneratedApprovals(oldWorkItemState);
		ArrayList<IApprovalDescriptor> requiredApprovalsToExist = new ArrayList<IApprovalDescriptor>();
		// add all the old approvals to the required list except
		// if the approval is configured to be removed, do not add it to the required list
		for (IApprovalDescriptor descriptor : oldApprovalDescriptors) {
			ApprovalContainer configuredApproval = ApprovalsHelper.getConfiguredApproval(advisorConfiguration, newWorkItemState, descriptor.getName(), workItemCommon);

			IWorkflowInfo workflowInfo = workItemCommon.findWorkflowInfo(newWorkItemState, monitor);
			if(configuredApproval == null || !wasConfiguredToRemove(configuredApproval, oldWorkItemState, workflowInfo, repositoryItemService)){
				requiredApprovalsToExist.add(descriptor);
			}
		}

		boolean nameChanged = false;
		boolean typeChanged = false;
		boolean detailsChanged = false; // currently only the due date	
		boolean approversAdded = false;
		boolean approversRemoved = false;
		boolean approverStateModifiedWhenNotAllowed = false;

		for(IApprovalDescriptor requiredApprovalDescriptor : requiredApprovalsToExist){
			IApprovalDescriptor newApprovalDescriptor = ApprovalsHelper.findApprovalDescriptorByName(newWorkItemState, requiredApprovalDescriptor.getName());
			ApprovalContainer configuredApproval = null;
			IWorkflowInfo workflowInfo = null;

			if(newApprovalDescriptor == null){
				nameChanged = true; // could not find an approval with the required name
			}else{
				// the approval was found, so check if the due date, type, or approvers have changed
				if((requiredApprovalDescriptor.getDueDate() == null && newApprovalDescriptor.getDueDate() != null) || 
						(requiredApprovalDescriptor.getDueDate() != null && requiredApprovalDescriptor.getDueDate().equals(newApprovalDescriptor.getDueDate()) == false)){
					detailsChanged = true;
				}
				if(requiredApprovalDescriptor.getTypeIdentifier().equals(newApprovalDescriptor.getTypeIdentifier()) == false){
					typeChanged = true;
				}

				IApprovals oldApprovals = oldWorkItemState.getApprovals();
				IApprovals newApprovals = newWorkItemState.getApprovals();
				List<IApproval> requiredApproverList = oldApprovals.getContents(requiredApprovalDescriptor);
				List<IApproval> newApproverList = newApprovals.getContents(newApprovalDescriptor);

				configuredApproval = ApprovalsHelper.getConfiguredApproval(advisorConfiguration, newWorkItemState, newApprovalDescriptor.getName(), workItemCommon);
				workflowInfo = workItemCommon.findWorkflowInfo(newWorkItemState, monitor);

				// first make sure all required approvers still exist
				for(IApproval requiredApprover : requiredApproverList){
					if(ApprovalsHelper.findApprover(newApproverList, requiredApprover) == null){
						approversRemoved = true;
						break;
					}
				}

				// now check to see if any new approvers were added
				// its okay to add approvers that belong to the list
				ApproverListContainer approverListContainer = ApprovalsHelper.getApproverList(operation.getProcessArea(), newApprovalDescriptor.getName(), newWorkItemState, auditableCommon, workItemCommon);
				Set<String> configuredApproverList;
				if(approverListContainer == null){
					configuredApproverList = new HashSet<String>();
				} else{
					// Here we are getting the list of approvers configured to be added so we don't throw an error when the CreateSTGApprovals follow-up
					// action adds the configured approvers. The change info will contain the save parameters from the CreateSTGApprovals
					WorkItemChangeInfo changeInfo = WorkItemChangeInfo.getWorkItemChangeInfo(saveParameter, workItemCommon, monitor, null);
					configuredApproverList = approverListContainer.getChangedApprovers(changeInfo, workItemCommon, monitor);
				}
				for(IApproval newApprover : newApproverList){
					if(ApprovalsHelper.findApprover(requiredApproverList, newApprover) == null){
						// this approver was added, see if it is in the configuration list
						String approverId = ((IContributor)auditableCommon.resolveAuditable(newApprover.getApprover(), ItemProfile.createFullProfile(IContributor.ITEM_TYPE), monitor)).getUserId();
						if(configuredApproverList.contains(approverId) == false){
							approversAdded = true; // an approver was added!
							break;
						}
					}
				}

				//Now check if an individual approver state is allowed to be modified
				//if not, determine if the user modified the approver state
				List<ApprovalWorkItemCondition> conditions = getCurrentMatchingConditions(configuredApproval, oldWorkItemState, saveParameter);
				if (!isAllowedToModifyApproverState(conditions, oldWorkItemState)) {
					for (IApproval oldApproval : requiredApproverList) {
						if (oldApproval.getApprover().sameItemId(auditableCommon.getUser())) {							
							IApproval newApproval = ApprovalsHelper.findApprover(newApproverList, oldApproval);
							if (newApproval != null && !newApproval.getStateIdentifier().equals(oldApproval.getStateIdentifier())) {
								//the state of the approval changed - error
								approverStateModifiedWhenNotAllowed = true;
								break;
							}							
						}
					}
				}
			}

			if(nameChanged == true){
				IAdvisorInfo info = collector.createProblemInfo("Approval Error", "An auto-generated approval cannot be manually removed or renamed.", ID);
				info.setProblemObject(newWorkItemState.getItemHandle());
				collector.addInfo(info);
				return;
			}

			if(typeChanged == true){
				IAdvisorInfo info = collector.createProblemInfo("Approval Error", "An auto-generated approval cannot be changed to a different type.", ID);
				info.setProblemObject(newWorkItemState.getItemHandle());
				collector.addInfo(info);
				return;
			}
			
			if (approverStateModifiedWhenNotAllowed) {
				String approvalStateName = WorkItemApprovals.getState(requiredApprovalDescriptor.getCumulativeStateIdentifier()).getDisplayName();
				IAdvisorInfo info = collector.createProblemInfo("Approval State Change Not Allowed", "The work item is in a state that requires the auto-generated approval " + requiredApprovalDescriptor.getName() + " to be " + approvalStateName + ".", ID);
				info.setProblemObject(newWorkItemState.getItemHandle());
				collector.addInfo(info);
				return;
			}

			if(detailsChanged == true){
				//if the due date was changed:  fail if the user does not have permission to modify the details of the approval
				List<ApprovalWorkItemCondition> conditions = getCurrentMatchingConditions(configuredApproval, oldWorkItemState, saveParameter);
				if(!isAllowedToModifyDueDates(conditions, oldWorkItemState, operation.getProcessArea(), auditableCommon)) {
					List<ApprovalWorkItemCondition> allConditions = new ArrayList<ApprovalWorkItemCondition>(configuredApproval.getAllConditionConfigs());
					// the user ever has permission to modify the due date, just not now, return a different error message
					if (isAllowedToModifyDueDates(allConditions, oldWorkItemState, operation.getProcessArea(), auditableCommon)) {
						IAdvisorInfo info = collector.createProblemInfo(
								"Approval Due Date Change Not Allowed",
								"The work item is in a state that requires the auto-generated approval "
										+ requiredApprovalDescriptor.getName() + " to be read-only.", ID);
						info.setProblemObject(newWorkItemState.getItemHandle());
						collector.addInfo(info);
						return;
					}
					// the user never has permission to modify the due date
					else {
						IAdvisorInfo info = collector.createProblemInfo("Permission Denied",
								"You don't have permission to modify the details of an auto-generated approval.", ID);
						info.setProblemObject(newWorkItemState.getItemHandle());
						collector.addInfo(info);
						return;
					}
				}
			}

			if(approversAdded == true){
				//if approvers were added:  fail if the user does not permission to add approvers to the approval
				List<ApprovalWorkItemCondition> conditions = getCurrentMatchingConditions(configuredApproval, oldWorkItemState,
						saveParameter);
				if (!isAllowedToAddApprovers(conditions, oldWorkItemState, operation.getProcessArea(), auditableCommon)) {
					List<ApprovalWorkItemCondition> allConditions = new ArrayList<ApprovalWorkItemCondition>(
							configuredApproval.getAllConditionConfigs());
					// the user ever has permission to add approvers, just not now, return a different error message
					if (isAllowedToAddApprovers(allConditions, oldWorkItemState, operation.getProcessArea(), auditableCommon)) {
						IAdvisorInfo info = collector.createProblemInfo(
								"Add Approvers Not Allowed",
								"The work item is in a state that requires the auto-generated approval "
										+ requiredApprovalDescriptor.getName() + " to be read-only.", ID);
						info.setProblemObject(newWorkItemState.getItemHandle());
						collector.addInfo(info);
						return;
					}
					// the user never has permission to add approvers
					else {
						IAdvisorInfo info = collector.createProblemInfo("Permission Denied",
								"You don't have permission to add approvers to an auto-generated approval.", ID);
						info.setProblemObject(newWorkItemState.getItemHandle());
						collector.addInfo(info);
						return;
					}
				}
			}

			if (approversRemoved == true) {
				// if approvers was removed: fail if the user does not permission to remove approvers from the approval
				List<ApprovalWorkItemCondition> conditions = getCurrentMatchingConditions(configuredApproval, oldWorkItemState,
						saveParameter);
				if (!isAllowedToRemoveApprovers(conditions, oldWorkItemState, operation.getProcessArea(), auditableCommon)) {
					List<ApprovalWorkItemCondition> allConditions = new ArrayList<ApprovalWorkItemCondition>(
							configuredApproval.getAllConditionConfigs());
					// the user ever has permission to remove approvers, just not now, return a different error message
					if (isAllowedToRemoveApprovers(allConditions, oldWorkItemState, operation.getProcessArea(), auditableCommon)) {
						IAdvisorInfo info = collector.createProblemInfo(
								"Remove Approvers Not Allowed",
								"The work item is in a state that requires the auto-generated approval "
										+ requiredApprovalDescriptor.getName() + " to be read-only.", ID);
						info.setProblemObject(newWorkItemState.getItemHandle());
						collector.addInfo(info);
						return;
					}
					// the user never has permission to remove approvers
					else {
						IAdvisorInfo info = collector.createProblemInfo("Permission Denied",
								"You don't have permission to remove approvers from an auto-generated approval.", ID);
						info.setProblemObject(newWorkItemState.getItemHandle());
						collector.addInfo(info);
						return;
					}
				}
			}
		}

		// now make sure there are no approvals that shouldn't belong
		boolean createdApproval = false;
		for(IApprovalDescriptor newApproval : newApprovalDescriptors){
			if(ApprovalsHelper.findApprovalDescriptorByName(oldWorkItemState, newApproval.getName()) == null){
				// new approval was created, see if it was configured
				ApprovalContainer configuredApproval = ApprovalsHelper.getConfiguredApproval(advisorConfiguration, newWorkItemState, newApproval.getName(), workItemCommon);
				if(configuredApproval == null){
					createdApproval = true;
					break;
				}else{
					// an approval was created and it is configured, so now see if it was configured to be created
					IWorkflowInfo workflowInfo = workItemCommon.findWorkflowInfo(newWorkItemState, monitor);
					if(!wasConfiguredToCreate(configuredApproval, oldWorkItemState, workflowInfo, repositoryItemService)){
						createdApproval = true;
						break;
					}
				}
			}
		}

		if(createdApproval == true){
			IAdvisorInfo info = collector.createProblemInfo("Approval Error", "An auto-generated approval cannot be manually created.", ID);
			info.setProblemObject(newWorkItemState.getItemHandle());
			collector.addInfo(info);
		}
	}


	/**
	 * Returns true iff two approval with the same name exist
	 * @param newApprovalDescriptors
	 * @return
	 */
	private boolean duplicateApprovals(List<IApprovalDescriptor> newApprovalDescriptors) {
		List<String> names = new ArrayList<String>();

		for(IApprovalDescriptor desc1 : newApprovalDescriptors){
			if(names.contains(desc1.getName())){
				return true;
			}else{
				names.add(desc1.getName());
			}
		}
		return false;
	}

	private boolean wasConfiguredToRemove(ApprovalContainer approval, IWorkItem workItem, IWorkflowInfo workflowInfo, IRepositoryItemService repositoryItemService) throws TeamRepositoryException{
		for(ApprovalWorkItemCondition condition : getPastMatchingConditions(approval, workItem, workflowInfo, repositoryItemService)){
			if(condition.isRemove()){
				return true;
			}
		}
		return false;
	}

	private boolean wasConfiguredToCreate(ApprovalContainer approval, IWorkItem workItem, IWorkflowInfo workflowInfo, IRepositoryItemService repositoryItemService) throws TeamRepositoryException{
		for(ApprovalWorkItemCondition condition : getPastMatchingConditions(approval, workItem, workflowInfo, repositoryItemService)){
			if(condition.isCreate()){
				return true;
			}
		}
		return false;
	}

	/**
	 * Matches conditions with from the past save, not the current save. This should be used when determining 
	 * if a follow-up action was configure to do something, not for determining this save's permissions. 
	 * The old work item state should be passed in.
	 * @param approval
	 * @param oldWorkItem
	 * @param workflowInfo
	 * @param repositoryItemService
	 * @return
	 * @throws TeamRepositoryException
	 */
	private List<ApprovalWorkItemCondition> getPastMatchingConditions(ApprovalContainer approval, IWorkItem oldWorkItem, IWorkflowInfo workflowInfo, IRepositoryItemService repositoryItemService) throws TeamRepositoryException{
		String fromState = null;
		String curState = null;
		String action = null;
		String resolution = null;

		if(oldWorkItem != null){
			curState = oldWorkItem.getState2().getStringIdentifier();
			Identifier<IResolution> resolutionId = oldWorkItem.getResolution2();
			resolution = (resolutionId == null) ? null : resolutionId.getStringIdentifier();

			IAuditableHandle predecessorState = oldWorkItem.getPredecessorState(); 
			if(predecessorState != null){
				IWorkItem predecessorWorkItemState = (IWorkItem) repositoryItemService.fetchState(oldWorkItem.getPredecessorState(), IRepositoryItemService.COMPLETE);
				fromState = predecessorWorkItemState.getState2().getStringIdentifier();
			}
		}

		action = WorkItemCommonTasks.findActionFromStates(curState, fromState, workflowInfo);

		return getMatchingConditions(approval, action, curState, fromState, resolution);
	}

	/**
	 * Matches conditions with the current work item state and the action being performed. This should be
	 * used for the current save permission checks, not for checking if a follow-up action did something 
	 * in the past save. The save parameter's old work item state should be passed in.
	 * @param approval
	 * @param oldWorkItem
	 * @param saveParameter
	 * @return
	 * @throws TeamRepositoryException
	 */
	private List<ApprovalWorkItemCondition> getCurrentMatchingConditions(ApprovalContainer approval, IWorkItem oldWorkItem, ISaveParameter saveParameter) throws TeamRepositoryException{
		String fromState = null;
		String curState = null;
		String action = null;
		String resolution = null;

		if(oldWorkItem != null){
			curState = oldWorkItem.getState2().getStringIdentifier();
			Identifier<IResolution> resolutionId = oldWorkItem.getResolution2();
			resolution = (resolutionId == null) ? null : resolutionId.getStringIdentifier();

		}

		action = saveParameter.getWorkflowAction();
		if(action != null){
			fromState = curState;
		}

		return getMatchingConditions(approval, action, curState, fromState, resolution);
	}

	private List<ApprovalWorkItemCondition> getMatchingConditions(ApprovalContainer approval, String action, String newState, String fromState, String resolution){
		List<ApprovalWorkItemCondition> matches = new ArrayList<ApprovalWorkItemCondition>();
		for(ApprovalWorkItemCondition condition : approval.getAllConditionConfigs()){
			if(condition.matches(action, newState, fromState, resolution)){
				matches.add(condition);
			}
		}
		return matches;
	}

//	private boolean isAllowedToAddApprovers(ApprovalContainer approval, IWorkItem workItem, ISaveParameter saveParameter, IProcessArea processArea, IAuditableCommon auditableCommon) throws TeamRepositoryException {
//		for(ApprovalWorkItemCondition condition : getCurrentMatchingConditions(approval, workItem, saveParameter)) {
//			if(isModificationAllowed(workItem, condition.getAddApproverRoles(), processArea, auditableCommon)){
//				return true;
//			}
//		}		
//		return false;
//	}
	
	private boolean isAllowedToAddApprovers(List<ApprovalWorkItemCondition> conditions, IWorkItem workItem, IProcessArea processArea,
			IAuditableCommon auditableCommon) throws TeamRepositoryException {
		for (ApprovalWorkItemCondition condition : conditions) {
			if (isModificationAllowed(workItem, condition.getAddApproverRoles(), processArea, auditableCommon)) {
				return true;
			}
		}
		return false;
	}

	// private boolean isAllowedToAddApprovers(ApprovalContainer approval, IWorkItem workItem, IProcessArea processArea, IAuditableCommon
	// auditableCommon) throws TeamRepositoryException {
	// for (ApprovalWorkItemCondition condition : approval.getAllConditionConfigs()) {
	// if(isModificationAllowed(workItem, condition.getAddApproverRoles(), processArea, auditableCommon)){
	// return true;
	// }
	// }
	// return false;
	// }

//	private boolean isAllowedToRemoveApprovers(ApprovalContainer approval, IWorkItem workItem, ISaveParameter saveParameter,
//			IProcessArea processArea, IAuditableCommon auditableCommon) throws TeamRepositoryException {
//		for (ApprovalWorkItemCondition condition : getCurrentMatchingConditions(approval, workItem, saveParameter)) {
//			if (isModificationAllowed(workItem, condition.getRemoveApproverRoles(), processArea, auditableCommon)) {
//				return true;
//			}
//		}
//		return false;
//	}

	private boolean isAllowedToRemoveApprovers(List<ApprovalWorkItemCondition> conditions, IWorkItem workItem, IProcessArea processArea,
			IAuditableCommon auditableCommon) throws TeamRepositoryException {
		for (ApprovalWorkItemCondition condition : conditions) {
			if (isModificationAllowed(workItem, condition.getRemoveApproverRoles(), processArea, auditableCommon)) {
				return true;
			}
		}
		return false;
	}

//	private boolean isAllowedToModifyDueDates(ApprovalContainer approval, IWorkItem workItem, ISaveParameter saveParameter,
//			IProcessArea processArea, IAuditableCommon auditableCommon) throws TeamRepositoryException {
//		for (ApprovalWorkItemCondition condition : getCurrentMatchingConditions(approval, workItem, saveParameter)) {
//			if (isModificationAllowed(workItem, condition.getModifyDueDateRoles(), processArea, auditableCommon)) {
//				return true;
//			}
//		}
//		return false;
//	}

	private boolean isAllowedToModifyDueDates(List<ApprovalWorkItemCondition> conditions, IWorkItem workItem, IProcessArea processArea,
			IAuditableCommon auditableCommon) throws TeamRepositoryException {
		for (ApprovalWorkItemCondition condition : conditions) {
			if (isModificationAllowed(workItem, condition.getModifyDueDateRoles(), processArea, auditableCommon)) {
				return true;
			}
		}
		return false;
	}

//	private boolean isAllowedToModifyApproverState(ApprovalContainer approval, IWorkItem workItem, ISaveParameter saveParameter)
//			throws TeamRepositoryException {
//		for (ApprovalWorkItemCondition condition : getCurrentMatchingConditions(approval, workItem, saveParameter)) {
//			// if the condition is found, only can modify the approver state if modify approvals is configured
//			if (condition.isModifyApprovals()) {
//				return true;
//			}
//		}
//		// there is no matching condition, cannot change the approver state
//		return false;
//	}

	private boolean isAllowedToModifyApproverState(List<ApprovalWorkItemCondition> conditions, IWorkItem workItem)
			throws TeamRepositoryException {
		for (ApprovalWorkItemCondition condition : conditions) {
			// if the condition is found, only can modify the approver state if modify approvals is configured
			if (condition.isModifyApprovals()) {
				return true;
			}
		}
		// there is no matching condition, cannot change the approver state
		return false;
	}

	private boolean isModificationAllowed(IWorkItem workItem, Set<String> roles, IProcessArea processArea, IAuditableCommon auditableCommon) throws TeamRepositoryException {
		for (String role : roles) {
			if (VALUE_WORK_ITEM_CREATOR.equals(role) && auditableCommon.getUser().sameItemId(workItem.getCreator())) {
				//special role value work item creator has permission
				//the user is the work item creator
				return true;
			}
			if (VALUE_WORK_ITEM_OWNER.equals(role) && auditableCommon.getUser().sameItemId(workItem.getOwner())) {
				//special role value work item owner has permission
				//the user is the work item owner
				return true;
			}
			if (PermissionsHelper.isUserRole(auditableCommon, processArea, role)) {
				//configured with a role, if the user has this role, they have permission
				return true;
			}
		}
		return false;
	}
}