package com.ibm.stg.rtc.ext.advisors;

import java.lang.reflect.Field;
import java.util.List;

import org.eclipse.core.runtime.IProgressMonitor;

import com.ibm.stg.rtc.ext.common.ConfigurationHelper;
import com.ibm.stg.rtc.ext.common.InternalRelease;
import com.ibm.stg.rtc.ext.common.Stream;
import com.ibm.team.process.common.IProcessConfigurationElement;
import com.ibm.team.process.common.IProjectArea;
import com.ibm.team.process.common.IProjectAreaHandle;
import com.ibm.team.process.common.ITeamArea;
import com.ibm.team.process.common.advice.AdvisableOperation;
import com.ibm.team.process.common.advice.IAdvisorInfoCollector;
import com.ibm.team.process.common.advice.IReportInfo;
import com.ibm.team.process.common.advice.runtime.IOperationAdvisor;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.repository.service.AbstractService;
import com.ibm.team.scm.common.IWorkspace;
import com.ibm.team.scm.common.process.ProcessOverride;
import com.ibm.team.scm.common.process.RunnableAdvisableOperation;
import com.ibm.team.scm.service.internal.process.WorkspaceOwnershipProcessOverride;
import com.ibm.team.workitem.common.IAuditableCommon;

/**
 * This operation advisor prevents stream renames, change to stream ownership,
 * and stream deletion of streams tracked in metadata.
 * 
 * @author Karen Paone
 * 
 */
public class PreventStreamEditAdvisor extends AbstractService implements IOperationAdvisor {

	private static final String ID = "com.ibm.stg.rtc.ext.advisors.PreventStreamEditAdvisor";

	public void run(AdvisableOperation operation, IProcessConfigurationElement advisorConfiguration,
			IAdvisorInfoCollector collector, IProgressMonitor monitor) throws TeamRepositoryException {

		if (operation == null) {
			return;
		}

		String[] actions = operation.getActions();
		if (actions == null) {
			return;
		}
	
		IWorkspace workspaceStream = null;

		// Using reflection to retrieve the stream since it is not accessible.
		// Change Owner uses a DeferredProcessOverride which does not contain a
		// reference to the stream so that operation cannot be blocked at this
		// time.
		if (operation instanceof RunnableAdvisableOperation) {
			try {
				ProcessOverride value = (ProcessOverride) getMemberFieldValue((RunnableAdvisableOperation) operation,
						"RunnableAdvisableOperation", "processOverride");
				if (!(value instanceof WorkspaceOwnershipProcessOverride)) {
					return;
				}

				if (value == null) {
					return;
				}
				workspaceStream = (IWorkspace) getMemberFieldValue(value, "WorkspaceOwnershipProcessOverride",
						"workspace");
				if (workspaceStream == null) {
					return;
				}
			}
			catch (Throwable e) {
				// error because of an unexpected error while retrieving the
				// stream
				IReportInfo info = collector
						.createProblemInfo(
								"Unexpected Error",
								"Unable to access the stream information from the operation to determine if it is a tracked stream.  Error:  "
										+ e.toString()
										+ "   Please open an ITHELP to report this error:  https://stgdoc.rchland.ibm.com/STGC/RTC_FAQ#rtcHelp",
								ID);
				collector.addInfo(info);
				return;
			}
		}

		IAuditableCommon auditableCommon = this.getService(IAuditableCommon.class);
		// get the internal release with this CR's literal value
		IProjectAreaHandle projectAreaHandle = null;
		if (operation.getProcessArea() instanceof IProjectArea) {
			projectAreaHandle = (IProjectAreaHandle) operation.getProcessArea();
		}
		else if (operation.getProcessArea() instanceof ITeamArea) {
			projectAreaHandle = operation.getProcessArea().getProjectArea();
		}
		else {
			return;
		}
		
		boolean isStreamTracked = false;
		// determine if the stream is being tracked
		// if the stream is in the configuration, then it is tracked
		List<InternalRelease> internalReleases = ConfigurationHelper.getAllInternalReleases(projectAreaHandle, auditableCommon);
		for (InternalRelease internalRelease : internalReleases) {
			if (isStreamTracked) {
				break;
			}
			for (Stream stream : internalRelease.getStreams()) {
				if (stream.getUuid().equals(workspaceStream.getItemId().getUuidValue())) {
					isStreamTracked = true;
					break;
				}
			}
		}

		if (!isStreamTracked) {
			// this stream can be modified
			return;
		}

		for (String action : actions) {
			if (action.equals("modify/stream/attributes/name")) {
				// error because a stream name is being edited
				IReportInfo info = collector
						.createProblemInfo(
								"Permission Denied",
								"To rename tracked streams you must submit an ITHelp Defect. Please see https://stgdoc.rchland.ibm.com/STGC/RTC_FAQ#whynotcreate for more information.",
								ID);
				info.setProblemObject(workspaceStream.getItemHandle());
				collector.addInfo(info);
			}
			else if (action.equals("modify/stream/attributes/owner")) {
				// error because a stream owner is being edited
				IReportInfo info = collector
						.createProblemInfo(
								"Permission Denied",
								"To change ownership of tracked streams you must submit an ITHelp Defect. Please see https://stgdoc.rchland.ibm.com/STGC/RTC_FAQ#whynotcreate for more information.",
								ID);
				info.setProblemObject(workspaceStream.getItemHandle());
				collector.addInfo(info);
			}
			else if (action.equals("modify/stream/delete")) {
				// error because a stream is being deleted
				IReportInfo info = collector
						.createProblemInfo(
								"Permission Denied",
								"To delete tracked streams you must submit an ITHelp Defect. Please see https://stgdoc.rchland.ibm.com/STGC/RTC_FAQ#whynotcreate for more information.",
								ID);
				info.setProblemObject(workspaceStream.getItemHandle());
				collector.addInfo(info);
			}
		}
	}

	private static Object getMemberFieldValue(Object object, String className, String fieldName)
			throws SecurityException, NoSuchFieldException, IllegalArgumentException, IllegalAccessException {
		Class clazz = object.getClass();
		Class theClazz = clazz;

		// find the Class that has the name
		while (!theClazz.getSimpleName().equals(className)) {
			if (theClazz.getSuperclass() != null) {
				theClazz = theClazz.getSuperclass();
			}
			else {
				return null;
			}
		}

		Field field = theClazz.getDeclaredField(fieldName);
		field.setAccessible(true);
		return field.get(object);
	}
}
