package com.ibm.stg.rtc.ext.advisors;

import java.util.List;

import org.eclipse.core.runtime.IProgressMonitor;

import com.ibm.stg.rtc.ext.common.ConfigurationHelper;
import com.ibm.stg.rtc.ext.common.WorkItemCommonTasks;
import com.ibm.team.links.common.IReference;
import com.ibm.team.links.common.registry.IEndPointDescriptor;
import com.ibm.team.links.common.registry.ILinkTypeRegistry;
import com.ibm.team.process.common.IProcessConfigurationElement;
import com.ibm.team.process.common.advice.AdvisableOperation;
import com.ibm.team.process.common.advice.IAdvisorInfoCollector;
import com.ibm.team.process.common.advice.IReportInfo;
import com.ibm.team.process.common.advice.runtime.IOperationAdvisor;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.repository.service.AbstractService;
import com.ibm.team.scm.common.links.ILinkConstants;
import com.ibm.team.workitem.common.ISaveParameter;
import com.ibm.team.workitem.common.IWorkItemCommon;
import com.ibm.team.workitem.common.model.IWorkItem;
import com.ibm.team.workitem.common.model.IWorkItemReferences;
import com.ibm.team.workitem.common.model.WorkItemLinkTypes;

/**
 * This advisor requires at least one change set (includes remote change set) to be added to a work item for certain states. To configure it,
 * first specify the work item type to require a change set on. Then optionally specify the states
 * to require a change set. If no states are specified, all states will require change sets.
 * 
 * Example:
 * 
		<precondition id="com.ibm.stg.rtc.ext.advisors.RequireChangeSetAdvisor" name="STG: Require Change Set" optional="false">
			<workItemType id="com.ibm.stg.workItemType.changeRecord">
				<workItemState id="com.ibm.stg.workflow.changeRecord.state.closed"/>
			</workItemType>
		</precondition>

 * @author Kyle Christianson
 *
 */
public class RequireChangeSetAdvisor extends AbstractService implements IOperationAdvisor {

	public static final String ID = "com.ibm.stg.rtc.ext.advisors.RequireChangeSetAdvisor";

	public void run(AdvisableOperation operation, IProcessConfigurationElement advisorConfiguration, IAdvisorInfoCollector collector, IProgressMonitor monitor) throws TeamRepositoryException {
		IWorkItemCommon workItemCommon = this.getService(IWorkItemCommon.class);
		WorkItemCommonTasks workItemCommonTasks = new WorkItemCommonTasks();
		ISaveParameter saveParameter = null;
		IWorkItem newWorkItemState = null;

		Object tempObj = operation.getOperationData();
		if (workItemCommonTasks.isSaveParamterType(tempObj)) {
			saveParameter = (ISaveParameter) tempObj;
		} else {
			return;
		}

		tempObj = saveParameter.getNewState();
		if (workItemCommonTasks.isIWorkItemType(tempObj)) {
			newWorkItemState = (IWorkItem) tempObj;
		} else {
			return;
		}

		// check to see if the work item type has this advisor configured, if not return
		List<String> states = ConfigurationHelper.getWIStatesConfigured(advisorConfiguration, newWorkItemState.getWorkItemType());
		if(states == null){
			// this work item type was not configured
			return;
		}

		if(states.size() > 0 && !states.contains(newWorkItemState.getState2().getStringIdentifier())){
			// this state was not configured
			return;
		}

		IWorkItemReferences references = workItemCommon.resolveWorkItemReferences(newWorkItemState, monitor);
		IEndPointDescriptor changeSetEndPoint = ILinkTypeRegistry.INSTANCE.getLinkType(ILinkConstants.CHANGESET_WORKITEM_LINKTYPE_ID).getSourceEndPointDescriptor();
		List<IReference> changeSetReferences = references.getReferences(changeSetEndPoint);

		// remote oslc change set link
		// the remote change set reference is the target end point of the TRACKS_CHANGES link type
		IEndPointDescriptor oslcChangeSetEndPoint = ILinkTypeRegistry.INSTANCE.getLinkType(WorkItemLinkTypes.TRACKS_CHANGES).getTargetEndPointDescriptor();
		List<IReference> oslcChangeSetReferences = references.getReferences(oslcChangeSetEndPoint);

		if(changeSetReferences.size() <= 0 && oslcChangeSetReferences.size() <= 0){
			String typeName = workItemCommonTasks.getTypeName(newWorkItemState, workItemCommon, monitor);
			String stateName = workItemCommonTasks.getStateName(newWorkItemState, workItemCommon, monitor);
			IReportInfo info = collector.createProblemInfo("Change set required", "At least one change set is required for " + typeName + " work items in the " + stateName + " state.", ID);
			info.setProblemObject(newWorkItemState.getItemHandle());
			collector.addInfo(info);

		}
	}
}
