package com.ibm.stg.rtc.ext.advisors;

import java.util.List;

import org.eclipse.core.runtime.IProgressMonitor;

import com.ibm.stg.rtc.ext.common.ConfigurationHelper;
import com.ibm.stg.rtc.ext.common.WorkItemCommonTasks;
import com.ibm.team.links.common.IReference;
import com.ibm.team.links.common.registry.IEndPointDescriptor;
import com.ibm.team.links.common.registry.ILinkTypeRegistry;
import com.ibm.team.process.common.IProcessConfigurationElement;
import com.ibm.team.process.common.advice.AdvisableOperation;
import com.ibm.team.process.common.advice.IAdvisorInfoCollector;
import com.ibm.team.process.common.advice.IReportInfo;
import com.ibm.team.process.common.advice.runtime.IOperationAdvisor;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.repository.service.AbstractService;
import com.ibm.team.scm.common.IChangeSet;
import com.ibm.team.scm.common.IChangeSetHandle;
import com.ibm.team.scm.common.links.ILinkConstants;
import com.ibm.team.workitem.common.ISaveParameter;
import com.ibm.team.workitem.common.IWorkItemCommon;
import com.ibm.team.workitem.common.model.IWorkItem;
import com.ibm.team.workitem.common.model.IWorkItemReferences;
import com.ibm.team.workitem.common.model.ItemProfile;
import com.ibm.team.workitem.common.model.WorkItemLinkTypes;

/**
 * This advisor requires that all change sets associated the work item are complete for certain states.
 * To configure it, first specify the work item type to require change set complete on. Then optionally
 * specify the states to require the change sets complete. If no states are specified, all states will 
 * require the change sets be complete.
 * 
 * Example:
 * 
		<precondition id="com.ibm.stg.rtc.ext.advisors.RequireChangeSetsCompleteAdvisor" name="STG: Require Change Sets Complete" optional="false">
			<workItemType id="com.ibm.stg.workItemType.changeRecord">
				<workItemState id="com.ibm.stg.workflow.changeRecord.state.closed"/>
			</workItemType>
		</precondition>

 * @author Kyle Christianson
 *
 */
public class RequireChangeSetsCompleteAdvisor extends AbstractService implements IOperationAdvisor {

	public static final String ID = "com.ibm.stg.rtc.ext.advisors.RequireChangeSetsCompleteAdvisor";

	public void run(AdvisableOperation operation, IProcessConfigurationElement advisorConfiguration, IAdvisorInfoCollector collector, IProgressMonitor monitor) throws TeamRepositoryException {
		IWorkItemCommon workItemCommon = this.getService(IWorkItemCommon.class);
		WorkItemCommonTasks workItemCommonTasks = new WorkItemCommonTasks();
		ISaveParameter saveParameter = null;
		IWorkItem newWorkItemState = null;

		Object tempObj = operation.getOperationData();
		if (workItemCommonTasks.isSaveParamterType(tempObj)) {
			saveParameter = (ISaveParameter) tempObj;
		} else {
			return;
		}

		tempObj = saveParameter.getNewState();
		if (workItemCommonTasks.isIWorkItemType(tempObj)) {
			newWorkItemState = (IWorkItem) tempObj;
		} else {
			return;
		}

		// check to see if the work item type has this advisor configured, if not return
		List<String> states = ConfigurationHelper.getWIStatesConfigured(advisorConfiguration, newWorkItemState.getWorkItemType());
		if(states == null){
			// this work item type was not configured
			return;
		}

		if(states.size() > 0 && !states.contains(newWorkItemState.getState2().getStringIdentifier())){
			// this state was not configured
			return;
		}

		IWorkItemReferences references = workItemCommon.resolveWorkItemReferences(newWorkItemState, monitor);
		IEndPointDescriptor changeSetEndPoint = ILinkTypeRegistry.INSTANCE.getLinkType(ILinkConstants.CHANGESET_WORKITEM_LINKTYPE_ID).getSourceEndPointDescriptor();
		List<IReference> changeSetReferences = references.getReferences(changeSetEndPoint);

		// remote oslc change set link
		// the remote change set reference is the target end point of the TRACKS_CHANGES link type
		IEndPointDescriptor oslcChangeSetEndPoint = ILinkTypeRegistry.INSTANCE.getLinkType(WorkItemLinkTypes.TRACKS_CHANGES).getTargetEndPointDescriptor();
		List<IReference> oslcChangeSetReferences = references.getReferences(oslcChangeSetEndPoint);


		// first check all local change sets for completeness
		for(IReference changeSetReference : changeSetReferences){
			Object resolved = changeSetReference.resolve();
			if(resolved == null || !(resolved instanceof IChangeSetHandle)){
				throw new TeamRepositoryException("A link of type " + ILinkConstants.CHANGESET_WORKITEM_LINKTYPE_ID + " did not resolve to a change set.");
			}
			IChangeSet changeSet = (IChangeSet)workItemCommon.getAuditableCommon().resolveAuditable((IChangeSetHandle)resolved, ItemProfile.createFullProfile(IChangeSet.ITEM_TYPE), monitor);
			if(changeSet.isActive()){
				// the change set is not complete!
				String typeName = workItemCommonTasks.getTypeName(newWorkItemState, workItemCommon, monitor);
				String stateName = workItemCommonTasks.getStateName(newWorkItemState, workItemCommon, monitor);
				IReportInfo info = collector.createProblemInfo("Change set not complete", "All associated change sets must be completed for a " + typeName + " to be in the " + stateName + " state.", ID);
				info.setProblemObject(newWorkItemState.getItemHandle());
				collector.addInfo(info);
				break; // only add the report once
			}
		}

		// now check all remote change sets for completeness
		/* TODO incomplete implementation of remote change sets
			IHttpClient httpClient = null;
			if(oslcChangeSetReferences.size() > 0){
				IHttpService httpService = this.getService(IHttpService.class);
				httpClient = httpService.getHttpClient(true);
			}
			for(IReference oslcChangeSetReference : oslcChangeSetReferences){
//				Experimental 
//				BootstrapConfig config = BootstrapConfig.getInstance();
//				OAuthConsumerProperties consumer = config.getOAuthConsumerProperties();
//
//				JazzHttpClient jazzHttpClient = HttpClientBuilder.buildOAuthProviderClient();
//				//Create a functional user context
//				HttpContext cntxt = jazzHttpClient.createDefaultHttpContext();
//				OAuthContextUtil.setOAuthAccessToken(cntxt,new OAuthAccessToken(HttpConstants.EMPTY, HttpConstants.EMPTY));
//				OAuthContextUtil.setOAuthConsumerProperties(cntxt,consumer);
//
//				//Make requests using this cntxt
//				HttpGet httpGet = new HttpGet(oslcChangeSetReference.createURI());
//				try {
//					HttpResponse jfsResponse = jazzHttpClient.execute(httpGet, cntxt);
//					jfsResponse.getEntity().getContentLength();
//				} catch (ClientProtocolException e) {
//					// TODO Auto-generated catch block
//					e.printStackTrace();
//				} catch (IOException e) {
//					// TODO Auto-generated catch block
//					e.printStackTrace();
//				}


				OSLCResource resource = new OSLCResource(httpClient, oslcChangeSetReference.createURI(), null);
				resource.getProperty("name");

			}*/
	}
}
