package com.ibm.stg.rtc.ext.advisors;

import java.util.Collections;
import java.util.List;
import java.util.Set;

import org.eclipse.core.runtime.IProgressMonitor;

import com.ibm.stg.rtc.ext.common.ConfigurationHelper;
import com.ibm.stg.rtc.ext.common.ChildWorkItemTypeStates;
import com.ibm.stg.rtc.ext.common.ChildWorkItemTypeStatesHelper;
import com.ibm.stg.rtc.ext.common.WorkItemCommonTasks;
import com.ibm.stg.rtc.ext.common.WorkItemCondition;
import com.ibm.team.process.common.IProcessConfigurationElement;
import com.ibm.team.process.common.advice.AdvisableOperation;
import com.ibm.team.process.common.advice.IAdvisorInfo;
import com.ibm.team.process.common.advice.IAdvisorInfoCollector;
import com.ibm.team.process.common.advice.runtime.IOperationAdvisor;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.repository.service.AbstractService;
import com.ibm.team.workitem.common.ISaveParameter;
import com.ibm.team.workitem.common.IWorkItemCommon;
import com.ibm.team.workitem.common.model.IState;
import com.ibm.team.workitem.common.model.IWorkItem;
import com.ibm.team.workitem.common.model.Identifier;
import com.ibm.team.workitem.common.workflow.IWorkflowInfo;

public class RequireChildrenStatesAdvisor extends AbstractService implements IOperationAdvisor {

	private static final String ID = "com.ibm.stg.rtc.ext.advisors.RequireChildrenStatesAdvisor";

	public void run(AdvisableOperation operation, IProcessConfigurationElement advisorConfiguration, IAdvisorInfoCollector collector, IProgressMonitor monitor) throws TeamRepositoryException {
		WorkItemCommonTasks workItemCommonTasks = new WorkItemCommonTasks();
		ISaveParameter saveParameter = null;
		IWorkItem newWorkItemState = null;
		IWorkItemCommon workItemCommon = this.getService(IWorkItemCommon.class);

		Object tempObj = operation.getOperationData();
		if (workItemCommonTasks.isSaveParamterType(tempObj)) {
			saveParameter = (ISaveParameter) tempObj;
		} else {
			return;
		}

		tempObj = saveParameter.getNewState();
		if (workItemCommonTasks.isIWorkItemType(tempObj)) {
			newWorkItemState = (IWorkItem) tempObj;
		} else {
			return;
		}


		// first check if there are requirements on children of the work item being saved
		IProcessConfigurationElement typeConfiguration = ConfigurationHelper.getWITypeConfiguration(advisorConfiguration, newWorkItemState.getWorkItemType());
		Set<WorkItemCondition> conditions = ConfigurationHelper.getAllConditions(typeConfiguration);

		for(WorkItemCondition condition : conditions){
			if(condition.matches(saveParameter)){
				// get all the children of this type and make sure they are in one of the states
				List<IWorkItem> children = workItemCommonTasks.getAllChildWorkItems(saveParameter.getNewReferences(), workItemCommon.getAuditableCommon(), monitor);
				checkChildrenStates(newWorkItemState, children, condition, workItemCommon, collector, monitor);
			}
		}


		// now check if this work item is a child with requirements on it
		// see if there is a parent
		IWorkItem parent = workItemCommonTasks.getParentWorkItem(saveParameter.getNewReferences(), workItemCommon.getAuditableCommon(), monitor);
		if(parent != null){
			// get the required children conditions of the parent
			IProcessConfigurationElement parentTypeConfiguration = ConfigurationHelper.getWITypeConfiguration(advisorConfiguration, parent.getWorkItemType());
			Set<WorkItemCondition> parentTypeConditions = ConfigurationHelper.getAllConditions(parentTypeConfiguration);

			for(WorkItemCondition condition : parentTypeConditions){
				// see if the parent is in the condition
				String parentState = parent.getState2() == null ? null : parent.getState2().getStringIdentifier();
				String parentResolution = parent.getResolution2() == null ? null : parent.getResolution2().getStringIdentifier();
				if(condition.matches(null, parentState, parentState, parentResolution)){
					// the parent is in the condition, so check the child's state
					checkChildrenStates(parent, Collections.singletonList(newWorkItemState), condition, workItemCommon, collector, monitor);
				}
			}
		}

		// now check to see if there is an old parent that has been detached
		// by using the IWorkItem parameter instead of saveParameter.getNewReferences(), it will get the parent as in the database before the save is done
		IWorkItem oldParent = workItemCommonTasks.getParentWorkItem(newWorkItemState, workItemCommon.getAuditableCommon(), workItemCommon, monitor);
		if(oldParent != null && oldParent.sameItemId(parent) == false){
			// there is an old parent that we need to check

			// get the required children conditions of the parent
			IProcessConfigurationElement parentTypeConfiguration = ConfigurationHelper.getWITypeConfiguration(advisorConfiguration, oldParent.getWorkItemType());
			Set<WorkItemCondition> parentTypeConditions = ConfigurationHelper.getAllConditions(parentTypeConfiguration);

			for(WorkItemCondition condition : parentTypeConditions){
				// see if the parent is in the condition
				String parentState = oldParent.getState2() == null ? null : oldParent.getState2().getStringIdentifier();
				String parentResolution = oldParent.getResolution2() == null ? null : oldParent.getResolution2().getStringIdentifier();
				if(condition.matches(null, parentState, parentState, parentResolution)){
					// the parent is in the condition, so check the child's state
					checkChildrenStates(oldParent, Collections.singletonList(newWorkItemState), condition, workItemCommon, collector, monitor);
				}
			}
		}
	}

	private void checkChildrenStates(IWorkItem parent, List<IWorkItem> allChildren, WorkItemCondition condition, IWorkItemCommon workItemCommon, IAdvisorInfoCollector collector, IProgressMonitor monitor) throws TeamRepositoryException{
		// get the child types which are required
		Set<ChildWorkItemTypeStates> requiredChildTypes;
		try{
			requiredChildTypes = ChildWorkItemTypeStatesHelper.getChildTypesAndStates(condition.getConditionElement());
		}catch(TeamRepositoryException e){
			IAdvisorInfo info = collector.createProblemInfo("Invalid Configuration", e.getMessage(), ID);
			info.setProblemObject(parent.getItemHandle());
			collector.addInfo(info);
			return;
		}

		// loop through the types of the children
		for(ChildWorkItemTypeStates requiredChildType : requiredChildTypes){
			List<IWorkItem> children = ChildWorkItemTypeStatesHelper.getChildrenOfType(allChildren, requiredChildType.getChildType());
			if(requiredChildType.isAllowNoChildren() == false && children.size() == 0){
				// there should be children of this type, but none, so raise an error
				String typeName = workItemCommon.findWorkItemType(parent.getProjectArea(), parent.getWorkItemType(), monitor).getDisplayName();
				String childTypeName = workItemCommon.findWorkItemType(parent.getProjectArea(), requiredChildType.getChildType(), monitor).getDisplayName();
				IWorkflowInfo workFlowInfo = workItemCommon.findWorkflowInfo(parent, monitor);

				IAdvisorInfo info = collector.createProblemInfo("Child work item required", "A " + typeName + " may not " + condition.toString(workFlowInfo) + " if it has no children of type " + childTypeName, ID);
				info.setProblemObject(parent.getItemHandle());
				collector.addInfo(info);
			}
			for(IWorkItem child : children){
				// for each child of this type check if it is in a valid state
				String childState = child.getState2().getStringIdentifier();
				List<String> childStates = requiredChildType.getChildStates();
				if(childStates.size() != 0 && !childStates.contains(childState)){
					// the child has an invalid state so report the error
					String typeName = workItemCommon.findWorkItemType(parent.getProjectArea(), parent.getWorkItemType(), monitor).getDisplayName();
					String childTypeName = workItemCommon.findWorkItemType(child.getProjectArea(), requiredChildType.getChildType(), monitor).getDisplayName();
					IWorkflowInfo workFlowInfo = workItemCommon.findWorkflowInfo(parent, monitor);
					IWorkflowInfo childWorkFlowInfo = workItemCommon.findWorkflowInfo(child, monitor);
					String stateNames = getStateNames(childStates, childWorkFlowInfo);
					IAdvisorInfo info = collector.createProblemInfo("Invalid child state", "A " + typeName + " may not " + condition.toString(workFlowInfo) + " unless all children of type " + childTypeName + " are in " + stateNames, ID);
					info.setProblemObject(parent.getItemHandle());
					collector.addInfo(info);
				}
			}
		}
	}

	/**
	 * @param requiredStates
	 * @return
	 */
	private String getStateNames(List<String> requiredStates, IWorkflowInfo workFlowInfo) {
		String stateNames = "";
		for(int i = 0; i < requiredStates.size(); i++){
			stateNames = stateNames + workFlowInfo.getStateName(Identifier.create(IState.class, requiredStates.get(i)));
			// if there are more, add the ',' and 'or'
			if(i < requiredStates.size() - 1){
				if(requiredStates.size() > 2){
					stateNames = stateNames + ", ";
				}else{
					stateNames = stateNames + " ";
				}
				if(i == requiredStates.size() - 2){
					stateNames = stateNames + "or ";
				}
			}
		}
		return stateNames;
	}
}
