package com.ibm.stg.rtc.ext.advisors;

import java.util.ArrayList;
import java.util.List;

import org.eclipse.core.runtime.IProgressMonitor;

import com.ibm.stg.rtc.ext.common.WorkItemCommonTasks;
import com.ibm.team.process.common.IProcessConfigurationElement;
import com.ibm.team.process.common.advice.AdvisableOperation;
import com.ibm.team.process.common.advice.IAdvisorInfo;
import com.ibm.team.process.common.advice.IAdvisorInfoCollector;
import com.ibm.team.process.common.advice.runtime.IOperationAdvisor;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.repository.service.AbstractService;
import com.ibm.team.workitem.common.ISaveParameter;
import com.ibm.team.workitem.common.IWorkItemCommon;
import com.ibm.team.workitem.common.internal.workflow.WorkflowInfo;
import com.ibm.team.workitem.common.model.IWorkItem;
import com.ibm.team.workitem.common.model.Identifier;
import com.ibm.team.workitem.common.workflow.IWorkflowAction;
import com.ibm.team.workitem.common.workflow.IWorkflowInfo;

public class RequireCommentAdvisor extends AbstractService implements IOperationAdvisor {

	/* The following is an example of how to configure this precondition in the project's process specification
	 * First you must specify the work item types you want to require comments for. Then for each work item type, 
	 * a comment is required for each of the actions. If fromState elements are specified under an action, a comment
	 * is only required when the user is in one of those states and performs the parent action. If fromState elements
	 * are not specified, the action requires a comment from all states.
		<precondition id="com.ibm.stg.stgdefect.advisors.RequireCommentAdvisor" name="Require comment on some STG Defect actions" optional="false">
			<workItemType id="com.ibm.stg.workItemType.stgDefect">
				<workItemAction id="com.ibm.stg.workflow.stgDefect.action.reject"/>
				<workItemAction id="com.ibm.stg.workflow.stgDefect.action.moreinfo"/>
				<workItemAction id="com.ibm.stg.workflow.stgDefect.action.reopen">
					<fromState id="com.ibm.stg.workflow.stgDefect.state.rejected"/>
					<fromState id="com.ibm.stg.workflow.stgDefect.state.moreinfo"/>
				</workItemAction>
				<workItemAction id="com.ibm.stg.workflow.stgDefect.action.closed"/>
			</workItemType>
		</precondition>
	 */

	public static final String ID = "com.ibm.stg.stgdefect.advisors.RequireCommentAdvisor";

	private static final String TAG_WORKITEM_TYPE = "workItemType";
	private static final String TAG_ACTION = "workItemAction";
	private static final String TAG_FROM_STATE = "fromState";
	private static final String ATTR_ID = "id";

	/**
	 * Requires the user to enter a comment on any action on STG Defects only
	 */
	public void run(AdvisableOperation operation, IProcessConfigurationElement advisorConfiguration, IAdvisorInfoCollector collector, IProgressMonitor monitor) throws TeamRepositoryException {
		ISaveParameter saveParameter = null;
		IWorkItemCommon workItemCommon = this.getService(IWorkItemCommon.class);
		WorkItemCommonTasks workItemCommonTasks = new WorkItemCommonTasks();
		IWorkItem newWorkItemState = null;
		IWorkItem oldWorkItemState = null;



		Object tempObj = operation.getOperationData();
		if (workItemCommonTasks.isSaveParamterType(tempObj)) {
			saveParameter = (ISaveParameter) tempObj;
		} else {
			return;
		}

		tempObj = saveParameter.getNewState();
		if (workItemCommonTasks.isIWorkItemType(tempObj)) {
			newWorkItemState = (IWorkItem) tempObj;
		} else {
			return;
		}

		tempObj = saveParameter.getOldState();
		if (workItemCommonTasks.isIWorkItemType(tempObj)) {
			oldWorkItemState = (IWorkItem) tempObj;
		} else if(tempObj == null){
			oldWorkItemState = null;
		} else{
			return;
		}


		// check to see if the work item type has required comment configured, if not return
		String workItemType = newWorkItemState.getWorkItemType();
		IProcessConfigurationElement workItemTypeConfiguration = getWITypeConfiguration(advisorConfiguration, workItemType);
		if(workItemTypeConfiguration == null){
			return;
		}

		IWorkflowInfo workflowInfo = workItemCommon.findWorkflowInfo(newWorkItemState, monitor);
		Identifier<IWorkflowAction> actionId = workItemCommonTasks.getActionIdentifierFromString(workflowInfo, saveParameter.getWorkflowAction());

		if(actionId == null){
			// the user did not invoke an action so we do not require a comment so just return
			return;
		}

		List<String> requiredActionIds = getRequireCommentActions(workItemTypeConfiguration);

		if(requiredActionIds == null || requiredActionIds.size() == 0){
			return;
		}

		if(requiredActionIds.contains(actionId.getStringIdentifier())){
			List<String> fromStateIds = getFromStateIds(workItemTypeConfiguration, actionId.getStringIdentifier());

			if(fromStateIds.size() == 0 && oldWorkItemState == null){
				// a new work item is being created and a comment is required
				int newNumComments = newWorkItemState.getComments().getContents().length;

				if(newNumComments == 0){
					IAdvisorInfo info= collector.createProblemInfo("Comment Required", "A new comment is required when creating an STG Defect.", ID);
					info.setProblemObject(newWorkItemState.getItemHandle());
					//info.setData();
					collector.addInfo(info);
				}
			}
			else if(fromStateIds.size() == 0 || fromStateIds.contains(oldWorkItemState.getState2().getStringIdentifier())){
				// we are in an action where a comment is required
				int newNumComments = newWorkItemState.getComments().getContents().length;
				int oldNumComments = oldWorkItemState.getComments().getContents().length;

				if(newNumComments == oldNumComments){
					String actionName = workflowInfo.getActionName(actionId);
					String stateName = workflowInfo.getStateName(oldWorkItemState.getState2());
					IAdvisorInfo info= collector.createProblemInfo("Comment Required", "A new comment is required on an STG Defect \"" + actionName + "\" action from the \"" + stateName + "\" state.", ID);
					info.setProblemObject(newWorkItemState.getItemHandle());
					collector.addInfo(info);
				}
			}
		}
	}

	private IProcessConfigurationElement getWITypeConfiguration(IProcessConfigurationElement advisorConfiguration, String workItemType){
		IProcessConfigurationElement[] configurationElements = advisorConfiguration.getChildren();
		// loop through all the child xml elements for this precondition
		for (IProcessConfigurationElement configurationElement : configurationElements) {
			// if the element is a work item type element, add the id to the list
			if (configurationElement.getName().equals(TAG_WORKITEM_TYPE)) {
				String type = configurationElement.getAttribute(ATTR_ID);
				if(workItemType.equals(type)){
					return configurationElement;
				}
			}
		}
		return null;
	}

	/**
	 * Looks up the list of all the actions which a comment is required for
	 * in the project configuration xml and returns the string ids
	 * @param advisorConfiguration
	 * @return
	 */
	private List<String> getRequireCommentActions(IProcessConfigurationElement advisorConfiguration) {
		List<String> actionIds = new ArrayList<String>();

		IProcessConfigurationElement[] configurationElements = advisorConfiguration.getChildren();

		// loop through all the child xml elements for this precondition
		for (IProcessConfigurationElement configurationElement : configurationElements) {
			// if the element is an action element, add the id to the list
			if (configurationElement.getName().equals(TAG_ACTION)) {
				actionIds.add(configurationElement.getAttribute(ATTR_ID));
			}
		}
		return actionIds;
	}

	/**
	 * Looks up the states in which the given action will require a comment from in the 
	 * project configuration xml and returns the list of string ids
	 * @param advisorConfiguration
	 * @param actionIdentifier
	 * @return
	 */
	private List<String> getFromStateIds(IProcessConfigurationElement advisorConfiguration, String actionIdentifier) {
		List<String> stateIds = new ArrayList<String>();

		IProcessConfigurationElement[] configurationElements = advisorConfiguration.getChildren();

		// loop through all the child xml elements for this precondition
		for (IProcessConfigurationElement configurationElement : configurationElements) {
			// if the element is the correct action element, look for child fromState elements
			if (configurationElement.getName().equals(TAG_ACTION) && configurationElement.getAttribute(ATTR_ID).equals(actionIdentifier)) {
				IProcessConfigurationElement[] actionChildElements = configurationElement.getChildren();
				for(IProcessConfigurationElement actionChildElement : actionChildElements){
					// if the child element is a fromState element, add the id to the list
					if(actionChildElement.getName().equals(TAG_FROM_STATE)){
						stateIds.add(actionChildElement.getAttribute(ATTR_ID));
					}
				}
			}
		}

		// convert state ids to their legacy ids. if this is not done, a state id of "s2" will return the string
		// identifier of "2" which does not match the configured states. so this method removes the "s" prefix
		// to legacy states. see com.ibm.team.workitem.common.internal.workflow.WorkflowInfo
		List<String> convertedStateIds = new ArrayList<String>();
		for(String stateId : stateIds){
			convertedStateIds.add(WorkflowInfo.stripOffPrefix(stateId, 's'));
		}
		return convertedStateIds;
	}
}
