package com.ibm.stg.rtc.ext.advisors;

import java.util.HashMap;
import java.util.Map;

import org.eclipse.core.runtime.IProgressMonitor;

import com.ibm.team.process.common.IProcessArea;
import com.ibm.team.process.common.IProcessAreaHandle;
import com.ibm.team.process.common.IProcessConfigurationElement;
import com.ibm.team.process.common.advice.AdvisableOperation;
import com.ibm.team.process.common.advice.IAdvisorInfo;
import com.ibm.team.process.common.advice.IAdvisorInfoCollector;
import com.ibm.team.process.common.advice.runtime.IOperationAdvisor;
import com.ibm.team.repository.common.IAuditableHandle;
import com.ibm.team.repository.common.IContributorHandle;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.repository.common.UUID;
import com.ibm.team.repository.service.AbstractService;
import com.ibm.team.repository.service.IRepositoryItemService;
import com.ibm.team.scm.common.IComponent;
import com.ibm.team.scm.common.IComponentHandle;
import com.ibm.team.scm.common.IRepositoryProgressMonitor;
import com.ibm.team.scm.common.internal.IScmQueryService;
import com.ibm.team.scm.common.process.IChangeHistoryOperandDelta;
import com.ibm.team.scm.common.process.IScmDeltaSource;

/**
 * This precondition runs on change set deliver. It is similar to the built-in Restrict Change Set Delivery precondition
 * except more generic. It will only allow users to deliver change sets if they are a member of the team
 * owning the component. This precondition will not factor in roles, and it intentionally does not inherit 
 * team members from parent teams.
 * 
 * @author Kyle Christianson
 *
 */
public class RequireComponentOwnerToDeliver extends AbstractService implements IOperationAdvisor {

	public static final String ID = "com.ibm.stg.rtc.ext.advisors.RequireComponentOwnerToDeliver";

	public void run(AdvisableOperation operation, IProcessConfigurationElement advisorConfiguration, IAdvisorInfoCollector collector, IProgressMonitor monitor) throws TeamRepositoryException {
		IScmDeltaSource scmDeltaSource;

		Object tempObj = operation.getOperationData();
		if (tempObj instanceof IScmDeltaSource) {
			scmDeltaSource = (IScmDeltaSource) tempObj;
		} else {
			return;
		}

		IRepositoryItemService repositoryItemService = this.getService(IRepositoryItemService.class);
		IScmQueryService scmQueryService = this.getService(IScmQueryService.class);

		// get components being delivered to
		IComponentHandle[] components = getComponents(scmDeltaSource);

		// get the component owners
		IAuditableHandle[] componentOwners = scmQueryService.findComponentOwners(components, IRepositoryProgressMonitor.ITEM_FACTORY.createItem(monitor));

		// determine if this user is part of the owning team of each component
		IContributorHandle user = getAuthenticatedContributor();
		for(int i = 0; i < components.length; i++){
			checkComponentOwner(components[i], componentOwners[i], user, collector, repositoryItemService);
		}
	}

	private IComponentHandle[] getComponents(IScmDeltaSource scmDeltaSource){
		// use a map to ensure there are no duplicate components
		Map<UUID,IComponentHandle> components = new HashMap<UUID, IComponentHandle>();
		for (IChangeHistoryOperandDelta delta : scmDeltaSource.getDeltas(IChangeHistoryOperandDelta.class)) {
			for(IComponentHandle handle : delta.getComponents()){
				components.put(handle.getItemId(), handle);
			}
		}
		return components.values().toArray(new IComponentHandle[components.size()]);
	}

	private void checkComponentOwner(IComponentHandle componentHandle, IAuditableHandle componentOwner, IContributorHandle user, IAdvisorInfoCollector collector, IRepositoryItemService repositoryItemService) throws TeamRepositoryException{
		if(componentOwner instanceof IContributorHandle){
			// the owner is a contributor
			if(user.sameItemId(componentOwner) == false){
				IComponent component = (IComponent)repositoryItemService.fetchItem(componentHandle, IRepositoryItemService.COMPLETE);
				String message = "You are not authorized to deliver to the component " + component.getName() + " because you are not its owner";
				IAdvisorInfo info= collector.createProblemInfo("Unauthorized component", message, ID);
				collector.addInfo(info);
			}
		}else if(componentOwner instanceof IProcessAreaHandle){
			// the owner is a project or team
			IProcessArea processArea = (IProcessArea) repositoryItemService.fetchItem(componentOwner, IRepositoryItemService.COMPLETE);
			boolean found = false;
			for(IContributorHandle memberHandle : processArea.getMembers()){
				if(user.sameItemId(memberHandle)){
					found = true;
				}
			}
			if(found == false){
				IComponent component = (IComponent)repositoryItemService.fetchItem(componentHandle, IRepositoryItemService.COMPLETE);
				String message = "You are not authorized to deliver to the component " + component.getName() + " because you are not a member of its owning team area " + processArea.getName();
				IAdvisorInfo info= collector.createProblemInfo("Unauthorized component", message, ID);
				collector.addInfo(info);
			}
		}else{
			// this should not happen
			throw new TeamRepositoryException("componentOwner is not an instance of IContributorHandle or IProcessAreaHandle");
		}
	}
}