package com.ibm.stg.rtc.ext.advisors;

import org.eclipse.core.runtime.IProgressMonitor;

import com.ibm.stg.rtc.ext.common.ConfigurationHelper;
import com.ibm.stg.rtc.ext.common.WorkItemCommonTasks;
import com.ibm.team.process.common.IProcessConfigurationElement;
import com.ibm.team.process.common.advice.AdvisableOperation;
import com.ibm.team.process.common.advice.IAdvisorInfo;
import com.ibm.team.process.common.advice.IAdvisorInfoCollector;
import com.ibm.team.process.common.advice.runtime.IOperationAdvisor;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.repository.service.AbstractService;
import com.ibm.team.workitem.common.ISaveParameter;
import com.ibm.team.workitem.common.IWorkItemCommon;
import com.ibm.team.workitem.common.model.AttributeTypes;
import com.ibm.team.workitem.common.model.IAttribute;
import com.ibm.team.workitem.common.model.IWorkItem;

/**
 * This precondition checks a field for the correct format from the configured regular expression.
 * All the configuration tags below are required. The attribute id must exist and be a string type.
 * This precondition only supports one attribute on one work item type at a time. Therefor if you 
 * wish to require the format of multiple fields, you must configure this precondition multiple times
 * (once for every field).
 * 
 * Here is an example configuration:
 * 
 * 		<precondition id="com.ibm.stg.rtc.ext.advisors.RequireFieldFormatAdvisor" name="Require Field Format" optional="false">
			<workItemType id="com.ibm.stg.workItemType.stgDefect">
				<attribute id="com.ibm.stg.attribute.hwMachineType"/>
				<message value="Please enter 4 characters containing only letters and/or numbers"/>
				<regex value="^[A-Za-z0-9]{4}$|^$"/>
			</workItemType>
		</precondition>

 * @author Kyle Christianson
 *
 */
public class RequireFieldFormatAdvisor extends AbstractService implements IOperationAdvisor {

	public static final String ID = "com.ibm.stg.rtc.ext.stgdefect.advisors.RequireFieldFormatAdvisor";

	private static final String TAG_ATTRIBUTE = "attribute";
	private static final String TAG_MESSAGE = "message";
	private static final String TAG_REGEX = "regex";
	private static final String ATTR_VALUE = "value";
	private static final String ATTR_ID = "id";

	public void run(AdvisableOperation operation, IProcessConfigurationElement advisorConfiguration, IAdvisorInfoCollector collector, IProgressMonitor monitor) throws TeamRepositoryException {
		WorkItemCommonTasks workItemCommonTasks = new WorkItemCommonTasks();
		IWorkItemCommon workItemCommon = this.getService(IWorkItemCommon.class);
		ISaveParameter saveParameter = null;
		IWorkItem newWorkItemState = null;

		Object tempObj = operation.getOperationData();
		if (workItemCommonTasks.isSaveParamterType(tempObj)) {
			saveParameter = (ISaveParameter) tempObj;
		} else {
			return;
		}

		tempObj = saveParameter.getNewState();
		if (workItemCommonTasks.isIWorkItemType(tempObj)) {
			newWorkItemState = (IWorkItem) tempObj;
		} else {
			return;
		}

		// check to see if this is the correct work item type
		IProcessConfigurationElement workItemTypeConfig = ConfigurationHelper.getWITypeConfiguration(advisorConfiguration, newWorkItemState.getWorkItemType());

		if(workItemTypeConfig == null){
			return;
		}

		// get and validate the configuration

		String attributeId = getAttribute(workItemTypeConfig, TAG_ATTRIBUTE, ATTR_ID, collector);
		String message = getAttribute(workItemTypeConfig, TAG_MESSAGE, ATTR_VALUE, collector);
		String regex = getAttribute(workItemTypeConfig, TAG_REGEX, ATTR_VALUE, collector);

		if(attributeId != null && message != null && regex != null){
			IAttribute attribute = workItemCommonTasks.getAttribute(attributeId, newWorkItemState.getProjectArea(), workItemCommon, monitor);
			// make sure the attribute exists
			if(attribute == null){
				IAdvisorInfo info = collector.createProblemInfo("Configuration Problem", "RequireFieldFormat Configuration Problem: The configured attribute " + attributeId + " does not exist in the project.", ID);
				info.setProblemObject(newWorkItemState.getItemHandle());
				collector.addInfo(info);
			}else{
				// make sure the attribute is a string type
				if(AttributeTypes.STRING_TYPES.contains(attribute.getAttributeType()) == false){
					IAdvisorInfo info = collector.createProblemInfo("Configuration Problem", "RequireFieldFormat Configuration Problem: The configured attribute " + attributeId + " is not a string type.", ID);
					info.setProblemObject(newWorkItemState.getItemHandle());
					collector.addInfo(info);
				}else{
					String value = (String) newWorkItemState.getValue(attribute);
					// check to see if the value matches the regular expression
					if(value.matches(regex) == false){
						String attributeDisplayName = attribute.getDisplayName();
						IAdvisorInfo info = collector.createProblemInfo("Invalid field format", "The " + attributeDisplayName + " field has an invalid format: " + message, ID);
						info.setProblemObject(newWorkItemState.getItemHandle());
						collector.addInfo(info);
					}
				}
			}
		}
	}

	private String getAttribute(IProcessConfigurationElement configuration, String elementTag, String attributeName, IAdvisorInfoCollector collector){
		String attributeValue = null;
		IProcessConfigurationElement element = ConfigurationHelper.getConfigurationElementByTagName(configuration, elementTag);
		if(element != null){
			attributeValue = element.getAttribute(attributeName);
		}

		if(attributeValue == null){
			IAdvisorInfo info = collector.createProblemInfo("Configuration Problem", "RequireFieldFormat Configuration Problem: The tag <" + elementTag + "> must be configured with the \"" + attributeName + "\" attribute.", ID);
			collector.addInfo(info);
		}
		return attributeValue;
	}
}
