package com.ibm.stg.rtc.ext.advisors;

import java.util.List;

import org.eclipse.core.runtime.IProgressMonitor;

import com.ibm.stg.rtc.ext.common.ConfigurationHelper;
import com.ibm.stg.rtc.ext.common.WorkItemCommonTasks;
import com.ibm.team.process.common.IProcessConfigurationElement;
import com.ibm.team.process.common.advice.AdvisableOperation;
import com.ibm.team.process.common.advice.IAdvisorInfoCollector;
import com.ibm.team.process.common.advice.IProcessReport;
import com.ibm.team.process.common.advice.IReportInfo;
import com.ibm.team.process.common.advice.runtime.IOperationAdvisor;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.repository.service.AbstractService;
import com.ibm.team.workitem.common.ISaveParameter;
import com.ibm.team.workitem.common.IWorkItemCommon;
import com.ibm.team.workitem.common.internal.workflow.WorkflowInfo;
import com.ibm.team.workitem.common.model.IResolution;
import com.ibm.team.workitem.common.model.IWorkItem;
import com.ibm.team.workitem.common.model.Identifier;
import com.ibm.team.workitem.common.workflow.IWorkflowAction;
import com.ibm.team.workitem.common.workflow.IWorkflowInfo;

/**
 * This precondition forces the user to make a resolution change. It can be configured for
 * certain work item types and states. The unassigned resolution is also set in the configuration.
 * 
 * 		<precondition id="com.ibm.stg.rtc.ext.advisors.RequireResolutionAdvisor" name="STG: Require Resolution" optional="false">
			<workItemType id="com.ibm.stg.workItemType.stgDefect">
				<unassignedResolution id="com.ibm.stg.workflow.stgDefect.resolution.-unassigned"/>
				<workItemState id="com.ibm.stg.workflow.stgDefect.state.working"/>
				<workItemState id="com.ibm.stg.workflow.stgDefect.state.rejected"/>
			</workItemType>
		</precondition>
 * 
 * @author Kyle Christianson
 *
 */
public class RequireResolutionAdvisor extends AbstractService implements IOperationAdvisor {

	private static final String TAG_UNASSIGNED_RESOLUTION = "unassignedResolution";
	private static final String ATTR_ID = "id";

	private static final String ID = "com.ibm.stg.rtc.ext.advisors.RequireResolutionAdvisor";

	public void run(AdvisableOperation operation, IProcessConfigurationElement advisorConfiguration, IAdvisorInfoCollector collector, IProgressMonitor monitor) throws TeamRepositoryException {
		WorkItemCommonTasks workItemCommonTasks = new WorkItemCommonTasks();
		ISaveParameter saveParameter = null;
		IWorkItem newWorkItemState = null;

		Object tempObj = operation.getOperationData();
		if (workItemCommonTasks.isSaveParamterType(tempObj)) {
			saveParameter = (ISaveParameter) tempObj;
		} else {
			return;
		}

		tempObj = saveParameter.getNewState();
		if (workItemCommonTasks.isIWorkItemType(tempObj)) {
			newWorkItemState = (IWorkItem) tempObj;
		} else {
			return;
		}

		if(saveParameter.getWorkflowAction() == null){
			// no action so the resolution cannot be changed
			return;
		}

		// check to see if the work item type has this advisor configured, if not return
		List<String> states = ConfigurationHelper.getWIStatesConfigured(advisorConfiguration, newWorkItemState.getWorkItemType());
		if(states == null){
			// this work item type was not configured
			return;
		}

		if(states.size() > 0 && !states.contains(newWorkItemState.getState2().getStringIdentifier())){
			// this state was not configured
			return;
		}

		IProcessConfigurationElement workItemTypeElement = ConfigurationHelper.getWITypeConfiguration(advisorConfiguration, newWorkItemState.getWorkItemType());
		String unassignedResolution = getUnassignedResolution(workItemTypeElement);
		if(unassignedResolution == null){
			// display warning and exit
			IReportInfo info = collector.createProblemInfo("Invalid Configuration", "The unassigned resolution was not found or is invalid", ID);
			info.setSeverity(IProcessReport.WARNING);
			collector.addInfo(info);
			return;
		}

		IWorkItemCommon workItemCommon = this.getService(IWorkItemCommon.class);
		IWorkflowInfo workflowInfo = workItemCommon.findWorkflowInfo(newWorkItemState, monitor);
		Identifier<IResolution>[] availableResolutions = workflowInfo.getResolutionIds(Identifier.create(IWorkflowAction.class, saveParameter.getWorkflowAction()));
		if(availableResolutions.length == 0){
			// there are no resolutions to set return
			return;
		}

		Identifier<IResolution> currentResolutionId = newWorkItemState.getResolution2();
		String currentResolution = (currentResolutionId == null) ? null : currentResolutionId.getStringIdentifier();

		if(unassignedResolution.equals(currentResolution)){
			// the user left the resolution as the unassigned resolution
			String resolutionName = workItemCommon.findWorkflowInfo(newWorkItemState, monitor).getResolutionName(Identifier.create(IResolution.class, currentResolution)); 
			IReportInfo info = collector.createProblemInfo("Resolution must be set", "The resolution must be set to something other than " + resolutionName, ID);
			info.setProblemObject(newWorkItemState.getItemHandle());
			collector.addInfo(info);
		}
	}



	private String getUnassignedResolution(IProcessConfigurationElement configuration){
		IProcessConfigurationElement[] children = configuration.getChildren();
		for(IProcessConfigurationElement child : children){
			if(TAG_UNASSIGNED_RESOLUTION.equals(child.getName())){
				String resolutionId = child.getAttribute(ATTR_ID);
				// convert resolution ids to their legacy ids. if this is not done, a resolution id of "r2" will return the string
				// identifier of "2" which does not match the configured states. so this method removes the "r" prefix
				// to legacy states. see com.ibm.team.workitem.common.internal.workflow.WorkflowInfo
				return WorkflowInfo.stripOffPrefix(resolutionId, 'r');
			}
		}
		return null;
	}

}
