package com.ibm.stg.rtc.ext.advisors;

import java.util.Iterator;
import java.util.List;
import java.util.Set;

import org.eclipse.core.runtime.IProgressMonitor;

import com.ibm.stg.rtc.ext.common.ApprovalContainer;
import com.ibm.stg.rtc.ext.common.ApprovalWorkItemCondition;
import com.ibm.stg.rtc.ext.common.ApprovalsHelper;
import com.ibm.stg.rtc.ext.common.WorkItemCommonTasks;
import com.ibm.stg.rtc.ext.common.WorkItemCondition;
import com.ibm.team.process.common.IProcessConfigurationElement;
import com.ibm.team.process.common.advice.AdvisableOperation;
import com.ibm.team.process.common.advice.IAdvisorInfo;
import com.ibm.team.process.common.advice.IAdvisorInfoCollector;
import com.ibm.team.process.common.advice.runtime.IOperationAdvisor;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.repository.service.AbstractService;
import com.ibm.team.workitem.common.ISaveParameter;
import com.ibm.team.workitem.common.IWorkItemCommon;
import com.ibm.team.workitem.common.model.IApprovalDescriptor;
import com.ibm.team.workitem.common.model.IWorkItem;
import com.ibm.team.workitem.common.model.Identifier;
import com.ibm.team.workitem.common.model.WorkItemApprovals;
import com.ibm.team.workitem.common.workflow.IWorkflowAction;
import com.ibm.team.workitem.common.workflow.IWorkflowInfo;

public class RequireSTGApprovalsAdvisor extends AbstractService implements IOperationAdvisor {

	public static final String ID = "com.ibm.stg.rtc.ext.advisors.RequireSTGApprovalsAdvisor";

	public void run(AdvisableOperation operation, IProcessConfigurationElement advisorConfiguration, IAdvisorInfoCollector collector, IProgressMonitor monitor) throws TeamRepositoryException {
		WorkItemCommonTasks workItemCommonTasks = new WorkItemCommonTasks();
		ISaveParameter saveParameter = null;
		IWorkItem newWorkItemState = null;
		IWorkItem oldWorkItemState = null;
		IWorkItemCommon workItemCommon = this.getService(IWorkItemCommon.class);

		Object tempObj = operation.getOperationData();
		if (workItemCommonTasks.isSaveParamterType(tempObj)) {
			saveParameter = (ISaveParameter) tempObj;
		} else {
			return;
		}

		tempObj = saveParameter.getNewState();
		if (workItemCommonTasks.isIWorkItemType(tempObj)) {
			newWorkItemState = (IWorkItem) tempObj;
		} else {
			return;
		}

		tempObj = saveParameter.getOldState();
		if (workItemCommonTasks.isIWorkItemType(tempObj)) {
			oldWorkItemState = (IWorkItem) tempObj;
		} else if(tempObj == null){
			oldWorkItemState = null;
		} else{
			return;
		}

		List<ApprovalContainer> approvals;
		try{
			approvals = ApprovalsHelper.getConfiguredApprovals(advisorConfiguration, saveParameter, newWorkItemState, workItemCommon);
		}catch(Exception e){
			IAdvisorInfo info = collector.createProblemInfo("Invalid advisor configuration", e.getMessage(), ID);
			collector.addInfo(info);
			return;
		}

		for(ApprovalContainer approval : approvals){
			for(ApprovalWorkItemCondition condition : approval.getConditionConfigs(saveParameter)){
				Set<String> requiredApprovalStates = condition.getRequiredApprovalStates();
				if(requiredApprovalStates.size() > 0){
					// an approval state is required so see if the work item has this approval in one of the required states
					String approvalName = approval.getName();
					IApprovalDescriptor approvalDescriptor = ApprovalsHelper.findApprovalDescriptorByName(newWorkItemState, approvalName);
					if(approvalDescriptor == null){

						IWorkflowInfo workflow = workItemCommon.findWorkflowInfo(newWorkItemState, monitor);
						String description = "One or more approvals are required to " + condition.toString(workflow) + ".";
						// see what states the approval should be created in
						Set<ApprovalWorkItemCondition> createConditions = approval.getCreateConditions();
						if(createConditions.size() == 0){
							description = description + " There are no configured conditions to create this approval.";
						}else if(approvalShouldExist(oldWorkItemState, createConditions)){
							description = description + " Try saving the work item without performing an action to create the approval.";
						}else{
							String action = getActionToCreateApproval(oldWorkItemState, createConditions, workflow);
							if(action != null){
								description = description + " Try performing the " + action + " action to create the approval.";
							}else{
								description = description + "To create the approval the work item must " + createConditions.iterator().next().toString(workflow);
							}
						}
						IAdvisorInfo info = collector.createProblemInfo("Approval Required", description, ID);
						info.setProblemObject(newWorkItemState.getItemHandle());
						collector.addInfo(info);
					}else if(!requiredApprovalStates.contains(approvalDescriptor.getCumulativeStateIdentifier())){
						// the work item does not have the approval in one of the required state
						StringBuilder sb = new StringBuilder();
						Iterator<String> i = requiredApprovalStates.iterator();
						while(i.hasNext()){
							sb.append(WorkItemApprovals.getState(i.next()).getDisplayName());
							if(i.hasNext()){
								sb.append(" or ");
							}
						}
						String requiredStateName = sb.toString();
						IWorkflowInfo workflow = workItemCommon.findWorkflowInfo(newWorkItemState, monitor);
						String description = "The approval '" + approvalName + "' must be " + requiredStateName + " to " + condition.toString(workflow);
						IAdvisorInfo info = collector.createProblemInfo("Approval Required", description, ID);
						info.setProblemObject(newWorkItemState.getItemHandle());
						collector.addInfo(info);
					}
				}
			}
		}
	}

	/**
	 * @param oldWorkItemState
	 * @param createConditions
	 * @param workflow
	 * @return
	 */
	private String getActionToCreateApproval(IWorkItem workItem, Set<ApprovalWorkItemCondition> createConditions, IWorkflowInfo workflow) {
		if(workItem == null){
			return null;
		}
		String fromState = workItem.getState2().getStringIdentifier();
		Identifier<IWorkflowAction>[] availableActions = workflow.getActionIds(workItem.getState2());

		for(WorkItemCondition condition : createConditions){
			// check to see if any of the available actions match
			for(Identifier<IWorkflowAction> action : availableActions){
				String resultState = workflow.getActionResultState(action).getStringIdentifier();
				// action matches
				if(condition.matches(action.getStringIdentifier(), resultState, fromState, null)){
					return workflow.getActionName(action);
				}
				// result state matches
				if(condition.matches(null, resultState, null, null)){
					return workflow.getActionName(action);
				}
			}
		}
		return null;
	}

	/**
	 * returns true if the work item passed in matches one of the conditions
	 */
	private boolean approvalShouldExist(IWorkItem workItem, Set<ApprovalWorkItemCondition> createConditions) {
		if(workItem == null){
			return false;
		}
		for(WorkItemCondition condition : createConditions){
			String state = workItem.getState2().getStringIdentifier();
			String resolution = workItem.getResolution2() == null ? null : workItem.getResolution2().getStringIdentifier();
			if(condition.matches(null, state, null, resolution)){
				return true;
			}
		}
		return false;
	}
}
