package com.ibm.stg.rtc.ext.advisors;

import org.eclipse.core.runtime.IProgressMonitor;

import com.ibm.stg.rtc.ext.common.ConfigurationHelper;
import com.ibm.stg.rtc.ext.common.STGCommonIDs;
import com.ibm.stg.rtc.ext.common.WorkItemCommonTasks;
import com.ibm.team.process.common.IProcessConfigurationElement;
import com.ibm.team.process.common.advice.AdvisableOperation;
import com.ibm.team.process.common.advice.IAdvisorInfoCollector;
import com.ibm.team.process.common.advice.IReportInfo;
import com.ibm.team.process.common.advice.runtime.IOperationAdvisor;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.repository.service.AbstractService;
import com.ibm.team.workitem.common.IAuditableCommon;
import com.ibm.team.workitem.common.ISaveParameter;
import com.ibm.team.workitem.common.IWorkItemCommon;
import com.ibm.team.workitem.common.model.IAttribute;
import com.ibm.team.workitem.common.model.IWorkItem;

public class UniversalIDAdvisor extends AbstractService implements IOperationAdvisor {
	/* This advisor (preconditions) calculates and adds the universal id to every work item.
	 * Note that advisors usually should not modify a work item, however this is a special 
	 * case because the universal id is only set once on create and we need to avoid sending
	 * multiple notifications. The calculation is done by converting the work item id to 
	 * base 31 and padding the number with '0' into a 6 character string. Then the prefix
	 * is added to the front. Six digits of base 31 gives us 887,503,681 ids. We will have
	 * to come up with a solution to change the prefix when we approach this max. 
	 * 
	 * 
	 * The following is an example of how to configure this precondition in the project's process specification
		<precondition id="com.ibm.stg.rtc.ext.advisors.UniversalIDAdvisor" name="Universal ID" optional="false">
			<forceNewUniversalIdOnCreate value="true"/>
			<universalID repositoryURI="https://jazz06.rchland.ibm.com:9443" prefix="AB"/>
			<universalID repositoryURI="https://jazz07.rchland.ibm.com:9443" prefix="BC"/>
			<universalID repositoryURI="https://jazz07.rchland.ibm.com:10443" prefix="CD"/>
			<workItemType id="com.ibm.stg.workItemType.stgDefect"/>
			<workItemType id="defect"/>
			<workItemType id="task"/>
			<workItemType id="enhancement"/>
			<workItemType id="com.ibm.team.apt.workItemType.story"/>
			<workItemType id="com.ibm.team.apt.workItemType.epic"/>
			<workItemType id="com.ibm.team.workItemType.buildtrackingitem"/>
			<workItemType id="com.ibm.team.workitem.workItemType.impediment"/>
			<workItemType id="com.ibm.team.workItemType.adoption"/>
			<workItemType id="com.ibm.team.workitem.workItemType.retrospective"/>
		</precondition>
	 */

	public static final String ID = "com.ibm.stg.rtc.ext.advisors.UniversalIDAdvisor";

	private static final String TAG_UNIV_ID = "universalID";
	private static final String TAG_FORCE_NEW_UNIVERSAL_ID_ON_CREATE = "forceNewUniversalIdOnCreate";
	private static final String ATTR_PRIFIX = "prefix";

	private static final String ENCODE = "0123456789BCDFGHJKLMNPQRSTVWXYZ";
	private static int BASE;
	private static final int UID_LENGTH = 6;

	static{
		BASE = ENCODE.length();
	}

	public void run(AdvisableOperation operation, IProcessConfigurationElement advisorConfiguration, IAdvisorInfoCollector collector, IProgressMonitor monitor) throws TeamRepositoryException {
		IWorkItemCommon workItemCommon = this.getService(IWorkItemCommon.class);
		WorkItemCommonTasks workItemCommonTasks = new WorkItemCommonTasks();
		ISaveParameter saveParameter = null;
		IWorkItem newWorkItemState = null;

		Object tempObj = operation.getOperationData();
		if (workItemCommonTasks.isSaveParamterType(tempObj)) {
			saveParameter = (ISaveParameter) tempObj;
		} else {
			return;
		}

		tempObj = saveParameter.getNewState();
		if (workItemCommonTasks.isIWorkItemType(tempObj)) {
			newWorkItemState = (IWorkItem) tempObj;
		} else {
			return;
		}

		// check to see if the work item type has this advisor configured, if not return
		if(ConfigurationHelper.getWITypeConfiguration(advisorConfiguration, newWorkItemState.getWorkItemType()) == null){
			// this work item type was not configured
			return;
		}

		IAttribute univIdAttr = workItemCommonTasks.getAttribute(STGCommonIDs.STG_ATTR_UNIVERSAL_ID, newWorkItemState.getProjectArea(), workItemCommon, monitor);

		boolean universalIdIsNull = !workItemCommonTasks.attributeHasNonNullValue(newWorkItemState, univIdAttr, workItemCommon.getAuditableCommon());

		// check to see if we need to force the universal id creation
		if(!universalIdIsNull && saveParameter.isCreation()){
			// a universal id is set, and this is a new work item, so check the isForceNewId flag
			// if TAG_FORCE_NEW_ID does not exist, this will default to false for backward compatibility
			boolean isForceNewUniversalIdOnCreate = ConfigurationHelper.getBooleanTagValue(advisorConfiguration, TAG_FORCE_NEW_UNIVERSAL_ID_ON_CREATE);
			if(isForceNewUniversalIdOnCreate){
				// setting universalIdIsNull to true will force the regeneration and setting of the universal id
				universalIdIsNull = true;
			}
		}

		// check to see if the work item has the universal id attribute set
		if(universalIdIsNull){
			// build the universal id from the prefix and work item id
			String univId;
			try{
				univId = buildUniversalID(advisorConfiguration, workItemCommon.getAuditableCommon(), newWorkItemState.getId());
			}catch(Exception e){
				IReportInfo info = collector.createProblemInfo("Universal ID creation failed", e.getMessage(), ID);
				collector.addInfo(info);
				return;
			}

			// if the attribute doesn't exist we need to add it
			if(!newWorkItemState.hasAttribute(univIdAttr)){
				newWorkItemState.addCustomAttribute(univIdAttr);
			}

			// set the value
			newWorkItemState.setValue(univIdAttr, univId);
		}
	}

	/**
	 * Gets the prefix from the process config and formats the universal id.
	 * Throws an exception if there is a problem getting the prefix.
	 * @param configuration
	 * @param workItemId
	 * @return
	 * @throws Exception
	 */
	private String buildUniversalID(IProcessConfigurationElement configuration, IAuditableCommon auditableCommon, int workItemId) throws Exception{

		String prefix = getUniversalIDPrefix(configuration, auditableCommon);

		String univId = prefix + padRight(encode(workItemId), UID_LENGTH, '0');
		return univId;
	}



	/**
	 * Returns the prefix specified for the universal ID.
	 * This prefix is configurable in the project's process specification xml (see example above)
	 * @param configuration
	 * @return
	 * @throws Exception 
	 */
	private String getUniversalIDPrefix(IProcessConfigurationElement configuration, IAuditableCommon auditableCommon) throws Exception {
		String prefix = null;



		IProcessConfigurationElement[] configurationElements = configuration.getChildren();

		// loop through all the child xml elements for this precondition
		for (IProcessConfigurationElement configurationElement : configurationElements) {
			// if the element is the UniversalIDPrefix element, get the value
			if (configurationElement.getName().equals(TAG_UNIV_ID)) {
				// if it is the correct repository
				if(ConfigurationHelper.hasThisRepositoryUriAttribute(configurationElement, auditableCommon)){
					prefix = configurationElement.getAttribute(ATTR_PRIFIX);
					break;
				}
			}
		}

		if(prefix == null || prefix.length() == 0){
			throw new Exception("The universal ID could not be created because the prefix was not specified correctly in the process config. <universalID repositoryURI=\"https://localhost:9443\" prefix=\"AB\"/>");
		}

		if(prefix.length() != 2){
			throw new Exception("The universal ID could not be created because the prefix was not in the correct format. The prefix must be exactly two characters.");
		}

		return prefix;

	}

	private String encode(long id) {
		if(id < 0){
			throw new IllegalArgumentException("Cannot ENCODE a negative id.");
		}
		StringBuffer encodeBuf = new StringBuffer();
		encodeBuf.setLength(0);
		long quotient  = id / BASE; 
		int remainder  = (int)(id % BASE);
		boolean quit  = false;
		while (!quit) {
			encodeBuf.insert(0, ENCODE.charAt(remainder));
			if (quotient < BASE) {
				encodeBuf.insert(0,ENCODE.charAt((int)quotient));
				quit = true;
			}
			remainder = (int)(quotient % BASE);
			quotient  = quotient / BASE;
		}
		return encodeBuf.toString();
	}     

	private String padRight(String str, int length, char padChar)
	{
		if (str.length() < length)
		{
			StringBuffer buf = new StringBuffer(str);
			for (int i = str.length(); i < length; i++)
				buf.insert(0, padChar);
			str = buf.toString();
		}
		return str;
	}
}
