package com.ibm.stg.rtc.ext.participants;

import java.util.ArrayList;
import java.util.List;

import org.eclipse.core.runtime.IProgressMonitor;

import com.ibm.stg.rtc.ext.common.ApprovalContainer;
import com.ibm.stg.rtc.ext.common.ApprovalWorkItemCondition;
import com.ibm.stg.rtc.ext.common.ApprovalsHelper;
import com.ibm.stg.rtc.ext.common.ConfigurationHelper;
import com.ibm.stg.rtc.ext.common.RepositoryCommonTasks;
import com.ibm.stg.rtc.ext.common.WorkItemCommonTasks;
import com.ibm.team.foundation.common.text.XMLString;
import com.ibm.team.process.common.IProcessConfigurationElement;
import com.ibm.team.process.common.advice.AdvisableOperation;
import com.ibm.team.process.common.advice.IProcessReport;
import com.ibm.team.process.common.advice.IReportInfo;
import com.ibm.team.process.common.advice.TeamOperationCanceledException;
import com.ibm.team.process.common.advice.runtime.IOperationParticipant;
import com.ibm.team.process.common.advice.runtime.IParticipantInfoCollector;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.repository.service.AbstractService;
import com.ibm.team.workitem.common.ISaveParameter;
import com.ibm.team.workitem.common.IWorkItemCommon;
import com.ibm.team.workitem.common.internal.WorkflowUtil;
import com.ibm.team.workitem.common.model.IApprovalDescriptor;
import com.ibm.team.workitem.common.model.IComment;
import com.ibm.team.workitem.common.model.IComments;
import com.ibm.team.workitem.common.model.IWorkItem;
import com.ibm.team.workitem.common.model.Identifier;
import com.ibm.team.workitem.common.model.WorkItemApprovals;
import com.ibm.team.workitem.common.workflow.IWorkflowAction;
import com.ibm.team.workitem.common.workflow.IWorkflowInfo;
import com.ibm.team.workitem.service.IWorkItemServer;

@SuppressWarnings("restriction")
public class PerformActionFromApprovalsParticipant extends AbstractService implements IOperationParticipant {

	public static final String ID = "com.ibm.stg.rtc.ext.participants.PerformActionFromApprovalsParticipant";
	public static final String NAME = "STG: Perform Action From Approvals";

	public void run(AdvisableOperation operation, IProcessConfigurationElement participantConfig, IParticipantInfoCollector collector, IProgressMonitor monitor) throws TeamRepositoryException {
		IWorkItemServer workItemServer = this.getService(IWorkItemServer.class);
		IWorkItemCommon workItemCommon = this.getService(IWorkItemCommon.class);
		WorkItemCommonTasks workItemCommonTasks = new WorkItemCommonTasks();
		ISaveParameter saveParameter = null;
		IWorkItem newWorkItemState = null;
		IWorkItem oldWorkItemState = null;

		Object tempObj = operation.getOperationData();
		if (workItemCommonTasks.isSaveParamterType(tempObj)) {
			saveParameter = (ISaveParameter) tempObj;
		} else {
			return;
		}

		tempObj = saveParameter.getNewState();
		if (workItemCommonTasks.isIWorkItemType(tempObj)) {
			newWorkItemState = (IWorkItem) tempObj;
		} else {
			return;
		}

		tempObj = saveParameter.getOldState();
		if (workItemCommonTasks.isIWorkItemType(tempObj)) {
			oldWorkItemState = (IWorkItem) tempObj;
		} else if(tempObj == null){
			oldWorkItemState = null;
		} else{
			return;
		}

		// no approval changes if this is a new work item
		if(saveParameter.isCreation()){
			return;
		}

		// don't perform an action if an action is already being performed
		if(saveParameter.getWorkflowAction() != null){
			return;
		}

		List<IApprovalDescriptor> changedStateApprovals = findChangedStateApprovals(newWorkItemState, oldWorkItemState);

		// if no approvals have changed states, return immediately
		if(changedStateApprovals.size() == 0){
			return;
		}
		IWorkflowInfo workflowInfo = workItemServer.findWorkflowInfo(newWorkItemState, monitor);
		String actionToPerform = null;
		String approvalName = null;
		String approvalStateName = null;
		boolean showWarnings = ConfigurationHelper.getBooleanTagValue(participantConfig, ConfigurationHelper.TAG_SHOW_WARNINGS);

		for(IApprovalDescriptor approvalDescriptor : changedStateApprovals){
			ApprovalContainer approvalContainer = ApprovalsHelper.getConfiguredApproval(participantConfig, newWorkItemState, approvalDescriptor.getName(), workItemCommon);
			if(approvalContainer != null){
				for(ApprovalWorkItemCondition condition : approvalContainer.getConditionConfigs(saveParameter)){
					if(condition.hasAutoAction(approvalDescriptor.getCumulativeStateIdentifier())){
						String action = condition.getAutoAction(approvalDescriptor.getCumulativeStateIdentifier());
						if(WorkflowUtil.validState(workflowInfo, newWorkItemState, action)){
							// there is a valid action configured
							if(actionToPerform != null && !actionToPerform.equals(action)){
								// more than one action was found to be performed so don't perform any
								if(showWarnings){
									IReportInfo info = collector.createInfo("Only one action can be performed", "Multiple actions are configured to be performed for the approval changes you just made, therefore no action has been performed.");
									info.setSeverity(IProcessReport.WARNING);
									collector.addInfo(info);
								}
								return;
							} else{
								actionToPerform = action;
								approvalName = approvalDescriptor.getName();
								approvalStateName = WorkItemApprovals.getState(approvalDescriptor.getCumulativeStateIdentifier()).getDisplayName();
							}
						}else if(showWarnings){
							String actionName = workflowInfo.getActionName(Identifier.create(IWorkflowAction.class, action));
							String message;
							if(actionName == null){
								message = "The configured action id '" + action + "' does not exist in the current workflow.";
							} else{
								String stateName = WorkItemApprovals.getState(approvalDescriptor.getCumulativeStateIdentifier()).getDisplayName();
								message = "The '" + actionName + "' action is configured to be run when the '" + approvalDescriptor.getName() + "' approval is '" + stateName + "', however this action does not exist for the current work item state."; 
							}
							IReportInfo info = collector.createInfo("Invalid Action", message);
							info.setSeverity(IProcessReport.WARNING);
							collector.addInfo(info);
						}
					}
				}
			}
		}

		if(actionToPerform != null){
			// a single action was found so perform that action
			// get a fresh working copy
			IWorkItem workItemWC = (IWorkItem)workItemServer.findWorkItemById(newWorkItemState.getId(), IWorkItem.FULL_PROFILE, null).getWorkingCopy();
			// add a comment
			String actionName = workflowInfo.getActionName(Identifier.create(IWorkflowAction.class, actionToPerform));
			String commentStr = "The action " + actionName + " was automatically performed because the approval '" + approvalName + "' was " + approvalStateName; //following condition was met: " + actionToPerformCondition.toString(workflowInfo);
			IComments comments = workItemWC.getComments();
			IComment newComment = comments.createComment(workItemServer.getAuditableCommon().getUser(), XMLString.createFromPlainText(commentStr));
			comments.append(newComment);
			try {
				workItemServer.saveWorkItem2(workItemWC, null, actionToPerform);
			} catch (TeamOperationCanceledException e) {
				RepositoryCommonTasks repoCommon = new RepositoryCommonTasks();
				if(showWarnings){
					IReportInfo info = collector.createInfo("Follow-up action failed", "The follow-up could not perform the action '" + actionName + "' because a precondition failed with the following message: \n" + repoCommon.getFirstErrorDescription(e));
					info.setProblemObject(workItemWC.getItemHandle());
					info.setSeverity(IProcessReport.WARNING);
					collector.addInfo(info);
				}
				// suppress the follow-up action save errors so this work item save can continue
				repoCommon.suppressAllErrors(NAME, e);
			}
		}
	}

	/**
	 * returns all auto generated approvals that are new or have changed states
	 * @param newWorkItem
	 * @param oldWorkItem
	 * @return
	 */
	private List<IApprovalDescriptor> findChangedStateApprovals(IWorkItem newWorkItem, IWorkItem oldWorkItem){
		List<IApprovalDescriptor> changed = new ArrayList<IApprovalDescriptor>();

		List<IApprovalDescriptor> generatedApprovals = ApprovalsHelper.getAllAutoGeneratedApprovals(newWorkItem);
		for(IApprovalDescriptor newDescriptor : generatedApprovals){
			IApprovalDescriptor oldDescriptor = ApprovalsHelper.findApprovalDescriptorByName(oldWorkItem, newDescriptor.getName());
			if(oldDescriptor == null){
				// a new auto generated approval was created
				changed.add(newDescriptor);
			} else if(!newDescriptor.getCumulativeStateIdentifier().equals(oldDescriptor.getCumulativeStateIdentifier())){
				// the approval state changed
				changed.add(newDescriptor);
			}
		}

		return changed;
	}
}
