package com.ibm.stg.rtc.ext.participants;

import java.util.List;

import org.eclipse.core.runtime.IProgressMonitor;

import com.ibm.stg.rtc.ext.common.ApprovalContainer;
import com.ibm.stg.rtc.ext.common.ApprovalWorkItemCondition;
import com.ibm.stg.rtc.ext.common.ApprovalsHelper;
import com.ibm.stg.rtc.ext.common.RepositoryCommonTasks;
import com.ibm.stg.rtc.ext.common.WorkItemCommonTasks;
import com.ibm.team.foundation.common.text.XMLString;
import com.ibm.team.process.common.IProcessConfigurationElement;
import com.ibm.team.process.common.advice.AdvisableOperation;
import com.ibm.team.process.common.advice.IProcessReport;
import com.ibm.team.process.common.advice.IReportInfo;
import com.ibm.team.process.common.advice.TeamOperationCanceledException;
import com.ibm.team.process.common.advice.runtime.IOperationParticipant;
import com.ibm.team.process.common.advice.runtime.IParticipantInfoCollector;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.repository.service.AbstractService;
import com.ibm.team.workitem.common.ISaveParameter;
import com.ibm.team.workitem.common.IWorkItemCommon;
import com.ibm.team.workitem.common.model.IApprovalDescriptor;
import com.ibm.team.workitem.common.model.IComment;
import com.ibm.team.workitem.common.model.IComments;
import com.ibm.team.workitem.common.model.IWorkItem;
import com.ibm.team.workitem.common.workflow.IWorkflowInfo;
import com.ibm.team.workitem.service.IWorkItemServer;

public class RemoveSTGApprovalsParticipant extends AbstractService implements IOperationParticipant {

	/* This follow-up action will remove all STG Auto-Generated approvals for the configured work items.
	 * Configure this follow-up action as shown below:
	 * 
	 	<removeApprovals workItemType="com.ibm.stg.workItemType.changeRecord" workItemState="com.ibm.stg.workflow.changeRecord.state.canceled"/>
	 	<removeApprovals workItemType="com.ibm.stg.workItemType.changeRecord" workItemState="com.ibm.stg.workflow.changeRecord.state.noChange"/>
	 	<removeApprovals workItemType="com.ibm.stg.workItemType.stgDefect" workItemState="com.ibm.stg.workflow.stgDefect.state.rejected"/>
	 */

	public static final String ID = "com.ibm.stg.rtc.ext.participants.RemoveSTGApprovalsParticipant";
	public static final String NAME = "STG: Remove STG Approvals";

	public void run(AdvisableOperation operation, IProcessConfigurationElement participantConfig, IParticipantInfoCollector collector, IProgressMonitor monitor) throws TeamRepositoryException {
		IWorkItemServer workItemServer = this.getService(IWorkItemServer.class);
		IWorkItemCommon workItemCommon = this.getService(IWorkItemCommon.class);
		WorkItemCommonTasks workItemCommonTasks = new WorkItemCommonTasks();
		ISaveParameter saveParameter = null;
		IWorkItem newWorkItemState = null;

		Object tempObj = operation.getOperationData();
		if (workItemCommonTasks.isSaveParamterType(tempObj)) {
			saveParameter = (ISaveParameter) tempObj;
		} else {
			return;
		}

		// if the work item is being created there are no approvals to remove so do nothing
		if(saveParameter.isCreation()){
			return;
		}

		tempObj = saveParameter.getNewState();
		if (workItemCommonTasks.isIWorkItemType(tempObj)) {
			newWorkItemState = (IWorkItem) tempObj;
		} else {
			return;
		}

		List<ApprovalContainer> configuredApprovals;
		try{
			configuredApprovals = ApprovalsHelper.getConfiguredApprovals(participantConfig, newWorkItemState, workItemCommon);
		}catch (Exception e){
			// problem reading the configured approvals
			IReportInfo info = collector.createInfo("Invalid approval configuration", e.getMessage());
			info.setSeverity(IProcessReport.ERROR);
			collector.addInfo(info);
			return;
		}

		// get a fresh working copy
		IWorkItem workItemWC = (IWorkItem)workItemServer.findWorkItemById(newWorkItemState.getId(), IWorkItem.FULL_PROFILE, null).getWorkingCopy();
		boolean approvalRemoved = false;

		for(ApprovalContainer approval : configuredApprovals){
			for(ApprovalWorkItemCondition condition : approval.getConditionConfigs(saveParameter)){
				if(condition.isRemove()){
					// configured and in the correct condition to remove this approval
					String name = approval.getName();
					// first, do not use the fresh working copy otherwise there could be conflicts with other follow-up actions
					IApprovalDescriptor approvalDescriptor = ApprovalsHelper.findApprovalDescriptorByName(newWorkItemState, name);
					// now, if an approval descriptor was found, get the working copy one
					if(approvalDescriptor != null){
						approvalDescriptor = ApprovalsHelper.findApprovalDescriptorByName(workItemWC, name);
					}

					if(approvalDescriptor != null){
						workItemWC.getApprovals().remove(approvalDescriptor);
						approvalRemoved = true;

						IWorkflowInfo workflowInfo = workItemServer.findWorkflowInfo(workItemWC, monitor);
						String commentStr = "The approval '" + name + "' has been removed because the following condition was met: " + condition.toString(workflowInfo);
						IComments comments = workItemWC.getComments();
						IComment newComment = comments.createComment(workItemServer.getAuditableCommon().getUser(), XMLString.createFromPlainText(commentStr));
						comments.append(newComment);
						break; // done with this approval, move to the next approval
					}
				}
			}
		}

		// if at least one approval was removed, save the work item
		if(approvalRemoved){
			try{
				workItemServer.saveWorkItem2(workItemWC, null, null);
			}catch(TeamOperationCanceledException e){
				RepositoryCommonTasks repoCommon = new RepositoryCommonTasks();
				IReportInfo info = collector.createInfo("Follow-up action failed", "The remove approvals follow-up could not complete because a precondition failed with the following message: \n" + repoCommon.getFirstErrorDescription(e));
				info.setProblemObject(workItemWC.getItemHandle());
				info.setSeverity(IProcessReport.WARNING);
				collector.addInfo(info);
				// suppress the create approvals errors so this save can continue
				repoCommon.suppressAllErrors(NAME, e);
			}
		}
	}
}
