package com.ibm.stg.rtc.ext.participants;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.eclipse.core.runtime.IProgressMonitor;

import com.ibm.stg.rtc.ext.common.ConfigurationHelper;
import com.ibm.stg.rtc.ext.common.RepositoryCommonTasks;
import com.ibm.stg.rtc.ext.common.WorkItemCommonTasks;
import com.ibm.team.process.common.IProcessConfigurationElement;
import com.ibm.team.process.common.advice.AdvisableOperation;
import com.ibm.team.process.common.advice.IProcessReport;
import com.ibm.team.process.common.advice.IReportInfo;
import com.ibm.team.process.common.advice.TeamOperationCanceledException;
import com.ibm.team.process.common.advice.runtime.IOperationParticipant;
import com.ibm.team.process.common.advice.runtime.IParticipantInfoCollector;
import com.ibm.team.repository.common.IAuditableHandle;
import com.ibm.team.repository.common.IItem;
import com.ibm.team.repository.common.TeamRepositoryException;
import com.ibm.team.repository.service.AbstractService;
import com.ibm.team.repository.service.IRepositoryItemService;
import com.ibm.team.scm.common.process.SaveClosedChangeSetOperationData;
import com.ibm.team.scm.common.process.SaveClosedChangeSetOperationData.ChangeType;
import com.ibm.team.workitem.common.IWorkItemCommon;
import com.ibm.team.workitem.common.internal.WorkflowUtil;
import com.ibm.team.workitem.common.model.IWorkItem;
import com.ibm.team.workitem.common.model.IWorkItemHandle;
import com.ibm.team.workitem.common.model.Identifier;
import com.ibm.team.workitem.common.workflow.IWorkflowAction;
import com.ibm.team.workitem.common.workflow.IWorkflowInfo;
import com.ibm.team.workitem.service.IWorkItemServer;

/**
 * This participant performs an action of the work item after it is associated to a change set. To configure it,
 * first specify the work item types and states to perform the action on. For each state you must specify the
 * action to perform for that state.
 * 
 * Note: The com.ibm.team.scm.server.modifyChangeset operation runs the preconditions configured for the
 * Change Set's Component owner. If the component is owned by a user, this will not run.
 * 
 * Example:
 * 
		<precondition id="com.ibm.stg.rtc.ext.participants.WIActionOnChangeSetAssociation" name="STG: WI Action on Change Set Association" optional="false">
			<showWarnings value="true"/>
			<workItemType id="com.ibm.stg.workItemType.changeRecord">
				<workItemState id="com.ibm.stg.workflow.changeRecord.state.open">
					<workItemAction id="com.ibm.stg.workflow.changeRecord.action.startWorking"/>
				</workItemState>
			</workItemType>
		</precondition>

	Notes:	- There may be 1 or more workItemType elements each containing 1 or more workItemState elements
	 		- Each workItemState element must contain exactly one workItemAction element

 * @author Kyle Christianson
 *
 */
public class WIActionOnChangeSetAssociation extends AbstractService implements IOperationParticipant {

	private static final String TAG_WORKITEM_ACTION = "workItemAction";

	public static final String ID = "com.ibm.stg.rtc.ext.participants.WIActionOnChangeSetAssociation";
	public static final String NAME = "STG: WI Action on Change Set Association";

	public void run(AdvisableOperation operation, IProcessConfigurationElement participantConfig, IParticipantInfoCollector collector, IProgressMonitor monitor) throws TeamRepositoryException {

		if(! (operation.getOperationData() instanceof SaveClosedChangeSetOperationData))
			return;
		SaveClosedChangeSetOperationData operationData = (SaveClosedChangeSetOperationData) operation.getOperationData();

		if(operationData.getChangeType() != ChangeType.ADD_LINK) {
			return;
		}

		boolean showWarnings = ConfigurationHelper.getBooleanTagValue(participantConfig, ConfigurationHelper.TAG_SHOW_WARNINGS);

		List<IWorkItemHandle> workItemHandles = new ArrayList<IWorkItemHandle>(); 
		// Note: operationData.getLinks() only returns the new work item links for this operation, which is what we want
		for (Iterator<?> iterator = operationData.getLinks().iterator(); iterator.hasNext(); ) {
			IAuditableHandle workItemHandle = (IAuditableHandle) iterator.next();
			if (workItemHandle instanceof IWorkItemHandle) {
				workItemHandles.add((IWorkItemHandle) workItemHandle);
			}
		}

		IRepositoryItemService itemService = getService(IRepositoryItemService.class);
		IWorkItemCommon workItemCommon = getService(IWorkItemCommon.class);
		IItem[] workItems = itemService.fetchItems(workItemHandles.toArray(new IWorkItemHandle[0]), null);



		for (int i = 0; i < workItems.length; i++) {
			IWorkItem workItem = (IWorkItem)workItems[i];
			IWorkflowInfo workflowInfo = workItemCommon.findWorkflowInfo(workItem, monitor);

			Identifier<IWorkflowAction> action = null;
			try{
				action = getConfiguredAction(participantConfig, workItem, workflowInfo);
			}catch(Exception e){
				IReportInfo info = collector.createInfo("Invalid Configuration", e.getMessage());
				info.setSeverity(IProcessReport.WARNING);
				collector.addInfo(info);
			}

			if(action != null){
				// there is an action to try to perform it

				IWorkItemServer workItemServer = this.getService(IWorkItemServer.class);

				// make sure the action is valid on the current state, if not saveWorkItem2 would "Submit" the defect
				if (WorkflowUtil.validState(workflowInfo, workItem, action.getStringIdentifier())) {
					IWorkItem workItemWorkingCopy = (IWorkItem)workItem.getWorkingCopy();
					try {
						workItemServer.saveWorkItem2(workItemWorkingCopy, null, action.getStringIdentifier());
					} catch (TeamOperationCanceledException e) {
						RepositoryCommonTasks repoCommon = new RepositoryCommonTasks();
						if(showWarnings){
							IReportInfo info = collector.createInfo("Follow-up action failed", "The follow-up action could not '" + workflowInfo.getActionName(action) + "' work item " + workItem.getId() + " because a precondition failed with the following message: \n" + repoCommon.getFirstErrorDescription(e));
							info.setProblemObject(workItem.getItemHandle());
							info.setSeverity(IProcessReport.WARNING);
							collector.addInfo(info);
						}
						// suppress the follow-up action work item save errors so this change set associate operation can continue  
						repoCommon.suppressAllErrors(NAME, e);
					}
				} else if(showWarnings){
					// the defect was not in a state where the desired action could be performed
					WorkItemCommonTasks workItemCommonTasks = new WorkItemCommonTasks();
					IReportInfo info = collector.createInfo("Follow-up action failed", "The follow-up action could not '" + workflowInfo.getActionName(action) + "' work item " + workItem.getId() + " because it is not a valid action for the " + workItemCommonTasks.getStateName(workItem, workItemCommon, monitor) + " state.");
					//info.setSeverity(IProcessReport.WARNING);
					info.setSeverity(IProcessReport.WARNING);
					collector.addInfo(info);
				}
			}
		}
	}

	/**
	 * returns the action configured for the given work item, or null if no action is configured
	 * @param participantConfig
	 * @param workItem
	 * @param workItemCommon
	 * @return
	 * @throws TeamRepositoryException 
	 * @throws IllegalStateException
	 */
	private Identifier<IWorkflowAction> getConfiguredAction(IProcessConfigurationElement participantConfig, IWorkItem workItem, IWorkflowInfo workflowInfo) throws IllegalStateException, TeamRepositoryException{

		String stateId = workItem.getState2().getStringIdentifier();
		stateId = ConfigurationHelper.convertLegacyId(stateId, 's');

		IProcessConfigurationElement stateElement = ConfigurationHelper.getWIStateConfiguration(participantConfig, workItem.getWorkItemType(), stateId);
		if(stateElement != null){
			// the work item is in the configured type and state
			// get the action
			List<String> actionList = ConfigurationHelper.getChildrenTagIds(stateElement, TAG_WORKITEM_ACTION);
			if(actionList == null || actionList.size() == 0){
				throw new IllegalStateException("The follow-up action could not be run because its xml configuration must contain a <" + TAG_WORKITEM_ACTION + "> element for the <" + ConfigurationHelper.TAG_WORKITEM_STATE + " " + ConfigurationHelper.ATTR_ID + "=\"" + stateId + "\"> element.");
			}else if(actionList.size() > 1){
				throw new IllegalStateException("The follow-up action could not be run because its xml configuration has multiple <" + TAG_WORKITEM_ACTION + "> elements for the <" + ConfigurationHelper.TAG_WORKITEM_STATE + " " + ConfigurationHelper.ATTR_ID + "=\"" + stateId + "\"> element.");
			}else{
				String action = actionList.get(0);

				Identifier<IWorkflowAction> actionId = (new WorkItemCommonTasks()).getActionIdentifierFromString(workflowInfo, action);
				return actionId;
			}
		}

		return null;
	}

}
