'''
Created on May 8, 2013

@author: jdevries
'''

import logging
import json
import jazzrest
from CLMException import CLMException
from errormessages import getmsg
from datetime import datetime,timedelta
import time

class Translator(object):
    '''
    classdocs
    '''
    session = None
    project = None
    removeemptylists = True
    
    attributes = {}
    workitems = {}
    literals = {}
    
    estimateattrs = ['rtc_cm:correctedEstimate','rtc_cm:estimate','rtc_cm:timeSpent']
    
    toattrstrip = ['dcterms:creator','oslc_cm:fixed','oslc_cm:status','oslc:shortTitle','oslc_cm:reviewed','dcterms:identifier',
                   'rtc_ext:com.ibm.stg.attribute.universalId','oslc_cm:approved','dcterms:modified','rtc_cm:resolvedBy',
                   'dcterms:created','rtc_cm:state','rtc_cm:modifiedBy','oslc_cm:closed','oslc_cm:verified','rtc_ext:archived',
                   'oslc_cm:inprogress','dcterms:type','rtc_cm:teamArea']
    fromattrstrip = ['rdf:about','rdf:type','prefixes','acp:accessControl','rtc_cm:type','rtc_ext:contextId',
                 'oslc:instanceShape','oslc:serviceProvider','rtc_cm:repository','rtc_cm:timeSheet','oslc_pl:schedule',
                 'rtc_cm:progressTracking']
    linkattrs = ['rtc_cm:com.ibm.team.connector.ccbridge.common.act2wi.s','rtc_cm:com.ibm.team.connector.ccbridge.common.ver2wi.s',
                 'rtc_cm:com.ibm.team.connector.scm.common.linkType.tracksUcmObject.t','rtc_cm:com.ibm.team.filesystem.workitems.change_set.com.ibm.team.scm.ChangeSet'
                 ]
    
    actionheaders = {'Accept': 'application/json'}
    
    attrexceptions = { 'com.ibm.team.workitem.workItemType.risk.impact': 'Risk Impact',
                  'com.ibm.team.workitem.linktype.resolvesworkitem.resolvedBy': 'Resolved By Link',
                  'rtc_cm:type': 'WI Type',
                  'rtc_ext:phasefoundcqdlink': 'Phase Found CQD',
                  'rtc_ext:subphasecqdlink': 'Subphase CQD',
                  'rtc_ext:universalidcqdlink': 'Universal ID CQD',
                  'rtc_ext:statecqdlink': 'State CQD'
                  }

    nslookup = {
        "acp": "http://jazz.net/ns/acp#",
        "dcterms": "http://purl.org/dc/terms/",
        "oslc": "http://open-services.net/ns/core#",
        "oslc_cm": "http://open-services.net/ns/cm#",
        "oslc_cmx": "http://open-services.net/ns/cm-x#",
        "oslc_pl": "http://open-services.net/ns/pl#",
        "rdf": "http://www.w3.org/1999/02/22-rdf-syntax-ns#",
        "rtc_cm": "http://jazz.net/xmlns/prod/jazz/rtc/cm/1.0/",
        "rtc_ext": "http://jazz.net/xmlns/prod/jazz/rtc/ext/1.0/"
    }

    def __init__(self,rtcsession,repository,project,cache,removeemptylists=True):
        '''
        Constructor
        '''
        self.rtcsession = rtcsession
        self.session = rtcsession.requestSession
        self.repository = repository
        self.project = project
        self.removeemptylists=removeemptylists
        
        self.logger = logging.getLogger('mainlogger')
        self.actionheaders.update(rtcsession.coreheaders)
        self.logger.debug('actionheaders: '+repr(self.actionheaders))
        '''
        load for specific project:
        '''
        key = str(repository+'/'+project)
        if key in cache:
            pcache = cache[key]
        else:
            pcache = {}
            cache[key]=pcache
        
        self.workitems = pcache.get('workitems',None)
        self.attributes = pcache.get('attributes',None)
        self.literals = pcache.get('literals',None)
        
        if self.workitems == None or self.attributes == None or self.literals == None:
            self.loadprojectdetails()
            pcache['workitems']=self.workitems
            pcache['attributes']=self.attributes
            pcache['literals']=self.literals


    def loadprojectdetails(self):
        self.logger.debug('Loading details: '+self.repository+'/'+self.project)
        url = self.repository+'/oslc/contexts/'+self.project+'/workitems/services'
        self.attributes = {}
        self.literals = {}
        self.workitems = {}
        witypes = {}
        winames = {}
        sr = self.rtcsession.get(url, headers=self.rtcsession.jsonheaders)
        services = json.loads(sr.text).get("oslc:service",[])
        # ignore: rdf:type, oslc:details, dcterms:title, prefixes, rdf:about, dcterms:publisher
        
        for service in services:
            
            factories = service.get("oslc:creationFactory",[])
            for fi in factories:
                tl = fi["oslc:resourceType"]
                typeName = None
                for tg in tl:
                    if tg["rdf:resource"].startswith(self.repository):
                        typeName = tg["rdf:resource"]
                if typeName != None:
                    self.logger.debug('starting: '+typeName)
                    #cfURL=fi["oslc:creation"]["rdf:resource"]
                    shapeURL=fi["oslc:resourceShape"]["rdf:resource"]
                    typeName = typeName.split('/')[-1]
                    attrs = self.shapeparsing(shapeURL,typeName)
                    tun = attrs['username']
                    if typeName not in witypes:
                        witypes[typeName]=attrs
                        if tun not in winames:
                            winames[tun]=typeName
                        elif winames[tun]!=typeName:
                            self.logger.debug('Type Name problem: '+tun+' == '+typeName+' | '+winames[tun])
        self.workitems['details']=witypes
        self.workitems['names']=winames

    def shapeparsing(self,shapeurl,typeName):
        attributes = self.attributes.get('details',None)
        if attributes == None:
            attributes = {}
            self.attributes['details']=attributes
        
        attrnames = self.attributes.get('names',None)
        if attrnames == None:
            attrnames = {}
            self.attributes['names']=attrnames

        wtid = typeName
        sr = self.rtcsession.get(shapeurl,headers=self.rtcsession.jsonheaders)
        
        shapes = json.loads(sr.text)
        uname = shapes["dcterms:title"]
        
        sprops = shapes['oslc:property']
        xprops = {} # map between literal and end-user
        statusinfo = None
        for p in sprops:
            ''' p is a dictionary '''
            ''' ignore rdf:type,  oslc:occurs, rdf:about = URL that might be useful for additional work
             use: dcterms:title=user name, oslc:name=literal name,  +oslc:readOnly, 
             unknown: oslc:propertyDefinition : might need this to differentiate stuff '''
            tl = p['oslc:name']
            l = self.attachns(tl,p['oslc:propertyDefinition']['rdf:resource'])
            if l in attributes:
                nd=attributes[l]
                atype=nd.get('avtype','')
                e = nd.get('username',None)
            else:
                te = self.attrexceptions.get(tl,None)
                if te == None:
                    te = self.attrexceptions.get(l,p['dcterms:title'])
                e = te
                a = p['rdf:about']
                nd = {'username': e}
                vt = p['oslc:valueType']['rdf:resource'].split('#')[-1]
                #rt = p.get('oslc:representation',{}).get('rdf:resource','').split('#')[-1]
                nd['type']=vt
                #nd['representation']=rt
                nd['about']=a
                nd['readonly']=p['oslc:readOnly']
                atype=''
                if 'oslc:allowedValues' in p:
                    av = p['oslc:allowedValues']['rdf:about'].replace(wtid,'{WITYPE}')
                    nd['allowedValuesURL'] = av
                    if l == "rtc_cm:state":
                        (atype, avl) = self.loadallowedvalues(l, wtid,loadurl=av) # getallowedvalues(p['oslc:allowedValues']['oslc:allowedValue'],self.project,l)
                        nd['allowedValues'] = avl
                        nd['avtype']=atype
                    #mastertypes.add(atype)
            
            if l not in xprops:
                if atype=='Status':
                    statusinfo = (l,nd)
                xprops[l]=nd
    
            if e not in attrnames:
                attrnames[e]=l
            else:
                if attrnames[e]!=l:
                    self.logger.warning('Name mismatch: '+e+' == '+attrnames[e]+' <> '+l)
     
            
            if l not in attributes and atype!='Status' and l!='rtc_cm:resolution':
                attributes[l]=nd
            else:
                # could compare attribute values to see if it's all the same
                pass
            
        
        actions = {}
        actionuri = None
        resolutions = {}
        resolutionuri = None
        
        if statusinfo != None:
            av = statusinfo[1].get('allowedValues',{})
            for n in av.keys():
                if n.startswith(self.repository):
                    actionuri = n.replace('/states/','/actions/')
                    actionuri = actionuri.rsplit('/',1)[0]
                    resolutionuri = n.replace('/states/','/resolutions/')
                    resolutionuri = resolutionuri.rsplit('/',1)[0]

                    #self.logger.debug(actionuri)
                    #self.logger.debug(resolutionuri)
                    break
            if actionuri != None:
                r = self.rtcsession.get(actionuri, headers=self.actionheaders)
                al = json.loads(r.text)
                details = {}
                titles = {}
                for x in al:
                    resultState = x['rtc_cm:resultState']['rdf:resource']
                    actionid = x['dc:identifier']
                    title = x['dc:title']
                    
                    details[actionid]={'resultState':resultState,'title':title}
                    titles[title]=actionid
                
                actions['details']=details
                actions['names']=titles
            if resolutionuri != None:
                r = self.rtcsession.get(resolutionuri, headers=self.actionheaders)
                al = json.loads(r.text)
                details = {}
                titles = {}
                for x in al:
                    resolutionid = x['rdf:resource']
                    title = x['dc:title']
                    
                    details[resolutionid]={'title':title}
                    titles[title]=resolutionid
                
                resolutions['details']=details
                resolutions['names']=titles
        
        shape = {'username':uname,'attributes':xprops,'actions':actions,'actionuri':actionuri,'resolutions':resolutions,'resolutionuri':resolutionuri}
        
        return shape

    def attachns(self,base, propdef):
        for (ns,nsu) in self.nslookup.items():
            if propdef.startswith(nsu):
                return ':'.join([ns,base])
        self.logger.debug('not found URI: '+propdef)
        return base



    def reconstituteuri(self,uri,r,p):
        return uri
    
    def fromwitype(self,witype):
        uname = self.workitems['details'].get(witype,None)
        if uname == None:
            if witype in self.workitems['names']:
                # this already is a human name:
                uname = witype
            else:
                raise CLMException(*getmsg("RJ03112",witype=witype))
        return uname
    
    def towitype(self,name):
        witype = self.workitems['names'].get(name,None)
        if witype == None:
            if name in self.workitems['details']:
                # this already is a witype
                witype = name
            else:
                raise CLMException(*getmsg("RJ03113",winame=name))
        return witype
    
    def fromactionid(self,witype,actionid):
        uname = self.workitems['details'].get(witype,{}).get('actions',{}).get('details',{}).get(actionid,None)
        if uname == None:
            if actionid in self.workitems['details'].get(witype,{}).get('actions',{}).get('names',{}):
                # already a name
                uname = actionid
            else:
                raise CLMException(*getmsg("RJ03114",actionid=actionid))
        return uname
    
    def toactionid(self,witype,name):
        actionid = self.workitems['details'].get(witype,{}).get('actions',{}).get('names',{}).get(name,None)
        if actionid == None:
            if name in self.workitems['details'].get(witype,{}).get('actions',{}).get('details',{}):
                actionid = name
            else:
                raise CLMException(*getmsg("RJ03115",actionname=name))
        return actionid
    
    def persontouri(self,person):
        mp  = person.replace('@','%40')
        personuri = self.rtcsession.jtsurl+'/users/'+mp
        r = self.rtcsession.get(personuri, headers=self.rtcsession.rdfheaders)
        if r.status_code != 200:
            raise CLMException(*getmsg("RJ03116",person=person))
        return personuri
    
    def uritoperson(self,uri):
        tid = uri.split('/')[-1]
        return tid.replace('%40', '@')
 
    def islinktype(self,attr,ad):
        rval = attr.upper().find('.LINKTYPE.')>-1 or ad.get('about','').upper().find('.LINKTYPE.')>-1 or attr in self.linkattrs
        return rval
    
    def valuefromoslc(self,attr,value,witype):
        #self.logger.debug('valuefromoslc: '+attr)
        removeattr = False
        nvalue = value
        ad = self.attributes['details'].get(attr,{})

        if self.islinktype(attr,ad):
            # format a linktype
            nvalue = []
            for x in value:
                nx = {}
                nx['title'] = x.get('dcterms:title','')
                nx['url'] = x.get('rdf:resource','')
                nvalue.append(nx)
            if len(nvalue)==0 and self.removeemptylists:
                removeattr = True
        elif attr == 'rtc_cm:subscribers':
            nvalue = []
            for x in value:
                nvalue.append(self.uritoperson(x.get('rdf:resource','')))
        else:
            if attr not in self.literals:
                self.loadallowedvalues(attr, witype)
            #self.logger.debug('attribute info: '+str(ad))
            ad = self.attributes['details'].get(attr,{})
            avtype = ad.get('avtype','')
            rdftypeperson = str(value).find("http://xmlns.com/foaf/0.1/Person") > -1
            if avtype =='Literal':
                # get the values from the literals list
                tvalue = value.get('rdf:resource','')
                tkey = tvalue.split('/enumerations/')[-1]
                key = self.repository+'/oslc/enumerations/'+self.project+tkey[tkey.find('/'):]
                #self.logger.debug('Checking literal key: '+key)
                nvalue = self.getliteralvalue(witype,attr,key,value)
                #nvalue = self.literals.get(attr,{}).get(key,value)
            elif avtype == 'Person' or rdftypeperson:
                nvalue = self.uritoperson(value.get('rdf:resource',''))
            elif avtype in ['Project','Category','Iteration','Deliverable'] or attr=='rtc_cm:teamArea':
                ruri = value.get('rdf:resource',None)
                if ruri != None:
                    r = self.rtcsession.get(ruri,headers=jazzrest.RTCSession.jsonheaders)
                    #self.logger.debug(r.text)
                    rjs = json.loads(r.text)
                    nvalue = rjs.get('dcterms:title',None)
                        
            elif attr=='oslc:discussedBy':
                curi = value.get('rdf:resource',None)
                if curi != None:
                    r = self.rtcsession.get(curi,headers=self.rtcsession.jsonheaders)
                    cdoc = json.loads(r.text)
                    clist = cdoc.get('oslc:comment')
                    nvalue = self.loadcomments(clist)
                else:
                    nvalue = []
            else:
                pass
                #self.logger.debug('No luck: '+attr)
        return (removeattr,nvalue)

    def getliteralvalue(self,witype,attr,key,value):
        self.logger.debug('get literal value '+attr+' '+key)
        if attr not in self.literals:
            self.loadallowedvalues(attr, witype)
        
        if attr in self.literals:
            return self.literals[attr].get(key,value)
        else:
            return value

    def loadcomments(self,clist):
        rlist = []
        #self.logger.debug(clist)
        #comments: follow URL, dc:description is comment text, dc:creator has rdf:resource link to author, dc:created has create time
        for d in clist:
            try:
                #self.logger.debug(x)
                #self.logger.debug(x['rdf:resource'])
                v = {}
                v['comment']=d['dcterms:description']
                v['author']=self.uritoperson(d['dcterms:creator']['rdf:resource'])
                v['timestamp']=d['dcterms:created']
                v['ref']=d['rdf:about']
                rlist.append(v)
            except:
                rlist.append(d)
        
        return rlist

    
    def attrfromoslc(self,attr,value,witype):
        nattr = None
        nvalue = None
        removeattr = False
        if attr in self.fromattrstrip:
            removeattr = True
        elif attr in self.attributes['details']:
            ad = self.attributes['details'][attr]
            nattr = ad['username']
            atype = ad['type']
            nvalue = value
            if atype=='Resource' and attr != 'rtc_ext:com.ibm.stg.attribute.references':
                (removeattr,nvalue) = self.valuefromoslc(attr,value,witype)
            elif attr in self.estimateattrs:
                nvalue = self.formatestimate(value)
            elif atype in ['boolean','XMLLiteral','dateTime','integer','long']:
                pass
        elif attr=='rtc_cm:resolution' and witype != None and witype in self.workitems['details']:
            ad = self.workitems['details'][witype]['attributes'].get(attr,{})
            nattr = ad.get('username','Resolution')
            resuri = value['rdf:resource']
            #key = "{REPOSITORY}/oslc/workflows/{PROJECT_UUID}/states/"+stateuri.split('/states/')[-1]
            self.logger.debug("resolution: "+resuri)
            key = resuri #self.repository+"/oslc/workflows/"+self.project+"/states/"+stateuri.split('/states/')[-1]
            resolutions = self.workitems['details'][witype]['resolutions'].get('details',{})
            #self.logger.debug(resolutions)
            nvalue = resolutions.get(key,{}).get('title',value)          
        elif attr=='rtc_cm:state' and witype != None and witype in self.workitems['details']:
            ad = self.workitems['details'][witype]['attributes'].get(attr,{})
            nattr = ad.get('username','Status')
            stateuri = value['rdf:resource']
            #key = "{REPOSITORY}/oslc/workflows/{PROJECT_UUID}/states/"+stateuri.split('/states/')[-1]
            self.logger.debug("state: "+stateuri)
            key = stateuri #self.repository+"/oslc/workflows/"+self.project+"/states/"+stateuri.split('/states/')[-1]
            nvalue = ad.get('allowedValues',{}).get(key,{}).get('title',value)
        else:
            nattr=attr
            nvalue=value
        
        if removeattr:
            nattr=None
        
        return (nattr,nvalue)
    
    def fromoslc(self,wi):
        xlate = {}
        witype = None
        if "rtc_cm:type" in wi:
            witype = wi["rtc_cm:type"]["rdf:resource"].split("/")[-1]
        for (oa,ov) in wi.items():
            (na,nv) = self.attrfromoslc(oa,ov,witype)
            if na != None:
                xlate[na]=nv
        return xlate

    def formatestimate(self,est):
        if est == None:
            return None
        v = est/1000 # get rid of milliseconds
        w = int(v/144000) # 144000 seconds in a 40 hour week
        v = v%144000 
        d = int(v/28800) # 28800 seconds in an 8 hour day
        v = v%28800
        h = int(v/3600) # 3600 seconds in an hour
        v = v%3600
        m = int(v/60) # 60 seconds in a minute
        s = v%60

        rval = ''
        if w>0:
            rval=rval+str(w)+' w '
        if d>0:
            rval=rval+str(d)+' d '
        if h>0:
            rval=rval+str(h)+' h '
        if m>0:
            rval=rval+str(m)+' m '
        if s>0:
            rval=rval+str(s)+' s'
        
        return rval.rstrip()    

    def parseestimate(self,est):
        if est == None:
            return None
        mt = {'W':144000,'D':28800,'H':3600,'M':60,'S':1}
        rval = 0
        mest = est.upper().rstrip()
        
        cval = ''
        
        for x in mest:
            if x.isdigit():
                cval = cval + x
            elif x.isalpha():
                rval = rval + (int(cval) * mt.get(x,0))
                cval = ''

        if len(cval)>0:
            if rval > 0:
                rval = rval + int(cval)
            else:
                rval = int(cval)*mt['H']
        
        return rval*1000

    def idtourl(self,ident):
        if ident.startswith('http'):
            return ident
        else:
            return self.repository+'/resource/itemName/com.ibm.team.workitem.WorkItem/'+ident

    def parsedatetime(self,dt):
        outformat="%Y-%m-%dT%H:%M:%S.%fZ"
        try:
            datetime.strptime(dt,outformat)
            # it met the format, just return what we have
            return dt
        except ValueError:
            pass
        
        fmtoptions = [
                      "%Y-%m-%d",
                      "%Y/%m/%d",
                      "%Y-%m-%dT%H:%M",
                      "%Y-%m-%dT%H:%M:%S",
                      "%Y-%m-%dZ",
                      "%Y/%m/%dZ",
                      "%Y-%m-%dT%H:%MZ",
                      "%Y-%m-%dT%H:%M:%SZ"
                      ]
        
        tdelta=timedelta(seconds=time.timezone) # timezone offset, everything we send is in UTC
        for fmt in fmtoptions:
            try:
                
                cv = datetime.strptime(dt,fmt)
                if not fmt.endswith('Z'):
                    cv = cv+tdelta
                rv = datetime.strftime(cv,outformat)
                self.logger.debug("Converted datetime to: "+rv)
                return rv
            except ValueError:
                pass
        
        return dt
    
    def valuetooslc(self,attr,value,witype):
        nvalue = value
        self.logger.debug('valuetooslc: '+attr+' '+witype)
        ad = self.attributes['details'].get(attr,{})
        atype = ad.get('type',None)
        if self.islinktype(attr,ad):
            # format a linktype
            nvalue = []
            self.logger.debug(type(value))
            if isinstance(value,list):
                for x in value:
                    nx = {}
                    if isinstance(x,dict):
                        nx = {}
                        if 'title' in x:
                            nx['dcterms:title'] = x['title']
                        if 'url' in x:
                            nx['rdf:resource'] = x['url']
                    else:
                        self.logger.debug('Type is string: '+x)
                        nx['rdf:resource'] = self.idtourl(x)
                    
                    nvalue.append(nx)
            else:
                nx = {'rdf:resource':self.idtourl(value)}
                nvalue.append(nx)
        elif atype == 'dateTime':
            nvalue = self.parsedatetime(value)
            
        elif attr == 'rtc_cm:subscribers':
            nvalue = []
            for x in value:
                nvalue.append({'rdf:resource':self.persontouri(x)})
        elif attr == 'rtc_cm:resolution':
            resval = self.workitems.get('details',{}).get(witype,{}).get('resolutions',{}).get('names',{}).get(value,None)
            if resval == None:
                if value in self.workitems.get('details',{}).get(witype,{}).get('resolutions',{}).get('details',{}):
                    resval = value
                else:
                    raise CLMException(*getmsg("RJ03300",attribute=attr,value=value))
            self.logger.debug('resolution: '+resval)
            nvalue = {'rdf:resource':resval}
        else:
            avtype = ad.get('avtype',None)
            if avtype == None:
                self.loadallowedvalues(attr, witype)
                avtype = ad.get('avtype','')
            self.logger.debug('avtype: '+avtype)
            if avtype =='Literal':
                self.logger.debug('get value for Literal')
                # get the values from the literals list
                tvalue = self.getliteralvalue(witype,attr,value,None)
                if tvalue!=None:
                    tvalue = self.reconstituteuri(tvalue, self.repository, self.project)
                    nvalue = {'rdf:resource':tvalue}
            elif avtype == 'Person':
                nvalue = {'rdf:resource':self.persontouri(value)}
            elif avtype in ['Project','Category','Iteration','Deliverable']:
                ruri = ad.get('allowedValuesURL',None)
                if ruri!=None:
                    ruri = self.reconstituteuri(ruri, self.repository, self.project).replace('{WITYPE}',witype)
                    r = self.rtcsession.get(ruri,headers=self.rtcsession.jsonheaders)
                    #self.logger.debug(r.text)
                    rjs = json.loads(r.text)
                    tvalue = None
                    for x in rjs.get('oslc:allowedValue',[{}]):
                        avr = self.rtcsession.get(x['rdf:resource'],headers=self.rtcsession.jsonheaders)
                        avj = json.loads(avr.text)
                        if value == avj.get('dcterms:title',None):
                            tvalue = avj['rdf:about']
                            break
                        elif value == avj.get('rdf:about',None):
                            tvalue = value
                            break
                    if tvalue==None:
                        # TODO: can you get cleartext attribute name?
                        raise CLMException(*getmsg("RJ03300",attribute=attr,value=value))
                else:
                    tvalue = value
                nvalue = {'rdf:resource':tvalue}
                if avtype == 'Deliverable':
                    nvalue['rdf:type']=[{"rdf:resource": "http://jazz.net/xmlns/prod/jazz/rtc/cm/1.0/Deliverable"}]

            elif attr=='oslc:discussedBy':
                comment = None
                if value == str(value):
                    comment = value
                else:
                    for x in value:
                        self.logger.debug('Comments '+str(x))
                        if x.get('ref','new')=='new':
                            comment=x['comment']
                nvalue = {'comment':comment}
            elif attr in self.estimateattrs:
                nvalue = self.parseestimate(value)
            else:
                #self.logger.debug('No luck: '+attr)
                pass

        
        return nvalue

    def attrtooslc(self,attr,value,witype):
        nattr = self.attributes.get('names',{}).get(attr,None)
        self.logger.debug("nattr: "+str(nattr))
        if nattr == None:
            if attr in self.attributes.get('details',{}):
                nattr = attr
            else:
                raise CLMException(*getmsg("RJ03301",attribute=attr))
        if nattr not in self.toattrstrip:
            
            nvalue = self.valuetooslc(nattr,value,witype)
        
            return (nattr,nvalue)
        else:
            return (None,None)

    def tooslc(self,witype,wi):
        xlate = {}
        
        for (oa,ov) in wi.items():
            (na,nv) = self.attrtooslc(oa,ov,witype)
            if na != None:
                xlate[na]=nv
        
        return xlate

    def loadallowedvalues(self,attribute,witype,loadurl=None):
        self.logger.debug('loadallowedvalues: '+attribute+' '+witype)
        allowedvalues = {}
        ad = self.attributes['details'].get(attribute,{})
        if loadurl!=None:
            ruri=loadurl
        else:
            ruri = ad.get('allowedValuesURL',None)
        avtype = None
        if ruri!=None:
            types = set()
            ruri = self.reconstituteuri(ruri, self.repository, self.project).replace('{WITYPE}',witype)
            self.logger.debug(ruri)
            r = self.rtcsession.get(ruri,headers=jazzrest.RTCSession.jsonheaders)
            #self.logger.debug(r.text)
            rjs = json.loads(r.text)
            for x in rjs.get('oslc:allowedValue',[{}]):
                ttype = x['rdf:type'][0]['rdf:resource']
                typen = ttype.split('/')[-1]
                types.add(typen)
                #self.logger.debug('typen: '+typen)
                if typen in ['Literal','Category','Iteration','Deliverable']:
                    avr = self.rtcsession.get(x['rdf:resource'],headers=self.rtcsession.jsonheaders)
                    avj = json.loads(avr.text)
                    
                    n = avj['dcterms:title']
                    k = avj['rdf:about']
                    allowedvalues[n]=k
                    allowedvalues[k]=n
                elif typen == 'Status':
                    #self.logger.debug('loading a status value')
                    dr = self.rtcsession.get(x['rdf:resource'],headers=self.rtcsession.jsonheaders)
                    desc = json.loads(dr.text)
                    t = desc['dcterms:title']
                    lkey = x['rdf:resource']
            
                    allowedvalues[lkey]={'title':t}
                    allowedvalues[t]={'literal':lkey}
            
            avtype = types.pop()
            self.logger.debug('avtype for '+attribute+' '+avtype)
            ad['avtype']=avtype
        else:
            self.logger.debug('No allowed values url for: '+attribute)
        
        if avtype == 'Status':
            return (avtype,allowedvalues)
        if attribute != 'rtc_cm:filedAgainst' and attribute != 'rtc_cm:plannedFor' and avtype not in ['Deliverable','Category','Iteration']:
            self.literals[attribute]=allowedvalues
        return (avtype,allowedvalues)


    def getallowedvalues(self,attribute,witype='task'):
        allowedvalues = {}
        self.logger.debug('get allowed values: '+attribute+' '+witype)
        if attribute=='rtc_cm:resolution':
            tav = self.workitems.get('details',{}).get(witype,{}).get('resolutions',{}).get('names',{})
        elif attribute=='rtc_cm:state':
            self.logger.debug(attribute+' '+witype)
            tav = self.workitems.get('details',{}).get(witype,{}).get('attributes',{}).get(attribute,{}).get('allowedValues',{})    
        elif attribute not in self.literals:
            self.logger.debug('loading allowed values: '+attribute+' '+witype)
            (_,tav) = self.loadallowedvalues(attribute,witype)
        else: 
            self.logger.debug('already have attribute: '+attribute)
            tav = self.literals[attribute]
            
        for t in tav:
            if not t.startswith(self.repository):
                allowedvalues[t]=tav[t]
        return allowedvalues

    def getworkitemtypes(self):
        return self.workitems.get('names',{}).keys()

    def getattributes(self,witype):
        typeid = self.towitype(witype)
        self.logger.debug('getting attributes for: '+witype+' '+typeid)
        
        attrkeys = self.workitems.get('details',{}).get(typeid,{}).get('attributes',{})
        attrnames = set()
        attrdetails = self.attributes.get('details',{})
        for (a,v) in attrkeys.items():
            if 'username' in v:
                #if v.get('readonly',False): readonly = ' (readonly)'
                #else: readonly = ''
                linktype=''
                if self.islinktype(a,v): linktype=' (link)'
                attrinfo = v['username']+linktype
                attrnames.add(attrinfo)
            else:
                username = attrdetails.get(a,{}).get('username',None)
                #if attrdetails.get(a,{}).get('readonly',False): readonly=' (readonly)'
                #else: readonly=''
                if username != None:
                    attrnames.add(username)#+readonly)
                
        return sorted(attrnames)
''' translation design
oslc to human:

this works for work item representation... 

input: 
 p: project area
 t: work item type being processed
 wi: work item representation in dictionary
 hwi: work item representation after translation
 translator: method for doing transformations
 
for each key in wi:
 (hkey,hval) = translator(p,t,key,wi[key])
 
translator(p,t,key,val)
 hkey = attr2human(p,t,key) # this could be implemented as a dictionary
 vti = translationinfo(p,t,key)
 
 hval = vti.translate(val)
 
 return (hkey,hval)

vti.translate(val)
 hval = XLATE(val)
 return hval
 

 
 

'''
