#!/usr/bin/env python
'''
Created on Apr 12, 2013

@author: jdevries
'''

'''
TODO:
 - change state doesn't return error if state change doesn't actually happen
 - error handling
 
'''

'''
Future:
 - Work Item "mask" -- run a display command and only show the specified fields. need additional design for this.
 - Work Item display: formatted layout
 - full text search?
 - attachments?
 - unset CWE?
 - list users: project members, team members

Done:
 - ensure serialized stuff is stored in a way that only the owner can read/modify (check file permissions)
 - cwe: add verbosity, translation
 - advanced help messages (rtcwi help <subcommand>)
 - { } on command line... need to handle when they are attached to strings
 - validate action names for changestate
 - TEST ping to verify connection is still valid -- seems to be slow...
 - no prompt
 - list nicknames
 - add / remove subscribers
 - link / unlink links
 - add comment
 - merge help (-h, --help and help all go through the same path)
 - file parameters
 - changing a state
 - have login store some stuff for convenience: nickname to repository map, "encrypted" password
 - "current work environment mode" - remember the repository, project, etc that the user is actively in.
 - logging configurable: use environment variable RTCWI_LOGLEVEL
 - option to set config dir path: --config <dir>
 - serialization
 - list repo/project info (need to implement this in jazzrest first)
  - get list of actions for workitem type
  - get list of categories <REPOSITORY URI>/service/com.ibm.team.workitem.common.internal.rest.IWorkItemRestService/values?attributeName=category&projectAreaItemId=<PROJECT UUID>&typeId=task&includeArchived=false

'''

import pkg_resources
import rtccliinfo
from optparse import OptionParser, OptionGroup
from jazzrest import jazzrest
from jazzrest.CLMException import CLMException
import bz2
import getpass
import json
import sys
import os
import shelve
import logging
from jazzrest import translator
import webbrowser
from errormessages import getmsg
from urlparse import urlparse


class RTCWICommandLine:
    origmask = None
    reposessions = None
    nicks = None # nicknames dictionary
    cuser = None
    cpassword = None
    cwe = {} # current work environment dictionary
    translate = None
    xlateto = True
    xlatefrom = True
    verbosity = '2'

    HOMEDIR = os.path.expanduser('~')
    DEFAULT_CONFIGDIR = HOMEDIR+os.sep+'.rtcwicli'+os.sep
    CONFIGDIR = None
    NICKFILE = 'rtcwinicks'
    #SESSIONFILE = 'rtcwi.json' # maybe get rid of this if shelve works
    REPOFILE = 'reposhelf'
    CWEFILE = 'cwe.json'
    
    logger = None
    commandmap = {}


    def __init__(self):
        self.origmask=os.umask(0077)
        self.logger = logging.getLogger('mainlogger')
        loglvl = os.getenv('RTCWI_LOGLEVEL', 'ERROR')
        if len(loglvl)==0:
            loglvl = 'ERROR'
        lv = logging.getLevelName(loglvl)
        self.logger.setLevel(lv)
        ch = logging.StreamHandler()
        ch.setLevel(lv)
        formatter = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')

        # add formatter to ch
        ch.setFormatter(formatter)

        # add ch to logger
        self.logger.addHandler(ch)
        
        self.CWEFILE = os.getenv('RTCWI_CWEFILE', self.CWEFILE)
        
        self.commandalias= {
                          'dsp': 'display',
                          'mod': 'modify',
                          'lq': 'listqueries',
                          'rq': 'runquery',
                          'ls': 'list'
                            }
        
        self.commandmap= { #note: do not put aliases in here
                          'login': [self.cmd_login,'Login to the repository',True],
                          'help': [self.print_help,'Display help. Optionally specify the subcommand to display help for.',True],
                          'create': [self.cmd_createwi,'Create a work item',True],
                          'display': [self.cmd_display,'Displays a work item',True],
                          'modify': [self.cmd_modify,'Modify a work item',True],
                          #'action': [self.cmd_modify,'Change a work item state',True],
                          'whoami': [self.cmd_whoami,'Shows ID of user signed in to RTC',True],
                          'CLEAR': [self.cmd_removeconfigfiles,'Removes configuration files.',True],
                          'setcwe': [self.cmd_storecwe,'Set the default values to use for repository and project.',True],
                          'listqueries': [self.cmd_listqueries,'List queries in the project area.',True],
                          'runquery': [self.cmd_runquery,'Run query.',True],
                          'subscribe': [self.cmd_subscribe,'Subscribe yourself or other users to a work item.',True],
                          'unsubscribe': [self.cmd_subscribe,'Unsubscribe yourself from a work item.',True],
                          'link': [self.cmd_link,'Add a link from one work item to another.',True],
                          'unlink': [self.cmd_link,'Remove a link from a work item.',True],
                          'addcomment': [self.cmd_addcomment,'Add a comment to a work item.',True],
                          'list': [self.cmd_list,'List information about categories, actions, etc... See help for more details.',True],
                          'fun': [self.cmd_fun,'Just for fun',False],
                          'browser': [self.cmd_browser,'Open the specified work item in a web browser.',True]}
    
    def useroutput(self,content,level=1,channel=None):
        if level<=self.verbosity:
            if channel==None:
                if level>2:
                    channel = sys.stderr
                else:
                    channel = sys.stdout
            channel.write(content)
            channel.write(os.linesep)
    
    def setupParser(self): # set up the command line parameters parser
        usage = "Usage: %prog [options] subcommand"

        parser = OptionParser(usage=usage,add_help_option=False)
        # option groups:
        workitem = OptionGroup(parser,'Work Item Options','Options used on work item commands.')
        connections = OptionGroup(parser,'Connection Options','Specify what project area to work in.')

        parser.add_option('-v','--version',action='store_true',dest='version',help='Print version information',default=False)
        parser.add_option('--config',None,dest='config',help='Name of a file that contains a work item description.',default=self.DEFAULT_CONFIGDIR)
        parser.add_option('-h','--help',action='store_true',dest='help',help='Display help.',default=False)
        parser.add_option('--noprompt',None,action='store_true',dest='noprompt',help='Never prompt for input.',default=False)
        parser.add_option('-V','--verbosity',dest='verbosity',help='How much to output: 0=nothing, 1=minimal, 2=normal, 3=extra',choices=['0','1','2','3'])
        connections.add_option('-r','--repository',dest='repository',help='Repository URL')
        connections.add_option('-p','--project',dest='project',help='Project name')
        connections.add_option('-u','--user',dest='user',help='Username')
        connections.add_option('-n','--nickname',dest='nickname',help='Repository alias. Only used on login command.')
        connections.add_option('-P','--password',dest='password',help='Password')
        connections.add_option('-c','--cache',action='store_true',dest='cache',help='Caches password on prompt',default=False)
        workitem.add_option('-t','--wifile',dest='wifile',help='Input file that stores values for attributes used on the create, modify or change state actions.')
        workitem.add_option('-a','--action',dest='action',help='The action to perform on a work item modify command. This is used to change a work item state.')
        workitem.add_option('-x','--translate',dest='translate',help='Translate display values to and from internal values. Options: yes, no, to, from',choices=['yes','no','to','from'])
        #parser.add_option('-e','--excludelists',action='store_true',dest='excludelists',help='Exclude Empty Lists',default=False) # this wasn't enabled anywhere...
       
        parser.add_option_group(workitem)
        parser.add_option_group(connections)
        
        return parser
    
    def managedir(self,location):
        if not os.path.exists(location):
            os.mkdir(location, 0700)
            #os.mkdir(location)
    
    def loadnicknames(self): # read the nicknames file if it exists
        try:
            #nf = open(self.CONFIGDIR+self.NICKFILE,'r')
            self.nicks = shelve.open(self.CONFIGDIR+self.NICKFILE, writeback=True)#json.load(nf)
        except:
            self.logger.error('failed loading nicknames')
        else:
            pass
            #nf.close()
    
    def loadcache(self):
        cachefile = self.CONFIGDIR+'cache'
        try:
            cfp = open(cachefile,'r')
            self.cache = json.load(cfp)
            cfp.close()
        except:
            self.cache = {}
        #self.cache = shelve.open(cachefile,writeback=True)
        
    def storecache(self):
        cachefile = self.CONFIGDIR+'cache'
        cfp = open(cachefile,'w')
        json.dump(self.cache, cfp,sort_keys=True,indent=4,separators=(',',': '))
        cfp.close()
    
    def resolvenickname(self,options):
        if options.repository in self.nicks:
            return self.nicks[options.repository]['repository']
        else:
            return options.repository
        
    def updatenicknamefile(self,opts):
        nn = {}
        key = opts.repository
        if opts.nickname != None:
            key = opts.nickname
        nn['repository']=opts.repository
        nn['username']=self.cuser
        if opts.cache or opts.password != None:
            nn['password']=bz2.compress(self.cpassword)
        else:
            nn['password']=None
        self.nicks[key]=nn
        self.nicks.sync()
    
    def fields2JSON(self,flist):
        rval = ''
        for f in flist:
            if len(f)==0:
                rval = rval + '""'
            elif f == ':': # just in case : is separate, eclipse does this
                rval = rval+': '
            elif f in ['{','}','[',']']:
                rval = rval + f
            elif f[-1] == ':': # if : isn't separate, windows command line does this
                rval = rval+'"' + f[:-1] + '":'
            elif f == ',':
                rval = rval+', '
            elif f[-1] == ',': # if , isn't separate
                rval = rval+'"' + f[:-1] + '",'
            else: # double quote all values
                rval = rval+'"' + f + '"'
        
        if rval[0] not in ['{']: 
            rval = '{'+rval+'}' # curly bracket it to make it a JSON dict
        return rval
    
    def generateWI(self,flist, wifname):
        fwi = {}
        
        if wifname != None:
            try: 
                wifp = open(wifname,'r')
                fwi = json.load(wifp)
                wifp.close()
            except:
                raise CLMException(*getmsg("RW04001",inputfile=wifname))
        elif not sys.stdin.isatty():
            # read stdin
            try:
                lines = sys.stdin.readlines()
                content = ''.join(lines)
                content = content.strip()
            except:
                pass
            
            try:
                if len(content)>0:
                    fwi = json.loads(content)
            except:
                raise CLMException(*getmsg("RW04002"))
            
        if len(flist)>0:
            fstr = self.fields2JSON(flist)
            self.logger.debug('Command line values: '+fstr)
            try:
                cwi = json.loads(fstr)
            except:
                raise CLMException(*getmsg("RW04003",jsonval=fstr))
            
            fwi.update(cwi)
    
        return fwi
    
    def composeComment(self,comment, cfname):
        rc = ''
        
        if cfname != None:   
            wifp = open(cfname,'r')
            rc = wifp.read()
            wifp.close()
        elif not sys.stdin.isatty():
            # read stdin
            try:
                rc = sys.stdin.read()
            except:
                rc = ''
            
        if len(comment)>0:
            if len(rc)>0:
                rc = rc+' '
            rc = rc+comment
    
        return rc
    
    def loadreposessions(self):
        self.reposessions = shelve.open(self.CONFIGDIR+self.REPOFILE,writeback=True)
    
    def checkurl(self,url):
        pr = urlparse(url)
        rval = pr.scheme != None and len(pr.scheme)>0
        rval = rval and pr.netloc != None and len(pr.netloc)>0
        return rval
    
    def getrepokey(self,rname,uname):
        return str(rname)+'/'+str(uname)
    
    def getsession(self, options, doLogin): # get a RTCSession object that is signed in
        userepo = self.resolvenickname(options)
        if userepo == None:
            raise CLMException(*getmsg("RW03004"))
        if not self.checkurl(userepo):
            raise CLMException(*getmsg("RW03006",url=userepo))
        self.logger.debug('userepo: '+userepo)
        
        if options.repository in self.nicks:
            p = self.nicks[options.repository]
            if options.user == None:
                options.user = p['username']
                try:
                    options.password = bz2.decompress(p['password'])
                except:
                    options.password = None
                
            elif options.password == None and options.user == p['username']:
                tp = p['password']
                if tp!=None:
                    try:
                        options.password = bz2.decompress(tp)
                    except:
                        options.password = None
                    
        else:
            self.logger.debug('No info in nickname file')
        self.cuser = options.user
        self.cpassword = options.password

        # if doing a login command then we really have to log in
        self.loadreposessions()
        self.logger.debug('found reposessions: '+str(len(self.reposessions)))
        self.logger.debug(list(self.reposessions.keys()))
        if not doLogin:
            self.logger.debug('Not doing login: '+userepo)
            rkey = self.getrepokey(userepo,self.cuser)
            if rkey in self.reposessions:
                self.logger.debug('Found stored session: '+userepo)
                #   if s.restore(userepo,self.CONFIGDIR+self.SESSIONFILE):
                s = self.reposessions.get(rkey)
                s.requestSession.verify = False
                s.shell = self
                s.configdir = self.CONFIGDIR
                s.checkzlib()
                return s
                #if s.ping():
                #    return s
                #else:
                #    pass
            else:
                self.logger.debug('not saved: '+userepo)
                #s = jazzrest.RTCSession()

        s = jazzrest.RTCSession()
        
        self.logger.debug('Login needed')
        s.shell = self
        s.configdir = self.CONFIGDIR
        
        self.login(s,userepo,options.user,options.password)
        return s
    
    def login(self,s,repository,user,password):
                
        # get the password if it isn't set
        if user == None:
            raise CLMException(*getmsg("RW03002",repository=repository))
        if password == None:
            if self.inputprompt:
                password = self.getpassword(user)
                self.cpassword = password
            if password == None:
                raise CLMException(*getmsg("RW03003"))
        
        s.loginNew(repository, user, password)
        if s != None:
            self.reposessions[self.getrepokey(repository, user)]=s
            return s
        else:
            return None
    
    def loadcwe(self):
        try:
            cwefp = open(self.CONFIGDIR+self.CWEFILE,'r')
            self.cwe = json.load(cwefp)
            cwefp.close()
        except:
            self.cwe = {}
    
    def mergecwe(self,options):
        # set values from the current work environment table if the values aren't specified in command parameters
        if options.repository == None:
            t = os.getenv('RTCWI_REPOSITORY')
            if t == None and 'repository' in self.cwe:
                options.repository = str(self.cwe['repository'])
            else:
                options.repository = t
            
        if options.project == None:
            t = os.getenv('RTCWI_PROJECT')
            if t== None and 'project' in self.cwe:
                options.project = str(self.cwe['project'])
            else:
                options.project = t
        
        if options.user == None:
            t = os.getenv('RTCWI_USER')
            if t == None and 'user' in self.cwe:
                options.user = str(self.cwe['user'])
            else:
                options.user = t
        
        if options.verbosity == None:
            t = os.getenv('RTCWI_VERBOSITY')
            if t == None and 'verbosity' in self.cwe:
                options.verbosity = str(self.cwe['verbosity'])
            else:
                options.verbosity = t

        if options.translate == None:
            t = os.getenv('RTCWI_TRANSLATE')
            if t == None and 'translate' in self.cwe:
                options.translate = str(self.cwe['translate'])
            else:
                options.translate = t
    
    def cmd_storecwe(self,options,args):
        '''
Help for: setcwe
Purpose:
  Sets the current work environment. This stores the repository, project name and user name so that these parameters won't need to be specified on subsequent commands.

Usage: rtcwi.py setcwe [-r repository] [-p project] [-u userid] [-V verbosity]
'''        
        try:
            self.cwe = {}
            if options.repository != None:
                self.cwe['repository']=options.repository
            if options.project != None:
                self.cwe['project']=options.project
            if options.user != None:
                self.cwe['user']=options.user
            if options.verbosity != None:
                self.cwe['verbosity']=options.verbosity
            if options.translate != None:
                self.cwe['translate']=options.translate
            cwefp = open(self.CONFIGDIR+self.CWEFILE,'w')
            json.dump(self.cwe,cwefp)
            cwefp.close()
            self.listcwe(options)
        except:
            cwefp.close()
            
    def settranslate(self,options):
        translate = 'yes'
        if options.translate!=None:
            translate = options.translate
        
        if translate == 'yes':
            self.xlateto = True
            self.xlatefrom = True
        elif translate == 'no':
            self.xlateto = False
            self.xlatefrom = False
        elif translate == 'to':
            self.xlateto = True
            self.xlatefrom = False
        elif translate == 'from':
            self.xlateto = False
            self.xlatefrom = True
        else:
            raise CLMException(*getmsg("RW03202",parameter=options.translate))
            
    
    def startup(self):  
        parser = self.setupParser()
        (options, args) = parser.parse_args()

        if options.help or (len(args)>0 and args[0]=='help'):
            self.print_help(parser,options,args)
            return (None,None)
        
        self.CONFIGDIR = options.config
        if self.CONFIGDIR[-1] != os.sep:
            self.CONFIGDIR = self.CONFIGDIR+os.sep
        
        self.inputprompt = not options.noprompt
        self.loadcwe()
        self.mergecwe(options)
        iv = 2
        if options.verbosity != None:
            try:
                iv = int(options.verbosity)
            except:
                self.logger.debug("verbosity: "+options.verbosity)
                raise CLMException(*getmsg("RW03201",parameter=options.verbosity))
        self.verbosity = iv
            
        self.settranslate(options)
        self.logger.debug(options)
        self.logger.debug(args)
        self.logger.debug('config dir: '+self.CONFIGDIR)
        self.managedir(self.CONFIGDIR)

        return (options,args)
        
        
    def runcommand(self,options,args):
        if options.version:
            build = rtccliinfo.build
            
            try:
                pv = pkg_resources.get_distribution("rtcwi").version
            except:
                pv='[No package installed]'
            
            self.useroutput("rtcwi {0} {1}".format(pv,build))
            return 0
        
        if len(args)==0:
            self.useroutput('No command specified.')
            return 1

        rc = 0
        try:
            self.loadnicknames()
            self.loadcache()
            command = args[0]
            command = self.commandalias.get(command,command)
            func = self.commandmap.get(command,[self.unknowncmd,'',False])[0]
            rc = func(options,args)
        except CLMException as c:
            rc = 1
            message = "{0}: {1}".format(c.code,c.value)+os.linesep
            sys.stderr.write(message)
        except:
            self.logger.exception(sys.exc_info())
            rc = 1
            message,code = getmsg("RW00000")
            sys.stderr.write(code+": "+message+os.linesep)
        
        if self.reposessions != None:
            self.reposessions.close()
        if self.nicks != None:
            self.nicks.close()
        if self.cache != None:
            self.storecache()
        
        # no exceptions... return everything worked
        return rc
    
    def print_help(self,parser,options,args):
        command = None
        if len(args)>0 and args[0] != 'help' and (args[0] in self.commandmap or args[0] in self.commandalias):
            command = self.commandalias.get(args[0],args[0])
        elif len(args)>1 and (args[1] in self.commandmap or args[1] in self.commandalias):
            command = self.commandalias.get(args[1],args[1])
        
        if command != None:
            parts = self.commandmap[command]
            helptext = parts[0].__doc__
            if helptext == None:
                helptext = command+':\t'+parts[1]
            parser.print_help()
            
            self.useroutput(helptext, 1)
        else:
            parser.print_help()
            self.useroutput('\nAvailable Subcommands:')
            for (cmd,parts) in self.commandmap.items():
                if parts[2]:
                    self.useroutput('\t'+cmd+':\t'+parts[1])

    def cmd_login(self,options,args):
        '''
Help for: login
Purpose:
  Log in to the specified repository. This will store the session information so that subsequent commands can make use of this login until the session expires.
  Use the --cache or -c option to specify that the password should be cached.

Usage: rtcwi.py login [Connection Options]
'''
        s = self.getsession(options, True)
        if s != None:
            # update the nickname file
            self.updatenicknamefile(options)
        self.useroutput("Logged in to "+s.repository,2)
        return 0
            
    def cmd_createwi(self,options,args):
        '''
Help for: create
Purpose:
  Create a new work item of the specified type. Attribute values can be provided on the command line after the name of the work item type. 

Usage: rtcwi.py create [options] <type> [attribute values]
    type: Type of work item to create
    attribute values: Comma separated list of attribute: value pairs.

Examples: 
Create a new task with Summary set to "New task from command line" filed against "Category 1"
    rtcwi.py create Task Summary: "New task from command line", "Filed Against": "Category 1"

Create a new task with attribute values input from JSON formatted file named taskfile.json. Either use -t to specify the filename or the input can be put on stdin via redirection or pipe.
    rtcwi.py create Task -t taskfile.json
    rtcwi.py create Task < taskfile.json
'''
        if len(args)<2:
            raise CLMException(*getmsg("RW03100"))
        witype = args[1]
        newwi = self.generateWI(args[2:], options.wifile)
        s = self.getsession(options, False)
        xlator = None
        if self.xlateto or self.xlatefrom:
            xlator = self.gettranslator(s,options.project)
            
        if self.xlateto:
            witype = xlator.towitype(witype)
            newwi = xlator.tooslc(witype,newwi)
            
        (_,rwi) = s.createworkitem(newwi, options.project, witype)
        #self.logger.debug(code)
        if self.verbosity<2:
            # only display the ID
            ident = rwi.get("dcterms:identifier",'')
            self.useroutput(ident,1)
        elif self.verbosity>1:
            if self.xlatefrom:
                rwi = xlator.fromoslc(rwi)
            self.useroutput(json.dumps(rwi,sort_keys=True,indent=4,separators=(',',': ')),2)
            
        return 0
        
    def cmd_display(self,options,args):
        '''
Help for: display
Purpose:
  Display the contents of a work item.

Aliases: dsp
Usage: rtcwi.py [options] display <ID>
    ID: work item ID to display

Examples:
Display work item 21333:
    rtcwi.py display 21333
'''
        if len(args)<2:
            raise CLMException(*getmsg("RW03101",command=args[0]))
        wid = args[1]
        s = self.getsession(options, False)
        (_,rwi) = s.getworkitembyid(wid)
        if self.xlatefrom:
            pi = rwi.get("rtc_ext:contextId",None)
            if pi != None:
                self.logger.debug('session coreheaders: '+repr(s.coreheaders))
                xlator = translator.Translator(s,s.repository,pi,self.cache)
            else:
                xlator = self.gettranslator(s,options.project)
            rwi = xlator.fromoslc(rwi)
        self.useroutput(json.dumps(rwi,sort_keys=True,indent=4,separators=(',',': ')))
        return 0

    def cmd_modify(self,options,args):
        '''
Help for: modify
Purpose:
  Modify a work item. This command is also used to change the work item state with the -a (--action) option.

Aliases:
  Modify: mod

Usage: 
    rtcwi.py [options] modify <ID> [-a | --action Action] [attribute values]
    ID: ID of the work item
    Action: Name of the action to perform on the work item. Use this to change the work item state.
    attribute values: Comma separated list of attribute: value pairs.
Examples:
Modify the summary of work item 21333:
    rtcwi.py modify 21333 Summary: "new summary from command line"

Change state of task 21333 to "In Progress" by running the "Start Working" action and set estimate to 1 day:
    rtcwi.py modify 21333 -a "Start Working" Estimate: "1 d"

Modify also accepts JSON formatted attribute values from either standard input or with the -t option:
    rtcwi.py modify 21333 -t values.json
    rtcwi.py modify 21333 < values.json 
'''
        if len(args)<2:
            raise CLMException(*getmsg("RW03101",command=args[0]))
        s = self.getsession(options, False)
        wid = args[1]
        action = options.action
        self.logger.debug('Action: '+str(action))
        argindex = 2
        '''
        if args[0] == 'changestate' or args[0] == 'cs':
            if len(args)<3:
                raise CLMException(*getmsg("RW03102"))                
            action = args[2]
            self.logger.debug('action: '+action)
            argindex = 3
        '''
        (r,owi) = s.getworkitembyid(wid)
        etag = r.headers['etag']
        witype = self.getwitype(owi)
        
        newwi = self.generateWI(args[argindex:], options.wifile)
        if self.xlateto or self.xlatefrom:
            xlator = self.gettranslator(s,options.project)
            if action!=None:
                action = xlator.toactionid(witype, action)
                self.logger.debug('action: '+action)
            
        if self.xlateto:
            newwi = xlator.tooslc(witype,newwi)

        #newwi['dc:identifier']=wid
        (_,rwi) = s.modifyworkitem(newwi,wid,etag,action)
        if self.xlatefrom:
            rwi = rwi = xlator.fromoslc(rwi)
        self.useroutput(json.dumps(rwi,sort_keys=True,indent=4,separators=(',',': ')),2)
        return 0

    def cmd_subscribe(self,options,args):
        '''
Help for: subscribe, unsubscribe
Purpose:
  Add or remove subscribers from a work item.

Usage: 
    rtcwi.py [options] subscribe <ID> [userid1 userid2 userid3...]
    rtcwi.py [options] unsubscribe <ID> [userid1 userid2 userid3...]
    
    ID: ID of the work item
    If no userids are specified then the current user is subscribed or unsubscribed.
    
Subscribe yourself to work item 21333:
    rtcwi.py subscribe 21333

Subscribe testuser1@us.ibm.com and testuser2@us.ibm.com to work item 21333
    rtcwi.py subscribe 21333 testuser1@us.ibm.com testuser2@us.ibm.com

Unsusbscribe testuser1@us.ibm.com from work item 21333
    rtcwi.py unsubscribe 21333 testuser1@us.ibm.com
'''     
        if len(args)<2:
            raise CLMException(*getmsg("RW03101",command=args[0]))
        s = self.getsession(options, False)
        wid = args[1]
        
        subscribers = []
        if len(args)>2:
            subscribers = args[2:]
        xlator = None
        if self.xlateto or self.xlatefrom:
            xlator = self.gettranslator(s, options.project)

        if self.xlateto:
            nsub = []
            for sub in subscribers:
                nsub.append(xlator.persontouri(sub))
            subscribers = nsub
        
        #newwi['dc:identifier']=wid
        if args[0]=='subscribe':
            (code,rwi) = s.subscribe(wid,subscribers)
        elif args[0]=='unsubscribe':
            (code,rwi) = s.unsubscribe(wid,subscribers)
        
        self.logger.debug('Return code: '+str(code.status_code))

        
        if self.xlatefrom:
            ds = rwi.get('rtc_cm:subscribers',[])
            nwi = []
            for xs in ds:
                ov = xs.get('rdf:resource','')
                nwi.append(xlator.uritoperson(ov))
            rwi = nwi
        self.useroutput(json.dumps(rwi,sort_keys=True,indent=4,separators=(',',': ')),2)
        
        return 0

    def cmd_addcomment(self,options,args):
        '''
Help for: addcomment
Purpose:
  Add a comment to a work item.

Usage: rtcwi.py [options] addcomment <ID> [comment]
    ID: ID of the work item to add the comment to
    comment: Text of the comment to add. This can have a restricted set of HTML tags in it for rich text formatting. bold: <b> italics: <i>

Examples:
Add a comment to work item 21333
    rtcwi.py addcomment 21333 "This is a comment from the command line."
    rtcwi.py addcomment 21333 < commenttext.txt
    rtcwi.py addcomment 21333 -t commenttext.txt
'''
        if len(args)<2:
            raise CLMException(*getmsg("RW03101",command=args[0]))
        s = self.getsession(options, False)
        wid = args[1]
        comment = ''
        if len(args)>2:
            comment = args[2]
        comment = self.composeComment(comment, options.wifile)
        #newwi['dc:identifier']=wid
        if len(comment)>0:
            (_,_) = s.addcomment(wid,comment)
        else:
            raise CLMException(*getmsg("RW03103"))
        return 0
    

    def cmd_link(self,options,args):
        '''
Help for: link, unlink
Purpose:
  Add or remove links of the specified type on a work item.

Usage: rtcwi.py [options] link <ID> <link type> <work item ids>
    ID: work item ID
    link type: Type of the link to add or remove
    workitem ids: Space separated list of work item ids.

Examples:
Remove a Blocks link to 21333 for work item 21444
    rtcwi.py unlink 21333 Blocks 21444

Add children 21444, 21445 to work item 21333:
    rtcwi.py link 21333 Children 21444 21445
'''
        alen = len(args)
        if alen<2:
            raise CLMException(*getmsg("RW03101",command=args[0]))
        elif alen < 3:
            raise CLMException(*getmsg("RW03104",command=args[0]))
        elif alen < 4:
            raise CLMException(*getmsg("RW03105",command=args[0]))
 
        mode = args[0]
        s = self.getsession(options, False)
        wid = args[1]
        
        linktype = args[2]
        linkdest = args[3:]
        xlator = None
        if self.xlateto or self.xlatefrom:
            xlator = self.gettranslator(s,options.project)
        
        if self.xlateto:
            linktype = xlator.attributes.get('names',{}).get(linktype,linktype)
            nld = []
            for l in linkdest:
                if l.startswith('http'):
                    nld.append(l)
                else:
                    nld.append(s.repository+'/resource/itemName/com.ibm.team.workitem.WorkItem/'+l)
            linkdest = nld
        
        #newwi['dc:identifier']=wid
        if mode=='link':
            (_,_) = s.link(wid,linktype,linkdest)
            msg = "Links added"
        elif mode == 'unlink':
            (_,_) = s.unlink(wid,linktype,linkdest)
            msg = "Links removed"
 
        '''
        if self.xlatefrom:
            #self.logger.debug(json.dumps(rwi,sort_keys=True,indent=4,separators=(',',': ')))
            rwi = rwi = xlator.fromoslc(rwi)
        '''
        self.useroutput(msg,2)
        
        return 0

            

    def cmd_whoami(self,options,args):
        '''
Help for: whoami
Purpose:
  List the ID of the currently signed in user.

Usage: rtcwi.py [options] whoami
'''
        s = self.getsession(options, False)
        whoami = s.whoami()
        if self.xlatefrom:
            tid = whoami.split('/')[-1]
            whoami = tid.replace('%40', '@')
        self.useroutput(whoami)
        return 0
        
    def cmd_removeconfigfiles(self,options,args):
        '''
Help for: CLEAR
Purpose:
  Remove configuration files. By default the configuration files are kept in the .rtcwicli folder in the user's home directory but this can be changed with the --config option.

Usage: rtcwi.py [options] CLEAR [mode]
    mode: Specify what to remove. Either no value or one of:
        cwe: Remove the Current Work Environment file
        sessions: Remove the current sessions file
        nicknames: Remove the nicknames and credentials file
        cache: Remove the project area cache
        all: All config files
    If no value is provided for mode then cwe, sessions and nicknames will be removed
'''
        mode = 'default'
        if len(args)>1:
            mode = args[1]
        modes = {'cwe': [self.CONFIGDIR+self.CWEFILE],
                 'sessions': [self.CONFIGDIR+self.REPOFILE],
                 'nicknames': [self.CONFIGDIR+self.NICKFILE],
                 'cache': [self.CONFIGDIR+'cache'],
                 'all': [self.CONFIGDIR+self.CWEFILE,self.CONFIGDIR+self.REPOFILE,self.CONFIGDIR+self.NICKFILE,self.CONFIGDIR+'cache'],
                 'default': [self.CONFIGDIR+self.CWEFILE,self.CONFIGDIR+self.REPOFILE,self.CONFIGDIR+self.NICKFILE]
                 }
        rfiles =  modes.get(mode,None)
        self.logger.debug('Removing mode: '+mode+' '+str(rfiles))
        if mode in ['all','cache']:
            self.cache = None
        if rfiles == None:
            raise CLMException(*getmsg("RW03112",clearmode=mode))
        pfiles = []
        
        for fn in rfiles:
            try:
                if os.path.exists(fn):
                    os.remove(fn)
            except:
                pfiles.append(fn)
        
        if len(pfiles)>0:
            raise CLMException(*getmsg("RW09001",files=','.join(pfiles)))
        return 0
    
    def cmd_listqueries(self,options,args):
        '''
Help for: listqueries
Purpose:
  List the queries available in the current project area.

Aliases: lq
Usage: rtcwi.py [options] listqueries [personal]
    personal: If personal is specified after listqueries then only the current user's personal queries are listed.
'''
        scope = 'shared'
        if len(args)>1:
            scope = args[1]
        
        s = self.getsession(options, False)
        (_,rwi) = s.listqueries(options.project,None,scope)
        if self.xlatefrom:
            rwi = self.fromquerylist(rwi)
        self.useroutput(json.dumps(rwi,sort_keys=True,indent=4,separators=(',',': ')))
        return 0
    
    def cmd_runquery(self,options,args):
        '''
Help for: runquery
Purpose:
  Runs the specified query.

Aliases: rq
Usage: rtcwi.py [options] runquery <query name> [personal]
    query name: Name of the query to run.
    personal: Use this if you own a query with the same name as another shared query. This will run your personal version of the query instead of the shared.
'''
        if len(args)<2:
            raise CLMException(*getmsg("RW03106"))
        query = args[1]
        
        scope = 'shared'
        if len(args)>2:
            scope = args[2]
        
        s = self.getsession(options, False)
        (_,rwi) = s.runquery(options.project,query,scope)

        self.useroutput(json.dumps(rwi,sort_keys=True,indent=4,separators=(',',': ')))
        return 0

    def cmd_browser(self,options,args):
        '''
Help for: browser
Purpose:
  Open the specified work item or URL in a web browser.

Usage: rtcwi.py [options] browser <item>
    item: Either a work item ID or a URL.
'''
        if len(args)<2:
            raise CLMException(*getmsg("RW03101",command=args[0]))
        
        url = args[1]
        if not url.startswith('http'):
            if options.repository == None:
                raise CLMException(*getmsg("RW03004"))
            url = options.repository+'/resource/itemName/com.ibm.team.workitem.WorkItem/'+url
        rv = webbrowser.open_new_tab(url)
        if rv != True:
            raise CLMException(*getmsg("RW09002",url=url))
        return 0
            
        
    def cmd_list(self,options,args):
        '''
Help for: list
Purpose:
  List information about the RTC CLI environment or project area.

Aliases: ls
Usage: rtcwi.py [options] list <topic> [qualifiers...]
    topic: Topic to list information about. See below for more details.
    qualifiers: Additional information about what to list. See below for more details.
    
Topics and qualifiers:

Environment:
    cwe: Show the current work environment. Qualifiers: none
    credentials: Show current credential information. Qualifiers: none
    sessions: Show current sessions. Qualifiers: "check" this will validate whether the session is active.

Project information:
    types: Show the work item types available within the project area.
    actions: Show the actions available for the specified work item type. Qualifier: work item type to list.
    attributes: Show the attributes registered with the specified work item type. Qualifier: work item type to list.
    values: Show valid values for the specified attribute. This may not be available for all attributes. Qualifier: attribute to list.
    
Examples:
List actions available for STG Defect:
    rtcwi.py list actions "STG Defect"

List all stored sessions and check whether they are still open:
    rtcwi.py list sessions check
'''

        topic = None
        if len(args)>1:
            topic = args[1]
        else:
            raise CLMException(*getmsg("RW03107"))
        
        if topic=='actions':
            if len(args)>2:
                witype = args[2]
                s = self.getsession(options,False)
                self.useroutput('Actions for '+witype+':')
                actns = self.getactions(witype,s,options.project)
                for x in actns:
                    self.useroutput('\t'+x)
            else:
                raise CLMException(*getmsg("RW03109"))
        elif topic=='values':
            if len(args)<3:
                raise CLMException(*getmsg("RW03110"))
            attr = args[2]
            s = self.getsession(options,False)
            winame = None
            if len(args)>3:
                winame = args[3]

            av = self.getallowedvalues(attr,s,options.project,winame)
            if av==None:
                self.useroutput("Couldn't find allowed values for: "+attr)
            else:
                self.useroutput("Allowed values for: "+attr)
                if len(av)==0:
                    self.useroutput('\tNone')
                else:
                    for a in av:
                        self.useroutput('\t'+a)
            pass
        elif topic=='cwe':
            self.listcwe(options)
        elif topic=='credentials':
            self.listcredentials()
        elif topic=='sessions':
            docheck = len(args)>2 and args[2]=='check' 
            self.listsessions(docheck)
        elif topic=='types':
            s = self.getsession(options,False)
            xlator = self.gettranslator(s, options.project)
            nl = xlator.getworkitemtypes()
            if len(nl)>0:
                self.useroutput('Work Item types for '+options.project)
                for win in nl:
                    self.useroutput('\t'+win)
            else:
                self.useroutput('No Work Item types for '+options.project)
        elif topic=='attributes':
            if len(args)<3:
                raise CLMException(*getmsg("RW03111"))
            s = self.getsession(options,False)
            xlator = self.gettranslator(s, options.project)
            nl = xlator.getattributes(args[2])
            if len(nl)>0:
                self.useroutput('Attributes for '+args[2])
                for a in nl:
                    self.useroutput('\t'+a)
            else:
                self.useroutput('No attributes for '+args[2])
        else:
            raise CLMException(*getmsg("RW03108",topic=topic))
        return 0

    def listcwe(self,options):
        #cwe = self.cwe.items()
        #if len(cwe)>0:
        self.useroutput("Current work environment settings:")
        self.useroutput('\tRepository: '+str(options.repository))
        self.useroutput('\tProject: '+str(options.project))
        self.useroutput('\tUser ID: '+str(options.user))
        self.useroutput('\tVerbosity: '+str(options.verbosity))
        self.useroutput('\tTranslate: '+str(options.translate))

    def getactions(self,witype,session,project):
        x = self.gettranslator(session, project)
        tk = x.towitype(witype)
        self.logger.debug('WI Type key: '+tk)
        alist = x.workitems.get('details',{}).get(tk,{}).get('actions',{}).get('names',{})
        
        actions = alist.keys()
        
        return actions

    def getallowedvalues(self,attribute,session,project,winame=None):
        x = self.gettranslator(session, project)
        attrk = x.attributes.get('names',{}).get(attribute,attribute)
        if winame != None:
            witype = x.towitype(winame)
            tav = x.getallowedvalues(attrk,witype)
        else:
            tav = x.getallowedvalues(attrk)
        if tav == None:
            return None
        return sorted(tav.keys())



    def listsessions(self,docheck):
        if self.reposessions == None:
            self.loadreposessions()
            for rs in self.reposessions.items():
                sk = rs[0]
                s = rs[1] 
                cval = ''
                if docheck:
                    try:
                        cval = s.ping()
                    except:
                        cval = False
                    
                    cval = 'Active: '+str(cval)
                self.useroutput("{0} {1}".format(sk,cval))
    
    def listcredentials(self):
        if self.reposessions == None:
            self.loadreposessions()
        
        if len(self.nicks)>0:
            for nk in self.nicks.items():
                ni = nk[1]
                repo = ni['repository']
                details = repo
                if nk[0]!=repo:
                    details = details+" ("+nk[0]+")"
                details = details+" User: "+ni['username']
                details = details+" Cached: "+str(ni['password']!=None)
                details = details+" Session: "+str(repo in self.reposessions)
                self.useroutput(details)
        else:
            self.useroutput("No credential information stored.")
    
    help_fun = 'Not used.'
    def cmd_fun(self,options,args):
        newwi = self.generateWI(args[1:], options.wifile)
        print(json.dumps(newwi,sort_keys=True,indent=4,separators=(',',': ')))
        return 0
    
    def unknowncmd(self,options,args):
        self.useroutput('No such command: '+args[0])
        return 1
    
    def gettranslator(self,session,project):
        if project == None:
            raise CLMException(*getmsg("RW03005"))
        pi = session.getprojectuuid(project)
        self.logger.debug('session coreheaders: '+repr(session.coreheaders))
        x = translator.Translator(session,session.repository,pi,self.cache)
        return x
    
    def getwitype(self,wi):
        if "rtc_cm:type" in wi:
            return wi["rtc_cm:type"]["rdf:resource"].split("/")[-1]
        return None

    def fromquerylist(self,qrl):
        nl = []
        for x in qrl['oslc_cm:results']:
            ne = {}
            ne['Name']=x['dc:title']
            ne['Description']=x['dc:description']
            #ne['URL']=x['rdf:resource']
            nl.append(ne)
        
        return nl
    
    def fromqueryresult(self,qrl):
        nl = []
        for x in qrl:
            ne = {}
            ne['Summary']=x['dc:title']
            ne['ID']=x['dc:identifier']
            nl.append(ne)
        return nl
    
    def getpassword(self,user):
        return getpass.getpass('Enter password for '+user+': ')

